/* bustermain.c
 * 
 * Port of Buster to Packet Island probe's Embedded Linux
 * 
 * Copyright 2003-2008 by Praemio Inc. All rights reserved.
 * Author: John Bartas - jbartas@praemio.com
 * This software is released under the terms of the LGPL.
 * See LGPL Notice at end of this file; Full license in ../license.txt
 */

#include "busterport.h"
#include "brnet.h"
#include "ip.h"
#include "brdb.h"

#include <sys/time.h>
#include <linux/if_packet.h>
#include <pcap.h>
#include <sys/ioctl.h>

#include "cap_sess.h"
//#include "cap_pkt.h

extern int  br_startup(void);
extern int  br_check(void);

#define CAP_SESS_FRAMES    256

u_long  cticks = 0;     /* increment TPS times a second */

int http_port = 88;     /* Listen port for GUI web server - configurable */

struct br_thread * pthreads[MAXTHREADS];

static char * iface_name;   /* Iface to watch */

static int    iface_index = -1;     /* From SIOCGIFINDEX */

extern  int br_rxpkt(int iface, PACKET rxpkt);

char * runtime_dir = "/tmp";

extern int  br_startdb(void);
extern int  br_startnet(void);


int br_system_shutdown = 0;     /* Flag; Non-zero if Buster is shutting down */

int    console_startup(void);   /* Start a menu thread and key processor */
static int    use_console = FALSE;

void** mutexlist[] = 
{
    &hosttab_mutex,
    &ip_hosttab_mutex,
    &control_ip_mutex,
    &ce_table_mutex,
    &br_comm_mutex,
    &br_dns_mutex,
    &br_tree_mutex,
    &control_ip_mutex,
    &largebufq.mutex,
    &smallbufq.mutex,
    &emptybufq.mutex,
    &alarms_free.mutex,
    &alarms_queued.mutex
};

#define NUMMUTII    (sizeof(mutexlist)/sizeof(void**))

pthread_mutex_t     mutii[NUMMUTII];  /* The actual mutex objects */


/* Create a thread for timer */
TH_OBJECT(to_timer);
TH_PROTO(te_timer);

struct br_thread timer_thread = 
{
    &to_timer,
    "Timer",
    te_timer,
    APP_PRIORITY,
    APP_STACK_SIZE,
};

void    usage()
{
    printf("discover --iface ifname\n");
}

static char macaddress[6];

int  set_macaddress()
{
    char *  cp;
    cp =  getenv("probe_entity_id");
    if(!cp)
        return ENP_PARAM;
    return (hex2bin(cp, macaddress, 6));
}


static long   sendsock;

int probe_raw_send(PACKET pkt)
{
    int     err;
    struct sockaddr_ll raw_sock_addr;

    if(iface_index == -1)   // Not ready yet?
        return ENP_RESOURCE;

    raw_sock_addr.sll_family = AF_PACKET;
    raw_sock_addr.sll_protocol = ARP_TYPE; 
    raw_sock_addr.sll_ifindex = iface_index;

    err = sendto(sendsock, pkt->pk_prot, pkt->pk_plen, 0, 
                 (struct sockaddr*)&raw_sock_addr, sizeof(raw_sock_addr));
    if(err < 1)
    {
        dtrap();
        err = errno;
        perror("raw send error");
        pkt->pk_net->txerrors++;
        return ENP_RESOURCE;
    }

    if(err == pkt->pk_plen)
    {
        pkt->pk_net->txbytes += err;
        pkt->pk_net->txpkts++;
    }
    else    /* really? */
    {
        dtrap();
        pkt->pk_net->txerrors++;
    }
    pk_free(pkt);

    return 0;
}

int
buster_main(int argc, char * argv[] )
{
    int         i;
    int         err;
    cap_sess_t  cap_sess;
    struct ifreq ifr;

    if(argc < 2)
    {
        usage();
        return -1;
    }

    for(i = 1; i < argc; i++)
    {
        if( strstr(argv[i], "-iface") )
        {
            iface_name = argv[i + 1];
            i++;
        }
        else if( strstr(argv[i], "-m") )
        {
            use_console = TRUE;
        }
        else
        {
            usage();
            return(-1);
        }
    }

    err = chdir(runtime_dir);
    if(err)
    {
        printf("Can't chdir to \"%s\"/n", runtime_dir);
        return -1;
    }

    if (!cap_sess_create( &cap_sess,
         iface_name,
         CAP_SESS_FRAMES,
         NULL,
         0 ))
    {
        printf("capture session for %s failed", iface_name );
        return(-1);
    }

    /* Create the mutexes pthreads will need */
    for(i = 0; i < NUMMUTII; i++)
    {
        err = pthread_mutex_init(&mutii[i], NULL);
        if(err)
        {
            printf("Mutex init error %d\n", err);
            return -1;
        }
        *mutexlist[i] = &mutii[i];  /* install the mutex */
    }

    /* Start the passive collection core */
    err = br_startdb();
    if(err)
    {
        dprintf("Fatal br_startdb error %d\n", err);
        return -1;
    }

    strncpy(nets[0]->name, iface_name, sizeof(nets[0]->name));
    strcpy(nets[0]->description, "Packet Island MA Ethernet");

    printf("Calling br_setipinfo()\n");

    /* get the IP address from the probe environment */
    err = br_setipinfo(0);
    if( err )
    {
        dprintf("Error getting IP info\n");
        return -1;
    }

    printf("Probe IP addr: %u.%u.%u.%u\n", PASS_IPADDR(nets[0]->n_ipaddr));
    printf("  subnet mask: %u.%u.%u.%u\n", PASS_IPADDR(nets[0]->n_mask));
    printf("       router: %u.%u.%u.%u\n", PASS_IPADDR(nets[0]->n_router));

    if( set_macaddress())
    {
        dprintf("Error getting MAC address\n");
        return -1;
    }
    nets[0]->n_macaddr = macaddress;
    nets[0]->ifOperStatus = 1;
    nets[0]->ifAdminStatus = 1;

    ifNumber = 1;

    if(use_console)
    {
        if(console_startup())
        {
            dprintf("Error starting menu console\n");
            return -1;
        }
    }

    /* Set up a raw socket for sending gratuitous ARPs, etc */
    sendsock = socket(PF_PACKET, SOCK_RAW, ARP_TYPE);
    if(sendsock == -1)
    {
        dprintf("Can't open raw socket for sending\n");
        return -1;
    }

    memset(&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, iface_name, sizeof(ifr.ifr_name));

    /* get index for pass to bind, send */
    err = ioctl(sendsock, SIOCGIFINDEX, &ifr);
    if(err < 0)
    {
        dprintf("Can't index raw socket for sending\n");
        return -1;
    }
    iface_index = ifr.ifr_ifindex;
    nets[0]->n_send = probe_raw_send;

    /* Start the Buster's interfaces */
    err = br_startnet();
    if(err)
    {
        dprintf("Fatal br_startnet error %d\n", err);
        return -1;
    }
    
    while(1)
    {
        void *                  frame;
        struct tpacket_hdr *    tphdr;
        PACKET                  pkt;

        frame = cap_sess_recv( &cap_sess );
        if (frame)
        {
            tphdr = (struct tpacket_hdr *)frame;

            pkt = pk_alloc(0);
            if(!pkt)
            {
                /* Count errors here later???? */
                usleep( 100 );  /* wait, maybe it will get better */
                continue;
            }
            pkt->pk_buff = frame + tphdr->tp_mac;
            pkt->pk_blen = tphdr->tp_snaplen;
            pkt->pk_prot = pkt->pk_buff + 14;
            pkt->pk_plen = pkt->pk_blen - 14;
            pkt->pk_net = nets[0];

            br_rxpkt( 0, pkt);  /* pass packet w/Iface index 0 */

            cap_sess_release( &cap_sess, frame );
        }
        else        
        {
            usleep( 2 );
        }
    }

	return 0;
}


TH_ENTRY(te_timer)
{
    while(1)
    {
        usleep( 1000000/TPS );     /* Sleep for one TPS-tick */

        cticks++;

        /* Once a second call the main timer loop */
        if((cticks % TPS) == 0)
            br_check();
    }
    return NULL;
}


 void
LinuxTrap(char * file, int line) // embedded break point
{
    printf("dtrap: file %s, line %d\n", file, line);
}


/* br_nbiosocket( long socket )
 * 
 * Portability routine to set a socket into non-blocking mode. BSD 
 * sockets botched this by using unportable "ioclt()", so it's different   
 * on just about every system.
 *
 * Returns 0 if OK, else negative error ENP_ code. 
 */

#include <fcntl.h>

int
br_nbiosocket( long socket )
{
    int err = -1;

    err = fcntl(socket, F_SETFL,  O_NONBLOCK );
    if(err)
    {
        dprintf("Error %d setting socket to non-blocking mode\n", errno);
    }
    // should return 0 if OK 
    return err;
}


void
panic(char * msg)
{
    printf("panic: %s\n", msg);
    exit( -1 );
}




void    TH_SLEEP( u_long ticks )
{
    usleep( (ticks * 1000)/TPS );
}

/* HACK: - In this port the ponly thing we even wake if the alert task, which 
 * runns with sleeps rather than blocks anyway. Just no-op this out. 
 */

void    TH_WAKE( long taskId )
{
    USE_ARG(taskId);
    return;
}

static int threads = 0;

int     TH_NEWTASK(struct br_thread * p)
{
    int err;

    /* Finish constructure of br_thread */
    p->th_ptr = &p->thread;     /* Set pointer to embedded pthread */
    err = pthread_create(p->th_ptr, NULL, p->entry, 0);

    if(threads < MAXTHREADS)
        pthreads[threads++] = p;
    
    return err;
}


u_long  br_gettime()
{
    return ( time( NULL ) );
}

void    br_user_alert(char * msg)
{
    //    dtrap();    /* Put msg on LDCs? */
}

void    br_flash(int itterate, int level)
{
    //    dtrap();    /* Flash an LED? */
}

void
br_lanchanged()
{
    dprintf("LAN seems to have changed");
}

#if defined(arch_arm)
unsigned short
cksum(void * iph, unsigned int ihl)
{
    unsigned int sum, tmp1;

    __asm__ __volatile__(
            "ldr    %0, [%1], #4                   \n\
            ldr     %3, [%1], #4                                    \n\
            sub     %2, %2, #5                                      \n\
            adds    %0, %0, %3                                      \n\
            ldr     %3, [%1], #4                                    \n\
            adcs    %0, %0, %3                                      \n\
            ldr     %3, [%1], #4                                    \n\
            1:      adcs    %0, %0, %3                                      \n\
            ldr     %3, [%1], #4                                    \n\
            tst     %2, #15                 @ do this carefully     \n\
            subne   %2, %2, #1              @ without destroying    \n\
            bne     1b                      @ the carry flag        \n\
            adcs    %0, %0, %3                                      \n\
            adc     %0, %0, #0                                      \n\
            adds    %0, %0, %0, lsl #16                             \n\
            addcs   %0, %0, #0x10000                                \n\
            mvn     %0, %0                                          \n\
            mov     %0, %0, lsr #16"
    : "=r" (sum), "=r" (iph), "=r" (ihl), "=r" (tmp1)
    : "1" (iph), "2" (ihl)
    : "cc", "memory");
    return sum;
}
#else   /* X86, we hope */
unsigned short
cksum(void * iph, unsigned wlen)
{
    u_short *   wordp;
    u_long      lsum;
    u_short     ssum;

    lsum = 0;
    wordp = (u_short *)iph;
    while(wlen-- > 0)
        lsum += *wordp++;

    ssum = lsum & 0x0000FFFF;
    do
    { 
        lsum >>= 16;
        lsum += (u_long)ssum;
        ssum = lsum & 0x0000FFFF;
    }
    while(lsum & 0xFFFF0000);

    return ssum;
}
#endif  /* arch_XXX */


/* br_ifsetup()
 *
 * Portability routine called to detect interface changes during 
 * Periods when Buster  was suspended. 
 * 
 * 
 * 
 * Returns -1 if system not ready yet, or
 * else returns number of interfaces.
 */

int   br_ifsetup()
{
    /* On PI probe it should be OK to always return 1. */
    return 1;
}

/* br_getenvip() - 
 * 
 * helper fuinction for br_setipinfo(int iface)
 */

int br_getenvip(char * envparm, ip_addr * ipaddrp)
{
    char *      cp;
    ip_addr     ipaddr;

    cp = getenv(envparm);
    if(!cp) 
    {
        dprintf("ENV parm \"%d\" not set\n", envparm); 
        dtrap();
        return ENP_RESOURCE;
    }
    cp = br_checkip(&ipaddr, cp);
    if(cp)
    {
        dprintf("ENV IP parm \"%s\" error \"%s\"", envparm, cp); 
        dtrap();
        return ENP_RESOURCE;
    }
    *ipaddrp = ipaddr;
    return 0;
}


/* br_setipinfo(int iface)
 *
 * - Called to re-read and interface's IP info from system
 *
 * The system should only update the info if it suspects the info
 * is currently inaccurate OR there is no old info.
 *
 * Returns 0 if OK, else EMP_ code
 */

int   br_setipinfo(int iface)
{
    ip_addr ipaddr, mask, router;

    if(br_getenvip("ipaddr", &ipaddr))
        goto bad_env_info;

    if(br_getenvip("netmask", &mask))
        goto bad_env_info;

    if(br_getenvip("gatewayip", &router))
        goto bad_env_info;

    /* Set info in the one & only iface */
    staticnets[iface].n_ipaddr = ipaddr;
    staticnets[iface].n_mask = mask;
    staticnets[iface].n_router = router;

    return 0;

bad_env_info:
    dtrap();    // Shouldn't happen
    return ENP_RESOURCE;
}

int     br_beep()
{
//    printf("beep!");
    return 0;
}

int     br_popup(struct alarm_type * type, char * errtext)
{
//    printf("popup: %s", errtext);
    USE_ARG(type);
    return 0;
}

/* Portability routine to write the UTP date/time passed.  */

int      br_writedatetime(u_long intime, char * out, int outlen)
{
    char * cp = ctime(&intime);
    if(!cp)
        return -1;
    strncpy(out, cp, outlen);
    return 0;    
}


/* Make a formatted time string for the log day changes */

static char rbuf[64];

char *  br_getlogdate()
{
    /* Should be OK on UNIX/Linux */
    br_writedatetime( time(NULL), rbuf, sizeof(rbuf));
    return rbuf;
}

char *  br_getlogtime()
{
    char * colon;
    br_writedatetime( time(NULL), rbuf, sizeof(rbuf));

    /* Extract the "xx:xx:xx" part */
    colon = strchr(rbuf, ':');
    if(colon)
    {
        *(colon + 9) = 0;
        return (colon - 2);
    }
    else
        return NULL;
}

/* br_portrunning()
 *
 * Port specific startup stuff.
 *
 */

void  br_portrunning()
{
    /* Nothing for PI Probe? */
    return;
}

/* br_portthreads()
 * 
 * Menu routine to print status of all threads
 * 
 */

void  br_portthreads(void * iodev, char * params)
{
    int i;

    fprintf(iodev, "Status of pthreads:\n");
    fprintf(iodev, "name           blocks   sleeps   wakes    yields\n");
    for(i = 0; i < MAXTHREADS; i++)
    {
        if( pthreads[i] == NULL)
            break;

        fprintf(iodev, "%15s: %9lu %9lu %9lu %9lu \n",
                pthreads[i]->name,
                pthreads[i]->blocks,
                pthreads[i]->sleeps,
                pthreads[i]->wakes,
                pthreads[i]->yields);
    }

    return;
}


/* br_portsuspendresume()
 *
 * Portability routine, called when the receiver is suspecnded
 * or resumed.
 *
 */

void
br_portsuspendresume(int disabled)
{
    USE_ARG(disabled);
    return;
}

#include <netdb.h>

/* Portability routine - Wrapper for gethostbyname() */

u_long   br_dnslookup(char * textname)
{
    struct hostent * hosts;

    hosts = gethostbyname(textname);
    if(hosts)
        return *(u_long*)(hosts->h_addr_list[0]);
    else
        return 0L;
}

void
br_startcticks()
{
    int     err;

    dprintf("Starting discover timer thread\n");
    err = TH_NEWTASK(&timer_thread);
    if( err)
    {
        panic("cticks thread");        
    }
    return;
}


#include "log.h"

int 
disc_log(char * string, ...)
{
    va_list ap;
     
    va_start (ap, string);
    vprintf (string, ap);
    va_end (ap);

    return 0;
}

extern int  br_menucmd(void * outdev, char * input);

extern char * prompt;

/* Create a thread for timer */
TH_OBJECT(to_console);

TH_PROTO(te_console)
{
    char    cmdbuf[256];

    printf( prompt );
    for(;;)
    {
        fgets(cmdbuf, sizeof(cmdbuf), stdin );
        br_menucmd( stdout, cmdbuf);
    }
}

struct br_thread console_thread = 
{
    &to_console,
    "Console",
    te_console,
    APP_PRIORITY,
    APP_STACK_SIZE,
};

int    console_startup()
{
    int     err;
    err = TH_NEWTASK(&console_thread);
    return err;
}

/*
Copyright (C) 2003-2008 by Praemio Inc. All rights reserved.

Author: John Bartas - jbartas@praemio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/



