/* braccess.c
 *
 * Buster control access routines. 
 *
 * Copyright 2003-2008 by Praemio Inc. All rights reserved.
 * Author: John Bartas - jbartas@praemio.com
 * This software is released under the terms of the LGPL.
 * See LGPL Notice at end of this file; Full license in ../license.txt
 *
 * All Buster's blocked objects (IP address, DNS name, URL path
 * ports, etc.) contain a ctdesc structure, aka a "control". The
 * fields for whitelisting and access timers are in this struct.
 * Taken together these fields are referred to as "access".
 * Changing access settings for some of the object types will
 * affect other objects, for example granting a 10-minute access
 * timer to a DNS name requires making the same access field
 * changes in any associated IP addresses.
 * This file contains the routines to handle these dependancies.
 * The extenal API to this is two routines: 
 *
 *		ctl_setacess() - set new access states
 *		ctl_getacess() - fetch access state
 *
 * ROUTINES: ctl_setacess(), ctl_getacess(), 
 * ROUTINES: br_setdntimer(), br_setdnacccess(), ctl_set1access(), 
 *
 */

#include "busterport.h"

#include "brnet.h"
#include "ip.h"
#include "brdb.h"



/* ctl_getaccess(struct ctdesc * ctl)
 *
 * Get the access state of a ctl structure. This is the 
 * external entry point for Buster code to get access info
 * about an object. It should be FAST!
 *
 * Returns values indicate blocking state of control's item
 *   0 - item is blacklisted w/o timer (no access) 
 *   1 - item is whitelisted (indefinite access)
 *   2 - item has an access timer set
 */

int
ctl_getaccess(struct ctdesc * ctl)
{
   if(ctl->action_mask & CTF_WHITELIST)
      return 1;      /* item is whitelisted */

   /* fall to here if CTF_WHITELIST bit is not set, see if 
    * a tmp access timer is set 
    */
   if(ctl->timeout)
   {
      /* This is the primary test for Buster to expire one of
       * these access timers. These timers should not be 
       * accessed directly unless a similar test is provided.
       */
      if(ctl->timeout <= cticks)  /* has access timer expired? */
      {
         ctl->timeout = 0;
         return 0;      /* no timer (anymore) */
      }

      /* valid access timer, still not expired */
      return 2;
   }
   else
      return 0;   /* Neither access timer or whitelist */
}



/* ctl_set1access() - set access for a single control
 *
 * Paramter "type" indicates type of access to set
 *   0 - item is to be blacklisted (no access) 
 *   -1 - item is whitelisted (indefinite access)
 *   >0 - type is number of minutes to allow access
 */

static void
ctl_set1access(struct ctdesc * ctl, int access)
{
   if(access > 0)      /* type is number of minutes for access */
   {
      ctl->timeout = cticks + (access * (60 * TPS));
   }
   else if(access == 0)      /* blacklist it */
   {
      ctl->action_mask &= ~CTF_WHITELIST;    /* clear whitelist bit */
      ctl->timeout = 0;                /* clear timer */
   }
   else if(access == -1)     /* whitelist it */
   {
      ctl->action_mask |= CTF_WHITELIST;  /* set whitelist bit */
      ctl->timeout = 0;                   /* clear timer */
   }
   else
   {
      dtrap();    /* bad access value */
   }
}



/* br_setdnacccess(dnp, int access)
 * set access in DNS blocklist entry and associated IP 
 * blocklist entries.
 *
 * access: -1 for whilelist, 0 for blacklist, positive value
 * (in ticks) for an access timer.
 */

static int
br_setdnacccess(struct control_dn * dnp, long access)
{
   struct control_ip * cip;

   ctl_set1access(&dnp->ctl, access);

   /* if there are IP blocks, loop though them; setting timer.*/
   for(cip = dnp->cu_cip; cip; cip = cip->ci_dnlist)
   {
      ctl_set1access(&cip->ctl, access);
   }
   return 0;
}



/* ctl_setaccess(struct ctdesc * ctl)
 *
 * Set the access state of a ctl structure, and make any required
 * adjustsments to related objects.
 *
 * Paramter "type" indicates type of access to set
 *   0 - item is to be blacklisted (no access) 
 *   -1 - item is whitelisted (indefinite access)
 *   >0 - type is number of minutes to allow access
 *
 * Parameter "object" is a pointer to the object containing 
 * the control. This is used if the control indicates an object 
 * type which requires adjusting other objects (eg. DNS name) 
 * Objects with no associates (eg. path) do not use this 
 * parameter, however it should be passed in case future needs 
 * requires it.
 *
 */

void
ctl_setaccess(struct ctdesc * ctl, long access, void * object)
{
   struct control_dn * dnp;
   struct control_ip * ipe;

   switch(ctl->action_mask & CTF_TYPEMASK)
   {
   case  CTF_TYPENAME:
      dnp = (struct control_dn *)object;
      if(&dnp->ctl != ctl)     /* sanity check */
      {
         /* Object may be an IP address associated with the domain name
          * rather than the name itself. If so we want to set access for
          * both the name and the IP address.
          */
         ipe = object;
         if(&ipe->ctl == ctl)
            goto set_ipaccess;

         dtrap(); /* object not IP control object, barf... */
         return;
      }
      br_setdnacccess(dnp, access);
      break;
   case  CTF_TYPEIP:
      ipe = (struct control_ip *)object;
      if(&ipe->ctl != ctl)     /* sanity check */
      {
         dtrap();
         return;
      }

set_ipaccess:
      /* If the IP address is associated with a DNS name then set
       * access for the DNS name and all it's associated IPs,
       * which will include this one.
       */
      if(ipe->ci_name)
         br_setdnacccess(ipe->ci_name, access);
      else     /* no DNS, set access for IP entry only */
         ctl_set1access(ctl, access);
      break;
   case  CTF_TYPEPORT:
   case  CTF_TYPEPATH:
   case  CTF_TYPEWORD:
   case  CTF_TYPEMAC:
      ctl_set1access(ctl, access);
      break;
   default:
      dtrap();    /* control type not set */
      return;
   }

   return;
}

/*
Copyright (C) 2003-2008 by Praemio Inc. All rights reserved.

Author: John Bartas - jbartas@praemio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/



