/*
 * brarp.c
 *
 * Copyright 2003-2008 by Praemio Inc. All rights reserved.
 * Author: John Bartas - jbartas@praemio.com
 * This software is released under the terms of the LGPL.
 * See LGPL Notice at end of this file; Full license in ../license.txt
 *
 * Buster ARP code, including gratuitous ARP. 
 *
 */


#include "busterport.h"
#include "brnet.h"
#include "ip.h"
#include "arp.h"
#include "dhcpdef.h"
#include "brdb.h"


struct arpentry * arp_table[MAXARPS];  /* the actual hash table */

/* Make hash for ARP entry - done by 6 byte MAC */

int arp_hash(ip_addr ip)
{
    int hash;

    hash = (ip & 0xFF) + (ip >> 8) + (ip >> 16) + (ip >> 24);
    return (hash % MAXARPS);
}


/* Lookup an IP address int he ARP table. IP is passed in 
 * NET endian 
 *
 * returns entry if found, else NULL.
 */

struct arpentry *   arp_lookup(ip_addr dest_ip, NET net)
{
    int     hash;
    struct arpentry *   entry;

    hash = arp_hash(dest_ip);
    for(entry = arp_table[hash]; entry; entry = entry->next)
    {
        if((entry->ipaddr == dest_ip) && (entry->net == net))
            return entry;
    }

    return NULL;
}


/* arp_mkentry()
 *
 * Make ARP hash atble entry. IP is passed in NET endian.
 *
 * Return new entry or NULL if no heap.
 */

struct arpentry *   arp_mkentry(ip_addr dest_ip, NET net)
{
    struct arpentry *   entry;
    int                 hash;

    if( (entry = arp_lookup(dest_ip, net)) != NULL)
    {
        dtrap();
        return entry;
    }
    entry = AT_ALLOC( sizeof(struct arpentry));
    if(!entry)
        return NULL;
    entry->ipaddr = dest_ip;
    entry->net = net;
    entry->createtime = entry->lasttime = cticks;

    hash = arp_hash(dest_ip);
    entry->next = arp_table[hash];
    arp_table[hash] = entry;

    return entry;
}


/* arp_rcv() - passed all received raw packets of Ethernet ARP type.
 *
 * Returns 0 if processed without incident, else ENP_ error
 */

int   arp_rcv(PACKET pkt)
{
    struct arp_hdr *    arph;
    struct arpentry *   entry;
    ip_addr             sender_ip;
    struct host_entry * host;
    int                 iface;   /* scratch */

    arph = (struct arp_hdr *)(pkt->pk_buff + MAC_HDRLEN);
    getlong(&arph->data[ARP_SRCIP], &sender_ip);

    if(arph->ar_hd != ARP_ETHTYPE)
        return ENP_PARAM;

    sender_ip = htonl(sender_ip);
    entry = arp_lookup( sender_ip, pkt->pk_net);

    if(entry)
    {
        /* Make sure the ARP entry has a matching MAC address. If is does
         * not it may have changed due to DHCP or similar activity. This 
         * also happens when the machine changes LANs; for example 
         * 192.168.1.1 may exist on both nets but with different MACs. 
         */
        if( MAC_EQ(entry->macaddr, &arph->data[ARP_SRCMACHA]) == FALSE )
        {
            if( MAC_EQ(entry->macaddr, nullmac) == FALSE )
            {
                dprintf("ARP: IP %u.%u.%u.%u on net %s on new MAC adress.\n",
                   PASS_IPADDR(sender_ip), pkt->pk_net->description );

                /* lookup host of MAC which send packet. */
                host = NULL;
                for(iface = 0; iface < (int)ifNumber; iface++)
                {
                   host = br_hostlookup(&arph->data[ARP_SRCMACHA], iface);
                   if(host)
                      break;
                }
                if(!host)
                {
                   dtrap();   /* host should have been created  by receiver */
                   return ENP_RESOURCE;
                }
                /* See if old MAC changed nets or if this is a different host */
                if( nets[iface] == pkt->pk_net)
                {
                   /* Nets are same, so IP changed */
                   host->iphost = br_iplookup(iface, sender_ip );
                }
                else    /* MAC changed nets */
                {
                   dtrap();
                }
            }
            else    /* entry MAC was not set - set it if ARP is reply */
            {
                if(arph->ar_op == ARP_REPCODE)
                    MEMCPY(entry->macaddr, &arph->data[ARP_SRCMACHA], 6);
            }
        }
        entry->lasttime = cticks;

        /* See if an outgoing IP packet is waiting for this reply */
        if(entry->pending)
        {
            entry->net->n_send(entry->pending);
            entry->pending = NULL;
        }
        return 0;   /* Done with ezisting entry, No error */
    }

    /* Fall to here if this is not in the ARP table. */
    entry = arp_mkentry(sender_ip, pkt->pk_net);
    if(!entry)
        return ENP_NOMEM;
    MEMCPY(entry->macaddr, &arph->data[ARP_SRCMACHA], 6);
    
    return 0;
}


/* send_via_arp()
 * 
 * Send an IP packet on ethernet when we have the IP address of the next hop, 
 * but not the MAC address. This resolves the MCA and sends the packet. 
 * If we need to send an ARP the get the MAC, we queue the packet for later
 * sending and return ENP_SEND_PENDING.
 *
 * Returns 0 if pkt sent, ENP_SEND_PENDING if we had to arp, else error.
 *
 */

int   send_via_arp(PACKET pkt, ip_addr nexthop)
{
    struct arpentry *   entry;
    struct eth_hdr *    eth;
    int                 err;

    entry = arp_lookup(nexthop, pkt->pk_net);
    if(entry)
    {
        eth = (struct eth_hdr *)pkt->pk_buff;
        MEMCPY(eth->e_dst, entry->macaddr, 6);
        eth->e_type = IP_TYPE;
        err = entry->net->n_send(pkt);
        return err;
    }

    if(pkt->pk_net == NULL)
        dtrap();

    entry = send_arp_net(pkt->pk_net, nexthop);
    if(!entry)
    {
        dprintf("Can't make ARP entry");
        dtrap();
        pk_free(pkt);
        return ENP_NOMEM;
    }
    if(entry->pending)
    {
       dprintf("dropping old pending ARP packet");
       pk_free(entry->pending);
    }
    entry->pending = pkt;

    return 0;
}

/* send_arp_net()
 * 
 * Send an ARP request on the net passed. This can be used for 
 * gratuitous arp.
 *
 * Returns ARP table entry for the new transaction is OK, else returns
 * NULL on error.
 */

struct arpentry *   send_arp_net(NET net, ip_addr dest_ip)
{
    struct arp_hdr *    arph;
    struct arpentry *   entry;
    struct eth_hdr *     eth;
    PACKET              pkt;
    int                 err;

    if(net->ifOperStatus != 1)
        return NULL;

    if( ! net->n_send )
    {
        dtrap();        /* porting error? */
        return NULL;
    }

    entry = arp_lookup(dest_ip, net);
    if(!entry)
    {
        /* no ARP table entry, make one */
        entry = arp_mkentry(dest_ip, net);
        if(!entry)
            return NULL;
    }
    pkt = pk_alloc(60);
    if(!pkt)
        return NULL;

    pkt->pk_net = net;

    eth = (struct eth_hdr *)pkt->pk_buff;
    arph = (struct arp_hdr *)(pkt->pk_buff + MAC_HDRLEN);

    /* Build braodcast MAC header for ARP */
    MEMSET(eth->e_dst, 0xFF, 6);
    MEMCPY(eth->e_src, net->n_macaddr, 6);
    eth->e_type = ET_ARP;

    /* Build ARP request header */
    MEMSET(arph, 0, sizeof(*arph));
    arph->ar_hd = ARP_ETHTYPE;
    arph->ar_op = ARP_REQCODE;
    arph->ar_pro = ARP_IPTYPE;
    arph->ar_hln = 6;
    arph->ar_pln = 4;
    MEMCPY( &arph->data[ARP_SRCMACHA], net->n_macaddr, 6);
    MEMCPY( &arph->data[ARP_SRCIP], &net->n_ipaddr, 4);
    MEMCPY( &arph->data[ARP_DESTIP], &dest_ip, 4);

    pkt->pk_prot = pkt->pk_buff;
    pkt->pk_plen = sizeof(struct eth_hdr) + sizeof(struct arp_hdr);

    err = net->n_send(pkt);
    if(err)
        return NULL;
    else
        return entry;
}


/* arp_scan()
 * 
 * Scan for next 5 hosts on the interface passed. ARP "scan" consists
 * of sending ARP requests to all 254 possible IP addresses on our
 * (hopefully) class C subnet. We only send 5 of these on each call 
 * to avoid flooding packet buffers, either in NDIS, mbufs, or internally. 
 * This does not wait for replies; those are received and logged by 
 * arp_rcv(). If this is called once a second, the process will take 
 * just under a minute. 
 * 
 * The variable arp_scanning is used as both a flag and container for the 
 * next IP address for which to ARP. When it's zero, we're done.
 * 
 * Returns 0 if next block of arps sent OK, else ENP_ code.
 */


int arp_scanning[MAXNETS];
int arp_scan_rate = 6;      /* Packets per sec for arp scan */

int
arp_scan(int iface)
{
    int         i;
    int         hostip;
    ip_addr     netip;
    ip_addr     sendip;
    struct arpentry *   entry;

    netip = nets[iface]->n_ipaddr & htonl(0xFFFFFF00);
    hostip = (ip_addr)arp_scanning[iface];

    for(i = 0; i < arp_scan_rate; i++)
    {
        hostip++;
        if(hostip >= 255)   /* done? */
        {
            hostip = 0;
            break;
        }
        sendip = (netip | htonl(hostip));
        entry = send_arp_net( nets[iface], sendip );
        if(entry == NULL)
            return ENP_SENDERR;
    }
    arp_scanning[iface] = hostip;
    return 0;
}

/* arp_cleanup()
 *
 * Called periodically to clean out old arp entries
 *
 *
 */

void
arp_cleanup()
{
    int     hash;
    struct arpentry * entry;
    struct arpentry * last;

    for(hash = 0; hash < MAXARPS; hash++)
    {
       last = NULL;
       for(entry = arp_table[hash]; entry; entry = entry->next)
       {
          if((entry->lasttime + (300 * TPS)) < cticks)
          {
             if(entry->pending)
                pk_free(entry->pending);
             entry->pending = NULL;
             if(last)
                last->next = entry->next;
             else
                arp_table[hash] = entry->next;

             AT_FREE(entry);
             break;
          }
          last = entry;
       }
    }
}

/*
Copyright (C) 2003-2008 by Praemio Inc. All rights reserved.

Author: John Bartas - jbartas@praemio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/



