/* brconfig.c
 *
 * Configuration related portions of Buster
 *
 * Copyright 2003-2008 by Praemio Inc. All rights reserved.
 * Author: John Bartas - jbartas@praemio.com
 * This software is released under the terms of the LGPL.
 * See LGPL Notice at end of this file; Full license in ../license.txt
 *
 */


#include "busterport.h"
#include "brnet.h"
#include "ip.h"
#include "brdb.h"


extern   void     br_nullterm(char * arg);
extern   struct  bk_motive * mot_byname(char * motive_text);

/* MAC image btree */
struct btnode * image_tree;

/* Bitmask of ctl types to be blocked. Deprecated? -JB- */
u_long   featuremask = 0xFFFFFFFF;

/* Internal to this file */
int   br_writeparms(FILE * fp);


/* Configuration. Includes times for aging out closed 
 * TCP or idle UDP connections 
 */

struct br_config br_cfg = 
{
   (30 * MINUTE),       /* Time to hold closed TCP connections  */
   (30 * MINUTE),       /* Time to keep idle UDP connections  */
   (20 * MINUTE),       /* Time to keep idle ICMP connections  */
   10000,               /* max. connections, total */
   1000,                /* max. connections per host */
   1000,                /* connections per host for alarm */
   (3 * HOUR),          /* cticks to keep idle remote host */
   (SECOND * 20),       /* interval between first two alarms of a given type */
   /* cflags - configuration boolean settings */
   BC_VIRUS_ALERT|BC_VIRUS_KILL|BC_DESPAM|BC_NOEXPERT,
   20,                  /* unacked syns per minute (to detect floods) */
   (120 * SECOND),      /* screen refresh interval */
   0x0000000f,          /* deletion log bitmask */
   64,                  /* Max. size of log in Kbytes */
   HP_BUSTER|HP_LOCAL,   /* hide Buster | hide locaL */
   'i',                 /* sortparm - sort by IP address */
   0,                   /* filterport 0==wildcard */
   FP_ALL,              /* filterparm - show all */
   600,                 /* udp_idle timeout, in seconds */
   2880,                /* mactmo, elapsed minutes to condider a MAC idle */
   240,                 /* update_interval, in minutes */
};


/* Application name keeps changing (fucking marketing). Try to automate this */
#ifdef NNSQUAD
char * br_appname = "NNAgent";
#else
char * br_appname = "Buster";
#endif


/* Version. There is both a string, which indicats config options, 
 * and a number which indicates code revisions.
 * Newer versions are always numerically higher than older ones.
 */

u_long   br_versnum = 0x02000006;

/* Windows exception handler requires the version string be a pointer
 * and not a char array. Weird....
 */
static char vbuffer[128];  /* e.g. "Praemio 2.0.0.3" or "NN-Squad 2.0.0.3" */
char *   br_version = vbuffer;      


struct laninfo * br_lans;

#ifdef NNSQUAD
char * parmsfile = "nnma.cfg";
#else
char * parmsfile = "buster.cfg";
#endif

static   int   userread = 0;     /* number users read from NV */
static   int   permread = 0;     /* number permissions read from NV */
static   int   passread = 0;     /* number passsord read from NV */

static char userbuf[MAXUSERLENGTH + 2];
static char passbuf[MAXUSERLENGTH + 2];

static char    lanname[MAXNAMELEN];    /* name of current LAN */
static u_char  lanmac[8];              /* MAC of current LAN */
static int     whoserverct = -1;       /* index of whois servers */

/* tmp community info. After name, file, url, priority are set a 
 * community is created. 
 */ 
static char cm_name[MAXFILENAME];
static char cm_filename[MAXFILENAME];
static char cm_url[MAXURL];
static int  cm_priority = 0;

int    nv_commvar(void);

struct parmdef  /* One of these for each parameter */
{
    char * format;
    void * pvalue;      /* pointer to value to read/write */
    /* optional read function */
    int   (*readfunc)(struct parmdef * def, char * parm);
    int   (*writefunc)(struct parmdef * def, void * parm);
};


int   brp_username(struct parmdef * def, char * parm);
int   brp_userpass (struct parmdef * def, char * parm);
int   brp_userpermit(struct parmdef * def, char * parm);
int   brp_alertmask(struct parmdef * def, char * parm);
int   brp_whoisserver(struct parmdef * def, char * parm);
int   brp_whoisport(struct parmdef * def, char * parm);
int   brp_whoistype(struct parmdef * def, char * parm);
int   br_commname(struct parmdef * def, char * parm);
int   br_commfile(struct parmdef * def, char * parm);
int   br_commpri(struct parmdef * def, char * parm);
int   br_commurl(struct parmdef * def, char * parm);
int   br_viewif(struct parmdef * def, char * parm);


struct parmdef  br_params[] =
{
   { "TCP timeout: %d\n", &br_cfg.keep_tcpconn, NULL },
   { "UDP timeout: %d\n", &br_cfg.keep_udpconn, NULL },
   { "ICMP timeout: %d\n", &br_cfg.keep_icmpconn, NULL },
   { "Max conns: %d\n", &br_cfg.max_connections, NULL },
   { "Host conns: %d\n", &br_cfg.max_connperhost, NULL },
   { "Alarm conns: %d\n", &br_cfg.alarm_connperhost, NULL },
   { "Idle host: %d\n", &br_cfg.keep_iphost, NULL },
   { "Alarm interval: %d\n", &br_cfg.br_alarminterval, NULL },
   { "Config flags: %xh\n", &br_cfg.cflags, NULL },
   { "SYN threshold: %d\n", &br_cfg.syns_per_minute, NULL },
   { "Screen refresh: %d\n", &br_cfg.screen_refresh, NULL },
   { "Logging mask: %xh\n", &br_cfg.logevents, NULL },
   { "Max Log size: %dKB\n", &br_cfg.logsize, NULL },
   { "Hide mask: %xh\n", &br_cfg.hide_parm, NULL },
   { "Sort parameter: %c\n", &br_cfg.sortparm, NULL },
   { "UDP Idle timeout: %d\n", &br_cfg.udp_idle, NULL },
   { "Update interval: %lu\n", &br_cfg.update_interval, NULL },

   /* The next three lines may itterate as a group */
   { "Username: %s\n", NULL, brp_username  },
   { "Userpass: %s\n", NULL, brp_userpass  },
   { "Permissions: %xh\n", NULL, brp_userpermit  },

   { "Alertmask: %s\n", NULL, brp_alertmask  },

   /* The next three lines may itterate as a group */
   { "Whois server: %s\n", NULL, brp_whoisserver  },
   { "Whois port: %u\n", NULL, brp_whoisport  },
   { "Whois type: %s\n", NULL, brp_whoistype  },

   { "View iface: %s\n", NULL, br_viewif  },
    
   { "MAC idle min: %d\n", &br_cfg.mactmo, NULL  },

#ifdef BR_COMMUNITIES
   /* The next four lines may itterate as a group */
   { "Community Name: %s\n", NULL, br_commname  },
   { "Community File: %s\n", NULL, br_commfile  },
   { "Community Priority: %d\n", NULL, &br_commpri },
   { "Community URL: %s\n", NULL, br_commurl  },
#endif   /* BR_COMMUNITIES */

   { "MyServer active: %s\n", &br_srcserving, NULL  },

};

static int num_params = sizeof(br_params)/sizeof(struct parmdef);


/* br_isbcast() - returns TRUE if IP address is a broadcast according
 * to the current configuration.
 */
 
int
br_isbcast(int iface, ip_addr ipaddr)
{
   if(ipaddr == 0xffffffff)
      return TRUE;

   if(nets[iface]->n_netbr == 0)
   {
      nets[iface]->n_netbr = nets[iface]->n_ipaddr & nets[iface]->n_mask;
   }
   if(ipaddr == nets[iface]->n_netbr )
      return TRUE;

   return FALSE;
}


/* Return TRUE if the passed IP address is a standard private address */

int
isprivate_ip(ip_addr ipaddr)
{
   ipaddr = htonl(ipaddr);
   if((ipaddr & 0xFFFF0000) == 0xc0a80000)   /* 192.168.X.X */
      return TRUE;
   if((ipaddr & 0xFF000000) == 0x0a000000)   /* 10.X.X.X */
      return TRUE;
   if((ipaddr & 0xFF000000) == 0xa900000)   /* 169.X.X.X */
      return TRUE;

   return FALSE;
}

/* islocal_ip() - return TRUE if passed IP is on local LAN */

int
islocal_ip(int iface, ip_addr ipaddr)
{
   u_long   localnet;

   /* If IP address or mask is not set, guess based on public/private 
    * aspects of passed IP address.
    */
   if(nets[iface]->n_mask == 0)     /* Can't do much w/o mask */
      return isprivate_ip(ipaddr);

   if((nets[iface]->n_ipaddr == 0) && (nets[iface]->n_router != 0))
   {
      /* Get here if we have a mask & gateway, but no IP address */
      localnet = nets[iface]->n_router & nets[iface]->n_mask;
   }
   else  if(nets[iface]->n_ipaddr == 0)
   {
      /* mask, but no IP or router */
      return isprivate_ip(ipaddr);
   }
   else /* we have an IP and mask set for this iface */
   {
      /* This is the "normal" path: */
      localnet = nets[iface]->n_ipaddr & nets[iface]->n_mask;
   }
   
   if(localnet == 0)
   {
//      dtrap();       /* Endian problem? */
      return FALSE;
   }

   /* test for local address per mask */
   if((ipaddr & nets[iface]->n_mask) == localnet)
      return TRUE;

   /* if it's a broadcast address, assume it's local */
   if(br_isbcast(iface, ipaddr))
      return TRUE;

   return FALSE;  /* not a local address */
}


/*********** DB read routines for MAC images ************/

/* MAC image btree test route */
int
bt_mactest(struct btnode * n, u_long key)
{
   int         i;
   u_char *    cp;
   u_char *    mac1;
   u_char *    mac2;
   u_long      mac_min;
   u_long      mac_max;
   u_long      mac_passed;
   struct image * imp;

   imp = (struct image *)n->data;
   mac1 = (u_char *)&imp->macbase[0];
   mac2 = (u_char *)key;

   /* first 3 bytes are IEEE assigned; straight compare */
   for(i = 0; i < 3; i++)
   {
      if(*mac1 > *mac2)
         return KEY_GT;
      if(*mac1 < *mac2)
         return KEY_LT;
      mac1++;
      mac2++;
   }

   /* rest of MAC may be covered by range, do a value test */
   cp = (u_char *)&imp->macbase[3];
   mac_min = (*cp << 16) + (*(cp + 1) << 8) + (*(cp + 2));
   mac_max = mac_min + imp->range;
   mac_passed = (*mac2 << 16) + (*(mac2 + 1) << 8) + (*(mac2 + 2));

   if(mac_min > mac_passed)
      return KEY_GT;
   if(mac_max < mac_passed)
      return KEY_GT;
   return KEY_EQ;
}

/* br_mkimage()
 *
 * Make database entry for image from passed text record.
 *
 * Returns NULL if OK, else error test.
 */


char *
br_mkimage(char * readbuf, struct comm_dbud * comm, int commit)
{
   struct image * img;
   char *   mac;
   char *   cp;
   char *   imagename;  /* name of image file */
   int      length;     /* generic */
   int      range;      /* Number of MACs to cover */
   u_char   macbuf[14];

   mac = br_getfield(readbuf, "mac_address");
   if(!mac)
      return "no MAC address";
   else
   {
      if(strcrlen(mac) != 12)
         return "MAC address wrong size";
      if(hex2bin(mac, &macbuf[0], 12 ))
         return "bad MAC format";
   }

   cp = br_getfield(readbuf, "mac_number");
   if(!cp)
      range = 1;
   else
   {
      range = atol(cp);
      if(range < 1)
         return "bad mac_number";
   }
   
   imagename = br_getfield(readbuf, "mac_address");
   if(!imagename)
      return "no image file name";

   length = strcrlen(imagename);
   if(length > sizeof(img->imagefile))
      return "image file name too long";

   /* If we're just checking syntax, return OK now */
   if(!commit)
      return NULL;

   /* If we got here then everything parsed OK, so allocate 
    * the structure and create the entry.
    */

   img = IMG_ALLOC(sizeof(*img));
   if(!img)
      return "alloc";
   img->btnp = BTN_ALLOC(sizeof(struct btnode));
   if(img->btnp == NULL)
      return "alloc";
   img->btnp->data = (struct btnode **)img;     /* set back pointer */

   /* Save image file name - length still has number of bytes */
   strncpy(img->imagefile, imagename, length);

   /* Save rest of item info in structure */
   MEMCPY(img->macbase, macbuf, sizeof(img->macbase));
   img->range = range;
   read_ctlinfo(readbuf, comm, CTF_TYPEMAC, &img->ctl);

   /* Install new image structure in btree */
   bt_insert( (struct btnode**)&image_tree, img->btnp, 
      (u_long)(&img->macbase), bt_mactest);

   return NULL;
}


/* tree helper routine to write MAC image tree recs */

void
br_imageout(struct btnode * n, long param, int depth)
{
   struct image * img;
   char *   cp;
   char     filebuf[256]; /* generic file io buffer */
   struct comm_dbud * comm = (struct comm_dbud *)param;
   FILE * fout = comm->cm_fd;

   img = (struct image *)n->data;

   /* skip if this item is not in current community */
   if(img->ctl.comm != comm)
      return;

   sprintf(filebuf, "type: HostImage\r\nfilename: %s\r\nmac_address: ",
      img->imagefile);

   cp = filebuf + strlen(filebuf);
   sprintf(cp, "%02X%02X%02X%02X%02X%02X", 
      img->macbase[0],
      img->macbase[1],
      img->macbase[2],
      img->macbase[3],
      img->macbase[4],
      img->macbase[5]);

   cp += strlen(cp);
   sprintf(cp, "\r\nmac_number: %lu\r\n", img->range);
      
   cp += strlen(cp);
   sprintf(cp, "filename: %s\r\n",  img->imagefile );

   append_ctl(&img->ctl, filebuf, sizeof(filebuf) );
   fprintf(fout, "%s\n", filebuf);

   USE_ARG(depth);
   return;
}

/* File Parameters system routines & support 
 *
 * The rest of ther file contains code to read/write a text file 
 * with formatted parameters from runtime memory. The file lines 
 * generally have the format "param_name: value". Each parameter
 * requires a parmdef structure table entry. Simeple strings/ints 
 * can read with a generic routine, more complex stuff (i.e.
 * itterated tables) require special optional read routines.
 *
 * All writing of parameters is done by a single monolithic routine.
 *
 */


int
brp_username(struct parmdef * def, char * parm)
{
    if(strlen(parm) > MAXUSERLENGTH)
    {
        dprintf("User name \"%s\" too long\n");
        return ENP_PARAM;
    }
    if(userread > permread)
    {
        dprintf("Missing password or permission for user %s\n", userbuf);
        return ENP_PARAM;
    }
    userread++;
    strncpy( userbuf, parm, MAXUSERLENGTH);

    return 0;
}

/* Read the MD5 encoded user password from a hex string */

int
brp_userpass (struct parmdef * def, char * parm)
{
    if(strlen(parm) > MD5PASSLENGTH)
    {
        dprintf("password \"%s\" too long\n");
        return ENP_PARAM;
    }
    passread++;
    if(passread != userread)
    {
        dprintf("NV password without matching user\n");
        return ENP_PARAM;
    }
    strncpy( passbuf, parm, MD5PASSLENGTH);
    return 0;
}

int brp_userpermit(struct parmdef * def, char * parm)
{
    u_long  tmp;
    char *  cp;
    int     err;

    permread++;
    if(permread != userread)
    {
        dprintf("NV permissions without matching user\n");
        return ENP_PARAM;
    }
    tmp = atoh(parm);
    if(strlen(passbuf) != 32)
    {
        cp = br_pwdigest(passbuf, tmp);
        strcpy(passbuf, cp);
    }
    err = add_user(userbuf, passbuf, tmp);
    /* add_user returns 0 if user was accepted, error if not */
    if(err)
        return ENP_RESOURCE;    /* User table must be full */
    return 0;
}

int brp_alertmask(struct parmdef * def, char * parm)
{
    char *  cp;
    int     i,j;
    
    cp = parm;
    for(i = 0; i < alertcount; i++)
    {
        for(j = 0; j < alert_methodct; j++)
        {
            if(*cp++ == '1')
                alarms[i]->actions |= a_methods[j].bitmask;
            else
                alarms[i]->actions &= ~a_methods[j].bitmask;
        }
    }
    return 0;
}

/* The code for brp_lanname and brp_lanrtmac just saves
 * the variables in static holders. The brp_lanrtip code 
 * below should be called last. It collects all three variables
 * and calls new_lan() the create the actual LAN entry. 
 */

int
brp_lanname(struct parmdef * def, char * parm)
{
   strncpy(lanname, parm, sizeof(lanname) );
   return 0;
}

int
brp_lanrtmac(struct parmdef * def, char * parm)
{
   hex2bin(parm, lanmac, 6);
   return 0;
}

int 
brp_lanrtip(struct parmdef * def, char * parm)
{
    char *  cp;
    ip_addr  lan_router;

    if(lanname[0] == 0)  /* test to make sure name was set first */
    {
        dprintf("Lan IP address must be preceeded by LAN name and router MAC address\n");
        return ENP_PARAM;
    }
    /* use space_server var for router IP address */
    cp = br_checkip(&lan_router, parm);
    if(cp)
    {
        dprintf("Bad LAN router IP address (%s) in NV file\n", parm);
        return ENP_PARAM;
    }
    if(new_lan(lanname, lanmac, lan_router) == NULL)
    {
        dprintf("Duplicate LAN name %s in NV file?\n", lanname);
        return ENP_PARAM;
    }
    lanname[0] = 0;   /* clear name for future test */

    return 0;
}


int brp_whoisserver(struct parmdef * def, char * parm)
{
    /* get server index this var is initialized to -1 */
    whoserverct++;
    if(whoserverct >= WHOSERVERS)    /* if it's too high, re-use last */
        whoserverct--;

    strncpy(whoservers[whoserverct].dnsname, parm, MAXDNSNAME);
    return 0;
}

int brp_whoisport(struct parmdef * def, char * parm)
{
    whoservers[whoserverct].whoport = atoi(parm);
    return 0;
}

int brp_whoistype(struct parmdef * def, char * parm)
{
    if(stricmp(parm, "IP") == 0)
        whoservers[whoserverct].flags |= WH_IPTYPE;
    if(stricmp(parm, "NAME")  == 0)
        whoservers[whoserverct].flags |= WH_NAMETYPE;

    /* If no type was set then default to name */
    if((whoservers[whoserverct].flags & (WH_NAMETYPE|WH_IPTYPE)) == 0)
        whoservers[whoserverct].flags |= WH_NAMETYPE;
    return 0;
}

int br_viewif(struct parmdef * def, char * parm)
{
    if(_strnicmp(parm, "Auto", 4) == 0)
        br_autoifview = TRUE;
    else
    {
        br_autoifview = FALSE;
        br_ifview = atoi(parm);
    }
    return 0;
}


int br_commurl(struct parmdef * def, char * parm)
{
    struct comm_dbud * comm;
    
    /* See if this is a known community */
    comm = br_cmlookup(parm);
    if(comm)
    {
        dprintf("NV file: Community URL %s already in use\n", parm);
        dtrap();
        return -1;
    }

    strcpy(cm_url, parm);   /* Save parm in tmp buffer */
    return( nv_commvar() ); /* Do we have enough info to make comm? */
}

int br_commname(struct parmdef * def, char * parm)
{
   /* New community name, clearm other community vars */
   cm_filename[0] = 0;
   cm_url[0] = 0;
   cm_priority = 0;
   strcpy(cm_name, parm);  /* record name, wait for other vars */

   return 0;
}

int br_commfile(struct parmdef * def, char * parm)
{
    strcpy(cm_filename, parm);
    return( nv_commvar() );
}

int br_commpri(struct parmdef * def, char * parm)
{
    cm_priority = atoi(parm);
    return 0;
}



static int
parmread(struct parmdef * parm, char * input)
{
    if(sscanf(input, parm->format,  parm->pvalue) != 1)
    {
        dprintf("params scanf error, input \"%s\", pattern \"%s\"\n",
            input, parm->format);
        dtrap();
        return -1;
    }
    else
        return 0;
}

struct parmdef * 
parm_lookup(char * name)
{
	int		i;
    char * colon;   /* valid parms are terminated by a colon */
    int     parmlen;

    colon = strchr(name, ':');
    if(colon)
        parmlen = colon - name;
    else
        parmlen = strlen(name);

    for(i = 0; i < num_params; i++)
    {
        if(strncmp(br_params[i].format, name, parmlen ) == 0)
            return &br_params[i];
    }
    return NULL;
}


int
br_readparms()
{
    FILE *  infile;
    char    buffer[128];
    int     lines;
    int     err;
    struct parmdef * parm;

    infile = fopen(parmsfile, "r");
    if(infile == NULL)
    {
        dprintf("unable to open parameters file %s\n", parmsfile);
#ifdef EMBEDDED_DEVICE
        dprintf("Using default settings:\n");
        br_writeparms( stdout );
        dprintf("End default settings.\n...");
        return 0;
#else   /* Windows or Linux */
        return ENP_FILEIO;
#endif /* EMBEDDED_DEVICE */
    }

    lines = 0;
    while(fgets(buffer, sizeof(buffer), infile))
    {
        lines++;
        if(*buffer == '\n')
            continue;
        if(*buffer == '\r')
            continue;
        if(*buffer == '#')
            continue;
        if(*buffer == ';')
            continue;
        parm = parm_lookup(buffer);
        if(!parm)
        {
            dprintf("unsupport parameter \"%s\" in file %s, line %d\n",
                buffer, parmsfile, lines);
            continue;
        }
        if(parm->readfunc)
        {
            char * end;
            char * cp = strchr(buffer, ':');
            if(!cp)
            {
                dprintf("Missing colon in \"%s\" in file %s, line %d\n",
                        buffer, parmsfile, lines);
                return -1;
            }
            cp++;   /* Bump past colon */
            while(*cp == ' ' )      /* bump past spaces */
                cp++;

            /* Strip trailing linefeeds */
            end = cp + strlen(cp) - 1;
            while((*end == '\n') || (*end == '\r'))
                *end-- = 0;

            err = parm->readfunc(parm, cp); /* past "foo: " */
        }
        else
            err = parmread(parm, buffer );
        if(err)
        {
            dprintf("Error %d reading parameter %s\n", err, buffer);
            return ENP_PARAM;
        }
    }
    return 0;
}


/* nv_commvar() - see if we have enough info to make a new community. 
 * If we do make a community, clear the info to prepare for the next 
 * community.
 */

int 
nv_commvar()
{
   char *   cp;
   int      i;
   struct comm_dbud * comm;

   if(cm_filename[0] && cm_priority && cm_url[0])
   {
      /* Support for older NV files, which had no seperate community names:
       * If we have the file name but no community name, then generate a fake
       * file name from the comm name. If there really is a file name further
       * down in the NV file then it will overwrite the fake one.
       */
      if(cm_name[0] == 0)
      {
         cp = cm_filename; 
         i = 0;
         while ((*cp > ' ') && (*cp != '.'))
            cm_name[i++] = *cp++;

         cm_name[i] = 0;   /* Null terminate */
      }

      /* Make new entry in communities list */
      comm = br_cmnew(cm_url, cm_name, cm_priority);
      cm_priority = 0; 
      cm_url[0] = 0;
      cm_name[0] = 0;

      if(!comm)
      {
         dprintf("Unable to create community \"%s\"\n");
         return ENP_PARAM;
      }
      strcpy(comm->cm_filename, cm_filename);
      cm_filename[0] = 0; 
   }
   return 0;
}


/* br_saveparms() - write generic NV parameters to NV file
 *
 * RETURNS: Returns 0 if OK, else ENP_ code
 */

int   br_saveparms()
{
    int     err;
    FILE *  fp;

    fp = fopen(parmsfile, "w");
    if(fp == NULL)
    {
        dprintf("unable to open parameters file %s\n", parmsfile);
        return ENP_FILEIO;
    }
    err = br_writeparms(fp);
    fclose(fp);
    return err;
}


/* br_writeparms() - write generic NV parameters to NV file
 *
 * RETURNS: Returns 0 if OK, else ENP_ code
 */

int   br_writeparms(FILE * fp)
{
   int      parm;
   int      row;
   char *   cp;
   char     alertbuf[128];
   struct comm_dbud *   comm;


   fprintf(fp, "# %s configuration file\n#\n\n", br_appname);

   for(parm = 0; parm < num_params; parm++)
   {
       /* Check each param for special case writing, else use final else */
       if(strncmp(br_params[parm].format, "Username", 8) == 0)
       {
           /* Get here to do iterrative user table */
           fprintf(fp, "\n");   /* LF for readability */

           /* Copy the user/password info into the file */
           for(row = 0; row < BR_MAXUSERS; row++)
           {
               if(users[row].username[0])
               {
                   fprintf(fp, br_params[parm].format,
                           users[row].username );
                   fprintf(fp, br_params[parm + 1].format,
                           users[row].password );
                   fprintf(fp, br_params[parm + 2].format,
                           users[row].usermask );
                   fprintf(fp, "\n");
               }
           }
           parm += 2;
       }
       else if(strncmp(br_params[parm].format, "Whois server", 12) == 0)
       {
           /* Get here to do iterrative Whois server table */
           fprintf(fp, "\n");   /* LF for readability */
           /* Save the whois server info */
           for(row = 0; row < WHOSERVERS; row++)
           {
               if(whoservers[row].dnsname[0])
               {
                   fprintf(fp, br_params[parm].format, 
                           whoservers[row].dnsname);
                   fprintf(fp, br_params[parm + 1].format, 
                           whoservers[row].whoport);

                   /* type can be NAME, IP, or both. Check flags */
                   if((whoservers[row].flags & (WH_IPTYPE|WH_NAMETYPE)) ==
                       (WH_IPTYPE|WH_NAMETYPE))
                   {
                       fprintf(fp, br_params[parm + 2].format,
                         "NAME,IP");
                   }
                   else if(whoservers[row].flags & WH_IPTYPE)
                       fprintf(fp, br_params[parm + 2].format, "IP");
                   else     /* default to name type */
                       fprintf(fp, br_params[parm + 2].format, "NAME");
               }
           }
           parm += 2;
           fprintf(fp, "\n");
       }
       else if(strncmp(br_params[parm].format, "Community Name", 14) == 0)
       {
           int i;
           /* Get here to do itterative URL table */
           fprintf(fp, "\n");   /* LF for readability */
           /* Iterate through community list, saving info */
           for(comm = br_communities;  comm; comm = comm->next)
           {
               /* Don't record the local community - it's all hardcoded.*/
               if(comm == &local_comm)
                   continue;

               if((comm->cm_filename[0] == 0) &&
                   (comm->cm_name[0] == 0))
               {
                   return ENP_PARAM;
               }

               /* Save community name */
               if(comm->cm_name[0] == 0)  /* if no name? */
               {
                   dtrap();
                   cp = comm->cm_filename; 
                   i = 0;
                   while ((*cp > ' ') && (*cp != '.'))
                       comm->cm_name[i++] = *cp++;
               }
               fprintf(fp, br_params[parm + 1].format, comm->cm_name);

               /* Save DB filename */
               if(comm->cm_filename[0] == 0)  /* if no filename then make one */
               {
                   dtrap();
                   cp = comm->cm_name; 
                   i = 0;
                   while (*cp)
                   {
                       if( !isdigit(*cp) && !isalpha(*cp) )
                           comm->cm_name[i++] = '_';
                       else
                           comm->cm_name[i++] = *cp;
                       if(++cp > &comm->cm_filename[MAXFILENAME-5])
                           break;
                   }
                   strcat(comm->cm_name, ".cdb");
               }
               fprintf(fp, br_params[parm + 1].format, comm->cm_filename);

               /* save the priority */
               fprintf(fp, br_params[parm + 2].format, comm->cm_priority);
               /* save communnity URL */
               fprintf(fp, br_params[parm + 3].format, comm->cm_url );
               fprintf(fp, "\n");
           }
           parm += 3;
       }
       else if(strncmp(br_params[parm].format, "MyServer active", 15) == 0)
       {
           fprintf(fp, br_params[parm].format, br_srcserving?"YES":"NO" );
       }
       else if(strncmp(br_params[parm].format, "View iface", 10) == 0)
       {
           if(br_autoifview)
               fprintf(fp, br_params[parm].format, "Auto");
           else
           {
               char number[12];
               sprintf(number, "%d", br_ifview);
               fprintf(fp, br_params[parm].format, number);
           }
       }
       else if(strncmp(br_params[parm].format, "Alertmask", 9) == 0)
       {
           int i, j;

           /* Print the alert table booleans in a "01" test sequence */
           MEMSET(alertbuf, 0, sizeof(alertbuf));
           cp = alertbuf;
           for(i = 0; i < alertcount; i++)
           {
               for(j = 0; j < alert_methodct; j++)
               {
                   if(alarms[i]->actions & a_methods[j].bitmask)
                       *cp++ = '1';
                   else
                       *cp++ = '0';
                   
                    if(strlen(alertbuf) >= sizeof(alertbuf))
                       panic("nv-alert");
               }
           }
           *cp = 0;       /* null terminate "01" mask string */
           fprintf(fp, br_params[parm].format,  alertbuf);
       }
       else
       {
           /* simple case, amenable to sprintf() */
           fprintf(fp, br_params[parm].format, *(int*)br_params[parm].pvalue );
       }
   }
   fprintf(fp, "\n\n");

   return 0;
}


int
br_initparms()
{
   int   err;

   /* reset global counters and index we use in Buster NV code */
   userread = 0;        /* number users read from NV */
   permread = 0;        /* number permissions read from NV */
   passread = 0;        /* number passsord read from NV */
   whoserverct = -1;    /* index of whois servers */

   err = br_readparms();
   return err;
}

/*
Copyright (C) 2003-2008 by Praemio Inc. All rights reserved.

Author: John Bartas - jbartas@praemio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


