/* brdb.c
 *
 * Buster database routines. 
 *
 * Copyright 2003-2008 by Praemio Inc. All rights reserved.
 * Author: John Bartas - jbartas@praemio.com
 * This software is released under the terms of the LGPL.
 * See LGPL Notice at end of this file; Full license in ../license.txt
 *
 * ROUTINES: mot_byname(), mot_bynum(), br_getcomment(), 
 *  br_assignid(), new_host(), host_delete((), br_hostlookup(), 
 *  iphost_new(), iphost_delete(), br_iplookup(), new_ipe(), 
 *  del_ipe(), new_cp(), del_cp(), new_patent(), 
 *  del_patent(), bt_dntest(), new_dn(), del_dname(), dn_lookup(), 
 *  br_cnlookup(), conn_delete(), 
 *  bt_pathtest(), new_path(), path_lookup(), del_path(), 
 *  br_blkpath(), br_nullterm(), br_blkip(), br_newpcl(), 
 *  br_delpcl(), br_lookuppcl(),  
 *  br_blkword(), new_word(), del_word(), word_lookup(), 
 *
 */

#include "busterport.h"
#include "brnet.h"
#include "ip.h"
#include "brdb.h"


struct ip_hostentry * nulliph;

/* Database Mutex pointers */
void *  control_ip_mutex;
void *  br_dns_mutex;
void *  ce_table_mutex;
void *  hosttab_mutex;
void *  ip_hosttab_mutex;

#ifdef _WINBASE_
/* Windows port helper routine to verify correct locking */
extern int   win_mutex_locked(void *p);
#endif

/***** database structures ******/

struct host_entry * hosttab[HASH_HOSTS];
struct conn_entry * ce_table[HASH_CONN_ENTRIES];
struct ip_hostentry * ip_hosttab[HASH_IPHOSTS];

/* bitmasks to indicate which patterns are present */
char pattern_bits[HASH_PATTERN_BITS];

/* the actual hash table */
struct pattern_entry * pat_table[HASH_PATTERN_ENTRIES];

struct btnode * ip_tree;
struct btnode * tcpp_tree;
struct btnode * udpp_tree;
struct btnode * dn_tree;      /* Domain name binary tree root */
struct btnode * path_tree;

void * br_tree_mutex;         /* protect access to btrees */

struct pop_client * pop_clients;


/****** Array of motives for blocking sites: *******/

struct bk_motive motives[] = 
{
   { OF_OTHER,      "Other" },    /* just felt like blocking it */
   { OF_ADVERTS,    "Ads" },      /* Source of ads or popups */
   { OF_POPUPS,     "Popups" },   /* Source of ads or popups */
   { OF_COOKIES,    "Cookies" },  /* 3rd party tracking cookies, etc. */
   { OF_SPYWARE,    "Spyware" },  /* Spyware site */
   { OF_FRAUD,      "Fraud" },    /* Earthlink Creditcard scam, etc. */
   { OF_CONTENT,    "Content" },  /* XXX content */
   { OF_ANTIVIRUS,  "Virus" },    /* Virus & trojans */
   { OF_HACKING,    "Hacking" },  /* ??? (from spy_ad)  */
   { OF_SPAM,       "Spam" },     /* ??? (from spy_ad)  */
   { 0, NULL },                /* end of table */
};

//#define NUM_MOTIVES  (sizeof(motives)/sizeof(struct bk_motive))
#define NUM_MOTIVES  10


/* Sometimes we need a way to signal the DB code not to delete anything
 * temorarily. For example when the web server is displaying a connection
 * list, we don't want the connections dissappearing. Here's the flag
 * for this:
 */
int   br_dbnodelete;


int   br_srcserving; /* True if content swapping is active */

/* mot_byname(char * cp2) - lookup block motive by name */

struct bk_motive *
mot_byname(char * cp2)
{
   int   i;
   char * cr;

   if(cp2 == NULL)
      goto bad_motive;

   /* null terminate string at CR, if any */
   cr = strchr(cp2, '\r');
   if(cr)
      *cr = 0;

   /* see if it's one of the motives */
   for(i = 0; i < NUM_MOTIVES; i++)
   {
      if(stricmp(motives[i].desc, cp2) == 0)
      {
         if(cr)
            *cr = '\r';
         return &motives[i];
      }
   }

   /* fall to here if passed motive text was not found in list */
   dtrap();    /* Bad HTML (SSI info) in form? */
   if(cr)
      *cr = '\r';

bad_motive:
   return &motives[0];
}

/* mot_bynum(char * cp2) - lookup block motive by code number */

struct bk_motive *
mot_bynum(int number)
{
   int   i;

   /* see if it's one of the motives */
   for(i = 0; i < NUM_MOTIVES; i++)
   {
      if(motives[i].code == number)
      {
         return &motives[i];
      }
   }
   return NULL;   /* not found */
}



/* br_clearconnblocks() - loop through connection table clearing
 * all reference to a block item (e.g. b_path). This prevents  
 * deleted block items from causing faults when the GUI tried 
 * to display info about why the connection was blacked.
 *
 * Returns number of connections actually cleared.
 */


int
br_clearconnblocks(void * block)
{
   int      i;
   int      cleared = 0;
   struct conn_entry * conn;

   /* loop through connection table looking for connections involving 
    * this block item.
    */
   for(i = 0; i < HASH_CONN_ENTRIES; i++)
   {
      conn = ce_table[i];
      while(conn)
      {
         /* see if we have a match */
         if(conn->block.ptr == block)
         {
            conn->block.type = 0;
            conn->block.ptr = NULL;
            cleared++;
         }
         conn = conn->next;
      }
   }

   return cleared;
}


/* br_dupobject()
 *
 * Handle situation where Duplicate object is detected. This should 
 * check if the new object is from a community with higher priority 
 * Thanks the older object; and if so replace to older object in the 
 * btree with the new one. In either case the lower priority object
 * should be retained in the dupctrl list.
 *
 * Duplicate objects from the same community generate an error.
 *
 * Retuns 0 if object ius inserted, else ENP code.
 */

int
br_dupobject(struct ctdesc * new_ctl, 
             struct btnode ** new_node,
             struct btnode ** old_node,
             struct ctdesc * old_ctl,
             void * new_data)
{
   struct ctdesc * tmp;

   if((new_ctl->comm == NULL) || (old_ctl->comm == NULL))
   {
      dtrap();             /* Object control missing community */
      return ENP_PARAM;    /* Drop malformed object */
   }
   if(new_ctl->comm == old_ctl->comm)
   {
      br_log(0, "Duplicate object in community %s\n", old_ctl->comm->cm_url);
      return ENP_PARAM;    /* Drop malformed object */
   }
   if((new_node == NULL) ||
      (old_node == NULL) ||
      (*old_node == NULL))
   {
      dtrap();
      return ENP_LOGIC;       /* Bad programming */
   }

   /* Figure out if the new item's community outranks the old */
   if(new_ctl->comm->cm_priority > old_ctl->comm->cm_priority)
   {
      /* New item's community has higher priority, insert new item
       * into the btree. Old item is attached to new item's dup list 
       */
      new_ctl->dupctrl = old_ctl;   /* put old in new node's dup list */
      *new_node = *old_node;        /* new item points to tree node tree */
      (*new_node)->data = new_data; /* tree node points to new item */
      *old_node = NULL;             /* Old node -> out of tree */
   }
   else  /* NO, just add new node to end of existing nodes duplicate list */
   {
      tmp = old_ctl;
      while(tmp->dupctrl)
         tmp = tmp->dupctrl;
      tmp->dupctrl = new_ctl;
   }
   return 0;
}


/***** database routines - create, delete, lookup ******/

/* pseudo entry for the broadcast address */
struct host_entry * br_bcasthost;

char br_dupip[] = {"Two computers both have IP address xxx.xxx.xxx.xxx "};


struct host_entry *
new_host(int iface, char * macaddr, ip_addr ipaddr, int type, long lease)
{
   struct host_entry * host;
   int   hashval;

   /* Sanity check - make sure host does not aleady exist. Duplicates 
    * are a serious error and should be prevented by the calling code!
    */
   if((host = br_hostlookup(macaddr, iface)) != NULL)
   {
      dtrap();    /* This should almost be a panic */
      dprintf("Attempttocreate duplicate MAC host, MAC %s\n", 
         br_prtmac((u_char*)macaddr));
      return host;      /* try to recover... */
   }

   host = HOST_ALLOC(sizeof(struct host_entry));
   if(host == NULL)
      return NULL;
   host->ip_lease = lease;

   /* If this is the broadcast address, make a pseudo entry for it */
   if(*macaddr == (char)(0xFF))
   {
      host->h_flags |= H_BCAST;  /* set broadcast flag */
      br_bcasthost = host;       /* remember where this is */
   }

   host->iphost = br_iplookup(iface, ipaddr);    /* Get IP host for this IP address */
   if(host->iphost == NULL)
   {
      /* IP host is unknown, make an entry for it. */
      host->iphost = iphost_new(iface, ipaddr);
      if(host->iphost == NULL)    /* alloc failure? */
      {
         HOST_FREE(host);
         return NULL;   
      }
      host->iphost->mac_host = host;      /* set back pointer */
   }
   else  /* found matching IP host entry */
   {
      struct host_entry * otherhost;

      otherhost = host->iphost->mac_host;

      /* See if this IP address is in use by another MAC address */
      if((otherhost != NULL) &&
         (otherhost != host) &&
         (host->iface != iface) &&
         (ipaddr != 0) &&     /* don't worry about null IP address (DHCP) */
         ((htonl(ipaddr) & 0xE0000000) != 0xE0000000))   /* or mcast or bcast */
      {
         /* If existing guy was read from the file at startup and and had no 
          * activity since, then assume IP changed while we were suspended.
          */
         if((otherhost->disctype & FF_FROMFILE) &&
            (otherhost->active == 0))
         {
            otherhost->iphost = nulliph;
            host->iphost->mac_host = host;      /* set back pointer */
         }
         else if((otherhost->ip_lease < 1) ||
            (otherhost->active != 0) ||
            (otherhost->ip_exp > cticks))
         {
            sprintf(br_dupip + 35, "%3u.%3u.%3u.%3u", PASS_IPADDR(ipaddr) );
            alarm_general(&alarm_text, NULL, br_dupip);
            host->h_flags |= H_DUPLIP;
            host->iphost->mac_host->h_flags |= H_DUPLIP;
         }
         else if(host->active != 0)
         {
            /* Other MAC which shares the IP has not been active 
             * since last restart or has an expired DHCP lease.. 
             * Assume we missed it loosing the lease and claim the 
             * IP for the new host.
             */
            dtrap(); /* can this happen anymore? */
            otherhost->iphost = nulliph;
            host->iphost->mac_host = host;      /* set back pointer */
         }
         else
         {
            host->iphost = nulliph;
         }
      }
   }

   host->disctype = type;              /* discovery type */
   host->age = cticks;
   host->iface = iface;
   MACCPY(host->macaddr, macaddr);

   /* Install host in hash table */
   hashval = HASH_HOST(macaddr);
   host->next = hosttab[hashval];   /* hash bucket for new host */
   hosttab[hashval] = host;


   /* See if this is an entry for ourself */
   if(MAC_EQ(macaddr, nets[iface]->n_macaddr))
      host->h_flags |= H_SELF;

   brstates[iface].machosts++;    /* count total MAC hosts */

   /* do we need to set an alarm on new local host? */
   if(newmac.actions)
   {
      /* Ignore Broadcast and Multicast addresses and hosts read
       * from a file.
       */
      if(((*macaddr & 0x01) == 0) &&    /* not a MAC multicast, and... */
         ((type & FF_FROMFILE) == 0))    /* not from a file */
      {
         alarm_general(&newmac, NULL, host);
      }
   }

   return host;
}


/* br_deleteWait() - wait until br_dbnodelete lock is clear.
 * 
 * Returns 0 when OK to delete, Else -1 if timeout.
 */

int   br_delete_delays = 0;

int
br_deleteWait(void)
{
   u_long   timeout;

   timeout = cticks + (10 * TPS);   /* set a large timeout */

   if(br_dbnodelete)
      br_delete_delays++;

   while(br_dbnodelete)
   {
      if(timeout < cticks)
      {
//         dtrap();    /* check this first few times */
         return -1;  /* return timeout code */
      }
      TH_SLEEP(1);
   }
   return 0;
}


/* host_delete(() - Destructor for MAC host entry
 *
 * Returns next entry in hash table bucket, else
 * returns NULL if there is none.
 */

struct host_entry *
host_delete(struct host_entry * del_host, int hashindex)
{
   int   iface;
   struct host_entry * machost;
   struct host_entry * last;

   if(br_deleteWait())  /* Wait until safe to delete */
      return del_host->next;

   /* Make sure Buster is not in received packet logic */
   if(br_inreceiver)
   {
      while(br_inreceiver)
      {
         TH_SLEEP(1);
      }
   }

   /* Search hash bucket for host to delete */
   TH_LOCK_MUTEX(hosttab_mutex);
   machost = hosttab[hashindex];
   last = NULL;
   while(machost)
   {
      int   rxstate;
      int   i;
      struct conn_entry * conn;

      if(machost == del_host)    /* found host to delete? */
      {
         brstates[del_host->iface].machosts--;    /* del from total */
         machost = del_host->next;  /* set return value */

         /* Unlink it from the hosts list. */
         if(last)
            last->next = machost; /* unlink */
         else
            hosttab[hashindex] = machost;

         /* Disable the receiver while we delete database items */
         rxstate = brstates[del_host->iface].rx_disabled;
         brstates[del_host->iface].rx_disabled = TRUE;

         /* clear the connections ram database of this hosts conns */
         for(i = 0; i < HASH_CONN_ENTRIES; i++)
         {
            conn = ce_table[i];
            while(conn)
            {
               if((del_host->iphost) && 
                  (conn->loc.ce_addr == del_host->iphost->ipaddr))
               {
                  conn = conn_delete(conn, i);
               }
               else
                  conn = conn->next;
            }
         }

         /* clear this host's iphost entry. First make sure IP host 
          * entry is not attached to a different MAC host: 
          */
         if((del_host->iphost) && 
            (del_host->iphost->mac_host == del_host))
         {
            struct host_entry * othermac;

            /* Make sure IP host is not being used by another MAC host */
            for(i = 0; i < HASH_HOSTS; i++)
            {
               othermac = hosttab[i];
               while(othermac)
               {
                  /* two hosts using the same IP? */
                  if((othermac->iphost) && 
                     (othermac->iphost == del_host->iphost))
                  {
                     dtrap(); /* really? */
                     /* Ideally this shouldn;t happen, but events like 
                      * a DHCP recycling to another host while Buster is off-line 
                      * make it hard. In these cases unhook the IP address from 
                      * the host being deleted, but don't delete the actual
                      * IP host.
                      */
                     del_host->iphost = NULL;   /* unlink it... */

                     /* Make sure IP host is not linked to del_host */
                     if(othermac->iphost->mac_host == del_host)
                        othermac->iphost->mac_host = othermac;
                     goto leave_iphost;         /* Don't delete it */
                  }
                  othermac = othermac->next;
               }
            }

            iphost_delete(del_host->iphost, HASH_IPH(del_host->iphost->ipaddr));
            del_host->iphost = NULL;
         }
leave_iphost:

         /* If this host is cached as a router, clear the router info 
          * in the iface status struct,  otherwise we leave a 
          * bogus reference - a GP fault waiting to happen.
          */
         for(iface = 0; iface < (int)ifNumber; iface++)
         {
            if(del_host == brstates[iface].br_router)
               br_clearrouter(iface);
         }

         /* restore receiver state */
         brstates[del_host->iface].rx_disabled = rxstate;

         HOST_FREE(del_host);
         TH_UNLOCK_MUTEX(hosttab_mutex);
         return machost;       /* non-error exit from routine */
      }
      last = machost;
      machost = machost->next;
   }
   dtrap();    /* host not found? */
   TH_UNLOCK_MUTEX(hosttab_mutex);
   return NULL;
}


/* Returns host_entry if one is found for passed MAC, else
 * returns NULL
 */

struct host_entry *
br_hostlookup(char * macaddr, int iface)
{
   struct host_entry * host;
   int   hashval;

   /* see if this bucket already has entrys. If so, then traverse to
    * end of bucket looking for the MAC.
    */
   hashval = HASH_HOST(macaddr);
   if(hosttab[hashval])
   {
      for(host = hosttab[hashval]; host; host = host->next)
      {
         if(iface != host->iface)   /* make sure iface matches */
            continue;

         /* check for our MAC */
         if(MAC_EQ(macaddr, host->macaddr))
            return host;
      }
   }
   return NULL;
}

/* constructor for IP host entry. 
 *
 * Returns new ip_hostentry if OK,
 *    NULL if memory ran out
 */

struct ip_hostentry *
iphost_new(int iface, ip_addr ipaddr)
{
   struct ip_hostentry * iphost;
   int      hashval;

   dprintf("iphost_new: ipaddr=%u.%u.%u.%u, iface:%d\n", 
      PASS_IPADDR(ipaddr), iface );

   iphost = IPH_ALLOC(sizeof(struct ip_hostentry));
   if(!iphost)
   {
      dtrap();
      return NULL;
   }
   hashval = HASH_IPH(ipaddr);
   iphost->next = ip_hosttab[hashval];
   ip_hosttab[hashval] = iphost;
   iphost->ipaddr = ipaddr;
   iphost->iface = iface;

   if(br_isbcast(iface, ipaddr))
   {
      iphost->flags |= IH_BROADCAST;
      strcpy(iphost->dnsname, "Broadcast");  /* Not really DNS, but... */
   }
   iphost->tm_detected = iphost->tm_lastpkt = iphost->tm_lastconn = cticks;

   return iphost;
}

/* iphost_delete(() - Destructor for IP host entry
 *
 * Returns next entry in hash table bucket, else
 * returns NULL if there is none.
 */

struct ip_hostentry * 
iphost_delete(struct ip_hostentry * to_del, int hashindex)
{
   struct ip_hostentry * iphost;    /* scratch */
   struct ip_hostentry * last;      /* previous host in list */

   dprintf("iphost_delete: ipaddr=%u.%u.%u.%u\n",
      PASS_IPADDR(to_del->ipaddr) );

   if(br_deleteWait())  /* Wait until safe to delete */
      return to_del->next;

   TH_LOCK_MUTEX(ip_hosttab_mutex);
   iphost = ip_hosttab[hashindex];
   last = NULL;
   while(iphost)
   {
      if(iphost == to_del)    /* found host to delete? */
      {
         iphost = to_del->next;  /* set return value */
         if(last)
            last->next = iphost; /* unlink*/
         else
            ip_hosttab[hashindex] = iphost;

         /* If this is IP host is the primary IP host for a 
          * MAC host, then clear the back pointer
          */
         if((to_del->mac_host != NULL) &&
            (to_del->mac_host->iphost == to_del))
         {
            to_del->mac_host->iphost = NULL;
         }

         IPH_FREE(to_del);
         TH_UNLOCK_MUTEX(ip_hosttab_mutex);
         return iphost;       /* non-error exit from routine */
      }
      last = iphost;
      iphost = iphost->next;
   }
   dtrap();    /* host not found? */
   TH_UNLOCK_MUTEX(ip_hosttab_mutex);
   return NULL;
}

/* br_iplookup()
 *
 * Lookup an IP host. requires a valid iface index, since there may by 
 * identidal IP addresses on different interfaces; especially when using. 
 * local addressed (e.g. routers addressed as 192.168.1.1) 
 *
 * For casual viewing you can use br_ipanfif().
 *
 * Returns IP host structure if found, else NULL. 
 */

struct ip_hostentry *
br_iplookup(int iface, ip_addr ipaddr)
{
   struct ip_hostentry * ip_host;
   int      hashval;

   hashval = HASH_IPH(ipaddr);
   for(ip_host = ip_hosttab[hashval]; ip_host; ip_host = ip_host->next)
   {
      if((ip_host->ipaddr == ipaddr) &&
         (ip_host->iface == iface))
      {
         return ip_host;
      }
   }

   /* fall to here if not found */
   return NULL;
}

/* br_ipanyif()
 *
 * Best-effort version of br_iplookup() - this tries to find an 
 * IP host matching the passed IP address. It first tries the 
 * viewable iface, but if that fails is tries the others.
 *
 * Returns IP host structure if found, else NULL. 
 */

struct ip_hostentry * 
br_ipanyif(ip_addr hostaddr)
{
   int   iface;
   struct ip_hostentry * iphost;

   iphost = br_iplookup(br_ifview, hostaddr);
   if(iphost)
      return iphost;

   for(iface = 0; iface < (int)ifNumber; iface++)
   {
      if(iface == br_ifview)  /* we already tried this */
         continue;

      iphost = br_iplookup(iface, hostaddr);
      if(iphost)
         return iphost;
   }

   return NULL;   /* Not found anywhere */
}


/* new_ipe() - constructor for creating a new control IP address 
 * entry. If the entry is being creating for a controled 
 * domain name entry then then caller must set the back 
 * pointer "->ci_name" to the dns entry.
 *
 * Optional third parameter is control struct to clone into new ipe.
 *
 */

struct control_ip *
new_ipe(ip_addr ipaddr, struct ctdesc * ctl, 
        int typeflag, struct comm_dbud * comm)
{
   struct control_ip *  new_ipe;
   struct control_ip *  old_ipe;


   /* Unlike most IP address in Buster, these ones are stored in 
    * local endian to facilitate GT and LT comparisons 
    */
#if (BYTE_ORDER == LITTLE_ENDIAN)
   ipaddr = htonl(ipaddr);
#endif

   /* Fall to here if we need to make a new entry */
   new_ipe = IPE_ALLOC(sizeof(struct control_ip) );
   if(!new_ipe) 
      return NULL;

   /* Set passed parameters */
   new_ipe->ci_flags = typeflag;
   new_ipe->ci_ipaddr = ipaddr;

   /* if caller passed a control structure, copy it into new ipe */
   if(ctl)
   {
      new_ipe->ctl.offense = ctl->offense;
      new_ipe->ctl.action_mask = ctl->action_mask & ~CTF_TYPEMASK;
      new_ipe->ctl.action_mask |= CTF_TYPEIP;
      new_ipe->ctl.comm = ctl->comm;
   }
   else     /* no control info passed, just set default */
   {
      /* Indicate the type of this object in the ctl structure */
      new_ipe->ctl.action_mask = CTF_TYPEIP;
   }

   new_ipe->ctl.comm = comm;
   comm->cm_currec++;         /* Bump community current rec count */

   /* see if we already have an entry for this value */
   old_ipe = ipe_lookup(ipaddr);

   TH_LOCK_MUTEX(control_ip_mutex);
   if(old_ipe)
   {
      int err;
      new_ipe->btnp = NULL;    /* New item not in btree */

      /* Call routine to chain duplicate objects together */
      err = br_dupobject(&new_ipe->ctl, &new_ipe->btnp, &old_ipe->btnp, 
         &old_ipe->ctl, new_ipe);
      if(old_ipe->btnp && (old_ipe->btnp->data != (struct btnode **)old_ipe))
         dtrap();
      if(new_ipe->btnp && (new_ipe->btnp->data != (struct btnode **)new_ipe))
         dtrap();
      if(err)
	  {
         TH_UNLOCK_MUTEX(control_ip_mutex);
         return NULL;
      }
   }
   else
   {
      new_ipe->btnp = BTN_ALLOC(sizeof(struct btnode));
      if(new_ipe->btnp == NULL)
      {
         IPE_FREE(new_ipe);
         TH_UNLOCK_MUTEX(control_ip_mutex);
         return NULL;
      }
      /* set btree node back pointer */
      new_ipe->btnp->data = (struct btnode **)new_ipe; 

      /* add new entry to btree */
      TH_LOCK_MUTEX(br_tree_mutex);
      bt_insert(&(struct btnode *)ip_tree, new_ipe->btnp, 
         (u_long)new_ipe->ci_ipaddr, bt_iptest);
      TH_UNLOCK_MUTEX(br_tree_mutex);
   }

   TH_UNLOCK_MUTEX(control_ip_mutex);
   return new_ipe;
}


/* br_highestcomm()
 *
 * Return control with the highest priority community from a passed 
 * list of duplicate controls. List may be just one item. This 
 * re-orders the list so the highest control is at the front.
 * Helper routine for delete ctl items with duplicate lists.
 */

struct ctdesc * 
br_highestcomm(struct ctdesc * ctllist)
{
   struct ctdesc * best;
   struct ctdesc * ctmp;

   /* search dup list for highest priority community */
   best = ctllist;      /* seed to head of list */
   for(ctmp = ctllist; ctmp; ctmp = ctmp->dupctrl)
   {
      if(ctmp->comm->cm_priority > best->comm->cm_priority)
         best = ctmp;
   }

   /* Now recorder list with best at head. First check head - if it's 
    * best then we are done.
    */
   if(best == ctllist)
      return best;

   dtrap();
   TH_LOCK_MUTEX(ctllist);
   for(ctmp = ctllist; ctmp; ctmp = ctmp->dupctrl)
   {
      /* See if next is best */
      if(ctmp->dupctrl == best)
      {
         ctmp->dupctrl = best->dupctrl;   /* unlink best */
         best->dupctrl = ctllist;         /* put best at head */
         TH_UNLOCK_MUTEX(ctllist);
         return best;
      }
   }
   dtrap();    /* should never get here */
   TH_UNLOCK_MUTEX(ctllist);
   return best;
}

/* br_ctlclear(struct ctdesc * ctl)
 *
 * Utility routine to free control's sub-fields prior to returning
 * freeing the control itself.
 *
 */

void
br_ctlclear(struct ctdesc * ctl)
{
#ifdef DESC_BLOAT

   /* free comment (description), if any */
   if(ctl->desc)
      CTD_FREE(ctl->desc);
   if(ctl->tags)
      CTT_FREE(ctl->tags);
#endif /* DESC_BLOAT */

   return;
}


void
del_ipe(struct control_ip * delipe, int verbose)
{
   struct control_ip * tmp;   /* for unlinking from name entry's list */
   struct control_ip * last;

   if(verbose)
   {
      br_log(0, "del_ipe: %u.%u.%u.%u at 0x%x\n", 
         PASS_IPADDR(delipe->ci_ipaddr), delipe);
   }

   TH_LOCK_MUTEX(control_ip_mutex);

   /* Make sure pointer is a Non-deleted control_ip structure */
   if((delipe->ctl.action_mask & CTF_TYPEMASK) != CTF_TYPEIP)
   {
      dprintf("del_ipe: bad type; mask:0x%x, IP:%u.%u.%u.%u\n",
         delipe->ctl.action_mask, 
         PASS_IPADDR(delipe->ci_ipaddr) );
      dtrap();
      //    should we ?? panic("del_ipe");
      TH_UNLOCK_MUTEX(control_ip_mutex);
      return;
   }

   /* clear action bits, set invalid type */
   delipe->ctl.action_mask = CTF_TYPEMASK;

   if(delipe->btnp)
   {
      /* Check for queued duplicate items */
      if(delipe->ctl.dupctrl)
      {
         struct ctdesc * bestctl;
         struct control_ip * bestip;

         /* reorder dup list with best at head */
         bestctl = br_highestcomm(delipe->ctl.dupctrl);
         if(!bestctl)
         {
            dtrap();
            TH_UNLOCK_MUTEX(control_ip_mutex);
            return;
         }

         /* Sanity check */
         if(bestctl == &delipe->ctl)
         {
            dtrap();
         }

         /* Get new best control_ip from bestctl */
         bestip = (struct control_ip *)((char*)bestctl - 
            ((char*)(&delipe->ctl) - (char*)(delipe)));

         /* sanity check: make sure best is not deleted */
         if((bestip->ctl.action_mask & CTF_TYPEMASK) != CTF_TYPEIP)
         {
            dtrap();
            TH_UNLOCK_MUTEX(control_ip_mutex);
            return;
         }

         /* Attach new best control to btree node */
         bestip->btnp = delipe->btnp;
         bestip->btnp->data = (struct btnode **)bestip;     /* set btree node back pointer */

         delipe->btnp = NULL;       /* Detach old control */
         /* fall to code to clean out connections, dns links */
      }
      else
      {
         /* get the IP entry (and it's tree node) out of the tree */
         /* br_tree_mutex should already be locked */
#ifdef _WINBASE_
         if(win_mutex_locked(br_tree_mutex) != 1)
            dtrap();
#endif
         bt_remove(&(struct btnode *)ip_tree, delipe->ci_ipaddr, bt_iptest);
      }
   }
   else
   {
      dtrap();    /* Trying to delete a control in a dup list? */
   }

   /* remove references to this control_ip from the connections list */
   br_clearconnblocks((void *)delipe);

   /* If it's attached to a DNS name, unlink it */
   if(delipe->ci_name)
   {
      last = NULL;      /* clear previous entry pointer */

      /* search DNS entry's list for this item; retaining a 
       * pointer to the previous entry for unlinking 
       */
      for(tmp = delipe->ci_name->cu_cip; tmp; tmp = tmp->ci_dnlist)
      {
         if(tmp == delipe)      /* found ourself? */
         {
            if(last == NULL)     /* no previous entry? */
               delipe->ci_name->cu_cip = delipe->ci_dnlist;
            else
               last->ci_dnlist = delipe->ci_dnlist;

            delipe->ci_name = NULL;    /* IP no longer in name list */
            break;
         }

         /* sanity check item in list */
         if((tmp->ctl.action_mask & CTF_TYPEMASK) != CTF_TYPEIP)
         {
            dprintf("del_ipe: bad tmp->type; mask:0x%x, IP:%u.%u.%u.%u\n",
               tmp->ctl.action_mask, tmp->ci_ipaddr);
            dtrap();
            TH_UNLOCK_MUTEX(control_ip_mutex);
            return;
         }
         last = tmp;
      }
      if(!tmp)
      {
         dtrap(); /* delipe entry not found in dns entry's list  */
      }
   }
   /* clear out control fields prior to freeing it  */
   br_ctlclear(&delipe->ctl);

   IPE_FREE(delipe);
   TH_UNLOCK_MUTEX(control_ip_mutex);
   return;
}


/* if we already have a port entry for this value then get it
 * and let the logic below update the flags; if not then make a
 * new control port entry.
 */

struct control_port *
new_cp(unshort portnum, int type, struct comm_dbud * comm)
{
   int   err;
   struct control_port *   new_cp;
   struct btnode *         btnp;
   struct btnode **        tree;

   switch(type)
   {
   case CPF_TCP:
      tree = &tcpp_tree;
      break;
   case CPF_UDP:
      tree = &udpp_tree;
      break;
   default:
      dtrap();       /* bad programming */
      return NULL;   /* bad parm */
   }

   /* Make a new DB entry fot this port value. */
   new_cp = CP_ALLOC(sizeof(struct control_port) );
   if(!new_cp) 
      return NULL;

   new_cp->cp_flags |= type;           /* set flag for TCP or UDP */
   new_cp->cp_port = portnum;          /* record port value */

   /* Indicate the type of this object in the ctl structure */
   new_cp->ctl.action_mask = CTF_TYPEPORT;
   new_cp->ctl.comm = comm;
   comm->cm_currec++;         /* Bump community current rec count */

   /* see if we already have an entry for port value */
   TH_LOCK_MUTEX(br_tree_mutex);
   btnp = bt_access(*tree, (u_long)portnum, bt_porttest);
   TH_UNLOCK_MUTEX(br_tree_mutex);
   if(btnp)
   {
      struct control_port * old_cp;

      new_cp->btnp = NULL;    /* New item not in btree */
      old_cp = (void*)btnp->data;

      /* Call routine to chain duplicate objects together */
      err = br_dupobject(&new_cp->ctl, &new_cp->btnp, &old_cp->btnp, 
         &old_cp->ctl, new_cp);
      if(new_cp->btnp && (new_cp->btnp->data != (struct btnode **)new_cp))
         dtrap();
      if(old_cp->btnp && (old_cp->btnp->data != (struct btnode **)old_cp))
         dtrap();
   }
   else  /* No duplicate, insert new item into tree */
   {
      new_cp->btnp = BTN_ALLOC(sizeof(struct btnode));
      if(new_cp->btnp == NULL)
      {
         CP_FREE(new_cp);
         return NULL;
      }
      new_cp->btnp->data = (struct btnode **)new_cp;        /* set btree node back pointer */
      /* add new entry to btree */
      TH_LOCK_MUTEX(br_tree_mutex);
      bt_insert(tree, new_cp->btnp, (u_long)new_cp->cp_port, bt_porttest);
      TH_UNLOCK_MUTEX(br_tree_mutex);
   }

   return new_cp;
}


void
del_cp(struct control_port * del_cp, int type)
{
   if(del_cp->btnp)
   {
      /* Check for queued duplicate items */
      if(del_cp->ctl.dupctrl)
      {
         dtrap();
      }
      else
      {
         TH_LOCK_MUTEX(br_tree_mutex);
         if(type == CPF_TCP)
            bt_remove(&(struct btnode *)tcpp_tree, (u_long)del_cp->cp_port, bt_porttest);
         else
            bt_remove(&(struct btnode *)udpp_tree, (u_long)del_cp->cp_port, bt_porttest);
         TH_UNLOCK_MUTEX(br_tree_mutex);

         /* See if this ever gets called, and if so, does it produce a 
          * fault. It seems like the bt_remove call above actually releases
          * the del_cp->btnp member. -JB-
          */   
         dtrap();
         BTN_FREE(del_cp->btnp);
      }
   }

   /* free control sub-fields */
   br_ctlclear(&del_cp->ctl);

   CP_FREE(del_cp);
   return;
}



u_long   pattern_dups = 0;

struct pattern_entry * 
new_patent(u_char * pat, int patlen)     /* PATTERN_LENTH binary image */
{
   struct pattern_entry * newpep;
   int      hashindex;     /* struct table index */
   int      bit_index;    /* byte in pattern_bits[] with correct bit */
   int      hashx;        /* bitwise pattern table index */

   /* get index into pattern bitmask table */
   hashx = HASH_PATTERN(pat);

   /* make struct table index */
   hashindex = (hashx & (HASH_PATTERN_ENTRIES - 1));

   /* see if the pattern already exists in the tables */
   for(newpep = pat_table[hashindex]; newpep; newpep = newpep->next)
   {
      if(MEMCMP(pat, newpep->pattern, PATTERN_LENGTH) == 0)
      {
         pattern_dups++;
         break;
      }
   }

   /* Create new pattern entry */
   newpep = PEP_ALLOC(sizeof(struct pattern_entry));
   if(!newpep) return NULL;

   newpep->next = pat_table[hashindex];
   pat_table[hashindex] = newpep;
   newpep->pattern = pat;
   newpep->patlen = patlen;

   /* make a bitmask entry to aid fast lookups */
   bit_index = (hashx & ((HASH_PATTERN_BITS * 8) - 1));

   /* point to byte with right bit */
   pattern_bits[bit_index >> 3] |= (1 << (bit_index & 7));  /* set the bit */

   return newpep;
}


void
del_patent(struct  pattern_entry * pep)
{
   struct pattern_entry * tmppep;
   struct pattern_entry * last;
   int      hashindex;

dtrap();
   /* find the pattern in the hash table */
   hashindex = PAT_TABHASH(pep->pattern);
   last = NULL;
   for(tmppep = pat_table[hashindex]; tmppep; tmppep = tmppep->next)
   {

   }

}


/* Domain name binary tree test routine */
int
bt_dntest(struct btnode * n, u_long key)
{
   u_char   nodec, keyc;
   struct control_dn *  dnp;
   u_char * nodest;
   u_char * keyst = (u_char*)key;

   dnp = (struct control_dn *)n->data;
   nodest = (u_char*)dnp->cu_name;

   /* compare strings. While strings are the same continue; If
    * one ends then the other is "greater", else if the
    * characters differ the higher ascii value is greater.
    * Pretty much the same as stricmp, but we do it here for QA.
    */
   for(;;)
   {
      /* get next set of chars, mask to uppercase */
      nodec = (*nodest++) & ~0x20;
      keyc = (*keyst++) & ~0x20;

      if(nodec < keyc)
         return KEY_GT;
      if(nodec > keyc)
         return KEY_LT;

      /* fall to here if chars are equal */
      if(nodec == 0)   /* end of strings? */
         return KEY_EQ;
   }
}


struct control_dn *
new_dn(char * name, struct comm_dbud * comm)
{
   int   err;
   struct control_dn *  new_dn;
   struct btnode * btnp;

   new_dn = DN_ALLOC(sizeof(struct control_dn) );
   if(!new_dn)
      return NULL;

   /* Indicate the type of this object in the ctl structure */
   new_dn->ctl.action_mask = CTF_TYPENAME;

   new_dn->cu_name = DNM_ALLOC(strlen(name) + 1);
   if(!new_dn->cu_name)
   {
      DN_FREE(new_dn);
      return NULL;
   }
   strcpy(new_dn->cu_name, name);
   new_dn->ctl.comm = comm;
   comm->cm_currec++;         /* Bump community current rec count */

   /* see if the name is already in the tree */
   TH_LOCK_MUTEX(br_tree_mutex);
   btnp = bt_access(dn_tree, (u_long)name, bt_dntest);
   TH_UNLOCK_MUTEX(br_tree_mutex);
   if(btnp)
   {
      struct control_dn * old_dn;

      /* There is a duplicate. Higher priority community goes into tree */
      old_dn = (struct control_dn *)btnp->data;

      /* Call routine to chain duplicate objects together */
      err = br_dupobject(&new_dn->ctl, &new_dn->btnp, &old_dn->btnp, 
         &old_dn->ctl, new_dn);
      if(old_dn->btnp && (old_dn->btnp->data != (struct btnode **)old_dn))
         dtrap();
      if(new_dn->btnp && (new_dn->btnp->data != (struct btnode **)new_dn))
         dtrap();
   }
   else  /* No duplicate, insert new item into tree */
   {
      new_dn->btnp = BTN_ALLOC(sizeof(struct btnode));
      if(new_dn->btnp == NULL)
      {
         DNM_FREE(new_dn->cu_name);
         DN_FREE(new_dn);
         return NULL;
      }
      new_dn->btnp->data = (struct btnode **)new_dn; /* btree node back ptr */

      /* add new entry to Domain name tree */
      TH_LOCK_MUTEX(br_tree_mutex);
      bt_insert( &dn_tree, new_dn->btnp, (u_long)name, bt_dntest);
      TH_UNLOCK_MUTEX(br_tree_mutex);
   }
   return new_dn;
}

/* Delete a blacklisted domain name. TH_LOCK_MUTEX(br_tree_mutex); should
 * be called before this is to protect the btree. 
 */

void
del_dname(struct control_dn * del_dn)
{
   struct control_ip * cip;
   struct control_ip * cip_next;

   /* first, unlink from possible control-dns entry */
   if(del_dn->cu_cip)
      del_dn->cu_cip->ci_name = NULL;

   /* See if this item is in the btree */
   if(del_dn->btnp)
   {
      /* Check for queued duplicate items */
      if(del_dn->ctl.dupctrl)
      {
         struct ctdesc * bestctl;
         struct control_dn * bestdata;

         /* reorder dup list with best at head */
         bestctl = br_highestcomm(del_dn->ctl.dupctrl);
         if(!bestctl)
         {
            dtrap();
            return;
         }

         /* Sanity check */
         if(bestctl == &del_dn->ctl)
         {
            dtrap();
         }

         /* Get new best b_path from bestctl */
         bestdata = (struct control_dn *)((char*)bestctl - 
            ((char*)(&del_dn->ctl) - (char*)(del_dn)));

         /* sanity check: make sure best is not deleted */
         if((bestdata->ctl.action_mask & CTF_TYPEMASK) != CTF_TYPENAME)
         {
            dtrap();
            return;
         }

         /* Attach new best control to btree node */
         bestdata->btnp = del_dn->btnp;
         bestdata->btnp->data = (struct btnode **)bestdata; /* set btree node back pointer */

         del_dn->btnp = NULL;       /* Detach old control */

      }
      else
      {
         /* remove from btree */
         bt_remove( &dn_tree, (u_long)del_dn->cu_name, bt_dntest);
      }
   }

   /* remove references to this control_ip from the connections list */
   br_clearconnblocks( (void *)del_dn );

   /* free control sub-fields */
   br_ctlclear(&del_dn->ctl);

   /* If there are IP addresses attached, delete them too. */
   for(cip = del_dn->cu_cip; cip; cip = cip_next)
   {
      cip_next = cip->ci_dnlist;    /* save rest of list */
      if((cip->ci_name != NULL) &&
         (cip->ci_name != del_dn))
      {
         dtrap();    /* corrupt IP list? */
      }
      /* clear DNS so del_ipe() does not try to free it */
      cip->ci_name = NULL;       
      del_ipe(cip, FALSE);       /* delete head of list */
   }
   del_dn->cu_cip = NULL;        /* clear ptr in DNS struct */

   DN_FREE(del_dn);     /* free the heap memory */
}

/* dn_lookup(char * name) - lookup name in the DNS name blacklist */

struct control_dn *
dn_lookup(char * name)
{
   char *   cp;
   struct control_dn * cdn;
   struct btnode * btnp;

   /* see if the name is already in the tree */
   TH_LOCK_MUTEX(br_tree_mutex);
   btnp = bt_access(dn_tree, (u_long)name, bt_dntest);
   TH_UNLOCK_MUTEX(br_tree_mutex);
   if(btnp == NULL)
   {
      /* If specific name is not in the list, try general domain */
      cp = strrchr(name, '.');   /* get last dot */
      if(cp)
      {
         cp--;       /* back up before last dot */
         while((cp > name) && (*cp > '.'))
            cp--;    /* back up to previous dot */
         if(*cp == '.')
            cp++;
         TH_LOCK_MUTEX(br_tree_mutex);
         btnp = bt_access(dn_tree, (u_long)cp, bt_dntest);
         TH_UNLOCK_MUTEX(br_tree_mutex);
      }
   }
   if(btnp == NULL)
      return NULL;

   cdn = (struct control_dn *)btnp->data;
   return (cdn);
}



/*********** connection routines ************/

/* br_cnlookup() - find connection matching the tupple & type passed.
*/

struct conn_entry *
br_cnlookup(ip_addr localip, ip_addr remoteip,
         unshort localport, unshort remoteport, int type)
{
   int   hashindex;
   struct conn_entry * conn;

   hashindex = HASH_CE(localip, remoteip, localport, remoteport);

   for(conn = ce_table[hashindex]; conn; conn = conn->next)
   {
      if(conn->rem.ce_addr != remoteip)
         continue;
      if(conn->loc.ce_addr != localip)
         continue;
      if(conn->loc.ce_port != localport)
         continue;
      if(conn->rem.ce_port != remoteport)
         continue;
      if(conn->ce_type != type)
         continue;

      return conn;      /* found a perfect match */
   }
   return NULL;   /* not found */
}


/* conn_delete()  - delete a connection entry. 
 *
 * Returns pointer to next entry in hash bucket list, NULL
 * if end of list.
 */
 
struct conn_entry *
conn_delete(struct conn_entry * to_del, int hashindex)
{
   struct conn_entry *  conn;    /* tmp - dual use */
   struct conn_entry *  last;    /* ptr to last entry in bucket */

   if(br_deleteWait())  /* Wait until safe to delete */
      return to_del->next;

   /* decrement host connection counts */
   if(to_del->loc.ce_iphost)
   {
#ifdef NPDEBUG
      if(to_del->loc.ce_iphost->conns == 0)
      {
         dtrap();
      }
#endif   /* NPDEBUG */
      to_del->loc.ce_iphost->conns--;
   }
   if((to_del->rem.ce_iphost) && 
      (to_del->loc.ce_iphost != to_del->rem.ce_iphost))
   {
#ifdef NPDEBUG
      if(to_del->rem.ce_iphost->conns == 0)
      {
         dtrap();
      }
#endif   /* NPDEBUG */
      to_del->rem.ce_iphost->conns--;
   }

   TH_LOCK_MUTEX(ce_table_mutex);

   /* find entry to delete in hash bucket passed. Keep a pointer
    * to "last entry for unlinking the list.
    */
   conn = ce_table[hashindex];
   last = NULL;
   while(conn)
   {
      if(conn == to_del)      /* found entry to delete */
      {
         brstates[conn->ce_iface].total_conns--;
         conn = to_del->next;    /* set entry to return */
         if(last)    /* unlink */
            last->next = conn;
         else
            ce_table[hashindex] = conn;

         /* to_del in now unlinked and conn has return value */
         TH_UNLOCK_MUTEX(ce_table_mutex);

         /* clear all the dying c`onnectin's members to aid other 
          * routines (e.g. hostrow() in mapcgi.c) in detecting that
          * the conn structure is gone. 
          */
         MEMSET(to_del, 0, sizeof(struct conn_entry));
         CN_FREE(to_del);
         return conn;
      }
      last = conn;
      conn = conn->next;
   }
   TH_UNLOCK_MUTEX(ce_table_mutex);
   dtrap();    /* entry not found - error */
   return NULL;
}


/* pathname binary tree test routine */

int
bt_pathtest(struct btnode * n, u_long key)
{
   int   result;
   struct b_path * bpath = (struct b_path *)n->data;

   result = _strnicmp(bpath->pathname, (char *)key, bpath->length );

   if(result > 0)
      return KEY_GT;
   if(result < 0)
      return KEY_LT;
   return KEY_EQ;
}

/* new_path(char * pathname) - make a new b_path object.
 * b_path is the databse entry for a blocked URL path/name, eg
 * pathname of "*popup*" blocks URLs containing the pattern 
 * "popup". 
 * 
 * Returns object if OK, else null. 
 */


struct b_path *
new_path(char * pathname, struct comm_dbud * comm)
{
   struct b_path * newpath;
   struct b_path * oldpath;
   char *   pathbuf;
   char *   cp;
   int      length;

   /* Check parameters and allocate resources first. */
   if((comm == NULL) || (pathname == NULL))
   {
      dtrap();
      return NULL;
   }

   /* get buffer for path/name - first get length of pathname */
   for(cp = pathname; *cp > ' '; cp++)
      ;
   length = cp - pathname;
   pathbuf = PTH_ALLOC(length + 1);
   if(pathbuf == NULL)
      return NULL;               /* return failure */

   /* Parms look OK, allocate the DB structure */
   newpath = (struct b_path *)PB_ALLOC(sizeof(struct b_path));
   if(newpath == NULL)
   {
       PTH_FREE(pathbuf);
      return NULL;
   }
   newpath->ctl.comm = comm;
   comm->cm_currec++;         /* Bump community current rec count */

   /* check for wildcard chars */
   if(*pathname == '*')
   {
      newpath->flags |= BP_LEADWILD;
      pathname++;    /* remove '*' from string head */ 
      length--;
   }
   if(*(pathname + length - 1) == '*')
   {
      newpath->flags |= BP_TAILWILD;
      length--;
      *(pathname + length) = 0;     /* put zero over '*' character */
   }

   newpath->pathname = pathbuf;
   strncpy(newpath->pathname, pathname, length);    /* cp */
   newpath->length = length;

   /* Indicate the type of this object in the ctl structure */
   newpath->ctl.action_mask = CTF_TYPEPATH;

   /* see if we already have an entry for this path */
   oldpath = path_lookup(pathname);
   if(oldpath)
   {
      int err;
      newpath->btnp = NULL;    /* New item not in btree */

      /* Call routine to chain duplicate objects together */
      err = br_dupobject(&newpath->ctl, &newpath->btnp, &oldpath->btnp, 
         &oldpath->ctl, newpath);
      if(oldpath->btnp && (oldpath->btnp->data != (struct btnode **)oldpath))
         dtrap();
      if(newpath->btnp && (newpath->btnp->data != (struct btnode **)newpath))
         dtrap();
      if(err)
         return NULL;
   }
   else  /* no duplicate, make new btree entry */
   {
      /* get btree support structure */
      newpath->btnp = BTN_ALLOC(sizeof(struct btnode));
      if(newpath->btnp == NULL)
      {
         PB_FREE(newpath);
         PTH_FREE(pathbuf);
         return NULL;
      }
      newpath->btnp->data = (struct btnode **)newpath;         /* set btree node back pointer */
      TH_LOCK_MUTEX(br_tree_mutex);
      bt_insert( &(struct btnode *)path_tree, newpath->btnp,
         (u_long)pathname, bt_pathtest);
      TH_UNLOCK_MUTEX(br_tree_mutex);
   }

   return newpath;
}


struct b_path *
path_lookup(char * path)
{
   struct btnode * btnp;

   TH_LOCK_MUTEX(br_tree_mutex);
   btnp = bt_access( (struct btnode *)path_tree, (u_long)path, bt_pathtest );
   TH_UNLOCK_MUTEX(br_tree_mutex);
   if(btnp)
      return((struct b_path *)btnp->data);
   else
      return NULL;
}

void
del_path(struct b_path * bpath)
{
   if(bpath->btnp)
   {
      /* Check for queued duplicate items */
      if(bpath->ctl.dupctrl)
      {
         struct ctdesc * bestctl;
         struct b_path * bestpath;

         /* reorder dup list with best at head */
         bestctl = br_highestcomm(bpath->ctl.dupctrl);
         if(!bestctl)
         {
            dtrap();
            return;
         }

         /* Sanity check */
         if(bestctl == &bpath->ctl)
         {
            dtrap();
         }

         /* Get new best b_path from bestctl */
         bestpath = (struct b_path *)((char*)bestctl - 
            ((char*)(&bpath->ctl) - (char*)(bpath)));

         /* sanity check: make sure best is not deleted */
         if((bestpath->ctl.action_mask & CTF_TYPEMASK) != CTF_TYPEPATH)
         {
            dtrap();
            return;
         }

         /* Attach new best control to btree node */
         bestpath->btnp = bpath->btnp;
         bestpath->btnp->data = (struct btnode **)bestpath;     /* set btree node back pointer */

         bpath->btnp = NULL;       /* Detach old control */
      }
      else
      {
         TH_LOCK_MUTEX(br_tree_mutex);
         bt_remove( &(struct btnode *)path_tree, (u_long)(bpath->pathname), 
            bt_pathtest);
         TH_UNLOCK_MUTEX(br_tree_mutex);
      }
   }
   br_clearconnblocks( (void *)bpath );

   if(bpath->pathname)
       PTH_FREE(bpath->pathname);

   /* free control sub-fields */
   br_ctlclear(&bpath->ctl);

   PB_FREE(bpath);
}


/* br_nullterm() - NULL terminate and arg string. */

void
br_nullterm(char * arg)
{
   char * end;
   for(end = arg; *end > ' '; end++)
      ;
   *end = 0;
   return;
}




/********** POP3 client object constructor & destructor *********/

#ifdef POP3_MONITOR

struct pop_client * 
br_newpcl(char * username, struct endpoint * client, ip_addr server)
{
   struct pop_client * pcl;

   if(!client)
   {
      dtrap();       /* not supposed to happen */
      return NULL;
   }

   pcl = POPC_ALLOC( sizeof(struct pop_client) );
   if(!pcl)
      return NULL;

   pcl->pcl_client = client->ce_iphost;
   pcl->pcl_server = server;
   strncpy(pcl->pcl_name, username, MAXPOPLEN);

   pcl->unstable = 1;     /* Init as slightly unstable timing */

   if(br_cfg.cflags & BC_DESPAM)
      pcl->pflags |= POPF_FILTER;

   /* add to linked list */
   pcl->next = pop_clients;
   pop_clients = pcl;

   return pcl;
}


int   
br_delpcl(struct pop_client * pcl)
{
   struct pop_client * tmp;
   struct pop_client * last;

   /* scan pop client list for client to unlink */
   last = NULL;
   for(tmp = pop_clients; tmp; tmp = tmp->next )
   {
      /* see if this is the list item to unlink */
      if((pcl->pcl_client == tmp->pcl_client) &&
         (pcl->pcl_server == tmp->pcl_server) &&
         (strcmp(pcl->pcl_name, tmp->pcl_name) == 0))
      {
         /* found it; unlink */
         if(last)
            last->next = tmp->next;
         else
            pop_clients = tmp->next;

         POPC_FREE(pcl);
         return 0;
      }
      last = tmp;
   }
   dtrap();    /* client not in list? */
   return 0;
}



struct pop_client * 
br_lookuppcl(char * username, struct conn_entry * conn)
{
   struct pop_client * tmp;

   /* Sanity check parameters */
   if(conn->loc.ce_iphost == NULL)
   {
      dprintf("br_lookuppcl: null conn->loc\n");
      return NULL;
   }
   if(conn->rem.ce_iphost == NULL)
   {
      dprintf("br_lookuppcl: null conn->rem\n");
      return NULL;
   }

   /* scan pop client list for client to unlink */
   for(tmp = pop_clients; tmp; tmp = tmp->next )
   {
      /* see if this is the list item to return */
      if((conn->loc.ce_iphost == tmp->pcl_client) &&
            ((conn->rem.ce_iphost->ipaddr == tmp->pcl_server) || 
             (tmp->pcl_server == 0)) &&
         (strncmp(username, tmp->pcl_name, MAXPOPLEN ) == 0))
      {
         return tmp;
      }
   }
   return NULL;
}

#endif  /* POP3_MONITOR */

/*********** Content blocking word entries *****************/

/* hash table containg all the words in the system. */
struct vword * wordhashtab[VOCABHASHSIZE];

/* filtering flag, TRUE if we are filtering web [pages based on words
 * was active_vocabs 
 */
int   filter_web = FALSE;


/* wordcpy() - copy dest word to source. This is different than 
 *        strcpy in the following ways
 * - stops at first non-printable char (ie. space)
 * - converts chars to lowercase
 * - limits string length to MAXVWORD-1
 */

void
wordcpy(char * dest, char * src)
{
   int   len;

   len = 0;
   while(*src > ' ')
   {
      if(len++ > (MAXVWORD-1))
      {
         *dest = 0;
         return;
      }
      *dest++ = (*src++) | 0x20;
   }
   return;
}


/* wordcmp() - Compare two words This is different than 
 *        strcmp in the following ways
 * - stops at first non-printable char (ie. space) in either word
 * - converts chars to lowercase
 * - limits string length to MAXVWORD-1
 *
 * Returns 0 if words match within these parameters, 
 * 1 (one) if first word has higher alphabetic value, or
 * -1 if send word has has higher alphabetic value.
 */

int
wordcmp(char * word1, char * word2)
{
   while((*word1 >= 'A') && (*word1 <= 'z'))
   {
      if((*word1++ | 0x20) == (*word2++ | 0x20))
         continue;

      /* fall to here if lowercased chars didn't match */
      if(*(word1 - 1) < *(word2 - 1))
         return -1;
      else
         return 1;
   }

   /* get to here of word1 ended. If BOTH words have ended then
    * they match and we return 0.
    */
   if(*word2 <= ' ')
      return 0;   /* words match */

   return -1;     /* words not same length == no match */
}


/* new_word - create a new word entry */

struct vword *
new_word(char * word,  struct comm_dbud * comm)
{
   struct vword * newword;
   struct vword * oldword;
   int      hashval;

   /* Check if the word is already set */
   oldword = word_lookup(word);
   if(oldword)
   {
      if(oldword->ctl.comm == comm)
      {
         dtrap();
         dprintf("Duplicate word %s in community %s\n", 
            word, br_prtcomm(comm));
         return NULL;
      }
   }

   newword = (struct vword *)WD_ALLOC(sizeof(struct vword));
   if(!newword)
      return NULL;

   comm->cm_currec++;         /* Bump community current rec count */

   /* Fill in new word entry */
   wordcpy(newword->word, word);
   newword->ctl.comm = comm;

   /* add to hash table */
   hashval = wordhash(word);
   newword->nexthash = wordhashtab[hashval];
   wordhashtab[hashval] = newword;

   /* If the word is already in the system from another community, the
    * words need to be ranked per community priotity. 
    */
   if(oldword)
   {
      dtrap();    /* duplicate word entry, put in ctl queue */
      return NULL;
   }

   return newword;
}

/* del_word() - delete a word entry */

void
del_word(struct vword * wordp)
{
   struct vword * tmp;
   struct vword * last;
   struct vword * phrase = NULL;
   int      hashval;

   last = NULL;
   hashval = wordhash(wordp->word);
   for(tmp = wordhashtab[hashval]; tmp; tmp = tmp->nexthash)
   {
      if(tmp == wordp)
      {
         if(last)
            last->nexthash = wordp->nexthash;   /* unlink */
         else
            wordhashtab[hashval] = wordp->nexthash;

         /* If word is part of a phrase, remember link */
//         if(wordp->phrase)
//            phrase = wordp;

         /* free control sub-fields */
         br_ctlclear(&wordp->ctl);
         WD_FREE(wordp); /* free word memory */
         break;         /* exit for() loop */
      }
   }

   /* delete remainder of phrase, if any */
   if(phrase)
      del_word(phrase);

   return;
}


/* br_blkword() - Add word to vocalulary list
 *
 * Called form DB engine with text record describing word. 
 *
 * Returns NULL if word was added OK, else returns pointer to
 * error message.
 */

char *
br_blkword(char * readbuf, struct comm_dbud * comm, int commit)
{
   char *   word;
   char *   cp;            /* pointer to input line */
   int      weight;        /* read from input line */
   struct vword * wordp;


   word = br_getfield(readbuf, "word");
   if((word == NULL) || (*word <= ' '))
      return "no word field";

   cp = br_getfield(readbuf, "weight");
   if(*cp <= ' ')
      return "no weight in input";

   if((*cp == 'n') && (*(cp+1) == 'o'))
   {
      weight = WT_SHOWSTOPPER;
   }
   else
   {
      weight = atoi(cp);
      if((weight < 1) || (weight > 10))
         return "weight value not 1-10 or 'no'";
   }

   /* If we're just checking syntax, return OK now */
   if(!commit)
      return NULL;

   wordp = new_word(word, comm);
   if(!wordp)
      return "new_word() failed";

   wordp->weight = weight;

   /* extract generic ctl info */
   read_ctlinfo(readbuf, comm, CTF_TYPEWORD, &wordp->ctl);
   wordp->ctl.offense = mot_bynum(OF_CONTENT);

   return NULL;      /* OK return */
}


/* word_lookup() - Look up a word in content blocking database. 
 *
 * NOTE - this is called for every word of scanned web pages, so
 * is must be really FAST!!!
 *
 * Returns NULL if no matching words found.
 */

struct vword *
word_lookup(char * word)
{
   struct vword * wordp;
   struct vword * result;
   int      hashval;

   hashval = wordhash(word);
   if(wordhashtab[hashval] == NULL)
      return NULL;

   result = NULL;
   for(wordp = wordhashtab[hashval]; wordp; wordp = wordp->nexthash)
   {
      if(wordcmp(word, wordp->word) == 0)
      {
         /* add matching word to the result list */ 
         wordp->answerlist = result;
         result = wordp;
      }
   }
   return result;
}

/* wordhash() - make hash value for content blocking word */

int
wordhash(char * word)
{
   int hashnum;

   hashnum = 0;
   while((*word >= 'A') && (*word <= 'z'))
      hashnum += ((*word++)  | 0x20);     /* uppercase */
   hashnum &= (VOCABHASHSIZE - 1);
   return hashnum;
}

/*
Copyright (C) 2003-2008 by Praemio Inc. All rights reserved.

Author: John Bartas - jbartas@praemio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/



