/* brdbio.c
 *
 * Database file IO routines. 
 *
 * Copyright 2003-2008 by Praemio Inc. All rights reserved.
 * Author: John Bartas - jbartas@praemio.com
 * This software is released under the terms of the LGPL.
 * See LGPL Notice at end of this file; Full license in ../license.txt
 *
 * ROUTINES: br_nextfield(), br_getfield(), br_fieldlen(), 
 *  strcrlen(), read_ctlinfo(), br_blkport(), br_blkdns(), 
 *  br_blkpath(), br_blkip(), append_ctl(), br_blkdnsout(), 
 *  br_blkipout(), br_blkpathout(), br_blkportout(), br_readrec(), 
 *  br_parseitem(), br_parsefirst(), br_md5comm(), br_readdbfile(), 
 *  br_dbread(), br_dbwrite(), br_dbcheck(), br_cmlookup(), 
 *  br_cmbyfile(), br_cmbypriority(), br_cmnew(), br_cmdel(), 
 *  br_cmupdate(),
 *
 */

#include "busterport.h"
#include "brnet.h"
#include "ip.h"
#include "brdb.h"
#include "brmd5.h"  // Actually just plain old RSA MD5 */


void *  br_comm_mutex;

char *   br_nextfield(char * rec);  /* get next field in record */

extern   void  comm_cleanup(struct comm_dbud * comm);

static   char     filebuf[2048];


/* Define the highest priority a community can have */
#define COMM_MAX_PRIORITY   0x7FFF  

/* The local community (local to us). This has no URL,
 * file is always "local.cdb", priority always highest.
 */

struct comm_dbud local_comm =
{
   NULL,    /* No next */
   { 0 },   /* no URL */
   NULL,    /* no pointer to URL path */
   IDLE,    /* state always idle */
   0L,      /* IP server - unused */
   0,       /* port - unused */
   COMM_MAX_PRIORITY,   /* highest priority */
   { "local" },         /* community name */
   { "local.cdb" },     /* local CDB file name */

   /* rest is zero */
};

/* Linked list of all communities (except local, below) */
struct comm_dbud * br_communities = &local_comm;



/******** Record read input routines (and helpers) for data classes *********/

/* br_nextfield(rec)  
 *
 * Get next field in record. Stops at end of record, which can be
 * either <CR/LF/CR/LF>, or NULL. 
 *
 * Helper routine for br_getfield.
 *
 * Returns pointer to next field if there is one, else null.
 */

char *
br_nextfield(char * rec)
{
   int   crlfs;
   while(*rec++ >= ' ')
         ;
   /* got control char, should be crlf or null */
   if(*rec == 0)
      return NULL;      /* End of record */

   /* Next chars should be CR and LF */
   crlfs = 0;
   while((*rec == '\r') || (*rec == '\n'))
   {
      rec++;
      crlfs++;
   }
   /* There MUST be one or two of these */
   if((crlfs < 1) || (crlfs > 2))
      return NULL;

   if(*rec >= ' ')
      return rec;    /* next field */
   else
      return NULL;  /* end of record? */
}


/* br_getfield(char * rec, char * field) 
 * 
 * rec - IN - pointer to DB text record
 * field - IN - name of field to extract, eg "type"
 *
 * Returns pointer to text for field, NULL if not found. Pointer may
 * point to NULL or CR if field has no data.
 */

char *
br_getfield(char * rec, char * field) 
{
   char * cp;
   int   len;

   len = strlen(field);
   for(cp = rec; cp; cp = br_nextfield(cp) )
   {
      if(_strnicmp(cp, field, len) == 0)
      {
         cp += len;
         if(*cp != ':')
            return NULL;   /* missing colon */
         cp++;
         while(*cp == ' ') /* get past spaces */
            cp++;
         return cp;
      }
   }
   return NULL;
}


/* br_fieldlen( char * field) 
 * 
 * Returns length of the passed field. Field may end with either CR or NULL.
 */

int 
br_fieldlen(char * desc)
{
   int len = 0;

   while(*desc++ >= ' ')
      len++;

   return len;
}

/* strcrlen(char * text) - return length of text up to first 
 * Control char (CR, LF, NULL, etc.)
 */

int
strcrlen(char * text)
{
   char * cp = text;
   while(*cp++ >= ' ')
      ;
   return ((cp - text) - 1);
}



/* read_ctlinfo()
 *
 * Extract the control info (ID, comment, other "-" tags)
 * from a text buffer (DB file or net update).
 * Info is returned in the "ctl" passed.
 *
 * NOTES:
 * - Comment buffers are alloced, caller should ensure that they
 * are eventually freed.
 * - Control passed is ZEROed out, and data in it is lost.
 *
 * Returns nothing, this is "best effort".
 */

void
read_ctlinfo(char *  readbuf,          /* IN - text buffer which may have info */
            struct comm_dbud * comm,   /* IN - community */
            int      ctltype,
            struct ctdesc * ctl)       /* OUT - filled in ctl */
{
   char * cp;
#ifdef DESC_BLOAT
   char * desc;
   char * tags;
   int   len;
#endif
   char * reason;

   MEMSET(ctl, 0, sizeof(struct ctdesc) );   /* clear output structure */

   ctl->action_mask = ctltype;      /* set type */

   /* if Status: is whitelist, set control bit */
   cp = br_getfield(readbuf, "status");
   if(cp && (stricmp(cp, "Whitelist") == 0))
   {
      ctl->action_mask |= CTF_WHITELIST;
   }

   ctl->comm = comm;

   /* If buf has comment, copy it into ctl */
#ifdef DESC_BLOAT
   desc = br_getfield(readbuf, "description");
   if(desc)
   {
      len = br_fieldlen(desc);
      ctl->desc = CTD_ALLOC(len + 1); /* get buffer */
      if(ctl->desc)
         strncpy(ctl->desc, desc, len);
   }

   /* Ditto with tags list */
   tags  = br_getfield(readbuf, "tags");
   if(tags)
   {
      len = br_fieldlen(tags);
      ctl->tags = CTT_ALLOC(len + 1); /* get buffer */
      if(ctl->tags)
         strncpy(ctl->tags, tags, len);
   }
#endif /* DESC_BLOAT */

   /* Look for and save "reason" for this item */
   reason = br_getfield(readbuf, "reason");
   if(reason)
   {
      if( (ctl->offense = mot_byname(reason)) == NULL )
      {
         br_log(0, "bad DB reason %s in community %s\n",
            reason, comm->cm_url);
      }

   }
   else  /* no reason given */
   {
      ctl->offense = mot_byname("Other");
   }

   return;
}


/* br_blkport() - make a blocked port entry from a line of text
 * 
 * Return NULL if OK, else error text;
 */

char *
br_blkport(char * readbuf, struct comm_dbud * comm, int commit)
{
   struct control_port * cpp;
   unshort  port;
   char *   cp;
   char *   direction;

   cp = br_getfield(readbuf, "port");
   if(!cp)
      return ("no port: field");

   port = (unshort)atoi(cp);
   if((port == 0) && (*cp != '0'))
      return("bad port number");

   cp = br_getfield(readbuf, "protocol");
   if(!cp)
      return ("no protocol: field");

   if((*cp != 'U') && (*cp != 'T'))
      return ("bad protocol");

   direction = br_getfield(readbuf, "direction");
   if(!direction)
      return ("no direction");

   /* If we're just checking syntax, return OK now */
   if(!commit)
      return NULL;

   if(*cp == 'T')
      cpp = new_cp(port, CPF_TCP, comm);
   else
      cpp = new_cp(port, CPF_UDP, comm);

   if(cpp == NULL)
      return ("alloc failed");    /* out of memory */

   /* See if there is an image file name */
   cp = br_getfield(readbuf, "Image");
   if(cp)
   {
      int   len = br_fieldlen(cp);
      if(len > MAXFILENAME)
      {
         dtrap();
         len = MAXFILENAME;
      }
      strncpy(cpp->imagefile, cp, len);    /* copy in file name */
   }

   /* Set flags for applying port to local or remote host, or
    * (default) both. Since this may be an update, we first clear 
    * any existing REMOTE/LOCAL flags 
    */
   if(*direction == 'I')
      cpp->cp_flags |= CPF_REMOTE;
   else if(*direction == 'O')
      cpp->cp_flags |= CPF_LOCAL;
   else if(*direction == 'B')
      cpp->cp_flags |= (CPF_LOCAL | CPF_REMOTE);

   read_ctlinfo(readbuf, comm, CTF_TYPEPORT, &cpp->ctl);   /* extract generic ctl info */

   return NULL;
}



/* br_blkdns() - make a blocked DNS entry from a line of test
 * 
 * Return NULL if OK, else error text;
 */

static char dupdns_msg[MAXDNSNAME + 21] = {"Duplicate DNS name "};

char *
br_blkdns(char * readbuf, struct comm_dbud * comm, int commit)
{
   char *   dname;
   char *   cr;
   char *   slash;
   char *   cp;
   int      is_ipaddr;
   ip_addr  ipaddr;
   struct control_dn * dnp;
   struct control_ip * ipe;

   dname = br_getfield(readbuf, "domain_name");
   if(!dname)
      return ("missing \"domain_name:\" field");

   /* Some buggy servers leave URL stuff prepended to name */
   if(strncmp(dname, "http://", 7) == 0)
      dname += 7;

   cr = strchr(dname, '\r');  /* Null terminate string */
   if(cr)
      *cr = 0;

   slash = strchr(dname, '/');   /* Also, remove any path */
   if(slash)
      *slash = 0;

   /* See if the "domain name" is really an IP addess */
   cp = br_checkip(&ipaddr, dname);
   if(!cp)     /* cp == NULL if danme parsed as IP address */
   {
      is_ipaddr = -1;   /* Set flag to TRUE */

      /* Make sure IP is not already in the database */
      ipe = ipe_lookup(htonl(ipaddr));
      if(ipe && ipe->ctl.comm == comm)
      {
         strncpy(&dupdns_msg[19], dname, MAXDNSNAME);
         return dupdns_msg;
      }
      dnp = NULL;    /* stifle compiler warnings about this being unset */
   }
   else  /* Proper name, not IP address */
   {
      is_ipaddr = 0;    /* Clear flag */

      /* See if we have an identical DNS name already. Since dn_lookup 
       * will return master domains (e.g. passing "foo.bar.com" can return
       * "bar.com") we check for exact string match.
       */
      dnp = dn_lookup(dname);
      if(dnp && 
         (dnp->ctl.comm == comm) &&
         (strcmp(dnp->cu_name, dname) == 0))    /* Names are identical */
      {
         strncpy(&dupdns_msg[19], dname, MAXDNSNAME);
         return dupdns_msg;
      }
      ipe = NULL;    /* stifle compiler warnings about this being unset */
   }

   /* If we're just checking syntax, return OK now */
   if(!commit)
      return NULL;

   if(is_ipaddr)
      ipe = new_ipe(ipaddr, NULL, CIF_COMM, comm);
   else
      dnp = new_dn(dname, comm);     /* make for domain entry */

   /* put back anything we clobbered */
   if(cr)
      *cr = '\r';
   if(slash)
      *slash = '/';

   if(is_ipaddr)
   {
      if(ipe == NULL)            /* check for out of memory */
         return("alloc failed");
      read_ctlinfo(readbuf, comm, CTF_TYPEIP, &ipe->ctl);   /* extract generic ctl info */
   }
   else
   {
      if(dnp == NULL)            /* check for out of memory */
         return("alloc failed");
      read_ctlinfo(readbuf, comm, CTF_TYPENAME, &dnp->ctl);   /* extract generic ctl info */
   }

   return NULL;      /* OK return */
}


/* br_blkpath() - make a blocked path entry from a line of test
 * 
 * Return NULL if OK, else error text;
 */

char *
br_blkpath(char * readbuf, struct comm_dbud * comm, int commit)
{
   struct b_path * bp;
   char *   path;
   char *   cr;

   path = br_getfield(readbuf, "pattern");
   if(!path)
      return ("no path.");

   cr = strchr(path, '\r');
   if(cr)
      *cr = 0;

   /* If we're just checking syntax, return OK now */
   if(!commit)
      return NULL;

   /* now that we have path & offense, create the db entry */
   bp = new_path(path, comm);
   if(cr)
      *cr = '\r';
   if(!bp)   /* URL path database entry */
      return ("alloc failed");

   read_ctlinfo(readbuf, comm, CTF_TYPEPATH, &bp->ctl);  /* extract generic ctl info */
   return NULL;      /* OK return */
}



/* br_blkip(char * buf) - make a control IP entry from a line of text.
 *
 * input string is of the format:  "ipaddr1 ipaddr2 [domain]" 
 *
 * Returns NULL if OK or pointer to error description text.
 */

static char dupip_msg[40] = {"Duplicate IP address "};

char *
br_blkip(char * buf, struct comm_dbud * comm, int commit)
{
   char * cp;
   char * cr;
   char * domain_name;
   char * iptxt;  /* Text for IP address */
   ip_addr  ipaddr;
   struct control_ip * ipe;
   struct control_dn * domain;

   iptxt = br_getfield(buf, "ip_address");
   if(!iptxt)
      return ("no ip_address");
   cr = strchr(iptxt, '\r');
   if(cr)
   {
      while(*cr == ' ')    /* trailing spaces upset br_checkip */
         cr--;
      *cr = 0;
   }

   /* get first IP address */
   cp = br_checkip(&ipaddr, iptxt);
   if(cr)
      *cr = '\r';
   if(cp)
      return(cp);

   /* Get optional domain name or NULL */
   domain_name = br_getfield(buf, "domain_name");

   /* Make sure IP is not already in the database */
   ipe = ipe_lookup(htonl(ipaddr));
   if(ipe && ipe->ctl.comm == comm)
   {
      cp = print_ipad(ipaddr);
      strcpy(&dupip_msg[21], cp); 
      return dupip_msg;
   }

   /* If we're just checking syntax, return OK now */
   if(!commit)
      return NULL;

   /* Create new blocked IP struct */
   ipe = new_ipe(ipaddr, NULL, CIF_COMM, comm);
   if(ipe == NULL)      /* check for out of memory */
     return ("can't make blocked IP address entry");

   /* Associate domain with the IP address (or range) */
   if(domain_name)
   {
      /* install this IP address range in DN list */
      cr = strchr(domain_name, '\r');
      if(cr)
         *cr = 0;
      domain = dn_lookup(domain_name);
      if(cr)
         *cr = '\r';
      if(domain == NULL)
      {
         dtrap();
      }
      else
         br_insertcip(domain, ipe);
   }

   /* read the control info */
   read_ctlinfo(buf, comm, CTF_TYPEIP, &ipe->ctl);   /* extract generic ctl info */

   return 0;
}

/******** output routines for data classes write to file *********/

/* The next few routines assemble the line in readbuf[], and then write 
 * to file. This assumes we won't read & write the file 
 * simultaineously. 
 */

 
/* Helper for write routine - this takes the ID, comment and other 
 * fields from a ctl struct and appends them in databse text format 
 * to a database-bound text buffer.
 *
 * Does length checking via the end of buffer pointer passed.
 */

#define CTL_OVERHEAD  60

void
append_ctl(struct ctdesc * ctl, char * buffer, int bufsize)
{
   char *   cp;

   /* figure out where free portion of buffer starts */
   while(*buffer > 0)
   {
      buffer++;
      bufsize--;
   }

   /* Check length. All fields (excluding description) have a reliable 
    * maximum size, which totals to about CTL_OVERHEAD.
    */
   if(bufsize < CTL_OVERHEAD)
   {
      dtrap();
      return;
   }

   /* Status field - Handle whitelisted items */
   if( WHITELISTED(ctl) )
      cp = "Whitelist";
   else
      cp = "Blocked";
   sprintf(buffer, "Status: %s\r\n", cp);

   if(ctl->offense)
   {
      cp = buffer + strlen(buffer);
      sprintf(cp, "Reason: %s\r\n", ctl->offense->desc );
   }

#ifdef DESC_BLOAT
   if(ctl->desc)
   {
      char * end;
      cp = buffer + strlen(buffer);

      /* If description won't fit, truncate it. */
      length = strlen(ctl->desc) + 16;
      end = buffer + bufsize;
      if((cp + length) > end)
      {
         *(ctl->desc + (end - (cp + length))) = 0;
      }
      sprintf(cp, "description: %s\r\n", ctl->desc);
   }
   if(ctl->tags)
   {
      char * end;
      cp = buffer + strlen(buffer);

      /* If description won't fit, truncate it. */
      length = strlen(ctl->tags) + 8;
      end = buffer + bufsize;
      if((cp + length) > end)
      {
         *(ctl->tags + (end - (cp + length))) = 0;
      }
      sprintf(cp, "tags: %s\r\n", ctl->tags);
   }
#endif /* DESC_BLOAT */

   return;
}


/* tree helper routine to write dns block recs */

void
br_blkdnsout(struct btnode * n, long param, int depth)
{
   struct control_dn *  dnp;
   struct ctdesc *      ctrlp;
   struct comm_dbud *   comm = (struct comm_dbud *)param;
   FILE *               fout = comm->cm_fd;

   dnp = (struct control_dn *)n->data;

   /* skip if this item is not in current community, but trace 
    * down the list of duplicates for a member that IS in this comm.
    */
   for(ctrlp = &dnp->ctl; ctrlp; ctrlp = ctrlp->dupctrl)
   {
      if(ctrlp->comm == comm)
         break;
   }
   if(!ctrlp)
      return;

   sprintf(filebuf, "type: DomainName\r\ndomain_name: %s\r\n", dnp->cu_name);
   append_ctl(&dnp->ctl, filebuf, sizeof(filebuf) );
   comm->cm_currec++;         /* Bump community current rec count */

   fprintf(fout, "%s\n", filebuf);

   USE_ARG(depth);
   return;
}


/* tree helper routine to write IP block recs */
void
br_blkipout(struct btnode * n, long param, int depth)
{
   struct control_ip *  cip;
   struct ctdesc *      ctrlp;
   struct comm_dbud *   comm = (struct comm_dbud *)param;
   FILE *               fout = comm->cm_fd;
   char *               cp;

   cip = (struct control_ip *)n->data;

   /* skip if this item is not in current community, but trace 
    * down the list of duplicates for a member that IS in this comm.
    */
   for(ctrlp = &cip->ctl; ctrlp; ctrlp = ctrlp->dupctrl)
   {
      if(ctrlp->comm == comm)
         break;
   }
   if(!ctrlp)
      return;

   /* Also skip IPs with timeouts - they came from DBS, and 
    * should be re-obtained from DNS info after the file is 
    * read back later on.
    */
   if(cip->ci_timeout)
      return;

   memset(filebuf, ' ', 16);

   /* format the IP addrinto first space */
   cp = print_ipad(htonl(cip->ci_ipaddr));

   sprintf(filebuf, "type: IPAddress\r\nip_address: %s\r\n", cp);
  
   /* If there's dns info associated with these addresses, append
    * the domain name on the line.
    */

   if(cip->ci_name)
   {
       cp = filebuf + strlen(filebuf);
      sprintf(cp, "domain_name: %s\r\n", cip->ci_name->cu_name);
   }

   append_ctl(&cip->ctl, filebuf, sizeof(filebuf) );
   comm->cm_currec++;         /* Bump community current rec count */

   fprintf(fout, "%s\n", filebuf);

   USE_ARG(depth);
   return;
}


void
br_blkpathout(struct btnode * n, long param, int depth)
{
   struct b_path * bp;
   char *   cp;
   struct ctdesc * ctrlp;
   struct comm_dbud * comm = (struct comm_dbud *)param;
   FILE * fout = comm->cm_fd;


   bp = (struct b_path *)n->data;

   /* skip if this item is not in current community, but trace 
    * down the list of duplicates for a member that IS in this comm.
    */
   for(ctrlp = &bp->ctl; ctrlp; ctrlp = ctrlp->dupctrl)
   {
      if(ctrlp->comm == comm)
         break;
   }
   if(!ctrlp)
      return;

   sprintf(filebuf, "type: FilePath\r\npattern: ");

   /* printf path name text, with wildcard ("*") markers */
   cp = filebuf + strlen(filebuf);
   if(bp->flags & BP_LEADWILD)
      strcat(cp, "*");
   strcat(cp, bp->pathname);
   if(bp->flags & BP_TAILWILD)
      strcat(cp, "*");

   cp = cp + strlen(cp);
   *cp++ = '\r';     /* End of path field */
   *cp++ = '\n';
   *cp++ = 0;

   append_ctl(&bp->ctl, filebuf, sizeof(filebuf) );
   comm->cm_currec++;         /* Bump community current rec count */

   fprintf(fout, "%s\n", filebuf);

   USE_ARG(depth);
   return;
}

/* shared code for writing TCP and UDP ports */

void
br_blkportout(struct btnode * n, long param, int depth)
{
   struct control_port * cpp;
   char *   prot;
   char *   cp;
   char *   rl;
   struct ctdesc * ctrlp;
   struct comm_dbud * comm = (struct comm_dbud *)param;
   FILE *   fout = comm->cm_fd;

   cpp = (struct control_port *)n->data;

   /* skip if this item is not in current community, but trace 
    * down the list of duplicates for a member that IS in this comm.
    */
   for(ctrlp = &cpp->ctl; ctrlp; ctrlp = ctrlp->dupctrl)
   {
      if(ctrlp->comm == comm)
         break;
   }
   if(!ctrlp)
      return;

   if(cpp->cp_flags & CPF_TCP)
      prot = "TCP";
   else if(cpp->cp_flags & CPF_UDP)
      prot = "UDP";
   else
   {
      dtrap();    /* Bogus protocol type */
      return;
   }

   sprintf(filebuf, "type: Port\r\nport: %u\r\nprotocol: %s\r\n", 
      cpp->cp_port, prot);

   /* set remote/local string. Blank is both */
   if((cpp->cp_flags & CPF_REMOTE) &&
      (cpp->cp_flags & CPF_LOCAL))
   {
      rl = "both   ";     /* both remote and local */
   }
   else if(cpp->cp_flags & CPF_RSONLY)
      rl = "inbound";
   else if(cpp->cp_flags & CPF_LOCAL)
      rl = "local  ";
   else if(cpp->cp_flags & CPF_REMOTE)
      rl = "remote ";
   else 
      rl = "       ";    /* entry is just informational */

   cp = filebuf + strlen(filebuf);
   sprintf(cp, "direction: %s\r\n", rl);
 
   if(cpp->imagefile[0])    /* see if entry has associated icon image */
   {
       cp = filebuf + strlen(filebuf);
      sprintf(cp, "image: %s\r\n", cpp->imagefile);
   }

   append_ctl(&cpp->ctl, filebuf, sizeof(filebuf) );
   comm->cm_currec++;         /* Bump community current rec count */

   fprintf(fout, "%s\n", filebuf);

   USE_ARG(depth);
   return;
}

/* br_blkwordout() - Save community word list to data file */

void
br_blkwordout(struct comm_dbud * comm)
{
   int      i;
   FILE * fout = comm->cm_fd;
   struct vword * wordp;

   for(i = 0; i < VOCABHASHSIZE; i++)     /* Scan hash table */
   {
      for(wordp = wordhashtab[i]; wordp; wordp = wordp->nexthash)
      {
         if(wordp->ctl.comm != comm)
            continue;

         /* Conpose a file entry for this word */
         if(wordp->weight == WT_SHOWSTOPPER)
         {
             sprintf(filebuf, "type: Word\r\nword: %s\r\nweight: no\r\n", 
               wordp->word);
         }
         else
         {
             sprintf(filebuf, "type: Word\r\nword: %s\r\nweight: %u\r\n", 
               wordp->word, wordp->weight);
         }

         /* tags/description are handled herein */
         append_ctl(&wordp->ctl, filebuf, sizeof(filebuf) );
         comm->cm_currec++;         /* Bump community current rec count */
         fprintf(fout, "%s\n", filebuf);
      }
   }

   return;
}


/******** Generic type Whole-file oriented routines follow *******/


/* br_readrec() - Read a record "item" from a community DB file 
 *
 * Returns number of lines in record, 0 if EOF, negative if error.
 */ 

int
br_readrec(char * recbuf,     /* OUT - buffer with next record */
           int    recbuflen,  /* IN - length of recbuf */
           FILE * fp)
{
   char *   cp;
   char *   next;
   int      remains;
   int      reclen;
   int      reclines;

   next = recbuf;
   remains = recbuflen;
   reclines = 0;

   /* Read the file until we get a double CRLF - end of rec */
   for(;;)
   {
      cp = fgets(next, remains, fp);

      if(cp != next)        /* End of file */
         break;

      /* If we got no text, it's end of record */
      if((*cp == 0) || (*cp == 10) || (*cp == 13))
      {
         /* If the record is empty, try reading some more - it
          * may be a triple crlf, which we accept .
          */
         if(reclines == 0)
            continue;
         break;
      }

      reclen = strcrlen(next);
      /* check for buffer overflow */
      if((remains - reclen) < 8)
      {
         return -1;
      }

      next += reclen;
      remains -= reclen;
      reclines++;

      /* Put CRLF at end of field */
      *next++ = '\r';
      *next++ = '\n';
      remains -= 2;
   }

   *next = 0;     /* Null terminate record */

   return reclines;
}

/* br_parseitem() - parse data record from a community DB stream 
 *
 * Param 3: boolean - TRUE if item should be committed to database,
 * FALSE if all we want is to check the format.
 *
 * Returns NULL if parse OK, else text describing error .
 */

char *
br_parseitem(char * readbuf, struct comm_dbud * comm, int commit)
{
   char * type;   /* type of record */
   char * cp;
   int   code;

   type = br_getfield(readbuf, "type");
   if(!type)
      return ("missing type field");

   code = (*(type)) << 24;
   code += (*(type + 1)) << 16;
   code += (*(type + 2)) << 8;
   code += *(type + 3);

   cp = NULL;        /* clear error indicator */

   /* pass the readbuf to the correct routine for the type of 
    * data in this section. Each of these routines returns a
    * NULL if the parsing was OK, else text describing the problem.
    */
   switch(code)
   {
#ifdef VIRUS_CHECK
   case 0x42696E61:  /* "Bina"ry */
      cp = br_blkbinary(readbuf, comm, commit);
      break;
#endif  /* VIRUS_CHECK */
   case 0x446F6D61:  /* "Doma"in name  */
      cp = br_blkdns(readbuf, comm, commit);
      break;
   case 0x49504164:  /* "IPAd"dress */
      cp = br_blkip(readbuf, comm, commit);
      break;
   case 0x496D6167:  /* "Imag"e */
      cp = br_mkimage(readbuf, comm, commit);
      break;
   case 0x46696C65:  /* "File"Path */
      cp = br_blkpath(readbuf, comm, commit);
      break;
   case 0x506f7274:  /* "Port" */
      cp = br_blkport(readbuf, comm, commit);
      break;
   case 0x4D414341:  /* "MACA"ddress */
      dtrap(); /* Unused? */
      br_readhost(readbuf, NULL);
      break;
   case 0x576F7264:  /* Word */
      cp = br_blkword(readbuf, comm, commit);
      break;
   default:
      cp = "Unknown type field";
      break;
   }
   return cp;
}

/* br_parsefirst() - parse the first record in a community DB stream */

char *
br_parsefirst(char * readbuf, 
              int * revision, 
              int * records, 
              int * nextud)
{
   char *   cp;
   int      tmp;

   cp = br_getfield(readbuf, "revision");
   if(!cp)
      return ("Missing DB revision");

   tmp = atoi(cp);
   if((tmp < 0) ||
      ((tmp == 0) && (*cp != '0')))
   {
      return ("Bad DB revision number");
   }

   *revision = tmp;
   
   cp = br_getfield(readbuf, "records");
   if(!cp)
      return ("Missing DB record count");

   tmp = atoi(cp);
   if((tmp == 0) && (*cp != '0'))
      return ("Bad number of records");

   *records = tmp;

   cp = br_getfield(readbuf, "update_interval");
   if(cp)
   {
      *nextud = atoi(cp);
      if((*nextud == 0) && (*cp != '0'))
         return ("Bad Update Interval time");
   }
   else
      *nextud = -1;
   
   return NULL;
}


/******** Whole-community oriented routines follow *******/

/* br_md5comm()
 *
 * MD5 sum a community file. passed file is open. The passed file 
 * should be completely written, with the possible exception of the 
 * digest value in the last record. This seeks to the head of 
 * the file before starting, and leaves at the digest record if it 
 * exists, or at the end of the file.. 
 *
 * Digest is returned in passed structure, failure is detected by 
 * incorrect digest value.
 */

void
br_md5comm(FILE * fp, MD5_CTX * digestp, char * filebuf, int bufsize)
{

   fseek(fp, 0, SEEK_SET);    /* Set file back to start */

   MD5Init(digestp);  
   while( br_readrec(filebuf, bufsize, fp) != 0)
   {
      /* Skip last record */
       if((br_getfield(filebuf, "digest")) || 
          (strncmp(filebuf, "EOF", 3) == 0))
      {
         break;
      }
      MD5Update(digestp, (u_char*)filebuf, strlen(filebuf) );
   }
   MD5Final(digestp);

   return;
}

/* br_readdbfile()
 *
 * Read an open CDB file into a community. This should not 
 * close the file or make any assumptions other than that 
 * the file is open and readable.
 *
 * Returns 0 if OK, else ENP_ code 
 */

char *
br_readdbfile(FILE * fp, struct comm_dbud * comm)
{
   int      actual_recs;
   int      revision;
   int      header_recs;
   int      interval;
   int      rlines;  /* Lines in current rec */
   int      line;    /* Line count (> record count) */
   char *   cp;

   fseek(fp, 0, SEEK_SET);    /* Set file back to start */

   /* read the file into memory structures */
   actual_recs = line = revision = header_recs = 0;
   comm->cm_currec = 0;

   while( (rlines = br_readrec(filebuf, sizeof(filebuf), fp)) != 0)
   {
      actual_recs++;
      line += (rlines + 1);   /* rec lines plus blank */
      if(actual_recs == 1)
      {
          cp = br_parsefirst(filebuf, &revision, &header_recs, &interval);
         if(cp)
         {
            fclose(fp);
            return (cp);
         }

         if(interval != -1)
            comm->cm_interval = interval * TPS;

         /* If we already have the data, return error */
         if(comm->cm_revision == (unsigned)revision)
            return ("Already have the revision");

         continue;
      }

      cp = br_parseitem(filebuf, comm, TRUE);
      if(cp)            /* Did we get a create error message? */
      {
          if(br_getfield(filebuf, "digest"))
            break;
          if(strstr(filebuf, "EOF"))
            break;

         fclose(fp);
         return (cp);
      }
   }

   actual_recs -= 2;     /* Subtract first & last rec */
   if((u_long)actual_recs != comm->cm_currec)
   {
      dtrap(); /* Programmer forgot to count record? */
   }

   if(actual_recs != header_recs)
   {
      dtrap();    /* This should have been caught in pre-checking */
      dprintf("Record count mismatch, actual: %d, header: %d\n",
         actual_recs, header_recs);
      return ("record count mismatch");
   }

   comm->cm_revision = revision;
   comm->cm_records += actual_recs;
   return NULL;
}


/* br_dbread() - read a database text file into our DB structures
 *
 * Returns 0 if OK, else ENP_ error code.
 */

static char hexdigest[MAXUSERLENGTH + 2];
static int  in_dbread = 0;

int
br_dbread(struct comm_dbud * comm)
{
   int      err;     /* return value */
   char *   cp;      /* scratch */
   FILE *   fp;      /* nv file */
   MD5_CTX  xdigest;

   if(in_dbread++)
   {
      in_dbread--;
      dprintf("dbreda: reentry, ct:%d\n", in_dbread);
      dtrap();
      while(in_dbread);
         TH_SLEEP(TPS);
   }

   err = 0;    /* Set return value - no error */

   fp = fopen(comm->cm_filename, "rb");
   if(!fp)
   {
      dprintf("Error opening community file\n");
      err = ENP_FILEIO;
      goto error_exit;
   }

   dprintf("Verifying database file %s...", comm->cm_filename );

   br_md5comm(fp, &xdigest, filebuf, sizeof(filebuf) );

   br_printdigest(hexdigest, xdigest.digest, 16);

   cp = br_getfield(filebuf, "digest");
   if(!cp)
   {
      dprintf("Missing digest line in file %s\n", comm->cm_filename);
      dtrap(); /* Send alert? */
      fclose(fp);
      err = ENP_PARAM;
      goto error_exit;
   }

   if(strncmp(cp, &hexdigest[0], 32))
   {
      dprintf("Corrupted CDB file %s\n", comm->cm_filename);
      dtrap(); /* Send alert? */
      fclose(fp);
      err = ENP_PARAM;
      goto error_exit;
   }

   cp = br_readdbfile(fp, comm);
   if(cp)
   {
      sprintf(brmsg, "DB file error (%s) in community %s\n", 
         cp, br_prtcomm(comm));

      comm_update.parm = comm;
      alarm_general(&comm_file, &comm->cm_alarm, brmsg);
      fclose(fp);
      goto error_exit;
   }

   fclose(fp);
   comm->cm_update = br_gettime();
   dprintf("Done.\n");

error_exit:
   in_dbread--;
   return err;
}


/* br_dbwrite() - Write the in-ram database for a passed community
 * (including any recent updates from community server) into a local 
 * file. Filename is set by the passed community. 
 * 
 * Returns 0 if OK, else ENP_ error code.
 */

int
br_dbwrite(struct comm_dbud * comm)
{
   int      line;    /* input line count */
   FILE *   fout;    /* out - new db file */
   MD5_CTX  xdigest;
   int      tries = 0;

   if(comm->cm_records == 0)
   {
      dprintf("Not writing community file %s - zero records.\n",
         comm->cm_filename);
      return ENP_PARAM;
   }

   br_log(0, "(re)writing database file for %s", comm->cm_url);

tryAgain:
   fout = fopen(comm->cm_filename, "w+b");          /* new output file */
   if(!fout)
   {
      dprintf("Error opening file %s\n",  comm->cm_filename);
      return ENP_FILEIO;
   }
   line = 0;

   fprintf(fout, "revision: %lu\r\nrecords: %lu\r\n\r\n", 
           comm->cm_revision, comm->cm_records);

   /* pass file descriptor to write routines in space struct */
   comm->cm_fd = fout;
   comm->cm_currec = 0;    /* Clear count of actual records written */

   /* traverse the btrees, writing info into file */
   TH_LOCK_MUTEX(br_tree_mutex);
   bt_depthfirst((struct btnode *)dn_tree, br_blkdnsout, (u_long)comm, 0);
   TH_UNLOCK_MUTEX(br_tree_mutex);

   TH_LOCK_MUTEX(br_tree_mutex);
   bt_depthfirst((struct btnode *)ip_tree, br_blkipout, (u_long)comm, 0);
   TH_UNLOCK_MUTEX(br_tree_mutex);
   
   TH_LOCK_MUTEX(br_tree_mutex);
   bt_depthfirst((struct btnode *)path_tree, br_blkpathout, (u_long)comm, 0);
   TH_UNLOCK_MUTEX(br_tree_mutex);
   
   TH_LOCK_MUTEX(br_tree_mutex);
   bt_depthfirst((struct btnode *)tcpp_tree, br_blkportout, (u_long)comm, 0);
   TH_UNLOCK_MUTEX(br_tree_mutex);
   
   TH_LOCK_MUTEX(br_tree_mutex);
   bt_depthfirst((struct btnode *)udpp_tree, br_blkportout, (u_long)comm, 0);
   TH_UNLOCK_MUTEX(br_tree_mutex);

//   bt_depthfirst((struct btnode *)image_tree, br_imageout, (u_long)space, 0);

   /* Write words (hash table) to file */
   br_blkwordout(comm);

#ifdef VIRUS_CHECK
   {
        int err;
        err = br_filesigs(fout, comm);
        if( err < 0 )
            dprintf("Error (%d) writing virus sigs \n", err);
   }
#endif /* VIRUS_CHECK */

   if(comm->cm_records != comm->cm_currec)
   {
      dtrap();    /* programmer didn't maintain record count */
      if(tries)   /* Don't loop here */
         return ENP_LOGIC;

      tries++;
      comm->cm_records = comm->cm_currec; /* try to fix */
      fclose(fout);
      goto tryAgain;
   }

   /* Now get the MD5 of the community DB tmp file */ 
   br_md5comm(fout, &xdigest, filebuf, sizeof(filebuf));
   fseek(fout, 0, SEEK_END);       /* Set file back to end */
   br_printdigest(hexdigest, xdigest.digest, 16);

   fprintf(fout, "digest: %s\r\nEOF\r\n\r\n", hexdigest);

   if(fclose(fout) != 0)
   {
      dprintf("Error closeing DB file %s \n", comm->cm_filename);
      return ENP_FILEIO;
   }

   return 0;
}

/* br_dbcheck()
 *
 * check the passed open file for sanity as a DB file. 
 *
 * Returns NULL if OK, else brief description of problem
 */

char *
br_dbcheck(FILE * fp, struct comm_dbud * comm)
{
   int      header_recs;
   int      actual_recs;
   int      revision;
   int      interval;
   int      reclen;
   char *   cp;
   MD5_CTX  xdigest;

   fseek(fp, 0, SEEK_SET);    /* Set file back to start */

   br_readrec(filebuf, sizeof(filebuf), fp);
   cp = br_parsefirst(filebuf, &revision, &header_recs, &interval);
   if(cp)
      return cp;

   actual_recs = 1;
   while( (reclen = br_readrec(filebuf, sizeof(filebuf), fp)) > 0)
   {
      /* Skip last record */
       if(((cp = br_getfield(filebuf, "digest")) != NULL) || 
            (strncmp(filebuf, "EOF", 3) == 0))
      {
         break;
      }

      actual_recs++;

      /* Call parse routine */
      cp = br_parseitem(filebuf, comm, FALSE);
      if(cp)            /* Did we get a create error message? */
      {
         if(br_getfield(filebuf, "digest"))
         {
            /* If last record had a digest, check it */
            if(cp)
            {
               br_md5comm(fp, &xdigest, filebuf, sizeof(filebuf));
               br_printdigest(hexdigest, xdigest.digest, 16);
               if(strncmp(cp, &hexdigest[0], 32))
                  return ("Bad digest");
            }
            break;
         }
         if(strstr(filebuf, "EOF"))
            break;

         return cp;
      }
   }

   /* Check for local buffer overflow */
   if(reclen < 0)
   {
      dtrap();
      return("Record too long");
   }

   if(actual_recs != (header_recs + 1))
      return ("incorrect record count");

   return NULL;   /* Can;t find anything wrong */
}



/******* Community basic object management *******/

#ifdef BR_COMMUNITIES

/* br_cmlookup(char * url)
 *
 * Lookup community by URL
 *
 * Returns community or NULL if not found.
 */

struct comm_dbud *
br_cmlookup(char * url)
{
   struct comm_dbud * tmp;

   for(tmp = br_communities; tmp; tmp = tmp->next)
   {
      if(stricmp(tmp->cm_url, url) == 0)
         return tmp;
   }
   return NULL;
}


/* br_cmbyfile(char * filename)
 *
 * Lookup community by passed file name. This should also work if 
 * the passed name from br_prtcomm().
 *
 * Returns community or NULL if not found.
 */

struct comm_dbud *
br_cmbyfile(char * filename)
{
   struct comm_dbud * tmp;
   int   len;
   char * cp;

   for(tmp = br_communities; tmp; tmp = tmp->next)
   {
      len = strlen(tmp->cm_filename);
      cp = strrchr(tmp->cm_filename, '.');
      if(cp)
         len = cp - tmp->cm_filename;
      if(strncmp(tmp->cm_filename, filename, len) == 0)
         return tmp;
   }
   return NULL;
}

/* br_cmbypriority()
 *
 * Look up a community by priority. Opcode is used as follows:
 *
 *   opcode == 0 - return community exactly matching priority
 *   opcode  > 0 - return community with next higher priority
 *   opcode  < 0 - return community with next lower priority
 *
 * Returns best matching community, NULL if none.
 */

struct comm_dbud *
br_cmbypriority(int priority, int opcode)
{
   struct comm_dbud * tmp;
   struct comm_dbud * best;

   if(br_communities == NULL)
      return NULL;

   if(opcode == 0)
      best = NULL;
   else
      best = br_communities;

   for(tmp = br_communities; tmp; tmp = tmp->next)
   {
      if(opcode > 0)
      {
         if((tmp->cm_priority > priority) &&
            (tmp->cm_priority < best->cm_priority))
         {
            best = tmp;
         }
      }
      else /* opcode < 0 */
      {
         if((tmp->cm_priority < priority) &&
            (tmp->cm_priority > best->cm_priority))
         {
            best = tmp;
         }
      }
   }
   return best;
}


/* br_cmnew(char * url, char * name)
 *
 * Constructor for community object. Caller shoud fill in 
 * filename and priority ASAP after theis returns. Defaults
 * are set, but they may not be viable.
 *
 *
 * Returns pointer to new communoity, or NULL if error/
 */

struct comm_dbud *
br_cmnew(char * url, char * name, int priority)
{
   char *   end;
   int      namelen;
   struct comm_dbud * newcm;
   struct comm_dbud * tmpcm;

   /* Check for serious screw-ups in passed parms */
   if(url == NULL)
   {
      dtrap();       /* Programming error */
      return NULL;
   }

   if(br_cmlookup(url))
   {
      dtrap();       /* Programming error? */
      dprintf("br_newcm: url already in use\n");
      return NULL;
   }

   if(strlen(url) >= MAXURL)
   {
      dprintf("br_newcm: url %s too long (over %d)\n", url, MAXURL);
      return NULL;
   }

   if(name)
   {
      namelen = strlen(name);
   }
   else
   {
      name = strstr(url, "fetch");  /* find level before name */
      if(!name)
      {
         dprintf("br_newcm: missing 'fetch in url %s \n", url);
         return NULL;
      }
      name += 6;  /* point past fetch/ */
      end = strchr(name, '/');
      if(end)
         namelen = end - name;
      else
         namelen = strcrlen(name);
   }

   if((namelen) < 3)
   {
      dprintf("br_newcm: name too short (< 4)\n");
      return NULL;
   }
   if((namelen + 6) > MAXFILENAME)
   {
      dprintf("br_newcm: name too long (> %d)\n", MAXFILENAME - 6);
      return NULL;
   }

   newcm = (struct comm_dbud *)CD_ALLOC(sizeof(struct comm_dbud));
   if(!newcm)
      return NULL;

   strcpy(newcm->cm_url, url);
   /* If trailing slash is missing, fix it: */
   if(*(url + strlen(url) - 1) != '/')
      strcat(newcm->cm_url, "/");
   strcpy(newcm->cm_name, name);
   newcm->cm_priority = priority;

   /* Insert new community in list, in priority order */
   if(br_communities == NULL)
   {
      /* new community is only one */
      br_communities = newcm;
   }
   else if(newcm->cm_priority > br_communities->cm_priority)
   {
      /* new community goes at top of list */
      newcm->next = br_communities;
      br_communities = newcm;
   }
   else  /* scan list to find insertion point */
   {
      for(tmpcm = br_communities; tmpcm; tmpcm = tmpcm->next)
      {
         if((tmpcm->next == NULL) || 
            (tmpcm->next->cm_priority < newcm->cm_priority))
         {
            newcm->next = tmpcm->next;
            tmpcm->next = newcm;
            break;
         }
      }
      if(tmpcm == NULL)
      {
         dtrap();
      }
   }

   /* Make default filenames, based on community name */
   strncpy(newcm->cm_filename, name, namelen);
   strcat(newcm->cm_filename, ".cdb");
   strncpy(newcm->cm_tmpname, name, namelen - 1);
   strcat(newcm->cm_tmpname, "0.tmp");

   return newcm;
}


/* Tree helper routines for cleaning out communitys */

/* counters - see comment in br_cmdel() */
static   int   ip_del;
static   int   dn_del;
static   int   path_del;
static   int   port_del;

/* Per node tree callback to delete controled IPs for community */
void
bt_ipe_delcomm(struct btnode * n, long param, int depth)
{
   struct control_ip * cip;
   struct comm_dbud * comm;

   cip = (struct control_ip *)n->data;
   comm = (struct comm_dbud *)param;

   /* If not the target comnmunity just return */
   if(cip->ctl.comm  != comm)
      return;

   del_ipe(cip, FALSE);
   ip_del++;
   USE_ARG(depth);

   return;
}

/* Per node tree callback to delete controled DNSs for community */
void
bt_dn_delcomm(struct btnode * n, long param, int depth)
{
   struct control_dn * dnp;
   struct comm_dbud * comm;

   dnp = (struct control_dn *)n->data;
   comm = (struct comm_dbud *)param;

   /* If not the target comnmunity just return */
   if(dnp->ctl.comm  != comm)
      return;

   del_dname(dnp);
   dn_del++;
   USE_ARG(depth);

   return;
}

/* Per node tree callback to delete controled ports for community */
void
bt_pt_delcomm(struct btnode * n, long param, int depth)
{
   int   type;
   struct control_port * ptp;
   struct comm_dbud * comm;

   ptp = (struct control_port *)n->data;
   comm = (struct comm_dbud *)param;
   if(ptp->btnp != n)    /* Check for tree corruption */
   {
      dtrap();
   }


   /* return if not in passed community */
   if(ptp->ctl.comm != comm)
      return;

   /* extract type (tcp/udp) bits) */
   type = ptp->cp_flags & (CPF_TCP|CPF_UDP);

   /* Make sure only one bit is set */
   if(type == (CPF_TCP|CPF_UDP))
   {
      dtrap();
      USE_ARG(depth);
      return;
   }
   port_del++;
   del_cp(ptp, type);
   return;
}

/* Per node tree callback to delete controled paths for community */
void
bt_path_delcomm(struct btnode * n, long param, int depth)
{
   struct b_path * pathp;
   struct comm_dbud * comm;

   pathp = (struct b_path *)n->data;
   comm = (struct comm_dbud *)param;
   if(pathp->btnp != n)    /* Check for tree corruption */
   {
      dtrap();
   }

   /* return if not in passed community */
   if(pathp->ctl.comm != comm)
      return;

   del_path(pathp);
   path_del++;

   USE_ARG(depth);
   return;
}

int
br_cmdel(struct comm_dbud * comm)
{
#define  MAXLOOPS 100
   int   loops;      /* delete loop counter for tree iterations */
   struct comm_dbud * tmp;
   struct comm_dbud * last;

   /* Find community in list and unlink */
   last = NULL;
   for(tmp = br_communities; tmp; tmp = tmp->next)
   {
      if(tmp == comm)   /* found community to delete */
      {
         TH_LOCK_MUTEX(br_comm_mutex);

         /* unlink */
         if(last)
            last->next = comm->next;
         else
            br_communities = comm->next;

         comm_cleanup(comm);

         /* Need to delete ALL communities controls... */

         /* The btree bt_depthfirst() call does not do a complete job of 
          * deleting all the community items in a tree. This is because 
          * The delete operation may change the left/right pointers, thus
          * causing the depthforst logic to miss items. To get around this 
          * we call it itterativly until no more community iterms are found
          * by the delete routine. 
          */
         /* IP addresses */
         ip_del = 1;
         loops = 0;
         while((ip_del > 0) && (loops++ < MAXLOOPS))
         {
            ip_del = 0;
            TH_LOCK_MUTEX(br_tree_mutex);
            bt_depthfirst((struct btnode *)ip_tree, bt_ipe_delcomm, (long)comm, 0);
            TH_UNLOCK_MUTEX(br_tree_mutex);
         }
         if(loops >= MAXLOOPS)
            panic("cmdel-ip");

         /* Domain names */
         dn_del = 1;
         loops = 0;
         while((dn_del > 0) && (loops++ < MAXLOOPS))
         {
            dn_del = 0;
            TH_LOCK_MUTEX(br_tree_mutex);
            bt_depthfirst((struct btnode *)dn_tree, bt_dn_delcomm, (long)comm, 0);
            TH_UNLOCK_MUTEX(br_tree_mutex);
         }
         if(loops >= MAXLOOPS)
            panic("cmdel-dn");

         /* File types */
         path_del = 1;
         loops = 0;
         while((path_del > 0) && (loops++ < MAXLOOPS))
         {
            path_del = 0;
            TH_LOCK_MUTEX(br_tree_mutex);
            bt_depthfirst((struct btnode *)path_tree, bt_path_delcomm, 
                    (long)comm, 0);
            TH_UNLOCK_MUTEX(br_tree_mutex);
         }
         if(loops >= MAXLOOPS)
            panic("cmdel-path");

         /* Port blocks */
         port_del = 1;
         loops = 0;
         while((dn_del > 0) && (loops++ < MAXLOOPS))
         {
            port_del = 0;
            TH_LOCK_MUTEX(br_tree_mutex);
            bt_depthfirst((struct btnode *)tcpp_tree, bt_pt_delcomm, (long)comm, 0);
            bt_depthfirst((struct btnode *)udpp_tree, bt_pt_delcomm, (long)comm, 0);
            TH_UNLOCK_MUTEX(br_tree_mutex);
         }
         if(loops >= MAXLOOPS)
            panic("cmdel-port");

         /* Virus signatures */

         /* Get rid of the community file */
         unlink(comm->cm_filename);

         /* Write parameters file */
         br_saveparms();

         CD_FREE(comm);
         TH_UNLOCK_MUTEX(br_comm_mutex);

         return 0;
      }
      last = tmp;
   }
   dtrap();    /* not found */
   return ENP_PARAM;
}


/* br_cmupdate() - Initiate a community update.
 *
 * Returns NULL if OK, else description of problem.
 *
 */

char *
br_cmupdate(struct comm_dbud * comm)
{
   if(!comm)
      return ("no community");

   if(comm->cm_url[0] == 0)
      return ("no URL");

   /* If the previous transfer failed, move to the idle state */
   if(comm->cm_state == FAILED)
      comm->cm_state = IDLE;

   /* If not in idle (READING, etc) wait for it to finish */
   if(comm->cm_state != IDLE)
      return("update already in progress");

   comm_cleanup(comm);     /* Clear the decks... */

   /* Set timer to start transfer right now. */
   comm->cm_interval = 1;

   return NULL;
}


#else    /* not BR_COMMUNITIES, no community support */

/* stub routine for community-less system */
struct comm_dbud *
br_cmbyfile(char * filename)
{
   USE_ARG(filename);
   return NULL;
}

struct comm_dbud *
br_cmlookup(char * url)
{
   USE_ARG(url);
   return NULL;
}



struct comm_dbud *
br_cmnew(char * url, char * name, int priority)
{
   USE_ARG(url);
   USE_ARG(name);
   USE_ARG(priority);
   return NULL;
}

int
br_cmdel(struct comm_dbud * comm)
{
   USE_ARG(comm);
   return 0;
}

char *
br_cmupdate(struct comm_dbud * comm)
{
   USE_ARG(comm);
   return("Non-comminty version");
}

#endif   /* BR_COMMUNITIES */

/*
Copyright (C) 2003-2008 by Praemio Inc. All rights reserved.

Author: John Bartas - jbartas@praemio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/



