/* brdbud.c
 *
 * Copyright 2003-2008 by Praemio Inc. All rights reserved.
 * Author: John Bartas - jbartas@praemio.com
 * This software is released under the terms of the LGPL.
 * See LGPL Notice at end of this file; Full license in ../license.txt
 *
 * Buster database update file
 *
 * ROUTINES: br_initcomm(), comm_cleanup(), br_prtcomm(), 
 *  br_update(), br_requpgrade(), 
 *
 *
 */


#include "busterport.h"
#include "brnet.h"
#include "ip.h"
#include "brdb.h"
#include "brmd5.h"

#ifdef BR_COMMUNITIES   /* whole community support file can be removed */

#define  DEFAULT_UPDATE_INTERVAL (4 * 3600)    /* In seconds */

/* Task structure for the Community database update task */

TH_OBJECT(to_udpcb);
TH_PROTO(br_update);

struct br_thread comm_thread = 
{
   &to_udpcb,
   "Community update",
   br_update,
   APP_PRIORITY,
   APP_STACK_SIZE,
};


u_long  br_dbdate;     /* date of last community update */

int
br_initcomm()
{
   int err;
   struct comm_dbud * comm;

   /* read in database files for the various Communities */
   for(comm = br_communities; comm; comm = comm->next)
   {
      err = br_dbread(comm);
      if(err)
         dprintf("Problem reading database file %s\n", comm->cm_filename);

      comm->cm_interval = 60;   /* Set to update in X seconds */
   }

   /* Start the community update thread */
   err = TH_NEWTASK(&comm_thread);
   if(err)
   {
      dprintf("Can't start Communinty DB thread");
      return ENP_RESOURCE;
   }
   return 0;
}

void
comm_cleanup(struct comm_dbud * comm, int code)
{
   if(comm->cm_socket != INVALID_SOCKET)
   {
      socketclose(comm->cm_socket);
      comm->cm_socket = INVALID_SOCKET;
   }

   /* If Ok, then clean up tmp file, if any */
   if(comm->cm_tmp )
   {
      fclose(comm->cm_tmp);
      comm->cm_tmp = NULL;
   }
   if(code == 0)
      unlink(comm->cm_tmpname);

   /* Close data file, if open */
   if( comm->cm_fd )
   {
      fclose(comm->cm_fd);
      comm->cm_fd = NULL;
   }

   return;
}


/* br_prtcomm(comm)
 *
 * Return a printable community name.
 *
 */

char static commname[MAXFILENAME];

char *
br_prtcomm(struct comm_dbud * comm)
{
   int    length;
   char * lastdot = strrchr(comm->cm_filename, '.');

   if(lastdot == NULL)
      return "noname";

   length = lastdot - comm->cm_filename;
   strncpy(commname, comm->cm_filename, length);
   commname[length] = 0;

   return commname;
}


static void
cm_failed(struct comm_dbud * comm)
{
   comm_cleanup(comm, 1);  /* error cleanup leaves tmp file */
   comm->cm_interval = DEFAULT_UPDATE_INTERVAL;
   comm->cm_state = FAILED;
   return;
}

/* The update thread. Thus needs to "idle" at about once a second to see 
 * if a Community should be asking for an update. Once an update starts
 * this should run at a low priority until the update is finished. This 
 * could be a number of seconds, even minutes if a large DB is loading 
 * from scratch. It does a round-robin through all active communities,
 * using non-blocking socket reads, as opposed to using the more 
 * lugubrious select().
 */

TH_ENTRY(br_update)
{
   char *   urlhost;
   char *   urlport;
   char *   cp;
   u_char * ucp;
   int      length;
   int      err;
   int      inlen;            /* input read */
   int      active;           /* number of active communities */
   int      loadtime;         /* time download took (in ticks) */
   struct comm_dbud * comm;
   struct sockaddr_in sa;
   char     macaddress[64];
   char     version[32];

   /* wait until nets are up... */
   while(nets[0]->ifOperStatus != 1)
      TH_SLEEP(2);

   for(;;)
   {
      active = 0;
      for(comm = br_communities; comm; comm = comm->next)
      {
         if((comm->cm_state != IDLE) &&
            (comm->cm_state != FAILED))
         {
            active++;      /* Any community not in IDLE or FAILED is active... */
         }

         if(comm->cm_url[0] == 0)
            continue;

         switch(comm->cm_state)
         {
         case FAILED:
         case IDLE:
            /* See if it's time for a database update request */
            if(( br_gettime() - comm->cm_update) < 
               (comm->cm_interval * SECOND))
            {
               continue;
            }

            /* Fall to here if it's time to check server */
            active++;

            dprintf("Starting update for community %s\n", br_prtcomm(comm));

            comm->cm_deadman = cticks;    /* Init deadman timer */

            /* As each update try starts (that's here) reset the time */
            comm->cm_update = br_gettime();

            /* default interval to next update time */
            comm->cm_interval = DEFAULT_UPDATE_INTERVAL;

            if((strncmp(comm->cm_url, "http://", 6)) &&
               (strncmp(comm->cm_url, "https://", 7)))
            {
               sprintf(brmsg, "Community URL does not start with \"http\":\n %s\n",
                  comm->cm_url);
               comm_update.parm = comm;
               alarm_general(&comm_update, &comm->cm_alarm, brmsg);
               cm_failed(comm);
               break;
            }

            urlhost = strstr(comm->cm_url, "//");
            if(!urlhost)
            {
               sprintf(brmsg, "Malformed community URL: %s - no host\n",
                  comm->cm_url);
               comm_update.parm = comm;
               alarm_general(&comm_update, &comm->cm_alarm, brmsg);
               cm_failed(comm);
               break;
            }
            urlhost += 2;     /* Bump past '//' chars */
            urlport = strchr(urlhost, ':');
            if(urlport)
               comm->cm_port = atoi(urlport + 1);
            else
               comm->cm_port = 80;

            /* Extract pathname portion of community URL */
            comm->cm_urlpath = strstr(urlhost, "/");
            if(!comm->cm_urlpath)
            {
               sprintf(brmsg, "Malformed community URL: %s - no path\n",
                  comm->cm_url);
               comm_update.parm = comm;
               alarm_general(&comm_update, &comm->cm_alarm, brmsg);
               cm_failed(comm);
               break;
            }

            /* Null terminate Host name, save in seperate buf */
            cp = strchr(urlhost, '/');
            if(urlport)    /* This may still point to colon */
               *urlport = 0;
            if(cp)
               *cp = 0;
            strcpy(comm->cm_host, urlhost);
            if(cp)
               *cp = (char)('/');
            if(urlport)
               *urlport = ':';

            /* DNS lookup the host */
            comm->cm_ipaddr = br_dnslookup(comm->cm_host);
            if(comm->cm_ipaddr == 0L)
            {
               sprintf(brmsg, "Unable to resolve community URL server: %s\n",
                  comm->cm_host);
               comm_update.parm = comm;
               alarm_general(&comm_update, &comm->cm_alarm, brmsg);
               cm_failed(comm);
               break;
            }

            /* Open socket to Community server */
            comm->cm_socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
            if(comm->cm_socket == INVALID_SOCKET)
            {
               err = sock_errno(comm->cm_socket);
               sprintf(brmsg, "Unable to get 'socket' for DB update, error code %d\n",
                  err);
               comm_update.parm = comm;
               alarm_general(&comm_update, &comm->cm_alarm, brmsg);
               cm_failed(comm);
               break;
            }

            sa.sin_family = AF_INET;
            sa.sin_addr.s_addr = comm->cm_ipaddr;
            sa.sin_port = htons(comm->cm_port);
            err = connect(comm->cm_socket, (struct sockaddr*)&sa, sizeof(sa));
            if(err)
            {
               err = sock_errno(comm->cm_socket);
               sprintf(brmsg, "Unable to contact community %s for update, error code %d\n",
                  comm->cm_url, err);
               comm_update.parm = comm;
               alarm_general(&comm_update, &comm->cm_alarm, brmsg);
               cm_failed(comm);
               break;
            }
            if(comm->cm_flags & CM_NOMAC)
            {
               macaddress[0] = 0;   /* try with no MAC address in URL */
            }
            else
            {
               ucp = (u_char*)nets[0]->n_macaddr;
               sprintf(macaddress, "?mac=%02X%02X%02X%02X%02X%02X", 
               *ucp,
               *(ucp+1),
               *(ucp+2),
               *(ucp+3),
               *(ucp+4),
               *(ucp+5));
            }

            sprintf(version, "%u.%u.%u.%u", PUSH_VERSION(br_versnum));

            /* Send database request */
            sprintf(comm->cm_buf, 
            "GET %s/%lu%s HTTP/1.0\r\nHost: %s\r\n",
               comm->cm_urlpath,
               comm->cm_revision,
               macaddress,
               comm->cm_host);

            strcat(comm->cm_buf, "User-Agent: Praemio Buster version ");
            strcat(comm->cm_buf, version);
            strcat(comm->cm_buf, 
                   "\r\nAccept: text/html\r\nAccept-Charset: ISO-8859-1\r\n\r\n");

            length = strlen(comm->cm_buf);
            err = send(comm->cm_socket, comm->cm_buf, length, 0);
            if(err != length)
            {
               sprintf(brmsg, "Unable to send request to community server %s.\n",
                  comm->cm_url);
               comm_update.parm = comm;
               alarm_general(&comm_update, &comm->cm_alarm, brmsg);
               cm_failed(comm);
               break;
            }

            /* Set mode to non-blocking during reads */
            err = br_nbiosocket(comm->cm_socket);

            /* Open the file for write */
            comm->cm_tmp = fopen(comm->cm_tmpname, "w+b");
            if(comm->cm_tmp == NULL)
            {
               sprintf(brmsg, "Unable to open temporary file %s.\n",
                  comm->cm_tmpname);
               comm_file.parm = comm;
               alarm_general(&comm_file, &comm->cm_alarm, brmsg);
               cm_failed(comm);
               break;
            }
            comm->cm_data = 0;
            comm->cm_state = READ_HEADER;
            comm->cm_deadman = cticks;    /* update deadman timer */

         case READ_HEADER:
            inlen = recv(comm->cm_socket, comm->cm_buf + comm->cm_data, 
               sizeof(comm->cm_buf) - comm->cm_data, 0);
            if(inlen <= 0)
            {
               err = sock_errno(comm->cm_socket);
               if(err ==  EWOULDBLOCK)
                  break;

               sprintf(brmsg, "Network error %d waiting for reply from %s.\n",
                     err, br_prtcomm(comm) );
               comm_update.parm = comm;
               alarm_general(&comm_update, &comm->cm_alarm, brmsg);
               cm_failed(comm);
               break;
            }
            else     /* Got reply, better have the whole HTTP reply */
            {
               cp = strstr(comm->cm_buf, "\r\n\r\n");  /* find end of rec */
               if(!cp || ((cp - comm->cm_buf) > inlen))
               {
                  comm->cm_data = inlen;  /* record size of current data */
                  break;   /* try for the rest soon */
               }

               cp = strchr(comm->cm_buf, ' ');

               if((strncmp(comm->cm_buf, "HTTP", 4)) ||
                  (cp == NULL))
               {
                  sprintf(brmsg, "Malformed HTTP header reply from %s.\n",
                     br_prtcomm(comm) );
                  comm_update.parm = comm;
                  alarm_general(&comm_update, &comm->cm_alarm, brmsg);
                  cm_failed(comm);
                  break;
               }

               /* Extract reply code */
               err = atoi(cp);
               if((err < 100) || (err > 299))
               {
                  /* If we were trying for a URL with a MAC address, retry
                   * without a MAC address. 
                   */
                  if((comm->cm_flags & CM_NOMAC) == 0)
                  {
                     comm_cleanup(comm, 0);
                     comm->cm_flags |= CM_NOMAC;
                     comm->cm_state = IDLE;
                     comm->cm_interval = 0;  /* retry ASAP */
                     break;
                  }
                  sprintf(brmsg, "Community %s returned an HTTP 'error %d' "
                     "in response to our update request. Please contact the "
                     "server administrator and/or try again later.\n",
                     br_prtcomm(comm), err );

                  comm_update.parm = comm;
                  alarm_general(&comm_update, &comm->cm_alarm, brmsg);
                  cm_failed(comm);
                  comm->cm_flags &= ~CM_NOMAC;
                  break;
               }

               /* Get here if reply header was OK. Write any data to 
                * file and go into SAVE_UPDATE state.
                */
               cp = strstr(cp, "\r\n\r\n");  /* find end of rec */
               if(cp)
               {
                  cp += 4;    /* past double crlf */
                  inlen -= (cp - comm->cm_buf);
                  if(inlen > 0)
                  {
                     err = fwrite(cp, 1, inlen, comm->cm_tmp);
                     if(err != inlen)
                     {
                        goto write_error;
                     }
                  }
               }
               else  /* HTML header incomplete? */
               {
                  dtrap();
                  cm_failed(comm);
                  break;
               }
               dprintf("Receiving update for %s\n", br_prtcomm(comm));
               comm->cm_deadman = cticks;    /* update deadman timer */
               comm->cm_state = SAVE_UPDATE;
            }
         case SAVE_UPDATE:
            inlen = recv(comm->cm_socket, comm->cm_buf, sizeof(comm->cm_buf), 0);
            if(inlen > 0)     /* Got some data */
            {
               comm->cm_deadman = cticks;    /* update deadman timer */
               err = fwrite(comm->cm_buf, 1, inlen, comm->cm_tmp);
               if(err != inlen)
               {
write_error:
                  sprintf(brmsg, "Disk error in temporary  file %s.\n",
                     comm->cm_tmpname);
                  comm_file.parm = comm;
                  alarm_general(&comm_file, &comm->cm_alarm, brmsg);
                  cm_failed(comm);
                  break;
               }
            }
            else if(inlen == 0)  /* Done reading data */
            {
               socketclose(comm->cm_socket);     /* Close socket */
               comm->cm_socket = INVALID_SOCKET;

               /* Make sure tmp file is a properly formatted database file */
               cp = br_dbcheck(comm->cm_tmp, comm);
               if(cp)
               {
                  sprintf(brmsg, "Community %s download format problem %s.\n",
                        br_prtcomm(comm), cp );
                  comm_file.parm = comm;
                  alarm_general(&comm_file, &comm->cm_alarm, brmsg);
                  cm_failed(comm);
                  break;
               }

               dprintf("Received update for %s, parsing...\n", br_prtcomm(comm));
               comm->cm_state = PARSING;
               comm->cm_deadman = cticks;    /* update deadman timer */
            }
            else if(inlen < 0)    /* Error on recv() */
            {
               /* Fall to here when socket read fails */
               err = sock_errno(comm->cm_socket);
               if(err ==  EWOULDBLOCK)
               {
                  /* Kill this after 5 minutes or so... */
                  if((comm->cm_deadman + (5*MINUTE)) < cticks)
                  {
                     sprintf(brmsg, "Network timeout updating %s\n", 
                        br_prtcomm(comm));
                     comm_update.parm = comm;
                     alarm_general(&comm_update, &comm->cm_alarm, brmsg);
                     cm_failed(comm);
                  }

                  break;
               }

               sprintf(brmsg, "Socket error %d updating community %s.\n",
                     err, br_prtcomm(comm) );
               comm_update.parm = comm;
               alarm_general(&comm_update, &comm->cm_alarm, brmsg);
               cm_failed(comm);
               break;
            }
            break;

         case PARSING:
            /* First, check file semantics */
            cp = br_dbcheck(comm->cm_tmp, comm);
            if(cp)
            {
               sprintf(brmsg, "Update problem %s from community %s\n", 
                  cp, br_prtcomm(comm) );
               comm_update.parm = comm;
               alarm_general(&comm_update, &comm->cm_alarm, brmsg);
               cm_failed(comm);
               break;
            }

            /* This reads in the tmp file into the database. */
            cp = br_readdbfile(comm->cm_tmp, comm);
            if(cp)
            {

               /* If we already have the revision, just go to idle. */
               if(strncmp(cp, "Already", 7) == 0)
               {
                  /* If there's unwritten data fron a previous revision, write it now */
                  if(comm->cm_flags & CM_NEWDATA)
                     goto do_filing;

                  comm_cleanup(comm, 0);
                  comm->cm_state = IDLE;
               }
               else
               {
                  sprintf(brmsg, "Update problem %s from community %s\n", 
                     cp, br_prtcomm(comm) );
                  comm_update.parm = comm;
                  alarm_general(&comm_update, &comm->cm_alarm, brmsg);
                  cm_failed(comm);
               }
            }
            else if(comm->cm_currec == 0)    /* No new records? */
            {
               /* If there's unwritten data fron a previous revision, write it now */
               if(comm->cm_flags & CM_NEWDATA)
                  goto do_filing;

               comm_cleanup(comm, 0);
               comm->cm_state = IDLE;
               dprintf("No new records from community %s\n", 
                  comm->cm_name );
            }
            else  /* Got new information, advance to filing phase */
            {
               comm->cm_flags |= CM_NEWDATA;
do_filing:
               comm_cleanup(comm, 0);
               comm->cm_state = FILING;
               br_dbdate = br_gettime();     /* date of last community update */
               dprintf("Saving update from %s\n", br_prtcomm(comm));
            }
            break;

         case FILING:
            loadtime = (br_gettime() -  comm->cm_update);
            if((comm->cm_interval > 60) ||
               (loadtime > (60 * SECOND)) )
            {
               err = br_dbwrite(comm);
               if(err)
               {
                  sprintf(brmsg, "Problem (code %d) recording community update from %s.\n",
                     err, br_prtcomm(comm) );
                  comm_file.parm = comm;
                  alarm_general(&comm_file, &comm->cm_alarm, brmsg);
                  cm_failed(comm);
                  break;
               }
               comm->cm_update = br_gettime();  /* drag along the update time */
            }
            comm->cm_state = IDLE;
            break;

         default:
            dtrap();
            comm->cm_state = IDLE;
         }
      }

      /* If no communities were active, go back to sleep */
      if(active == 0)
      {
          TH_SLEEP(SECOND);     /* long sleep, to save cycles */
      }
      else
      {
          TH_SLEEP(1);   /* One tick, to throttle downloads */
      }
   }
   TH_RETURN_OK();
}
TH_EXIT("br_update")



/* br_requpgrade() - utility routine to request the upgrade server
* begin the file transmission. This creates a sockets, connects.
* and send the HTTP GET requrest.
*
* This should NOT be called in the context of br_newimage(), since
* br_autoupdate() is called in the context of the network receive
* thread; which will prevent the TCP SYN/ACK required for a socket
* connect from being processed.
*/

SOCKTYPE
br_requpgrade(ip_addr ipaddr, unshort port, char * pathname)
{
  int         err;
  int         sendlen;
  SOCKTYPE    sock;
  struct sockaddr_in sa;

  /* We'll begin the image download
   * via an HTTP connection, then turn it over to a "per-port" routine
   * which will finish the download and install the new image.
   */

  sock = socket(AF_INET, SOCK_STREAM, 0);
  if(!sock)
  {
     err = sock_errno(sock);
     dprintf("Upgrade: error creating socket.\n", err);
     return INVALID_SOCKET;
  }
  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = ipaddr;
  sa.sin_port = port;

  err = connect(sock, (struct sockaddr *)&sa, sizeof(struct sockaddr_in));
  if(err)
  {
     err = sock_errno(sock);
     if(err == ETIMEDOUT)
     {
        dprintf("Timed out connecting to upgrade server %u.%u.%u.%u port %u.\n",
           PASS_IPADDR(ipaddr), htons(port) );
     }
     else if(err == ECONNREFUSED)
     {
        dprintf("Upgrade connection refused by server %u.%u.%u.%u port %u.\n",
           PASS_IPADDR(ipaddr), htons(port) );
     }
     else
        dprintf("upgrade connect error %d\n", err);
     return INVALID_SOCKET;
  }

  /* Send get command to server */
  sprintf(brmsg, "GET %s HTTP/1.1\r\nUser-Agent: Praemio Buster/%u.%u.%u.%u\r\n\r\n",
     pathname, PUSH_VERSION(br_versnum) );

  sendlen = strlen(brmsg);
  err = send(sock, brmsg, sendlen, 0);
  if(err != sendlen)
  {
      err = sock_errno(sock);
     dprintf("Upgrade connection send error %s\n", err);
     socketclose(sock);
     return INVALID_SOCKET;
  }

  return sock;      /* OK return */
}

/*
Copyright (C) 2003-2008 by Praemio Inc. All rights reserved.

Author: John Bartas - jbartas@praemio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/

#endif  /* BR_COMMUNITIES - whole community support file can be removed */


