/* brdns.c
 *
 * Buster received DNS packet handler. 
 *
 * Copyright 2003-2008 by Praemio Inc. All rights reserved.
 * Author: John Bartas - jbartas@praemio.com
 * This software is released under the terms of the LGPL.
 * See LGPL Notice at end of this file; Full license in ../license.txt
 */


#include "busterport.h"
#include "brnet.h"
#include "ip.h"
#include "brdb.h"
#include "brdns.h"


unshort ip_id = 1;

u_long   dnsreplies = 0;   /* DNS replys for Buster names */
u_long   dnsrequests = 0;  /* DNS requests for IP addrs for blocked names */
u_long   dns_badhdrs = 0;  /* DNS pkts received with bad header */

int   br_dnsready = FALSE;

void * br_dns_mutex;

ip_addr  dns_servers[DNSSERVERS];


#ifdef NPDEBUG
char dnsdbg_name[MAXNAMELEN] = {"buydomains.com"};
#endif

/* helper routine for DNS reply encoding . These taje a pointer to the 
 * buffer where the item is to be inserted, and place it there in a
 * byte-wise maner. 
 *
 * both return a pointer to the next byte.
 */

char *
putshort(char * cp, unshort val)
{
   *cp++ = (val >> 8);
   *cp++ = (val & 0x00FF);
   return cp;
}

char *
putlong(char * cp, u_long val)
{
   *cp++ = (val >> 24);
   *cp++ = (val & 0x00FF0000) >> 16;
   *cp++ = (val & 0x0000FF00) >> 8;
   *cp++ = (val & 0x00FF);
   return cp;
}

/* DNS helper routines */
extern char * getshort(char * cp, unshort * val);
extern char * getlong(char * cp, u_long * val);


/* br_getipaddr()
 * 
 * br_getipaddr() is a special case of getlong for IP addresses - 
 * it preserves the endianness of the IP address; i.e. returns it
 * in NETWORK endian.
 *
 * Returns pointer to next bytes in stream after IP address
 */

char *
br_getipaddr(char * cp, ip_addr * ipptr)
{
   ip_addr   addr;

   getlong(cp, &addr);
   *ipptr = (ip_addr)htonl(addr);
   return (cp + 4);
}


/* br_adddnssvr()
 * 
 * Add a DNS server to the master list of DNS servers 
 */

void
br_adddnssvr(ip_addr dns_ipaddr)
{
   int   i;

   /* scan server list for an empty slot */
   for(i = 0; i < DNSSERVERS; i++)
   {
      if(dns_servers[i] == dns_ipaddr)
         break;      /* We already know this one */

      if(dns_servers[i]  == 0)
      {
         dns_servers[i] = dns_ipaddr;
         break;
      }
   }
   return;
}


void
br_dnsinit()
{
   int   i;

   /* Add the DNS servers to the routing table so DNS quries get
    * set out the right interfaces
    */
   for(i = 0; i < DNSSERVERS; i++)
   {
      if(dns_servers[i] == 0)
         continue;
   }
   br_dnsready = TRUE;
   return;
}

/* dns_copyout() - helper function for DNS routines */

int
dns_copyout(char * dest, char * src)
{
    int     namelen;
    char *  cp;  
    char *  field;
    char *  next;

    cp = dest;
    next = src;

    while (next) 
    {
        field = next;
        next = strchr(field, '.');
        if (next)
        {
            namelen = next - field;
            next++;
        }
        else
            namelen = strlen(field);

        *cp++ = (char)(namelen & 0xFF);
        MEMCPY(cp, field, namelen);
        cp += namelen;
    }
                *cp++ = 0;  /* null terminate string */
                return(cp - dest);
}


/* dns_copyin() - helper function for DNS routines */

int
dns_copyin(char * dest, char * src, struct dns_hdr * dns)
{
    unshort   namelen;
    char *   cp;        /* scratch */
    char *   field;
    char *   next;
    int      donelen;
    unshort  offset;

    cp = dest;
    next = src;
    donelen = 0;

    /* Loop through domain name string */
    while (next)
    {
        field = next;
        namelen = *field++;
        if (namelen == 0)
            break;   /* done */
        if ((namelen & 0xC0) == 0xC0)
        {
            field--;
            offset = (unshort)(*(u_char*)field); /* get first byte of offset code */
            field++;
            offset &= 0x3f;               /* mask our high two bits */
            offset <<= 8;                 /* make it high byte of word */
            offset += (unshort)(*(u_char*)field); /* add low byte of word */
            field = offset + (char *)dns;   /* ptr into domain name */
            namelen = *field++;
        }
        if (namelen + donelen > MAXDNSNAME) /* check for overflow */
            return 0;   /* error */
        MEMCPY(cp, field, namelen);
        donelen += (namelen+1); /* allow for dot/len byte */
        cp += namelen;
        *cp++ = '.';
        next = field + namelen;
    }

    *(cp-1) = 0;   /* null terminate string (over last dot) */
    return donelen + 1;  /* include null in count */
}


/* FUNCTION: br_dnsparse()
 * 
 * Parse received DNS Packets. This logs replies
 * to controlled domain names. 
 *
 * This NOT reentrant, but since it's called from the receive ISR
 * that shouldn't matter.
 *
 * PARAM1: PACKET pkt
 * PARAM2: struct ip * pip
 * PARAM3: struct conn_entry * conn
 * 
 * RETURNS: 1 if we answered the packet, 0 if not.
 */

char *   authserver = "ns1.clear2net.com";

char dnsname[DNSNAMELEN];     /* buffer for DNS name text */

int
br_dnsparse(PACKET pkt, struct ip * pip, struct conn_entry * conn)
{
   char *   cp;
   struct udp *      pup;  /* UDP header */
   struct dns_hdr *  dns;     /* incoming dns header */
   int      namelen;

   pup = (struct udp *)(ip_data(pip));
   dns = (struct dns_hdr *)(pup + 1);

   /* If this packet is from a server and the Iniche stack 
    * has < MAX DNS servers, then save server IP for stack.
    */
   if(pup->ud_srcp == htons(53))
      br_adddnssvr(pip->ip_src);

   cp = (char *)(dns + 1); /* point to DNS records */

   /* Get DNS query name from packet */
   namelen = dns_copyin(dnsname, cp, dns);
   if(namelen == 0)
      return 0;

   /* Log all DNS replies */
   if(dns->flags & htons(DNSF_QR))  /* bit set means response */
   {
      /* extract answer, add to IP hosts list */
      br_dnsreply(dns, conn, (unshort)htons(pup->ud_len) );
      pk_free(pkt);
      return 1;      
   }

#ifdef SELF_DNS
   /* see if the DNS query is for this Buster */
   if( strstr(dnsname, bustername ) )
   {
      int      iface = 0;
      ip_addr  tmp;
      unshort  offset;
      char *   cp2;
      u_long   timeout;    /* time to live for spoofed DNS we send */

      /* send a DNS reply in the same packet */
      /* turn on response & recursion bits */
      dns->flags |= htons(0x8180);
      dns->ancount = htons(1);         /* send back one answer */
      dns->nscount = htons(1);         /* one auth record */

      cp = pkt->pk_prot + pkt->pk_plen;   /* end of old pkt */

      offset = sizeof(struct dns_hdr);    /* point to name in query */

      /* Build the answer record. Start by setting offset 
       * mask (0xC0) & value into 16 bits.
       */
      *cp++ = (char)(0xC0 | (u_char)(offset >> 8));
      *cp++ = (char)(offset & 0x00FF);

      cp = putshort(cp, DNS_TYPE_IPADDR);    /* DNS type */
      cp = putshort(cp, 1);    /* class - inet */

      /* Set the Time and data-length */
      if(brstates[iface].evflags & EV_IPSTATIC) /* static IP */
      {
         timeout = 0x00015180;     /* 0x00015180 seconds == 1 day */
//         timeout = 0x0000001e;       /* 30 sec, for testing */
//         timeout = 0x000000708;        /* half hour, in seconds */
      }
      else     /* We're doing DHCP, use lease value */
      {
         if(brstates[iface].br_lease == -1)
         {
            timeout = 0x00015180;     /* 0x00015180 seconds == 1 day */
         }
         else if((brstates[iface].br_lease == 0) ||
            (brstates[iface].br_leasestart == 0))
         {
            /* Incomplete lease info */
            timeout = 0x00000708;   /* half hour? */
         }
         else  /* valid, non-infinte lease value */
         {
            timeout = brstates[iface].br_lease -
               brstates[iface].br_leasestart;
         }
      }
      cp = putlong(cp, timeout);
      cp = putshort(cp, 4);               /* length of record data */

      /* paste in IP address */
      cp = putlong(cp, htonl(pkt->net->n_ipaddr));

      /* build the auth record */
      cp = putshort(cp, (unshort)(0xC000 | offset)); /* offset to name */
      cp = putshort(cp, DNS_TYPE_AUTHNS); /* DNS type */
      cp = putshort(cp, 1);               /* class - inet */
      cp = putlong(cp, 0x00000e10);       /* TTL - 0x0e10 is one hour */
      cp2 = cp;                  /* remember where to put auth length */
      cp += 2;                   /* skip auth-length field for now */
      namelen = dns_copyout(cp, authserver);  /* auth server name */
      cp += namelen;
      putshort(cp2, (unshort)namelen);

      pkt->fhost = pip->ip_src;        /* send it back to client */
      pkt->pk_plen = cp - (char*)dns;  /* length of packet */
      pkt->pk_prot = (char *)dns;

      /* build the lower layer headers and send the reply */
      udp_bldhead(pkt, pup->ud_srcp, htons(DNS_PORT));    /* UDP header */
      tmp = pip->ip_dest;           /* swap src and dest IP address */
      pip->ip_dest = pip->ip_src;
      pip->ip_src = tmp;
      mk_iphdr(pkt, ip_id++, UDP_PROT, 0);       /* IP header */
      send_via_arp(pkt, pip->ip_dest);
      return 1;
   }
#endif   /* SELF_DNS */

   /* packet not for us - return without any action */
   return 0;
}

/* dns_extractip(dnsname) - helper routine to extract the IP adderss 
 * from a DNS PTR name string.
 *
 * Returns IP address, or 0 if no IP address could be parsed.
 */

ip_addr
dns_extractip(char * name)
{
   char * cp;
   char * msg;
   ip_addr  ipaddr;

   /* null terminate name so we can get back the original IP address that
    * was in the request.
    */
   cp = strstr(name, ".in-addr.arpa");
   if(!cp)
   {
      dtrap();    /* malformed DNS reply? */
      return 0;   /* can't find Ip address text */
   }

   *cp = 0;      /* null terminate it IP portion of text */
   if((msg = br_checkip(&ipaddr, dnsname)) != NULL)
   {
      dtrap();
      return 0;
   }
   *cp = '.';        /* restore dot we clobbered with null */
   return (htonl(ipaddr));
}

/* br_insertcip() - Insert control_ip structure with blocked IP range into a 
 * blocked control_dn (dns) entry list.
 *
 * returns 0 if OK, else error code (-1)
 */

int
br_insertcip( struct control_dn * dnp, struct control_ip * new_cip)
{
   struct control_ip * this_cip;
   struct control_ip * next_cip;

   /* Set back pointer form IP to DNS entry */
   new_cip->ci_name = dnp;

   /* simplest case - new cip is the only entry */
   if(dnp->cu_cip == NULL)
   {
      dnp->cu_cip = new_cip;
      return 0;
   }

   /* insert into ordered DNS list. Start by checking if 
    * it goes before first entry. 
    */
   if(new_cip->ci_ipaddr < dnp->cu_cip->ci_ipaddr)
   {
      /* insert at head of list */
      new_cip->ci_dnlist = dnp->cu_cip;
      dnp->cu_cip = new_cip;
      return 0;
   }

   /* scan list for spot to insert new entry. */
   TH_LOCK_MUTEX(br_dns_mutex);
   this_cip = dnp->cu_cip;
   next_cip = this_cip->ci_dnlist;
   while(this_cip)
   {
      /* see if this is a duplicate entry */
      if(new_cip->ci_ipaddr == this_cip->ci_ipaddr)
         break;   /* not an error */

      next_cip = this_cip->ci_dnlist;
      if((next_cip == NULL) || (new_cip->ci_ipaddr < next_cip->ci_ipaddr))
      {
         this_cip->ci_dnlist = new_cip;
         new_cip->ci_dnlist = next_cip;
         break;      /* OK return */
      }
      this_cip = this_cip->ci_dnlist;
   }
   TH_UNLOCK_MUTEX(br_dns_mutex);

   if(this_cip)      /* Make sure we found the slot */
      return 0;

   dtrap();    /* didn't find insertion point (list damaged?) */
   return -1;
}


/* FUNCTION: brdn_ckdnp()
 * 
 * Helper routine for br_dnsreply() This is called for each answer
 * record in the datagram IF the dgram's question subject machine matches 
 * one of the entries in Buster's blocked DNS name list.
 * 
 * PARAM1: struct dns_hdr * dns - IN - matching blocked DNS entry
 * PARAM2: ipaddr local_ip - IN - answer record IP address, in local endian
 *
 * RETURNS: 0 if OK, else neg. error.
 */
            
int
brdn_ckdnp(struct control_dn * dnp, ip_addr local_ip, u_long ttl)
{
   struct control_ip * cip;
   u_long   answer_ip;

   cip = dnp->cu_cip;
   if(cip)      /* Name already has some IP info? */
   {
      /* make new cip, add to dns entry list */
      answer_ip = htonl(local_ip);
      cip = new_ipe(answer_ip, &dnp->ctl, CIF_DNSREP, dnp->ctl.comm);
      if(cip == NULL)
      {
         /* Non-creation could be because IP address would have been duplicated
          * by the DNS-induced entry. This is not an error.
          */
         return 0;
      }
      if(br_insertcip(dnp, cip))
      {
         dtrap();
         return -1;
      }
   }
   else     /* name entry does not yet have IP info */
   {
      /* see if the IP address is already blacklisted */
      cip = ipe_lookup(local_ip);
      if(cip)
      {
         /* It's OK if the IP entry was administrativly added without 
          * a domain name, but make sure the IP entry is NOT attached
          * to another domain name.
          */
         if((cip->ci_name != NULL) && (cip->ci_name != dnp))
         {
            /* If we got here, the DNS name in the received reply is in
             * our controlled name database, and does not yet have any
             * IP info associated with it; but the IP address in the received DNS
             * reply (which we would normally attach the the DNS name's entry)
             * is already attached to another controlled DNS name. Since the 
             * IP is already controled, we do nothing.....
             */
//            dtrap();    /* maybe we should move the IP to current DNS entry ?*/
            dprintf("brdn_ckdnp: cross-matched IP address %u.%u.%u.%u\n",
               PASS_IPADDR(local_ip) );
            dprintf("brdn_ckdnp: dnp name %s  cip name %s \n",
                dnp->cu_name, cip->ci_name->cu_name );

            return 0;   /* Is this an error? */
         }
         else
            dnp->cu_cip = cip;
      }
      else    /* First we've heard of this IP, make new control IP entry */
      {
         answer_ip = htonl(local_ip);
         dnp->cu_cip = new_ipe(answer_ip, &dnp->ctl, CIF_DNSREP, dnp->ctl.comm);
         dnp->cu_cip->ci_name = dnp;   /* set back pointer to name */
      }
   }

   /* see if address TTL just got longer */
   if((ttl > dnp->cu_time) && (dnp->cu_time != TTL_FOREVER))
      dnp->cu_time = ttl;

   /* Set a timeout to delete this later. */
   if(ttl == 0)
   {
      /* ttl of zero means this is a volatile DNS entry and should
       * not be cached at all. Since this would prevent us from 
       * blocking it, we assign is a very short timer. 
       */
      dnp->cu_cip->ci_timeout = (30 * TPS); /* 30 seconds */
   }
   else  /* ttl is non-zero */
   {
      /* It's OK if if we keep it longer than the DNS
       * client, but we must not delete it early; so set 
       * our timer to more than double the DNS ttl.
       */
      dnp->cu_cip->ci_timeout = MINUTE + (ttl * TPS * 2);
   }

   return 0;
}


/* getoffset()
 *
 * Get offset to DNS name text from a DNS reply.
 *
 */

char *  getoffset(char * cp, struct dns_hdr * dns, unshort * offsetp)
{
    unshort  offset;
    char * dnshdr = (char*)dns;

    if ((*cp & 0xC0) == 0xC0)
    {
        /* cp points to encoded offset, decode it */
        offset = (unshort)(*cp++);
        offset &= 0x3f;
        offset <<= 8;
        offset += (unshort)(*cp++);
    }
    else
    {
        /* cp points to actual name, make offset */
        offset = (unshort)(cp - dnshdr);
        while (*cp++ != 0);
    }
    *offsetp = offset;      /* return offset */
    return cp;
}

/* FUNCTION: br_dnsreply()
 * 
 * br_dnsreply() - Process a received DNS reply. These are received
 * promiscuously, so they usually won't match our requests.
 *
 * PARAM1: struct dns_hdr * dns
 * PARAM2: struct conn_entry * conn
 * PARAM3: unshort dnslen
 *
 * RETURNS: 0 if OK, else ENP_ error code 
 */

int 
br_dnsreply(struct dns_hdr * dns, struct conn_entry * conn, unshort dnslen)
{
   unshort  flags;      /* dns->flags in local endian */
   unshort  rcode;      /* response code */
   unshort  queries;    /* number of question records in reply */
   unshort  answers;    /* number of answer records in reply */
   unshort  records;    /* total records in reply */
   unshort  type;       /* various fields from the reply */
   unshort  netclass;   /* class of net (1 == inet) */
   u_long   ttl;        /* records time to live */
   unshort  rdlength;         /* length of record data */
   unshort  offset;           /* scratch offset to domain name text */
   unshort  nameoffset = 0;   /* offset to name in query */
   unshort  authoffset = 0;   /* offset to first auth server name */
   char *   cp;               /* scratch */
   char *   end;              /* end of of received datagram */
   ip_addr  tmpipaddr;        /* scratch */
   ip_addr  answer_ip = 0;    /* IP address from answer record */
   int      i;                /* record index */
   struct control_dn * dnp;
   struct ip_hostentry * ipe;


   flags = htons(dns->flags);      /* extract data fromheader */
   rcode = flags & DNSF_RCMASK;    /* Response code */
   queries = htons(dns->qdcount);  /* number of questions */
   answers = htons(dns->ancount);  /* number of answers */
   records = queries + answers + 
      htons(dns->nscount) + htons(dns->arcount);

   /* set marker to end of packet */
   end = (char*)dns;
   end += dnslen;

   if (rcode != DNSRC_OK)
      return(ENP_PARAM); /* bad response code */

   if (answers < 1)
      return(ENP_PARAM); /* no answer records */

   cp = (char*)(dns+1);    /* start after DNS entry */
   ipe = NULL;             /* use as flag for unset */

#if (DNSNAMELEN < 256)
   /* first bytes should hold total length of name */
   if((unshort)(*cp) > DNSNAMELEN)  /* check for overflow */
   {
      dns_badhdrs++;
      return ENP_PARAM;
   }
#endif

   /* Get DNS query name from 1st record */
   if(dns_copyin(dnsname, cp, dns) == 0)
      return ENP_PARAM;

#ifdef NPDEBUG
   /* the the debug name is set, and it matches the received name, 
    * then hook the debugger.
    */
   if((dnsdbg_name[0] != 0) &&
      (strstr(dnsname, dnsdbg_name)))
   {
      dtrap();
   }
#endif

   /* See if name is in our block list. Leave dnp either NULL or set to
    * the blocked domain name entry.
    */
   dnp = dn_lookup(dnsname);

   /* loop through remaining records - answers, auth, and RRs */
   for (i = 0; i < records; i++)
   {
      /* Get offset to record's name */
      cp = getoffset(cp, dns, &offset);

      /* get records type and class from packet bytes */
      cp = getshort(cp, &type);
      cp = getshort(cp, &netclass);
      if (netclass != 1)
      {
         dns_badhdrs++;
         return(ENP_PARAM); /* bad class type (not internet) */
      }

      if(i < queries)      /* just skip if echos question record */
         continue;

      /* Get the Time and data-length */
      cp = getlong(cp, &ttl);       /* 4 byte time to live field */
      cp = getshort(cp, &rdlength); /* length of record data */

      /* see if record runs off end of packet I've see actually happen
       * when ancount has high bits set. -JB-
       */
      if((cp + rdlength) > end)
      {
         dns_badhdrs++;
         return(ENP_PARAM); /* bad class type (not internet) */
      }

      switch(type)
      {
      case DNS_TYPE_IPADDR:
         if (rdlength != 4)
            return ENP_PARAM; /* IP address is 4 bytes */

         dnsreplies++;
         if(i < (queries + answers))
         {
            br_getipaddr(cp, &answer_ip);

            /* Add this host to our IP host database. If it already
             * exists then update the name.
             */
            ipe = br_iplookup(conn->ce_iface, answer_ip);
            if(!ipe)
            {
               ipe = iphost_new(conn->ce_iface, answer_ip);
               if(!ipe)
                  return ENP_NOMEM; /* no memory, logged in iphost_new() */
            }
            strncpy(ipe->dnsname, dnsname, MAXNAMELEN);
            pl_newitem(ipe, PL_NEWNAME);

            /* see if this name is on our controlled name list */
            if(dnp) 
               brdn_ckdnp(dnp, htonl(answer_ip), ttl);
         }
         else  /* auth record or additional record */
         {
            if(offset == nameoffset)   /* another addr for query? */
            {
               /* expand the IP address entry */
               dtrap();
            }
            else if(offset == authoffset) /* auth server IP address */
            {
               if(type == DNS_TYPE_IPADDR)
               {
                  // dtrap();    /* Make IP entry for auth server? */
               }
            }
         }
         break;

      case DNS_TYPE_ZONE:     /* Start of zone authority */

         /* If we got this authority RR reply with no answer, it means the 
          * name we asked for does not exist (i.e. is unknown) in DNS. 
          * In this case, find the ip_host entry (if it's ours) and set
          * the DNS retry time out to the TTL.
          */

         if(answers == 0)
         {
            u_long dns_nexttick;    /* ctick for TTL expiration */
            tmpipaddr = dns_extractip(dnsname);
            if(tmpipaddr == 0L)
               break;
            ipe = br_iplookup(conn->ce_iface, tmpipaddr);
            if(!ipe)
               break;
            dns_nexttick = cticks + (ttl * TPS);
            if(dns_nexttick > ipe->tm_nextdns)
               ipe->tm_nextdns = dns_nexttick;
         }
         break;

      case DNS_TYPE_PTR:      /* reverse lookup answer */

         tmpipaddr = dns_extractip(dnsname);
         if(tmpipaddr == 0L)
            break;
         ipe = br_iplookup(conn->ce_iface, tmpipaddr);
         if(!ipe)
         {
            /* This can happen if we were not the source of the reverse lookup */
            ipe = iphost_new(conn->ce_iface, tmpipaddr);
            if(!ipe)
            {
               dtrap();    /* out of memory */
               break;
            }
         }
         dns_copyin(ipe->dnsname, cp, dns);
         break;

      case DNS_TYPE_AUTHNS: /* authoritative name server */
         /* What we really want for the name server is it's IP address,
          * however this record only contains a name, ttl, etc. We save 
          * the offset to the name, hoping that one of the additional 
          * records will have the IP address matching this name.
          */
         if(authoffset == 0)  /* only save first one */
            authoffset = cp - (char*)dns;
         break;

      case DNS_TYPE_ALIAS:  /* alias */
         break;

      default:       /* unhandled record type, ignore it. */
         break;
      }

      cp += rdlength;   /* bump past trailing data to next record */
   }

   /* record information in the connection structure */
   if(conn)
   {
      if(answer_ip)
      {
         sprintf(conn->data, "Name \"%s\" is IP %u.%u.%u.%u", 
            dnsname, PASS_IPADDR(answer_ip) );
      }
      else
         sprintf(conn->data, "Name \"%s\" is unknown to DNS system", dnsname);
   }
   return 0;
}


/* send a DNS request for the name associated with the IP 
 * address passed. Replies are processed by br_dnsreply()
 *
 * Returns ENP_ error, or 0 if OK.
 */


int
dnc_namereq(struct ip_hostentry * iphost)
{
   int      i;       /* counter for nets */
   int      err;
   int      len;
   ip_addr  server;
   ip_addr  target;
   PACKET   pkt;
   char *   cp;      /* scratch pointer for building question field */
   char *   ipcp;    /* pointer to IP address text */
   char     name[32];      /* DNS name; dot notation plus "in-addr.arpa" */
   struct dns_hdr *  dns;  /* outgoing dns header */


   if(br_dnsready == FALSE)
      br_dnsinit();

   /* figure out which server to try */
   for (i = 0; i < DNSSERVERS; i++)
   {
      if (dns_servers[i] == 0L)
         break;
   }
   if(!i)    /* No DNS servers known. */
   {
      dtrap();
      return ENP_RESOURCE;
   }
   i = dnsrequests % i;
   server = dns_servers[i];

   target = htonl(iphost->ipaddr);  /* put address backwards for DNS encoding */

   /* Don't send DNS requests for silly addresses */
   if(((target & 0xFF000000) == 0x0A000000) ||     /* 10.0.0.0 net*/
      ((target & 0xFFFF0000) == 0xC0A80000) ||     /* 192.168.0.0 net*/
      ((target & 0xE0000000) == 0xE0000000))       /* Multicast */
   {
      return ENP_PARAM;
   }

   /* allocate packet for DNS request */
   if ((pkt = pk_alloc(512 + IP_OVERHEAD)) == NULL)
      return ENP_RESOURCE;

   /* The requests encode the IP address as text, and send a "PTR" 
    * type request. Opcode in flags field is zero, the standard query
    * opcode. The PRT type is indicated by a type of 0x000c in the 
    * question type field.
    */

   /* fill in DNS header. Most fields are cleared to 0s */
   dns = (struct dns_hdr *)pkt->pk_prot;
   MEMSET(dns, 0, sizeof(struct dns_hdr));
   dns->id = htons((unshort)dnsrequests);
   dns->flags = htons(DNSF_RD);  /* Recursion Desired */
   dns->qdcount = htons(1);          /* 1 question */

   /* format question field after header */
   cp = (char*)(dns + 1);  /* point at next byte past header */

   /* convert the IP address to text */
   ipcp = print_ipad(target);       /* convert backwards addres to text */
   strcpy(name, ipcp);              /* start building PTR target */
   strcat(name, ".in-addr.arpa");   /* add the DNS suffix */

   /* convert to DNS-ish format and copy to packet */
   len = dns_copyout(cp, name);
   cp += len;

   cp = putshort(cp, 0x000C);    /* DNS query type for PTR */
   cp = putshort(cp, 1);         /* class - inet */

   /* send the packet via standard UDP */
   pkt->pk_plen = cp - pkt->pk_prot;
   err = udp_socksend(server, DNS_PORT, pkt);

   dnsrequests++;

   return err;
}

/*
Copyright (C) 2003-2008 by Praemio Inc. All rights reserved.

Author: John Bartas - jbartas@praemio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


