/* brlans.c
 *
 * Busters lists of various LANs it visits.
 *
 * Copyright 2003-2008 by Praemio Inc. All rights reserved.
 * Author: John Bartas - jbartas@praemio.com
 * This software is released under the terms of the LGPL.
 * See LGPL Notice at end of this file; Full license in ../license.txt
 *
 * ROUTINES: 
 * new_lan(), del_lan(), lanline(), br_filelan(), 
 * br_writelan(), br_readhost(), br_readlan(), br_checklan(), 
 * 
 */

#include "busterport.h"
#include "brnet.h"
#include "arp.h"
#include "ip.h"
#include "dhcpdef.h"
#include "brdb.h"


struct laninfo * br_lans;  /* List of LANs we know about */

/* For early startup; True when network.db has been read into ram */
int   lan_file_read = 0;

char * br_lanfile = "network.db";    /* File with all LANs info */

int   br_dellanf(char * lanname);   /* deletea lan from network.db file */

extern   u_long   atox(char * hexnum);


/* Micorsoft C has flaws which force us to use static instead of on-stack
 * buffer addresss. 
 */
static char msgbuf[250];


/* br_lanlookup() - lookup a lan structure in the list based on name.
 * 
 * Returns pointer to lan, or NULL if not found.
 */

struct laninfo * 
br_lanlookup(char * name)
{
   struct laninfo * ln;

   for(ln = br_lans; ln; ln = ln->next)
   {
      if(stricmp(ln->name, name) == 0)
      {
         return ln;
      }
   }
   return NULL;   /* named lan not found */
}



/* new_lan(char * name, u_char * router_mac, ip_addr router_ip)
 *
 * Creates a new LAN info entry, filled in with the passed info.
 * 
 * Returns laninfo struct ptr if OK, else NULL.
 */

struct laninfo * 
new_lan(char * name, u_char * router_mac, ip_addr router_ip)
{
   struct laninfo * ln;

   /* make sure name is not a duplicate */
   if(br_lanlookup(name))
   {
      dprintf("Tried to created duplicate LAN name %s\n", name);
      return NULL;
   }

   ln = (struct laninfo *)LAN_ALLOC(sizeof(struct laninfo));
   if(!ln)
      return NULL;

   strncpy(ln->name, name, MAXNAMELEN);
   MEMCPY(ln->router_mac, router_mac, sizeof(ln->router_mac) );
   ln->router_ip = router_ip;

   /* insert new entry into list */
   ln->next = br_lans;
   br_lans = ln;

   return ln;
}


void 
del_lan(struct laninfo * dellan)
{
   struct laninfo * ln;
   struct laninfo * last;

   br_cleardbs();    /* Clear the in-ram database */

   /* rewrite network.db file without this lan */
   br_dellanf(dellan->name);

   /* remote LAN fromintornal linked list */
   last = NULL;
   for(ln = br_lans; ln; ln = ln->next)
   {
      if(ln == dellan)
         break;
      last = ln;
   }
   if(!ln)
      return;

   if(last)    /* unlink */
      last->next = ln->next;
   else
      br_lans = ln->next;

   LAN_FREE(ln);

   /* write the parameters to delete the reference to the LAN name */
   br_saveparms();

   return;
}


/* lanline() - 
 * Format info for a single LAN host into a text buffer. The
 * buffer is then set for writing to file or passing over the 
 * network database update system.
 * 
 * Returns 0 if OK else ENP_ error code.
 */

int
lanline(struct host_entry * host, char * readbuf, unsigned rblen)
{
   char *   cp;
   struct ip_hostentry * iphost;
   char *   name;
   char *   spacep;
   char *   image;
   u_long   ffmask;

   if(host->locname[0])
   {
      name = &host->locname[0];
      ffmask = host->h_flags & H_NAMEMASK;
      iphost = host->iphost;
   }
   else if(host->iphost)
   {
      iphost = host->iphost;
      if(host->iphost->dnsname[0])
      {
         name = &host->iphost->dnsname[0];
         ffmask = H_DNSNAME;
      }
      else
      {
         name = "unknown";
         ffmask = H_NONAME;
      }
   }
   else
   {
      dtrap();       /* can this happen? */
      iphost = NULL;
      name = "unknown";
      ffmask = H_NONAME;
   }

   ffmask |= (host->disctype & FF_TYPEMASK); /* save discover type info */

   /* Flag if this host is the router for this LAN */
   if(host == brstates[host->iface].br_router)
      ffmask |= FF_LANROUTER;

   /* borrow the read buffer for writing, put in name */
   strcpy(readbuf, name);

   /* If name has enbedded spaces, replace with plus signs for write to 
    * file. 
    */
   for(spacep = strchr(readbuf, ' '); spacep; spacep = strchr(spacep, ' '))
      *spacep++ = '+';

   /* Add the MAC address after the name */
   cp = readbuf + strlen(readbuf);
   *cp++ = ' ';
   while(cp < &readbuf[20])
      *cp++ = ' ';

   sprintf(cp, "%02x%02x%02x%02x%02x%02x  ",
      host->macaddr[0],
      host->macaddr[1],
      host->macaddr[2],
      host->macaddr[3],
      host->macaddr[4],
      host->macaddr[5] );

   /* Add IP address (0.0.0.0 if unknown) */
   cp = readbuf + strlen(readbuf);
   if(iphost)
      sprintf(cp, "%u.%u.%u.%u   ", PASS_IPADDR(iphost->ipaddr));
   else
      strcpy(cp, "0.0.0.0    ");

   /* print out iface where host was found */
   cp = readbuf + strlen(readbuf);
   sprintf(cp, "%u   ", host->iface );

   /* Add lease, -1 if infinite, 0 if static IP */
   cp = readbuf + strlen(readbuf);
   sprintf(cp, "%ld   ", host->ip_lease);

   /* Add image file name  */
   if(host->imagefile[0])
   {
       if( (strlen(cp) + strlen(host->imagefile) + 50) > rblen )
      {
         dtrap(); /* buffer overflow */
         return ENP_RESOURCE;
      }
      image = host->imagefile;
   }
   else     /* no image file name set */
   {
      image = "unknown.gif";
   }

   cp = readbuf + strlen(readbuf);
   sprintf(cp, "%s   ", image);

   /* Add flags fields */
   cp = readbuf + strlen(readbuf);
   if((ffmask & 0x0F00) == 0x0700)
   {
      dtrap();
   }
   sprintf(cp, "%08x   %08x \n", host->h_flags, (unsigned)ffmask );

   return 0;
}

/* br_filelan(FILE * fout, struct laninfo * lan)
 *
 * Write the passed LAN hosts info to an open file.
 * 
 * Returns 0 if OK else ENP_ error code.
 * 
 */

#define FBLEN   312     /* Size of buffers for Lan info file lines */
        
int
br_filelan(FILE * fout, struct laninfo * ln)
{
   int      i;
   int      err;
   int      buflen;
   char     filebuf[FBLEN]; /* generic file io buffer */
   struct host_entry * host;

   /* Loop through all in-ram hosts */
   for(i = 0; i < HASH_HOSTS; i++)
   {
      for(host = hosttab[i]; host; host = host->next)
      {
         if(host->h_flags & H_BCAST)
            continue;

         if((host->iphost == NULL) || (host->iphost->ipaddr == 0))
            continue;

         if(host->lan != ln)
            continue;

         lanline(host, filebuf, sizeof(filebuf) );

         buflen = strlen(filebuf);
         err = fwrite(filebuf, 1, buflen, fout);
         if(err != buflen)
         {
            dtrap();
            return ENP_FILEIO;
         }
      }
   }
   return 0;
}


/* br_writelan() - write the local LAN mac database to a file.
 *
 * parameter: char * lanname - name of current LAN
 *
 * Lan MAC line format is:
 * 
 *    "name macaddr ipaddr iface lease image hflags fflags lanname"
 *
 * RETURNS 0 if OK or ENP_ code
 */

static char * tmp_lanfile = "network.db0";

int
br_writelan()
{
   int      err;
   int      len;
   int      line;
   int      wrotecurrent;
   char *   cp;
   char *   endbracket;
   FILE *   fin;
   FILE *   fout;
   char     filebuf[FBLEN];
   struct laninfo * currentlan;
   struct laninfo * ln;

   dprintf("br_writelan called...");
   if( !(brstates[br_ifview].evflags & EV_NETNAMED))
   {
      dprintf("br_writelan: LAN not identified yet.\n");
      return 0;
   }

   wrotecurrent = FALSE;
   currentlan = br_lanlookup(brstates[br_ifview].current_lan);
   if(!currentlan)
   {
      dtrap();
      return ENP_LOGIC;
   }

   /* Don't write "temp" LAN info */
   if(currentlan->flags & LN_TEMP)
      return 0;

   /* Rename current LAN file to tmp file. We'll create a new 
    * lan file below and use the old one as input to guide
    * contents of the new one.
    */
   unlink(tmp_lanfile);
   err = rename(br_lanfile, tmp_lanfile);
   if(err)
   {
      /* If could be that the Network.db file does not exist yet. Try a
       * simple open to test for this - if it does not exist, create
       * the tmpfile.the logic below expects. An empty file is OK.
       */
      fin = fopen(tmp_lanfile, "rb");
      if(fin)
      {
         /* File existed - something else is wrong */
         fclose(fin);
         dprintf("Error renaming LAN file %s to %s \n", 
            br_lanfile, tmp_lanfile);
         return ENP_FILEIO;
      }
      /* create the tmp input file */
      fin = fopen(tmp_lanfile, "wb");
      if(fin == NULL)
      {
         dprintf("Error creating LAN file \n");
         return ENP_FILEIO;
      }
      fclose(fin);
   }

   /* open input & output files. */
   fin = fopen(tmp_lanfile, "rb");
   if(!fin)
   {
      dprintf("Error opening old LAN file %s\n", tmp_lanfile);
      rename(tmp_lanfile,  br_lanfile);      /* restore original DB file */
      return ENP_FILEIO;
   }
   fout = fopen(br_lanfile, "wb");          /* new output file */
   if(!fout)
   {
      dprintf("Error opening new LAN file %s\n",  br_lanfile);
      rename(tmp_lanfile,  br_lanfile);      /* restore original NV file */
      return ENP_FILEIO;
   }
   line = 0;

   /* read the tmp line by line making new file as we go */
   while(fgets(filebuf, sizeof(filebuf), fin) == filebuf)
   {
      len = strlen(filebuf);
      line++;

      if(len < 2)       /* skip empty lines */
         continue;

      if(filebuf[0] == '#')     /* copy comment lines */
      {
          fwrite(filebuf, 1, len, fout);
         continue;
      }

      /* See if this is a LAN name (eg "Home", "Traveling") */
      if(filebuf[0] == '[')
      {
          cp = &filebuf[1];

         /* In either case, write name to output file */
         fprintf(fout, "\n%s\n", filebuf);

         /* lookup this LAN - if we know it, write file from memory records */
         endbracket = strchr(cp, ']');
         if(endbracket == NULL)
         {
            dtrap();       /* No end bracket - Corrupt file ? */
            fclose(fout);  /* close new copy */
            fclose(fin);   /* close old copy */
            return err;
         }
         *endbracket = 0;        /* Null terminate name */
         ln = br_lanlookup(cp);  /* lookup terminated name */
         *endbracket = ']';      /* restore end bracket after name */
         if(ln)
         {
            err = br_filelan(fout, ln);
            if(err)
            {
               dtrap();
               fclose(fout);  /* close new copy */
               fclose(fin);   /* close old copy */
               return err;
            }
            fprintf(fout, "\n");

            /* Set flag if we wrote current LAN info */
            if(ln == currentlan)
               wrotecurrent = TRUE;
         }
      }
   }

   /* If current lan is not temp, and we didn't find it
    * in the file, then write it now.
    */
   if( !wrotecurrent && !(currentlan->flags & LN_TEMP) )
   {
      fprintf(fout, "\n[%s]\n\n",  brstates[br_ifview].current_lan );
      err = br_filelan(fout, currentlan);
   }

   fclose(fin);            /* close old copy */
   unlink(tmp_lanfile);    /* get rid of old copy file */
   fclose(fout);           /* close current copy */

   return 0;
}

/* br_dellanf()
 * 
 * delete the lan info for the LAN name passed from the lan records 
 * file (currently lnetwork.db);
 * 
 * 
 *
 * RETURNS 0 if OK or ENP_ code
 */

int
br_dellanf(char * lanname)
{
   int      err;
   int      len;        /* Length of line read from file */
   int      deleting;   /* flag TRUYE when in deleting portion */
   FILE *   fin;
   FILE *   fout;
   char     filebuf[FBLEN];

   /* Rename current LAN file to tmp file. We'll create a new 
    * lan file below and use the old one as input to guide
    * contents of the new one.
    */
   err = rename(br_lanfile, tmp_lanfile);
   if(err)
   {
      dprintf("br_dellanf: Error renaming LAN file %s to %s \n", 
         br_lanfile, tmp_lanfile);
      return ENP_FILEIO;
   }

   /* open input & output files. */
   fin = fopen(tmp_lanfile, "rb");
   if(!fin)
   {
      dprintf("br_dellanf: Error opening old LAN file %s\n", tmp_lanfile);
      rename(tmp_lanfile,  br_lanfile);      /* restore original DB file */
      return ENP_FILEIO;
   }
   fout = fopen(br_lanfile, "wb");          /* new output file */
   if(!fout)
   {
      dprintf("br_dellanf: Error opening new LAN file %s\n",  br_lanfile);
      rename(tmp_lanfile,  br_lanfile);      /* restore original NV file */
      return ENP_FILEIO;
   }
   deleting = FALSE;

   /* read throught the file line by line */
   while(fgets(filebuf, sizeof(filebuf), fin) == filebuf)
   {
       len = strlen(filebuf);

      /* If we found the lan name to delete, go into deleting mode */
       if(filebuf[0] == '[')
      {
          if(_strnicmp(lanname, &filebuf[1], strlen(lanname)) == 0)
         {
            deleting = TRUE;
         }
         else
         {
            deleting = FALSE;
         }
      }
      if(deleting)
         continue;

      /* Get here if NOT in deleting portion, write line */
      fwrite(filebuf, 1, len, fout);
   }

   dprintf("br_dellanf: removed lan %s from LAN file %s\n", lanname, br_lanfile);

   fclose(fin);            /* close old copy */
   unlink(tmp_lanfile);    /* get rid of old copy file */
   fclose(fout);           /* close current copy */

   return 0;
}


/* br_macnameset()
 * 
 * Set name for a MAC host 
 * 
 * Returns 0 if OK or ENP error code.
 */

int
br_macnameset(struct host_entry * host, char * name, int nametype)
{
   switch(nametype)
   {
   case H_NONAME:       /* there was no name info */
      break;
   case H_USERNAME:
   case H_NETBIOSNAME:  /* got name from netbios */
   case H_DHCPNAME:     /* got name from DHCP */
   case H_DNSNAME:  /* got name from DNS */
      if((host->locname[0] == 0) ||
         (nametype > (host->h_flags & H_NAMEMASK)))
      {
         strcpy(host->locname, name);
         host->h_flags &= ~H_NAMEMASK;
         host->h_flags |= nametype;
      }
      break;
   default:
      dtrap();    /* bad programming or corrupt DB file */
      return ENP_PARAM;
   }
   return 0;      /* OK return */
}

/* br_parsehost()
 *
 * Parse a host buffer and set the the passed pointers.to it's fields
 * This MUST find MAC, IP, LAN router,and the two flags fields or it
 * returns an erro. The other return pointers are set to zero if the 
 * fields are not found.
 *
 * Retuns 0 if OK, else ENP_ code.
 */

int
br_parsehost(char * inbuf,
   char **   name,
   char **   image,
   char *    macaddr,
   int *     iface,
   u_long *  hflags,
   u_long *  fflags,
   u_long *  lease,
   ip_addr * ipaddr)
{
   char *   cp;
   char *   next;
   char *   spacep;
   int      rblen;
   int      err;

   /* make sure text is of reasonable length */
   rblen = strlen(inbuf);
   if((rblen < 50) || (rblen >= FBLEN))
   {
      dtrap();
      return ENP_PARAM;
   }
   cp = inbuf;
   next = nextarg(cp);     /* find start of MAC address */
   br_nullterm(cp);        /* null terminate name */
   if(strlen(cp) > MAXNAMELEN)
   {
      dprintf("br_readhost: overlong name: %s\n", name);
      dtrap();
      return ENP_PARAM;
   }
   /* Replace plus signs ("+") with spaces in name string */
   for(spacep = strchr(cp, '+'); spacep; spacep = strchr(spacep, '+'))
      *spacep++ = ' ';

   *name = cp;
   cp = next;

   err = hex2bin(cp, (u_char*)macaddr, 6);
   if(err)
   {
      dprintf("br_readhost: bad MAC: %s\n", cp);
      dtrap();
      return ENP_PARAM;
   }

   cp = nextarg(cp);     /* find start of IP address */
   next = nextarg(cp);     /* find start of MAC address */
   br_nullterm(cp);        /* null terminate image file */
   spacep = br_checkip(ipaddr, cp);
   if(spacep)
   {
      dprintf("br_readhost: bad IP: %s (%s)\n", cp, spacep);
      dtrap();
      return ENP_PARAM;
   }

   cp = next;          /* get iface */
   *iface = atoi(cp);
   if((*iface == 0) && (*cp != '0'))
   {
      dprintf("br_readhost: bad iface number: %s\n", cp);
      dtrap();
      return ENP_PARAM;
   }

   cp = nextarg(cp);          /* save start of next parm (lease) */
   *lease = atoi(cp);
   if((*lease == 0) && (*cp != '0'))
   {
      dprintf("br_readhost: bad lease: \"%s\"\n", cp);
      dtrap();
      return ENP_PARAM;
   }

   cp = nextarg(cp);   /* find start of image file */
   if(!cp || !*cp)
   {
      dtrap();
      return ENP_PARAM;
   }
   next = nextarg(cp);  /* find host flags field */
   br_nullterm(cp);        /* null terminate image file */
   *image = cp;

   *hflags = atox(next);      /* read flags from hex */
   if((hflags == 0) && (*cp != '0'))
   {
      dtrap();
      return ENP_PARAM;
   }
   /* Clear flags that may not be valid in new environment */
   *hflags &= ~(H_DUPLIP | H_MULTILOCAL);

   cp = nextarg(next);       /* find file flags field */
   *fflags = atox(cp);      /* read flags from hex */
   if((*hflags == 0) && (*cp != '0'))
   {
      dtrap();
      return ENP_PARAM;
   }

   return 0;   /* good return */
}


/* br_readhost() - Read a line containing info about a local LAN
 * (MAC based) host. Line format is the same as that written by
 * br_writelen() above. Parses the line and places an entry in 
 * the RAM-based hosts list.
 *
 * The passed buffer is NOT preserved!
 *
 * RETURNS 0 if OK or ENP_ code
 */

int
br_readhost(char * filebuf, struct laninfo * ln)
{
   char *   name;
   char *   image;
   int      err;
   int      iface;
   u_char   macaddr[6];
   ip_addr  ipaddr;
   u_long   lease;
   u_long   hflags;
   u_long   fflags;
   struct host_entry * host;

   MEMSET(macaddr, 0, sizeof(macaddr) );

   err = br_parsehost(filebuf,
         &name,
         &image,
         (char*)&macaddr,
         &iface,
         &hflags,
         &fflags,
         &lease,
         &ipaddr);

   if(err)
      return err;

   /* If passed lan has no router MAC, install it */
   if((ln) && 
      (MAC_EQ(ln->router_mac, nullmac)) &&
      (fflags & FF_LANROUTER))
   {
      MEMCPY(ln->router_mac, macaddr, 6);
      ln->router_ip = ipaddr;
   }

   /* See if this is supposed to be an entry for ourself */
   if(hflags & H_SELF)
   {
      int   matched = 0;

      /* Mac address should match one of our iface macs */
      for(iface = 0; iface < (int)ifNumber; iface++)
      {
         if( MAC_EQ(macaddr, nets[iface]->n_macaddr))
         {
            matched++;
            break;
         }
      }

      if(!matched)
      {
         /* Didn't match - delete this entry, Alert user. */
         sprintf(msgbuf, "This computers MAC address seems to have changed.\n"
            "The old MAC was %s \n", br_prtmac(macaddr));

         alarm_general(&alarm_text, NULL, msgbuf);
         return ENP_PARAM;    /* Just don't make an entry for it */
      }
   }

   /* we've now read all the fields OK, allocate the host object. */
   br_hostupdate((char*)macaddr, ipaddr, 
         (hflags | FF_FROMFILE), iface);

   host = br_hostlookup((char*)macaddr, iface);
   if(!host)
   {
      dtrap();
      return ENP_PARAM;
   }

   host->lan = ln;

   /* update fields that were not set up in br_hostupdate(). 
    * Generally do not write filed info over new info.
    */
   if(!host->disctype)
      host->disctype = (fflags & FF_TYPEMASK);
   if(host->ip_lease == -1)
      host->ip_lease = lease;
   if(host->h_flags == 0)
      host->h_flags = hflags;

   /* We just read this from a file; clear activity */
   host->active = 0;

   /* figure out where (and if) to set the name */
   br_macnameset(host, name, fflags & H_NAMEMASK);

   /* figure out what to set as an image file name. First see
    * if the image file was known (e.g. not "unknown.gif")
    */
   if(strcmp(image, "unknown.gif"))
   {
      strcpy(host->imagefile, image);
      strcpy(host->iphost->imagefile, image);
   }

   /* Set router if this host is the router for the LAN */
   if(fflags & FF_LANROUTER)
   {
      /* Set router, but with modest confidence level */
      br_setrouter(iface, host, 0x40);
   }

   return 0;
}



/* br_checklan()
 *
 * Called when switching to a different LAN
 *
 *
 */


void
br_checklan()
{
   struct laninfo * ln;

   /* Don't search if we don't know the router yet */
   if( !(brstates[br_ifview].evflags & EV_ROUTERSET))
   {
      return;
   }

   /* Make sure we actually know the router's MAC address */
   if(brstates[br_ifview].br_router == NULL)
      return;

   if(lan_file_read == 0)
      return;

   /* see if we're on one of the known LANs */
   for(ln = br_lans; ln; ln = ln->next)
   {
      if(MAC_EQ(brstates[br_ifview].br_router->macaddr, ln->router_mac))
         break;
   }

   if(ln)      /* found a lan with matching router MAC address */
   {
      /* Set this LAN as current_lan */
      strcpy( brstates[br_ifview].current_lan, ln->name);
      brstates[br_ifview].evflags |= EV_NETNAMED;

      dprintf("On LAN \"%s\"\n", ln->name);
      return;
   }

   /* Create a temporary LAN. User can name & save it later. */
   if(brstates[br_ifview].br_router)
   {
      ln = new_lan("temp", brstates[br_ifview].br_router->macaddr,
         brstates[br_ifview].br_router->iphost->ipaddr);
   }
   else  /* create a LAN with no router information */
   {
      ln = new_lan("temp", nullmac, 0);
   }

   if(ln)
      ln->flags |= LN_TEMP;      /* Mark as temporary */

   /* Call alert routine to inform user of unknown LAN */
   if((brstates[br_ifview].evflags & EV_LANUKMSG) == 0)
   {
      sprintf(msgbuf,  
         "%s has detected that it's on a LAN which is not recorded"
         "in the database. You can name & save this LAN via the GUI\n",
         br_appname );

      alarm_general(&alarm_text, NULL, msgbuf );
      brstates[br_ifview].evflags |= EV_LANUKMSG;
   }

   return;
}


/* Initial read of LAN info file. This file will not exist on 
 * the first run, but once LAN info has been discovered and saved
 * the lan file should have historic info about LANs the system 
 * has been run on .
 */

int
br_readlans()
{
   char *   cp;
   char *   end;
   FILE *   fp;
   int      len;
   int      err;
   char		lanbuf[256];
   struct laninfo * ln = NULL;
   int      line = 0;

   fp = fopen(br_lanfile, "r");
   if(!fp)
      return ENP_RESOURCE;

   /* read the file into memory structures */
   while(fgets(lanbuf, sizeof(lanbuf), fp) == lanbuf)
   {
      len = strlen(lanbuf);
      line++;
      if(len < 3)       /* skip empty lines (\n, \r\n) */
         continue;
      if(lanbuf[0] == '#')     /* copy comment lines */
         continue;

      /* If we got brace, lookup lan name */
      if(lanbuf[0] == '[')
      {
         cp = &lanbuf[1];
         end = strchr(cp, ']');
         if(end)
            *end = 0;
         ln = br_lanlookup(cp);
         /* Skip if this is a known LAN */
         if(ln)
            continue;
         else
            ln = new_lan(cp, nullmac, 0);
      }
      else if(ln)     /* read the line into the current lan */
      {
         err = br_readhost(lanbuf, ln);
      }
      else
      {
         fclose(fp);
         dtrap();
         return -1;
      }
   }
   fclose(fp);
   lan_file_read = 1;   /* set global flag */
   return 0;
}

/*
Copyright (C) 2003-2008 by Praemio Inc. All rights reserved.

Author: John Bartas - jbartas@praemio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/



