/* brmsgs.c
 *
 * Buster to Buster messaging code.
 *
 * Copyright 2003-2008 by Praemio Inc. All rights reserved.
 * Author: John Bartas - jbartas@praemio.com
 * This software is released under the terms of the LGPL.
 * See LGPL Notice at end of this file; Full license in ../license.txt
 */


#include "busterport.h"
#include "brnet.h"
#include "ip.h"
#include "brdb.h"
#include "brmenus.h"
#include "brmd5.h"


/* Systems master network admin password */
char  br_password[MAXNAMELEN] = "evaluation";

char brmsg[MSGSIZE];  /* scratch message buffer */


/* Minimum time between routine (non-asserting) messages */
int bg_interval = MINUTE;

/* Time for regular polling of other Busters */
int bg_polltime = (MINUTE * 5);


/* Computers which may run two Busters at once (i.e. Windows) 
 * should make sure each instance of Buster has a unique 
 * instance number, so they can communicate with each other. 
 * How this is done is a porting issue. Devices which only support
 * one instance at a time (e.g. embedded appliances) can just 
 * leave this set to zero.
 */
u_long  br_instance = 0;

/* master list of known Busters nearby this segment */
struct other_agent * busterlist;

/* name number of this Buster (1 == "buster.1") */
int   br_nameber = 1;


enum br_modes br_mode = HOSTONLY;


/* UDP port for buster<->buster messages */
unshort  msgport = 628;


/* strings for printf mode names. MUST match enum br_modes  */

char * br_modenames[] = 
{
   "OFFLINE",        /* Power down, prog unloaded, etc. */
   "PASSIVE",        /* Monitors the segment, may go active */
   "STANDBY",        /* Only monitors the segment */
   "HOSTONLY",       /* Buster only acts on one host */
   "ASSERTING",      /* Wants to be active (e.g. just powered up.) */
   "ACTIVE",         /* Buster acts on the whole net segment */
};

SOCKTYPE    br_msgsocket;

/* br_sendmsg() - send a basic agent<->agent message. Message type
 * is passed as parameter.
 * 
 * Returns 0 if OK. else ENP_ error code 
 */

#define  AUTHDATASIZE   16

int
br_sendmsg(int opcode, ip_addr fhost)
{
   int      err;
   PACKET   pkt;
   struct bustergram * bgram;
   MD5_CTX  txdigest;

   pkt = pk_alloc(sizeof(struct bustergram) + IP_OVERHEAD);
   if(pkt == NULL)
      return ENP_NOBUFFER;

   bgram = (struct bustergram *)pkt->pk_prot;

   bgram->mode = htonl(opcode);
   bgram->version = htonl(br_versnum);
   bgram->number = htonl(br_nameber);
   bgram->instance = htonl(br_instance);

   /* Net endian GUI parameters */
   bgram->active_ip = nets[br_ifview]->n_ipaddr;
   bgram->activeport = htons(http_port);

   /* Clear authdata field before MD5-ing header */
   MEMSET(bgram->authdata, 0, AUTHDATASIZE);

   MD5Init(&txdigest);
   MD5Update(&txdigest, (u_char*)bgram, sizeof(struct bustergram) );
   MD5Update(&txdigest, (u_char*)br_password, strlen(br_password) );
   MD5Final(&txdigest);
   MEMCPY(bgram->authdata, txdigest.digest, AUTHDATASIZE );

   pkt->pk_plen = sizeof(struct bustergram);

   err = udp_socksend(fhost, msgport, pkt);

   return err;
}


/* br_brlookup() - lookup an Agent in our internal list. An Agent 
 * is made unique if MAC address, instance, and version are unique.
 * This lookup routine will allow one of the parameters to be 
 * a wildcard (0 or NULL) and return the first matching buster 
 * in the list.
 * 
 * returns NULL if no matching buster found.
 */

struct other_agent * 
br_brlookup(u_char * mac, u_long instance, u_long version)
{
   struct other_agent * obp;

   for(obp = busterlist; obp; obp = obp->next)
   {
      if(version && (obp->version != version))
         continue;
      if(instance && (obp->instance != instance))
         continue;
      if(mac && (MEMCMP(mac, obp->macaddr, 6)))
         continue;
      return obp;
   }
   return NULL;
}


/* br_newbp() - make a new agent entry from a bustergram header.
 *
 * parms are passed to newbp in network endian.
 *
 * returns new struct if OK, NULL ifi alloc fails. 
 */

struct other_agent * 
br_newbp(struct bustergram * bgram, ip_addr ipaddr, unshort port, int iface)
{
   struct other_agent * obp;

   obp = (struct other_agent *)OBP_ALLOC(sizeof(struct other_agent));
   if(!obp)
      return NULL;

   obp->ipaddr = ipaddr;
   obp->version = htonl(bgram->version);
   obp->instance = htonl(bgram->instance);
   obp->number = htonl(bgram->number);
   obp->mode = bgram->mode;
   obp->last_activity = cticks;
   obp->httpport = htons(port);
   MEMCPY(obp->macaddr, bgram->macaddr, 6);

   obp->next = busterlist;
   busterlist = obp;

   /* Set back pointer in other agent's IP structure */
   obp->iphost = br_iplookup(iface, ipaddr);
   if(obp->iphost)
      obp->iphost->buster = obp;
   else
      dtrap();    /* Can this happen? */
   
   return obp;
}


static int  mode_secs;

u_long   bg_lastpoll;


/* mode_callback() - Timer, called once a second */

void
mode_callback()
{
   int      err;

   /* See if it's time to poll the other local Agents */
   if((bg_lastpoll + bg_polltime) < cticks)
   {
      br_sendmsg(br_mode, 0xFFFFFFFF);
      bg_lastpoll = cticks;
   }

   if((br_mode != ASSERTING) && (br_mode != ACTIVE))
      return;

   mode_secs++;         /* count the seconds */
   if(mode_secs >= 5)
   {
      br_mode = ACTIVE;      /* 5 seconds, no objections */
      err = br_sendmsg(ACTIVE, 0xFFFFFFFF);
      dprintf("Mode set to %s\n", br_modenames[br_mode] );
   }
   else
   {
      /* Send message adverising that we want to go Active */
      err = br_sendmsg(ASSERTING, 0xFFFFFFFF);

      /* After Windows hibernation, we may come back up faster 
       * than the MAC system, especially on 802.11 devices;
       * so if sendmsg fails we delay the timeout process.
       */
      if(err)
      {
         mode_secs--;
      }
   }

   return;
}

/* br_setmode() - called to (try to) set Buster mode.
 *
 * Returns 0 if process OK so far, else ENP_ code
 */

int
br_setmode(enum br_modes mode)
{
   int   err;

   /* don't let system go straight into active mode - force
    * it through an assertion period first.
    */
   if(mode == ACTIVE)
      mode = ASSERTING;

   if(mode == ASSERTING)
      mode_secs = 0;

   br_mode = mode;
   err = br_sendmsg(mode, 0xFFFFFFFF);
   if(err)
      return err;

   dprintf("Mode set to %s\n", br_modenames[mode] );
   return 0;
}


/* bg_callback() - Called when we receive a Bustergram. */

int
bg_callback(PACKET pkt,struct ip * pip)
{
   enum br_modes  mode;       /* mode of sender */
   struct bustergram * bgram;
   struct other_agent * obp;
   struct host_entry * buster_host;
   MD5_CTX  tmpdigest;
   MD5_CTX  pktdigest;
   u_char   mac[6];        /* actual MAC address */
   ip_addr  ipaddr;        /* IP of sender */
   int      iface;         /* interface that received datagram */
   

   iface = if_netnumber(pkt->pk_net);

   bgram = (struct bustergram *)pkt->pk_prot;

   /* copy in the sender MAC address from MAC header */
   MEMCPY(mac, pkt->pk_buff + 6, 6);

   /* get IP address of sender */
   ipaddr = pip->ip_src;

   /* If this is an echo of one of our packets, punt it now */
   /* If we have a packet from our own MAC, it's probably an 
    * echo (from the drivers) of one we sent, but could be 
    * an upgrade Buster starting up on this machine. Use instance 
    * & version to tell the difference.
    */
   if(MEMCMP(mac, pkt->pk_net->n_macaddr, 6) == 0)
   {
      if(htonl(bgram->instance) == br_instance)
      {
         return ENP_NOT_MINE;    /* It's an echo, ignore it */
      }
   }

   /* verify MD5 authentication */
   MEMCPY(&pktdigest.digest, bgram->authdata, AUTHDATASIZE);
   MEMSET(bgram->authdata, 0, AUTHDATASIZE);

   MD5Init(&tmpdigest);
   MD5Update(&tmpdigest, (u_char*)bgram, sizeof(struct bustergram) );
   MD5Update(&tmpdigest, (u_char*)br_password, strlen(br_password) );
   MD5Final(&tmpdigest);

   if(MEMCMP(tmpdigest.digest, pktdigest.digest, 16) != 0)
   {
      dprintf("Bad digest in Buster message from %02X%02X%02X-%02X%02X%02X\n",
         mac[0], mac[1], mac[2], mac[3], mac[4], mac[5] );
      return ENP_LOGIC;
   }

   /* look up sending Buster. Remember: this may be NULL */
   obp = br_brlookup(bgram->macaddr, htonl(bgram->instance), 
      htonl(bgram->version));


   mode = htonl(bgram->mode);
   if(obp == NULL)
   {
      /* Pass parms to newbp in network endian */
      obp = br_newbp(bgram, ipaddr, bgram->activeport, iface);
      if(!obp)
         return ENP_RESOURCE;
      dprintf("Buster detected: IP:%u.%u.%u.%u.\n",
         PASS_IPADDR(obp->ipaddr) );

      br_sendmsg(br_mode, ipaddr);  /* Let new guy know we're here */
   }
   else  /* Previously known Agent, see if we should owe a reply */
   {
      if((obp->last_activity + bg_interval) <  cticks)
      {
         br_sendmsg(br_mode, ipaddr);
      }
      obp->last_activity = cticks;
      obp->mode = mode;
   }

   /* lookup the MAC host for this Buster */
   buster_host = br_hostlookup( (char*)(&mac[0]), iface );
   if(buster_host)
   {
      buster_host->h_flags |= H_AGENT;    /* set Buster bit flag */
   }

   /* We always have to reply to ASSERTING */
   if(mode == ASSERTING)
   {
//      dtrap();
      // change our mode???
//      br_sendmsg(br_mode, ipaddr);
      return 0;
   }

   /* If we were asserting and another guy is active, then back off */
   if((br_mode == ASSERTING) &&
      (mode == ACTIVE))
   {
      br_mode = STANDBY;
   }

   /* If he was the active guy and is going inactive, see if we 
    * should take over.
    */
   if((obp->mode == ACTIVE) &&
      (mode != ACTIVE) &&
      (br_mode == STANDBY))
   {
      br_setmode(ASSERTING);
   }

   return 0;
}

/*
Copyright (C) 2003-2008 by Praemio Inc. All rights reserved.

Author: John Bartas - jbartas@praemio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/



