/*
 * FILENAME: brnet.c
 *
 * Copyright 2003-2008 by Praemio Inc. All rights reserved.
 * Author: John Bartas - jbartas@praemio.com
 * This software is released under the terms of the LGPL.
 * See LGPL Notice at end of this file; Full license in ../license.txt
 *
 * Buster misc networking code. 
 *
 * ROUTINES:
 * set_brstate(), get_net(), br_check(), 
 * br_startup(), same_subnet(), br_shutdown(), 
 * cknet_callback(), br_powerresume(), br_checknet(), 
 *
 */


#include "busterport.h"
#include "brnet.h"
#include "ip.h"
#include "arp.h"
#include "dhcpdef.h"
#include "brdb.h"


/* NULL MAC address, for setting MAC addresses to all zeros */
u_char nullmac[8] = { 0,0,0,0,0,0,0,0 };


extern   struct menu_op br_menu[12];
extern   struct menu_op db_menu[12];

extern  struct br_thread alert_thread;


extern   void     br_shutdown(void);    /* precedl */
extern   void     br_portrunning(void); /* per-port initialization */
extern   int      dhc_discover(int iface);
extern   void     cknet_callback(void);
extern   int      sa_init(void);
extern   int      wi_init(void);


#ifdef SMTP_ALERTS
extern   int      smtpalert_check(void);
#endif

/* from misclib directory: */
extern int hex2bin(char * inbuf, unsigned char * outbuf, int bcount);

unshort   pingseq = 99;


/* mode before suspend message */
int      br_powerdown_mode = HOSTONLY;

u_long   br_downtime = 0;     /* time/date Buster went to hibernate */
u_long   br_sleeptime;        /* time of last hibernation in seconds */
int      br_ifview = 0;       /* Index of iface for viewing and mapping */
int      br_autoifview = 1;   /* bool for autoselect br_ifview */
unsigned ifNumber;            /* number of ifaces in  nets[] */

enum br_sysstates br_sysstate;      /* overall system state */
struct br_state brstates[MAXNETS];  /* per-net states info */

struct netiface staticnets[MAXNETS];

NET   nets[MAXNETS];

brqueue    largebufq;  /* Full size packet buffers */
brqueue    smallbufq;  /* small packet buffers */
brqueue    emptybufq;  /* empty buffers, for RX */


unsigned   emptybufs = 100;         /* number of bufferless pkts */
unsigned   smallbufs = 30;          /* number of small buffer pkts */
unsigned   smallbufsize = 128;      /* size of small buffers */
unsigned   largebufs = 20;          /* number of big buffer pkts */
unsigned   largebufsize = 1506;     /* size of big buffers */

u_long  br_lastresume;  /* time when br_powerresume() was last called */

char systemname[MAXNAMELEN] = "(unnamed)";

#ifdef WEBIO_GUI
#include "websys.h"
#include "webio.h"

extern  int   httpport;
extern  int   wi_thread(void);
extern  char * wi_rootfile;
extern  int   buster_execfunc(void * sess, char * args);

/* Task structure for the Community database update task */
TH_OBJECT(to_webio);
TH_PROTO(wi_entry);

TH_PROTO(wi_entry)
{
   wi_thread();
   panic("wi_entry");
   return 0;
}

struct br_thread webio_thread = 
{
   &to_webio,
   "Web server",
   wi_entry,
   APP_PRIORITY,
   APP_STACK_SIZE,
};
#endif /* WEBIO_GUI */


/* A timer thread should be provided by the porting code; but
 * we want to start it from here to keep the startup sequence
 * under control. 
 */
extern struct br_thread timer_thread;


/* This flag is used to see if the Buster is
 * busy checking out the current network after a startup or 
 * power resume (e.g. wake from Windows hibernation). 
 */
long  br_checknet_timer;


/* Called by Init logic when we're running OK */

void
br_running()
{
   int      err;

   /* send out the hello bustergram (Buster) packet */
   if(br_mode == STANDBY )
      err = br_setmode(ASSERTING);
   else
      err = br_setmode(br_mode);   /* Broadcast our mode */

   if(err)
      dprintf("Error sending %d initial Buster message\n", err);

   br_portrunning();     /* do per-port initialization */
   return;
}



void
br_autoselect()
{
   int   iface;

   if(ifNumber < 2)
      return;

   /* See if we need to update iface autoselection */
   if(br_autoifview)
   {
      for(iface = 0; iface < (int)ifNumber; iface++)
      {
         if(nets[iface]->ifOperStatus != 1)
         {
            /* If view iface is down, switch to first iface that is up */
            if(iface == br_ifview)
            {
               int   i;
               for(i = 0; i < (int)ifNumber; i++)
               {
                  if(nets[i]->ifOperStatus == 1)
                  {
                     br_setiface(&br_ifview, i);
                     break;
                  }
               }
            }
            continue;
         }

         if(nets[iface]->traffic > ((nets[br_ifview]->traffic * 4) + 16) )
         {
            dprintf("Autoselect: switching view iface\n");
            dprintf("   %s traffic: %d\n   %s traffic: %d\n",
               nets[iface]->description,
               nets[iface]->traffic,
               nets[br_ifview]->description,
               nets[br_ifview]->traffic);

            /* re-read the IP info for the interface we're switching to */
            br_setipinfo(iface);

            /* Switch view interface */
            br_setiface(&br_ifview, iface);

            /* Restart the "checknet" timer */
            br_checknet_timer = TRUE;
         }
      }
   }

   return;
}

void
bt_ipe_age(struct btnode * n, long param, int depth)
{
   struct control_ip * cip;

   cip = (struct control_ip *)n->data;

   /* If not timed, just retturn */
   if((cip->ci_flags & CIF_DNSREP) == 0)
      return;


   /* See if it went below 1 minute */
   if(cip->ci_timeout <=  MINUTE)
   {
      USE_ARG(depth);
      USE_ARG(param);

      del_ipe(cip, TRUE);
   }
   else
   {
      cip->ci_timeout -= MINUTE;
   }
   return;
}

u_long   br_suspended = 0;

/* br_suspendresume() - toggle suspend state */

int
br_suspendresume(int tickct)
{
   int   i;
   int   truefalse;

   br_suspended = tickct;

   if(tickct == 0)
      truefalse = FALSE;
   else
      truefalse = TRUE;

   for(i = 0; i < (int)ifNumber; i++)
      brstates[i].rx_disabled = truefalse;

   br_portsuspendresume(truefalse);
   return 0;
}


/* br_check() - once-a-second callback to check state */

int   stats_age = 0;
int   in_brtimer = 0;      /* timer routine re-entry guard */
int   switch_period = 30;  /* timer to spoof switch tables */

/* last host for which we sent DNS quesry */
struct ip_hostentry * dnshost = NULL;


void
br_check()
{
   int      i;                      /* index for connection table */
   int      err;
   struct conn_entry *     conn;    /* scratch */
   struct ip_hostentry *   iphost;
   struct host_entry *     host;

   /* Age out traffic ratios for the network interfaces */
   if((stats_age % 3) == 0)
   {
      for(i = 0; i < (int)ifNumber; i++)
         nets[i]->traffic >>= 1;
   }

   /* If receiver is suspended, see if it's time to resume. */
   if(br_suspended  && (br_suspended <= cticks))
   {
      br_suspendresume(0);
   }

   /* this timer tick may do some lengthy processing, and some 
    * timer systems (like windows2K) are not be proofed against 
    * re-entering timer routines; so guard against this here:
    */
   if(++in_brtimer != 1)    /* re-entry guard */
   {
      dtrap();          /* can this happen? */
      in_brtimer--;
      return;
   }

#ifdef SMTP_ALERTS
   smtpalert_check();
#endif

   br_autoselect();

   /* rotate through more CPU consuming operations based on the
    * stats_age counter.
    */
   if((stats_age % 10) == 0)
   {
      /* Age out arp table */
      arp_cleanup();

      /* Age out connections */
      for(i = 0; i < HASH_CONN_ENTRIES; i++)   /* loop through connections */
      {
         conn = ce_table[i];
         while(conn)    /* loop through hash buchet */
         {
            switch(conn->ce_type)
            {
            case H_TCP:
               /* Age out TCP entries based on TCP state */
               switch(conn->ce_state)
               {
               case TCPS_CLOSED:
                  conn = conn_delete(conn, i);
                  continue;
               case TCPS_CLOSING:
               case TCPS_CLOSE_WAIT:
               case TCPS_LAST_ACK:
               case TCPS_FIN_WAIT_1:
               case TCPS_FIN_WAIT_2:
               case TCPS_TIME_WAIT:
               case TCPS_RESET:
               case TCPS_KILLED:
                  /* see if conn has been closing for longer than keep time */
                  if((cticks - conn->ce_time) > br_cfg.keep_tcpconn)
                  {
                     conn = conn_delete(conn, i);
                     continue;
                  }
                  break;
               default:
                  dtrap();    /* did we miss any states? */
                  /* fall to established case */
               case TCPS_ESTABLISHED:
                  /* Established gets two hours; the default keepalive time */
                  if((cticks - conn->ce_time) > (2 * HOUR))
                  {
                     conn = conn_delete(conn, i);
                  continue;
                  }
                  break;
               case TCPS_SYN_SENT:
               case TCPS_SYN_RECEIVED:
                  /* see if sending syns and getting no reply */
                  if((cticks - conn->ce_time) > (30 * SECOND))
                  {
                     /* merge count of syn packets into IP sender */
                     if(conn->ce_flags & CEF_LOCALCLIENT)
                        iphost = br_iplookup(conn->ce_iface, conn->loc.ce_addr);
                     else if(conn->ce_flags & CEF_LOCALSERVER)
                        iphost = br_iplookup(conn->ce_iface, conn->rem.ce_addr);
                     else     /* assume lower port is server */
                     {
                        dtrap(); /* can't happen? */
                        iphost = 0;	/* stifle dingbat compiler warnings */
                     }
                     if(iphost)
                     {
                        long  syn_rate;      /* syns/minute */
                        iphost->tcp_syns += (conn->rem.ce_outpkts + conn->loc.ce_outpkts);
                        /* set syn_rate to minues host has been active */
                        syn_rate = (cticks - iphost->tm_detected)/MINUTE;
                        /* dive number of syns by minues avtive */
                        if(syn_rate)      /* avoid divide by xero */
                           syn_rate = iphost->tcp_syns / syn_rate;
                        if(syn_rate > br_cfg.syns_per_minute)
                        {
                           /* too many syns, send alarm */
                           alarm_general(&syn_flood, &iphost->synflood, iphost);
                        }
                     }
                     conn = conn_delete(conn, i);
                     continue;
                  }
                  break;
               }
               break;
            case H_UDP:    /* UDP and ICMP get removed if idle for too long */
            case H_ICMP:
            case H_OTHER:
            case H_NETBEUI:
               if((cticks - conn->ce_time) > br_cfg.keep_udpconn)
               {
                  conn = conn_delete(conn, i);
                  continue;
               }
               break;
            default:
               dtrap();
               break;
            }
            conn = conn->next;
         }
      }
   }

    /* Drive ARP scans */ 
   if( arp_scanning[br_ifview] )
       arp_scan(br_ifview);

   if((stats_age % 5) == 1)
   {
      int      dnssends = 0;     /* DNS reqs sent this second */
      int      dodns = FALSE;    /* TRUE if we are in dns portion */

      /* Age out idle IP hosts, and try to populate names in the rest */
      for(i = 0; i < HASH_IPHOSTS; i++)
      {
         iphost = ip_hosttab[i];
         while(iphost)
         {
            if((cticks - iphost->tm_lastpkt) > br_cfg.keep_iphost)
            {
               if(iphost->conns)    /* don't delete if it has connections */
               {
                  iphost = iphost->next;
                  continue;
               }
               if(iphost->mac_host) /* don't delete if it's MAC host info */
               {
                  iphost->tm_lastpkt = cticks + HOUR; /* Put expire time way out */
                  iphost = iphost->next;              /* set ptr for next loop */
                  continue;
               }
               /* If we're deleting the last host for which we sent a DNS req, 
                * then clear dnshost variable 
                */
               if(iphost == dnshost)
                  dnshost = NULL;
               iphost = iphost_delete(iphost, i);
               continue;
            }
            /* See if we're in the section of the database where we should be 
             * doing DNS for namesless hosts. Start when we find the last host
             * we DNS'ed for on the last tick.
             */
            if((dnshost == iphost) || (dnshost == NULL))
               dodns = TRUE;

#ifdef  DNS_REVLOOKUPS
            /* If we are in the dodns postion of the list and...
             * current host has no address and ...
             * this is not the last host we DNS'ed for, and... 
             * ipentry's DNS exponential backoff has expired...
             * then send a DNS request packet.
             */
            if((dodns) &&                          /* make sure feature is enabled */
               (iphost->dnsname[0] == 0) &&        /* skip if we already have name */
               (dnshost != iphost) &&              /* skip last guy we did */
               (cticks >= iphost->tm_nextdns) &&   /* skip if not time for retry yet */
               (iphost->ipaddr  != 0) &&           /* skip pre-DHCP nodes */
               ((iphost->ipaddr & htonl(0xFF000000)) != htonl(0x0A000000)) && /* skip 10 net */
               ((iphost->ipaddr & htonl(0xFFFF0000)) != htonl(0xC0A80000)) &&   /* skip 192.168 */
               ((iphost->flags & (IH_MULTICAST|IH_BROADCAST)) == 0))         /* skip Broadcast entry */
            {
               if(dnssends++ > 3)   /* only send three per tick */
               {
                  dodns = FALSE;    /* done for this tick */
                  break;            /* need a double break... */
               }

               dnc_namereq(iphost);          /* send DNS query */

               /* implement exponential backoff */
               iphost->tm_nextdns = (cticks + ((1 << (u_long)iphost->dns_sent) * TPS * 5));
               iphost->dns_sent++;           /* count DNS queries for each host */
               dnshost = iphost;             /* set as last DNS'ed host */
            }
#endif /* DNS_REVLOOKUPS */
            iphost = iphost->next;
         }
         if(dnssends > 3)     /* only send three per tick */
             break;             /* second part of double break */
      }
      /* If we sent no DNS requests then reset the dnshost */
      if(dnssends == 0)
         dnshost = NULL;
   }

   
//   if((stats_age % 3) == 1)
   {
      u_long  ageout;

      /* figure out ctick value is use for deleting idle mac hosts. */
      ageout = cticks - (br_cfg.mactmo * MINUTE);

      /* If it went negative, flaten it. */
      if(ageout & 0x80000000)
         ageout = 0;

      /* Age traffic stats in MAC hosts list */
      for(i = 0; i < HASH_HOSTS; i++)
      {
         u_long  pkts;
         u_long  bytes;

         host = hosttab[i];
         while(host)
         {

            /* See if host has aged out */
            if(host->active < ageout)
            {
               if(br_cfg.cflags & BC_DELIDLE)
               {
                  host = host_delete(host, i);
                  continue;
               }
               else  /* not deleting, just hide the host */
               {
                  host->h_flags |= H_HIDDEN;
               }
            }
            pkts = host->sec_pkts;
            host->sec_pkts = 0;
            bytes = host->sec_bytes;
            host->sec_bytes = 0;

            /* Bump the small interval index. If the index wrapped, fall into
             * logic to zero the index and handle the next larger interval.
             */

            host->sec_index++;
            if(host->sec_index >= 60)  /* second index wrapping? */
            {
               host->sec_index = 0;    /* zero second index */
               host->min_index++;
               if(host->min_index >= 60)     /* time for hour stuff? */
               {
                  host->min_index = 0;       /* zero minute index */
                  host->hour_index++;
                  if(host->hour_index >= 24) /* time to do day stuff */
                  {
                     host->hour_index = 0;   /* zero hour index */
                     host->day_index++;
                     if(host->day_index >= 7)   /* week stuff...*/
                     {
                        host->day_index = 0;
                     }
                     host->day_p[host->day_index] = 0;   /* zero counters */
                     host->day_b[host->day_index] = 0;
                  }
                  host->hour_p[host->hour_index] = 0;   /* zero counters */
                  host->hour_b[host->hour_index] = 0;
               }
               host->min_p[host->min_index] = 0;   /* zero old/new counters */
               host->min_b[host->min_index] = 0;
            }

            /* add current count to each current bucket */
            host->day_p[host->day_index] += pkts;
            host->day_b[host->day_index] += bytes;
            host->hour_p[host->hour_index] += pkts;
            host->hour_b[host->hour_index] += bytes;
            host->min_p[host->min_index] += pkts;
            host->min_b[host->min_index] += bytes;
            host->sec_p[host->sec_index] = pkts;
            host->sec_b[host->sec_index] = bytes;

            host = host->next;      /* do for all hosts */
         }
      }
   }

   /* Once a second, check for Iface changes */
   cknet_callback();

   stats_age++;

   /* A few times a minute check for stale LAN info */
   if((stats_age % 20) == 1)
   {
      /* Make sure Buster has figured out what LAN we're on by
       * checking the restart timer, br_checknet_timer.
       */
      if((br_checknet_timer == 0) &&
         (brstates[br_ifview].evflags & EV_LANSTALE))   /* and only if something changed */
      {
         brstates[br_ifview].evflags &= ~EV_LANSTALE;  /* clear stale flag */
         err = br_writelan();                   /* write lan file info */
      }

      /* Once a minute check for DNS cache aging */
      TH_LOCK_MUTEX(br_tree_mutex);
      bt_depthfirst((struct btnode *)ip_tree, bt_ipe_age, 0, 0);
      TH_UNLOCK_MUTEX(br_tree_mutex);
   }

   /* Check for Bustergram timeouts & retrys */
   mode_callback();

   if(--in_brtimer != 0)   /* clear re-entry guard */
   {
      dtrap();
   }

   return;
}



/* First Buster code to run, called at system startup */


int
br_startdb()
{
   int   err;
   int   i;

   /* Thread the staticnets[] into the pointer nets[] */
   for(i = 0; i < MAXNETS; i++)
       nets[i] = &staticnets[i];

   /* Init raw version string, may be modified by custom cfg files */
#ifdef NNSQUAD
      sprintf(br_version, 
         "NNSquad Network Measurement Agent Beta (%u.%u.%u.%u NNS-B1)", 
         PUSH_VERSION(br_versnum) );
#else /* Not NNSQUAD build, use consumer version */      
      /* append code version to parms string */
      sprintf(br_version, "Buster %u.%u.%u.%u",
         PUSH_VERSION(br_versnum) );
#endif         /* NNSQUAD or not */      

   err = alert_init();    /* init alert objects queue */
   if(err)
   {
       dprintf("Alert object init error %d\n", err);
       return err;
   }

   br_log(0, "********* Buster (%s) startup *********\n", systemname);
   err = pk_init();   /* init the packet queues */
   if(err)
   {
       char * pberr = "Can't init packet buffers\n";
       br_log(0, pberr);
       return err;
   }

   err = br_initparms();
   if(err)
   {
       char * cferr = "Can't read configuration file.\n";
       br_log(0, cferr);
       return err;
   }

   nulliph = iphost_new(0, 0);

#ifdef BR_COMMUNITIES
   err = br_initcomm();
   if(err)
   {
      dprintf("Error %d starting communities\n", err );
      return err;
   }
#endif  /* BR_COMMUNITIES */

#ifdef PI_SITEADVISOR
   err = sa_init();
   if(err)
   {
      dprintf("Error %d starting Site Advison plug-in\n", err );
      return err;
   }
#endif  /* PI_SITEADVISOR */

   /* Call portability routine to start the cticks timer */
   br_startcticks();

#ifdef WEBIO_GUI
   httpport = 88;
   wi_rootfile = "index.htm";
   wi_execfunc = buster_execfunc;
   err = wi_init();
   if(err)
   {
      dprintf("Error %d starting web server\n", err );
      return err;
   }
   /* start thread to drive server */
   err = TH_NEWTASK(&webio_thread);
   if(err)
   {
      dprintf("Can't start Web server thread");
      return ENP_RESOURCE;
   }
#endif

   return 0;
}


int
br_startnet()
{
   int   i;

   /* Start the interfaces, do initial arps for routers */
   for(i = 0; i < (int)ifNumber; i++)
   {
      br_setipinfo(i);
      if( nets[i]->n_router )
         send_arp_net( nets[i], nets[i]->n_router );
   }

#ifndef EMBEDDED_DEVICE
   br_check();
   br_powerresume();    /* start probing local LAN, etc. */
#endif /* EMBEDDED_DEVICE */

   return 0;
}


/* returns TRUE if the two IP addresses are on the same subnet */

int
same_subnet(ip_addr ip1, ip_addr ip2)
{
   u_long subnet;    /* CIDR subnet of 1st address */

   subnet = htons(ip1);
   if(subnet >= 192)       /* class C address? */
      subnet = 0xFFFFFF00;
   else if(subnet >= 128)  /* class B address? */
      subnet = 0xFFFF0000;
   else                    /* must be class A address */
      subnet = 0xFF000000;

   /* compare the two IP's net numbers based on the subnet of the first
    * address. If the addresses are in different classes, then this test
    * will always return FALSE, else the subnet is valid for both IPs
    * and the test will be valid.
    */
   if((subnet & htons(ip1)) == (subnet & htons(ip2)))
      return TRUE;
   else
      return FALSE;
}

void
br_shutdown(void)
{
   br_log(0, "%s shutting down...\n", br_appname);

   br_setmode(OFFLINE);

   return;
}


/* cknet_callback()
 * 
 * Timer calback routine for checking for a change in LAN setup
 */

static int reentery = 0;

void
cknet_callback()
{
    int      check;

    /* This routine can take over one second, (especially under
     * the debugger! so guard against re-entry.
     */
    if(reentery)
        return;
    reentery++;

    /* The "timer" is set if we're in a short interval after startup or
     * hibernation. If we're outside of the timer, don;t bother with this.
     * The porting code is responsible for setting the timer after the
     * appropriate events.
     */
    if( !br_checknet_timer )
    {
        reentery--;
        return;
    }

    /* Don't worry about this if there is no IP set yet */
    br_setipinfo(br_ifview);

    if(nets[br_ifview]->n_ipaddr == 0)
    {
        reentery--;
        return;
    }

    /* call the checknet function - see if LAN has chaged */
    check = br_checknet(br_ifview);    /* check iface */

    if(br_sysstate == COLDSTART)
    {
        if(check != 0)
        {
            check = 0;     /* No stale LAN to get rid of */
        }
    }
    else   /* See if the LAN has changed */
    {
        /* if we cannot tell, wait till next tick and try again */
        if(check == 1)
        {
            reentery--;
            return;
        }
    }

    /* We got a result (changed or not) either way kill the timer */
    br_checknet_timer = 0;              /* clear timer/flag */

    /* see if we know the router for the IP net yet */
    if( !(brstates[br_ifview].evflags & EV_ROUTERSET))
    {
        reentery--;
        return;
    }

    /* If network has NOT changed just return */
    if(check == 0)
    {
        reentery--;
        return;
    }

    br_lanchanged();  /* call portability routine for net restart */

    /* Fall to here if we're on a new network */
    br_clearrouter(br_ifview);

    /* clear LAN name */
    brstates[br_ifview].current_lan[0] = 0;      
    brstates[br_ifview].evflags &= ~EV_NETNAMED;

    /* One more try selecting the best iface */
    br_autoselect();

    reentery--;
    return;
}



/* br_powerresume()
 * 
 * Called on program startup, and also after power is resumed 
 * on a system that can "hibernate",
 * such as a laptop or cell phone.
 */

void
br_powerresume()
{
   u_long   uptime;
   int      iface;

   /* MS-Windoze tends to spray us with multiple duplicate power 
    * resume messages, so Make sure we have not done this in the 
    * last few minutes.
    */
   uptime = br_gettime();
   if(br_lastresume && ((uptime - br_lastresume) < SYS_STARTUP_LAG))
   {
      return;
   }
   br_lastresume = uptime;

   if(br_downtime == 0)    /* Zero "downtime" means fresh startup */
   {
      br_sysstate = COLDSTART;
      br_sleeptime = 0;
   }
   else    /* Nonzero "downtime" means we're resuming */
   {
      br_sysstate = WAKEUP;

      /* This is a resume, (not a fresh program start) so adjust 
       * cticks forward for the time we were out. 
       */
      br_sleeptime = uptime - br_downtime;
      cticks += (br_sleeptime * TPS);
   }

   /* Reduce the confidence of the routers. If the router assigned by 
    * DHCP is different than before the power cycle we want the new 
    * one to replace the old.
    */
   for(iface = 0; iface < (int)ifNumber; iface++ )
   {
      if(brstates[iface].br_rtconf > 2)
         brstates[iface].br_rtconf--;
   }

   /* Re-read the system's IP addressing info - it may have
    * changed during hibernbation or restart.
    */
   br_setipinfo(iface);

   if(ifNumber > 1)
   {
      /* See if we should shift our attention to another interface. */
      br_autoselect();
   }

   /* Flag that we have moved to a new network. */
    br_checknet_timer = TRUE;
   
    br_running();  /* Start up messaging system, etc. */

   return;
}



/* br_checknet() - Check LAN against the local database to see if 
 * we have moved to another LAN. This is called after mobile 
 * devices (Windows laptops) wake from "hiberbnation" modes.
 *
 * Returns: 0 if we apprear to be on the same old LAN 
 *         -1 if apparently NOT on the old LAN
 *          1 if the network is (still) not up
 */

int
br_checknet(int iface)
{
   struct arpentry * arptab;
   u_long   uptime;     /* Time since start/resume, in cticks */
   ip_addr  tmp;
   int      err;

   /* Ensure a router IP address is already known */
   if(nets[iface]->n_router == 0)
      return 1;      /* We didn't have router info */ 

   /* If we didn't have a router previously then we don't
    * know if we're on the same old LAN or not. 
    * Assume it changed:
    */
   if((brstates[iface].br_router == NULL) ||
      (brstates[iface].br_router->iphost == NULL))
   {
      /* Send arp to probe for new router  */
      arptab = send_arp_net(nets[iface], nets[iface]->n_router);
      return -1;
   }

   /* If something (dhcp?) has changed the router IP address, assume it's a 
    * different LAN 
    */
   if((nets[iface]->n_router != brstates[iface].br_router->iphost->ipaddr))
   {
      return -1;
   }

   /* ARP won't work until we have an IP address (i.e. we're
    * doing DHCP), so hold off.
    */
   if(nets[iface]->n_ipaddr == 0)
      return 1;

   /* If we just resumed, recheck the IP address on the iface - Windows may 
    * still be messing with it. 
    */
   uptime = br_gettime() - br_lastresume;

arpagain:
   /* arp for the pre-existing router. */
   arptab = send_arp_net(nets[iface], brstates[iface].br_router->iphost->ipaddr);
   if(arptab == NULL)      /* coundn't send the ARP request? */
   {
      return 1;      /* network probably not up */
   }

   TH_SLEEP(TPS);    /* give router time to reply */

   /* See if the router still has the expected MAC address.
    * If the ARP entry has no hardware address (all zeros)
    * it means no ARP reply, which means router has changed. 
    * (or is just MIA). A new MAC addres means a different 
    * router which has the same IP address. In either case, 
    * assume we're on a new LAN.
    */
   if(MAC_EQ(arptab->macaddr, brstates[iface].br_router->macaddr))
      return 0;      /* MACs match, router is the same one */

   /* Update - sometimes routers just won't ARP (glitch?). If
    * this happens DON'T assume new network.
    */
   if(MAC_EQ(arptab->macaddr, nullmac))
      return 1;

   if(uptime < (3 * 60))   /* uptime is in seconds */
   {
      tmp = nets[iface]->n_ipaddr;        /* Save existing IP */
      err = br_setipinfo(iface);          /* Re-read system info */
      if(err)                             /* Can't read it, retry later */
         return 1;
      if(tmp != nets[iface]->n_ipaddr)    /* did it change? */
         goto arpagain;                   /* Try again with new address */
      return 1;
   }

   dprintf("checknet: Changed LAN. Uptime: 0x%x, br_lastresume: 0x%x\n", 
      uptime, br_lastresume);   

   return -1;
}


/* br_setiface()
 *
 * Set the passed iface newif to be the current active display interface.
 */

void
br_setiface(int * setif, int newif)
{

   if((newif < 0) || (newif > (int)ifNumber))
   {
      dtrap();
      br_log(0, "br_setiface: bad newif %d.", newif);
      return;
   }
   if(newif == *setif)     /* Just re-setting the current value? */
   {
      dtrap();
      return;
   }

   /* If we're about to change the view interface, then save existing 
    * LAN data, flush in-ram DB 
    */
   if(setif == &br_ifview)
   {
      if(brstates[br_ifview].evflags & EV_LANSTALE)   /* and only if something changed */
      {
         brstates[br_ifview].evflags &= ~EV_LANSTALE;  /* clear stale flag */
         br_writelan();    /* write lan file info */
      }

      /* Start the LAN discovery process on new iface */
      br_checknet_timer = TRUE;
   }

   /* set the Interface */
   *setif = newif;

   /* recheck the IP parameters on this iface */
   br_setipinfo(newif);

   /* Verify LAN info for the new network */
   br_checklan();

   /* write the non-volitile parameters file */
   br_saveparms();

   return;
}

static char piret[18];

char *  print_ipad(unsigned long ipaddr)
{
    sprintf( piret, "%u.%u.%u.%u", PASS_IPADDR(ipaddr) );
    return piret;
}

PACKET   pk_alloc(unsigned psize)
{
    PACKET  pkt = NULL;

    if(psize > largebufsize)    /* check for oversize */
        return NULL;
    
    if(psize == 0)
    {
        pkt = getq(&emptybufq);
        if(pkt)
            pkt->pk_flags |= PF_EMPTY;
    }
    if(!pkt && (psize < smallbufsize))
    {
        pkt = getq(&smallbufq);
    }
    if(!pkt)
        pkt = getq(&largebufq);

    if(!pkt)
        return NULL;

    pkt->pk_prot = pkt->pk_buff;
    pkt->pk_flags &= ~PF_FREE;

    return pkt;
}

void    pk_free(PACKET  pkt)
{
    pkt->pk_type = 0;
    pkt->pk_tstamp = 0;
    pkt->pk_net = NULL;

    if(pkt->pk_flags & PF_FREE)
    {
       dtrap();
       return;
    }
    pkt->pk_flags |= PF_FREE;

    if(pkt->pk_flags & PF_EMPTY)
    {
        pkt->pk_buff = NULL;
        pkt->pk_blen = 0;
        putq(&emptybufq, pkt);
        return;
    }
    if(pkt->pk_blen == largebufsize)
    {
        putq(&largebufq, pkt);
        return;
    }
    if(pkt->pk_blen == smallbufsize)
    {
        putq(&smallbufq, pkt);
        return;
    }
    dtrap();    /* Bad size in packet - corrupt? */
    return;
}


/* list of all packets, for debugging leaks */
PACKET *    br_pktbufs;
static int  firstlarge;  /* for debug */

int
pk_init()
{
    unsigned   i;
    PACKET     pkt;
    int        pkts = 0;

    if(largebufq.q_min)
    {
        dtrap();    /* double init call */
        return 0;
    }

    br_pktbufs = (PACKET*)npalloc( 
       (emptybufs + smallbufs + largebufs) * sizeof(PACKET));
    if(!br_pktbufs)
       return ENP_NOMEM;

    /* Populate free queue of empty structures */
    for(i = 0; i < emptybufs; i++)
    {
        pkt = PK_ALLOC(sizeof(*pkt));
        if(!pkt)
            return ENP_NOMEM;
        br_pktbufs[pkts++] = pkt;
        pkt->pk_flags |= (PF_EMPTY | PF_FREE);
        putq(&emptybufq, pkt);
    }
    emptybufq.q_min = emptybufs;

    /* Populate free queue of small packets */
    for(i = 0; i < smallbufs; i++)
    {
        pkt = PK_ALLOC(sizeof(*pkt));
        if(!pkt)
            return ENP_NOMEM;
        br_pktbufs[pkts++] = pkt;
        pkt->pk_buff = PKS_ALLOC(smallbufsize);
        if(!pkt->pk_buff)
            return ENP_NOMEM;
        pkt->pk_blen = smallbufsize;
        pkt->pk_flags |= PF_FREE;
        putq(&smallbufq, pkt);
    }
    smallbufq.q_min = smallbufs;

    /* Populate free queue of large packets */
    firstlarge = pkts;  /* for debug */

    for(i = 0; i < largebufs; i++)
    {
        pkt = PK_ALLOC(sizeof(*pkt));
        if(!pkt)
            return ENP_NOMEM;
        br_pktbufs[pkts++] = pkt;
        pkt->pk_buff = PKS_ALLOC(largebufsize);
        if(!pkt->pk_buff)
            return ENP_NOMEM;
        pkt->pk_blen = largebufsize;
        pkt->pk_flags |= PF_FREE;
        putq(&largebufq, pkt);
    }
    largebufq.q_min = largebufs;

    return 0;
}

void mk_iphdr(PACKET pkt, unsigned pid, u_char prot, unshort fragword, int len)
{
    struct ip * pip;

    pip = (struct ip *)(pkt->pk_prot - sizeof(struct ip));
    pip->ip_ver_ihl = 0x45;
    pip->ip_flgs_foff = fragword;
    pip->ip_time = IP_TTL;
    pip->ip_id = htons(pid);
    pip->ip_len = htons(len);
    pip->ip_tos = 0;
    pip->ip_prot = prot;
    pip->ip_chksum = 0;     /* Clear before summing */
    pip->ip_chksum = ~cksum(pip, 10);

    return;
}

int
if_netnumber(NET net)
{
    int i;
    for(i = 0; i < MAXNETS; i++)
        if(nets[i] == net)
            return i;

    panic("netnum");
    return -1;    
}


/*
Copyright (C) 2003-2008 by Praemio Inc. All rights reserved.

Author: John Bartas - jbartas@praemio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/



