/* brout.c
 *
 * Buster packet output routines. 
 *
 * Copyright 2003-2008 by Praemio Inc. All rights reserved.
 * Author: John Bartas - jbartas@praemio.com
 * This software is released under the terms of the LGPL.
 * See LGPL Notice at end of this file; Full license in ../license.txt
 */


#include "busterport.h"
#include "brnet.h"
#include "ip.h"
#include "arp.h"
#include "dhcpdef.h"
#include "brdb.h"


int   rtdisc_sent = 0;     /* Router discovery counter */
int   rtdisc_tick = 0;     /* Time of last Router discovery */


/* prepend a UDP header to the passed packet.
 *
 * NOTES:
 * - This does NOT do a check of prepended space in pk_prot.
 * - UDP ports are passed in NET endian
 */

void
udp_bldhead(PACKET p, unshort fport, unshort lport)
{
   struct udp * pup;

   /* prepend UDP header to packet data */
   p->pk_prot -= sizeof(struct udp);
   p->pk_plen += sizeof(struct udp);

   pup = (struct udp*)p->pk_prot;

   pup->ud_srcp = lport;
   pup->ud_dstp = fport;
   pup->ud_len = htons(p->pk_plen);
   pup->ud_cksum = 0;      /* clear checksum field - no sum */

   return;
}



/* Send ICMP router discovery packet. Source IP address is Buster.
 *
 * Returns 0 if OK, or ENP_ error code
 */

int
br_rtdisc(int iface)
{
   PACKET         pkt;
   struct icmp *  icmp;
   struct ip *    pip;
   int            err;

   /* make sure the Buster iface IP address is set */
   if(nets[iface]->n_ipaddr == 0)
   {
      dtrap();
      return ENP_LOGIC;
   }

   pkt = pk_alloc(24);
   if(!pkt)
      return ENP_NOBUFFER;

   pip = (struct ip *)pkt->pk_prot;
   icmp = (struct icmp *)(pip + 1);

   icmp->type = 10;
   icmp->code = 0;
   icmp->chksum = ~cksum(icmp, 2);

   pip->ip_src = nets[iface]->n_ipaddr; /* from Buster IP */
   pip->ip_dest = 0xFFFFFFFF;     /* broadcast */

   /* Can't use mk_iphdr() because we need TTL==1 */
   pip->ip_ver_ihl = 0x45;       /* 2 nibbles; VER:4, IHL:5. */
   pip->ip_flgs_foff = 0; /* fragment flags and offset */
   pip->ip_time = 1;
   pip->ip_id = htons(ip_id);   /* IP datagram ID */
   pip->ip_len = htons(24);
   pip->ip_tos = 0;
   pip->ip_prot = 1;          /* install protocol ID (TCP, UDP, etc) */
   pip->ip_chksum = 0;      /* clear checksum field for summing */
   pip->ip_chksum = ~cksum(pip, 10);

   ip_id++;    /* bump private IP ID generator */

   pkt->pk_net = nets[iface];
   err = send_via_arp(pkt, 0xFFFFFFFF);   /* broadcast the packet */

   if(err == 0)
   {
      rtdisc_sent++;
      rtdisc_tick = cticks;
   }

   return err;
}

/* br_icmpquench() - Send a source quench to a TCP connection's remote host.
 *
 * Connection passed is TCP con to quench
 * rxpkt is last packet from node to quench
 *
 * Returns 0 if OK or ENP_ error.
 */

int
br_icmpquench(struct conn_entry * conn, PACKET rx_pkt)
{
   int         err = 0;
   int         iface;
   PACKET      pkt;
   struct ip * pip;
   struct quench *   icmp;

   iface = conn->ce_iface;

   /* make sure the Buster iface IP address is set */
   if(nets[iface]->n_ipaddr == 0)
   {
      dtrap();
      return ENP_LOGIC;
   }

   /* get buffer big enough for quench (72), IP header (20) & MAC: */
   pkt = pk_alloc(sizeof(struct ip) + sizeof(struct quench) + MAC_HDRLEN);
   if(!pkt)
      return ENP_NOBUFFER;

   pip = (struct ip *)pkt->pk_prot;
   icmp = (struct quench *)(pip + 1);

   icmp->ptype = 4;     /* ICMP quench packet type */
   icmp->pcode = 0;
   MEMCPY(icmp->data, ip_header(rx_pkt), ICMPQDATA);

   icmp->pchksum = ~cksum(icmp, ICMPQUENCH_SIZE/2);

#ifdef  NOTYET /* !!! */
   pip->ip_src = nets[iface]->n_ipaddr; /* from Buster IP */
   pip->ip_dest = 0xFFFFFFFF;     /* broadcast */

   mk_iphdr( );

   pkt->net = nets[iface];
   err = ip2mac(pkt, );   /* send the packet */
#endif

   return err;
}


/*
Copyright (C) 2003-2008 by Praemio Inc. All rights reserved.

Author: John Bartas - jbartas@praemio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/



