/* brrx.c
 *
 * Buster received packet handler. 
 *
 * Copyright 2003-2008 by Praemio Inc. All rights reserved.
 * Author: John Bartas - jbartas@praemio.com
 * This software is released under the terms of the LGPL.
 * See LGPL Notice at end of this file; Full license in ../license.txt
 */


#include "busterport.h"
#include "brnet.h"
#include "arp.h"
#include "ip.h"
#include "dhcpdef.h"
#include "brdb.h"
#include "brdns.h"


int   br_dnsparse(PACKET rxpkt, struct ip * pip, struct conn_entry * conn);
int   br_rxdhcp(struct host_entry * host, struct ip * pip, int iface);

/* recever thread flag & reentry guard */
int   br_inreceiver = 0;


/* Javascript sequence to do a page close w/o the html header */
char closeonly[] = {"<SCRIPT LANGUAGE=\"JavaScript\">\r\n<!--\r\n"
"if(window.opener != null) \r\n   window.close(self);\r\n // --></SCRIPT></html>"};

int closeonly_length = sizeof(closeonly);


/* find_opt()
 *
 * Helper routine to search a DHCP formatted options string for a 
 * particular option code.
 *
 * Returns a pointer to the code if found, else NULL
 */

u_char * find_opt(u_char code, u_char * options)
{
    u_char * end   =  options  +  DHCP_OPTSIZE;  /* limit scope of search */

    while (options < end)
    {
        if (*options == code)       /* succes */
            return options;
        if (*options == DHOP_END)   /* End of options; code not found */
            return NULL;            /* proper exit point */
        if (*options == DHOP_PAD)   /* PAD is 1 byte option */
            options++;
        else     /* all other options have length field */
            options += (*(options+1))+2;
    }
    dtrap();    /* This is a format error */
    return NULL;
}


/* Buster Promisc mode receive packet handler */

int
br_rxpkt(int iface, PACKET rxpkt)
{
   struct eth_hdr *   eth;
   struct ip *       pip;     /* IP header below this layer */
   struct udp *      pup;     /* UDP header */
   struct tcphdr *   ptcp;    /* TCP header */
   struct icmp *     icmp;    /* ICMP header*/
   struct arp_hdr * arp;     /* arp header */
   unshort           etype;   /* type from ethernet header */
   unshort           port;    /* from UDP or TCP header - scratch */
   int               fromremote;    /* TRUE if packet is from remote */
   struct host_entry *     sender;
   struct conn_entry *     conn;
   u_char *          data;    /* protocol data */
   int               dlen;    /* protocol length */

   /* Keep weighted average of recent traffic activity */
   nets[iface]->traffic++;
   nets[iface]->ifOperStatus = 1;
   nets[iface]->rxpkts++;    /* Keep Buster stats */

   /* set type of rxpkt */
   eth = (struct eth_hdr *)(rxpkt->pk_buff);
   etype = htons(eth->e_type);

   arp = NULL;    /* clear arp pointer means packet is not arp */
   pip = NULL;    /* clear IP pointer means packet is not IP */

   if(etype == 0x0800)     /* regular healthy IP type */
   {
      pip = ip_header(rxpkt);       /* Set IP header pointer */
      nets[iface]->rxip++;    /* Keep Buster stats */
   }
   else if(etype == 0x0806)    /* ARP */
   {
      arp = (struct arp_hdr *)(rxpkt->pk_buff + 14);
      nets[iface]->rxarp++;    /* Keep Buster stats */
      arp_rcv(rxpkt);
   }
   else if(etype < 0x600)    /* IEEE 802.3 - ughh! */
   {
      char * mymac = nets[iface]->n_macaddr;

      nets[iface]->rx8023++;    /* Keep Buster stats */

      /* See if it's an echo from ourslf */
      if((eth->e_src[5] != *(mymac + 5) ) &&
         (eth->e_src[4] != *(mymac + 4) ) &&
         (eth->e_src[4] != *(mymac + 3) ) &&
         (eth->e_src[4] != *(mymac + 2) ) &&
         (eth->e_src[4] != *(mymac + 1) ) &&
         (eth->e_src[4] != *(mymac + 0) ))
      {
         struct snap_hdr * barf;

         /* assume 802.5 snap header exactly follows 802.3 */
         barf = (struct snap_hdr *)(eth + 1);
         etype = htons(barf->type);

         if(barf->llc_etc[0] != 0xAA)
         {
            dprintf("Not SNAP, type==0x%02x \n", (u_char)barf->llc_etc[0] );

            /* sanity check length */
            if((etype >= 46) &&
               ((etype + 14) != (int)(rxpkt->pk_plen)))
            {
               dprintf("802.3 len:%d actual:%d\n", (etype + 14), rxpkt->pk_plen);
            }
            /* not SNAP, ignore it */
            pk_free(rxpkt);
            return 0;
         }

         if(etype == 0x0800)
         {
            pip = (struct ip *)(barf + 1);
            arp = NULL;
            nets[iface]->rxip++;    /* Keep Buster stats */
         }
         else if(etype == 0x0806)
         {
            arp = (struct arp_hdr *)(barf + 1);
            pip = NULL;
            nets[iface]->rxarp++;    /* Keep Buster stats */
         }
         else
         {
            /* not IP, ignore it */
            pk_free(rxpkt);
            nets[iface]->rxbad8023++;    /* Keep Buster stats */
            return 0;
         }
      }
      else     /* Echo of our own pkt (e.g. ARP?), ignore it. */
      {  
         pk_free(rxpkt);
         return 0;
      }
   }
   else if(etype == 0xFFFF)   /* Listed as "reserved" in specs... */
   {
      /* My Linksys/Cisco WRT54G sometimes sends out what appear to
       * be IP packets with FFFF in the type field instead of 0800.
       * These packets are often (usually?) part of a normal Ethernet-II 
       * data stream. WiFi wierdness???.
       */
      if(*(rxpkt->pk_buff + 14) == 0x45)
      {
         etype = 0x800;
         pip = ip_header(rxpkt);      /* Set IP header pointer */
      }
      nets[iface]->rxFFFF++;    /* Keep Buster stats */
   }
   else
   {
      nets[iface]->rxbadtype++;    /* Keep Buster stats */
   }

   if(br_inreceiver++)  /* bump reentry counter */
   {
//      dtrap();
   }

   /* Don't update tables for packets until we have an IP on this iface, 
    * otherwise  remote/local logic (et.al.) gets confused. Revisit this
    * idea later if we do an applicance. -JB-
    */
   if(nets[iface]->n_ipaddr == 0L)
   {
      nets[iface]->rx_noip++;    /* Keep Buster stats */
      alarm_general(&no_ipaddr, &brstates[iface].noip, 
         nets[iface]->description );
      pk_free(rxpkt);
      br_inreceiver--;
      return 0;
   }

   /* If Buster network receiving has been disabled on this
    * iface, just dump packet.
    */
   if(brstates[iface].rx_disabled)
   {
      nets[iface]->rx_disabled++;    /* Keep Buster stats */
      pk_free(rxpkt);
      br_inreceiver--;
      return 0;
   }

   sender = NULL;
   if(pip)
   {
      /* Sometimes the idiot Windows stack gratuitously changes IP 
       * addresses when coming back from hibernation.This could 
       * legitimatly be due to DHCP or user change, but it also happens 
       * in goofy ways - For example, sometimes my laptop boots with a 
       * 10.x.x.x net address on the Wifi in a 192.168.x.x environment. 
       * This is wierd because no iface has recently been assigned a 
       * 10 net either manually or via DHCP. Anyway, It switches over to 
       * the correct 192 address in a few seconds, but not before Buster 
       * has read and recorded the bogus address. 
       *
       * Try to detect this here.
       */
      if( MAC_EQ(nets[iface]->n_macaddr, eth->e_src))
      {
         /* This packet is sourced from our MAC, check IP address */
         if((pip->ip_src != 0) &&      /* Don'trigger on DHCP packets */
            (pip->ip_src != nets[iface]->n_ipaddr))
         {
            /* Skip TCP spoofs generated by Buster. If this is a normal TCP 
             * reset from the Winsock stack, then the IP should match (above)
             * and we won't get here.
             */
            if(pip->ip_prot == 6)
            {
               /* If it's a Buster reset to a remote, skip it. */
               ptcp = (struct tcphdr *)ip_data(pip);
               if(ptcp->th_flags & TH_RST)
               {
                  br_inreceiver--;
                  pk_free(rxpkt);   /* just free the TCP buffer */
                  return 0;
               }

               /* If it's a packet from this MAC and to this MAC, it's probably
                * one of the popup closer messages, skip those too.
                */
               if( MAC_EQ(eth->e_dst, eth->e_src))
               {
                  br_inreceiver--;
                  pk_free(rxpkt);   /* just free the TCP buffer */
                  return 0;
               }
            }

            /* Yep, our IP changed; or else we have two IP addresses */
            nets[iface]->rx_ipcheck++;    /* Keep Buster stats */
            br_setipinfo(iface);       /* Re-read system info */
         }
      }

      /* update MAC host stats (and maybe even create host) for IP 
       * packets. ARP packets will get done below.
       */
      switch(pip->ip_prot)
      {
      case UDP_PROT:
         pup = (struct udp *)(ip_data(pip));
         if((pup->ud_srcp == 0) || (pup->ud_dstp == 0))
         {
            dtrap();
            break;
         }

        /* get service port (lower numbered of the two */
        port = htons(pup->ud_srcp);
        if(port > htons(pup->ud_dstp))
            port = htons(pup->ud_dstp);

        switch(port)
        {
            case BOOTP_SERVER_PORT:
            case BOOTP_CLIENT_PORT:
                sender = br_hostupdate((char*)(&eth->e_src[0]), 
                    pip->ip_src, H_DHCP, iface);
                break;
             case 127:  /* Microsoft Netboogers */
             case 128:
             case 129:
                sender = br_hostupdate((char*)(&eth->e_src[0]), 
                pip->ip_src, H_NETBEUI, iface);
                break;
            default:
                sender = br_hostupdate((char*)(&eth->e_src[0]), 
                    pip->ip_src, H_UDP, iface);
                break;
         }
         break;
      case 6:     /* TCP: */
         sender = br_hostupdate((char*)(&eth->e_src[0]), pip->ip_src, H_TCP, iface);
         break;
      case ICMP_PROT:
         sender = br_hostupdate((char*)(&eth->e_src[0]), pip->ip_src, H_ICMP, iface);
         break;
      default:
         sender = br_hostupdate((char*)(&eth->e_src[0]), pip->ip_src, H_OTHER, iface);
         break;
      }
   }


   if(arp)     /* handle ARPs here */
   {
      u_long   src_ip;  /* src & dest IP addresses, net endian */
      u_long   dest_ip;

      /* show ARP packets to host table builder. First extract IP
       * address from mis-aligned ARP header. If the ARP packet is ARP
       * request then check only the senders IP, else if reply then check 
       * both IP addresses. 
       */

      /* Macro to get net-endian IP address from 16 bit boundary */
#if (BYTE_ORDER == LITTLE_ENDIAN)
#define EXTRACT_ARPIP(cp) (((u_long)*(u_char*)cp) + \
                        ((u_long)*(u_char*)(cp + 1) << 8) + \
                        ((u_long)*(u_char*)(cp + 2) << 16) + \
                        ((u_long)*(u_char*)(cp + 3) << 24))
#else
#define EXTRACT_ARPIP(cp) (((*(unshort*)cp) << 16) | (*(unshort*)(cp + 2)))
#endif   /* BYTE_ORDER */

      src_ip = EXTRACT_ARPIP(&arp->data[6]);
      dest_ip = EXTRACT_ARPIP(&arp->data[16]);

#ifdef WINDOWS_HOSTED
#ifdef USE_AUTOIP
         /* On windows systems, screen out echos of our own Upnp ARP */
         if(dest_ip == autoIPs[iface].try_address)
            break;
#endif   /* USE_AUTOIP */
#endif   /* WINDOWS_HOSTED */

      /* make sure this is not a reply to a proxy ARP */
      if(islocal_ip(iface, dest_ip))
      {
         sender = br_hostupdate((char*)(&eth->e_src[0]), src_ip, H_ARP, iface);

         if(arp->ar_op == htons(0x0002))  /* reply? */
         {
            /* see if this is in response to a discovery ARP to our router */
            if(src_ip == nets[iface]->n_router)
            {
               struct host_entry * rtmac;

               rtmac = br_hostlookup((char*)&eth->e_src[0], iface);
               if(rtmac)
               {
                  if(rtmac->iphost->ipaddr == nets[iface]->n_router)
                  {
                     br_setrouter(iface, rtmac, 220);
                  }
                  else
                  {
                     /* misidentified rtmac->iphost, or incorrect IP address 
                      * assignd to router 
                      */
                     if( islocal_ip(iface, rtmac->iphost->ipaddr) == FALSE )
                     {
                        struct ip_hostentry * iphost;

                        /* Find or make IP host table entry for router */
                        iphost = br_iplookup(iface, src_ip);
                        if(iphost == NULL)
                           iphost = iphost_new(iface, src_ip);

                        if(iphost)
                        {
                           rtmac->iphost = iphost;
                           br_setrouter(iface, rtmac, 120);
                        }
                     }
                  }
               }
               else
               {
                  dtrap();    /* out of heap memory? */
               }
            }
         }
      }
   }

   
   /* look up host, bump packet and byte counts. Sender is most often 
    * for IP and ARP packets in the above logic, but check first.
    */
   if(sender == NULL)
      sender = br_hostlookup((char*)&eth->e_src[0], iface);

   if(sender)
   {
      sender->sec_pkts++;
      sender->sec_bytes += rxpkt->pk_plen;
      sender->active = cticks;

      /* If sender IP host is known,make sure IP's back pointer to MAC is set */
      if((sender->iphost) && (sender->iphost->mac_host == NULL))
      {
         sender->iphost->mac_host = sender;
      }
   }


   if(pip)
   {

#ifdef NO_LOCAL_ECHO
         /* On windows systems, screen out echos of our own packets */
         if(pip->ip_src == nets[iface]->n_ipaddr)
         {
            pk_free(rxpkt);
            br_inreceiver--;
            return 0;
         }
#endif   /* NO_LOCAL_ECHO */

      switch(pip->ip_prot)
      {
      case 1:
         icmp = (struct icmp *)(ip_data(pip));
         br_cnupdate(pip, (char *)eth, icmp, H_ICMP, iface, &fromremote);
         if(icmp->type == ECHOREP)
         {
            /* see if this is a reply from the router - we may ping it at 
             * init time. If so, set the router's MAC entry.
             */
            if((pip->ip_src == nets[iface]->n_router) && 
               (brstates[iface].br_router == NULL))
            {
               struct host_entry * rtmac;
               rtmac = br_hostlookup((char*)&eth->e_src[0], iface);
               if(rtmac)
                  br_setrouter(iface, rtmac, 200);
               else
                  dtrap();
            }
         }
         break;

      case UDP_PROT:    /* UDP protocol */
         pup = (struct udp *)(ip_data(pip));
         if((pup->ud_srcp == 0) || (pup->ud_dstp == 0))
         {
            dtrap();
            break;
         }
         conn = br_cnupdate(pip, (char *)eth, pup, H_UDP, iface, &fromremote);

         if(pup->ud_dstp == htons(1026))    /* It's a Microsoft popup? */
         {
            br_log(0, "Microsoft popup, allegedly from IP %u.%u.%u.%u\n", 
               PASS_IPADDR(pip->ip_src) );
            break;
         }

         /* special case for DHCP packets */
         if((pup->ud_dstp == htons(67)) ||      /* DHCP packet sent to server port */
            (pup->ud_dstp == htons(68)))        /* DHCP packet sent to client port */
         {
            /* check for valid DHCP packet */
            if((pup->ud_srcp != htons(68)) &&
               (pup->ud_srcp != htons(67)) &&
               (iface == br_ifview))
            {
               sprintf(brmsg, "Bad source port (%d) in DHCP packet.\n "
                  "Sender %s; length:%d)\n", 
                  htons(pup->ud_srcp),
                  print_mac(eth->e_src),
                  rxpkt->pk_plen);

               bad_dhcp.parm = pip;
               alarm_general(&bad_dhcp, &baddhcp, brmsg);
               break;
            }
            /* Extract what information we can from packet */
            if(!sender)
               sender = br_hostlookup((char*)&eth->e_src[0], iface);
            if(sender)
               br_rxdhcp(sender, pip, iface);
         }

         if(pup->ud_dstp == htons(msgport))
         {
             rxpkt->pk_prot = (char*)(pup +1);
             bg_callback(rxpkt, pip);
         }

         /* special case for DNS packets */
         if((pup->ud_dstp == htons(53)) || /* DNS packets */
            (pup->ud_srcp == htons(53)))
         {
            /* run it by Buster DNS spoofer */
            if(br_dnsparse(rxpkt, pip, conn) == 1)
            {
               br_inreceiver--;
               return 0;      /* we replied, done with pkt */
            }
            break;      /* send other DNS pkts to stack */
         }
         break;
      case 6:     /* TCP protocol */

         ptcp = (struct tcphdr *)(ip_data(pip));
         conn = br_cnupdate(pip, (char *)eth, ptcp, H_TCP, iface, &fromremote);

         if(conn == NULL)     /* if no connection, quit now */
            break;

         /* rest of the TCP logic does not apply to packet to/from 
          * Buster's web server. 
          */
         if((pip->ip_src == nets[iface]->n_ipaddr) &&
            (ptcp->th_sport == htons(HTTP_PORT)))
         {
            break;    /* source is Buster web server */
         }
         if((pip->ip_dest == nets[iface]->n_ipaddr) &&
            (ptcp->th_dport == htons(HTTP_PORT)))
         {
            break;    /* dest is Buster web server */
         }

         if(ptcp->th_flags & TH_RST)   /* Don't do anything with resets */
            break;

         /* If the connection is one we're trying to kill, try again */
         if(conn->ce_flags & CEF_RESETLOCAL)
         {
            br_tcpreset(conn, 1);
            break;
         }
         if(conn->ce_flags & CEF_RESETREMOTE)
         {
            /* Don't send these in response to spoof packets from ourself. */
            if( MAC_EQ(eth->e_src, nets[iface]->n_macaddr) == FALSE )
            {
               br_tcpreset(conn, 0);
            }
            break;
         }

         /* If is from the local host, and we already sent FIN, try again */
         if((conn->ce_flags & CEF_CLOSEDWIN) &&
            (pip->ip_src == conn->loc.ce_addr))
         {
            br_spooftcp(conn, (u_char*)closeonly, closeonly_length, 1, 
                (TH_ACK|TH_PUSH|TH_FIN) );
            break;
         }

         /* see if the IP address or port is restricted */
         if(br_tcpblock(conn, fromremote) == 2)
            break;      /* break if we reset connection */

         /* get TCP data pointer and length */
         dlen = ptcp->th_doff;       /* get offset from ptcp to data */
         dlen >>= 2;                 /* convert to bytes */
         data = ((u_char*)ptcp) + dlen;      /* find tcp data */
         dlen = htons(pip->ip_len) - (dlen + 20);   /* set actual data length */

         /* If there is content and server is not whitelisted, then scan HTML content */
         if(dlen > 0)
         {
            struct control_ip * cip;

            cip = ipe_lookup(htonl(conn->rem.ce_addr));
            if((cip == NULL) || 
               (ctl_getaccess(&cip->ctl) == 0))    /* no Access bit */
            {

               if(ptcp->th_sport == htons(80))
               {
                  br_serverhttp(conn, data, dlen);
               }
               else if(ptcp->th_dport == htons(80))
               {
                  br_clienthttp(conn, data, dlen);
               }
#ifdef POP3_MONITOR
               else if((ptcp->th_sport == htons(POP3_PORT)) || 
                     (ptcp->th_dport == htons(POP3_PORT)))
               {  
                  br_pop3(conn, data, dlen, pip->ip_dest, pip->ip_src);
               }
#endif /* POP3_MONITOR */
            }
         }

         /* see if we need to adjust for TCP options */
         if((ptcp->th_doff & 0xF0) != 0x50)
         {
            int optlen = ((ptcp->th_doff & 0xF0) >> 2) - 20;
            data += optlen;
            dlen -= optlen;
         }
         if(dlen <= 0)
            break;

#ifdef VIRUS_CHECK
         /* scan data for virus */
         if((br_cfg.cflags & (BC_VIRUS_ALERT|BC_VIRUS_KILL)) && (dlen > 0))
         {
            struct   pattern_entry * pep;
            struct endpoint * ep;

            /* get endpoint which sent the packet, for last packet info */
            if(fromremote)
               ep = &conn->rem;
            else
               ep = &conn->loc;

            /* If we got a partial pattern match last time, check the rest: */
            if((ep->partial) && 
              (dlen >= ep->pep->patlen))   /* make sure we have enough data */
            {
               if(MEMCMP(data, ep->pep->pattern, ep->pep->patlen) == 0)
               {
                  found_virus(conn, ep->pep);
                  break;
               }
            }

            if((dlen + ep->pdlen) < PATTERN_LENGTH)
            {
               if((dlen + ep->pdlen) < 0)
               {
                  dtrap();
               }
               /* not enough data to bother with a search, */
               MEMCPY(&ep->pdata[ep->pdlen], data, dlen);
               ep->pdlen += dlen;
               break;
            }
            else  /* we have at least PATTERN_LENGTH bytes of packet data */
            {

               /* scan prefix and data */
               pep = br_patscan(data, dlen, ep);

               if(pep && (ep->partial == 0))
               {
                  switch(pep->pe_type)
                  {
                     case PET_VIRUS:
                        found_virus(conn, pep);
                        break;
                     default:
                        dtrap();    /* bad type in pattern hash table */
                  }
               }
               /* save last PATTERN_LENGTH bytes in connction */
               if(dlen >= PATTERN_LENGTH)
               {
                  /* all trailing data is in new packet */
                  MEMCPY(ep->pdata, data + (dlen - PATTERN_LENGTH), PATTERN_LENGTH);
               }
               else  /* trailing data is split between pdata & new packet */
               {
                  int preserve = PATTERN_LENGTH - dlen;

                  MEMMOVE(ep->pdata, ep->pdata + (ep->pdlen - preserve), preserve);
                  MEMCPY(ep->pdata + preserve, data, PATTERN_LENGTH - preserve);
               }
               ep->pdlen = PATTERN_LENGTH;
            }
         }
         break;
#endif  /* VIRUS_CHECK */

      default:
         break;
      }
   }

   pk_free(rxpkt);
   br_inreceiver--;
   return 0;
}


/* nameset() - helper routine for br_rxdhcp()
 *
 * Copies a name string from a DHCP options list into a host_entry
 * name field.
 */

static void
nameset(int opcode, int pkttype, struct host_entry * client, char * opts, int optlen)
{
   int      copylen;
   char     tmp;
   char *   namebuf;    /* pointer to iphost name buffer to use */

   /* Only do this for request packets */
   if(pkttype != DHCP_REQUEST)
      return;

   if((client->h_flags & H_NAMEMASK) > H_DHCPNAME)
      return;

   client->h_flags &= ~H_NAMEMASK;

   /* get name buffer based on opcode */
   switch(opcode)
   {
      case DHOP_NAME:
         namebuf = client->locname;
         client->h_flags |= H_DHCPNAME;
         break;
      case DHOP_DOMAIN:
         dtrap();
         namebuf = client->locname;
         client->h_flags |= H_DHCPNAME;
         break;
      case DHOP_CLASSID:   /* we already processed this - just advance pointer */
         namebuf = client->iphost->classId;
         break;
      case DHOP_DNSNAME:
         namebuf = client->iphost->dnsname;
         break;
      default:
         panic("nameset");
         return;
   }

   /* record if we're changing a name */
   if((*namebuf) && 
      (strncmp(namebuf, (char*)opts, optlen)))
   {
      tmp = *(opts + optlen);
      *(opts + optlen) = 0;      /* terminate  option for printing */

      dprintf("DHCP code %d replacing client name %s with %s\n",
         opcode, namebuf, opts);
      *(opts + optlen) = tmp;    /* restore trailing byte */
   }

   /* finally, save the name. */
   copylen = min(optlen, DNSNAMELEN);
   MEMSET(namebuf, 0, DNSNAMELEN);
   strncpy(namebuf, opts, copylen);

   return;
}


/* brrt_fromdhcp()
 *
 * On small LANs the DHCP server is often the router as well. Use
 * received DHCP packets as an opportunity to discover and set the router.
 *
 */

void
brrt_fromdhcp(int iface, struct host_entry * sender, struct bootp * bp)
{
   u_char * opts;       /* scratch options pointer */
   ip_addr  ipaddr;

   /* If there is a router IP already set, see if the sending IP 
    * address IS that router.
    */

   if(nets[iface]->n_router)
   {
      if(sender->iphost->ipaddr == nets[iface]->n_router)
      {
         br_setrouter(iface, sender, 85);
      }
   }
   else     /* no router known, see if one is being handed out */
   {
      opts = find_opt(DHOP_ROUTER ,&bp->options[4]);
      if (opts)
      {
         opts += 1;  /* past type and len bytes */
         EXTRACT_IP_ADDR(opts, ipaddr);
         if(ipaddr)
         {
            /* Set the IP extracted from the option as the router, and
             * probe it with an ARP request.
             */
            nets[iface]->n_router = ipaddr;
            send_arp_net(nets[iface], nets[iface]->n_router);
         }
      }
   }

   return;
}


/* br_rxdhcp(iface, sender, struct udp * pup);
 * 
 * Extract information from received DHCP packet
 *
 * Returns 0 if OK, else negative ENP_ error code.
 */


int
br_rxdhcp(struct host_entry * sender, struct ip * pip, int iface)
{
   struct bootp * bp;
   int      dhcptype =  0;    /* DHCP type - not valid if bootp */
   u_char * opts;             /* scratch options pointer */
   u_char * end;
   u_long   addr;
   u_char   optlen;
   struct host_entry * client;
   struct udp * pup;

   pup = (struct udp *)(ip_data(pip));
   bp = (struct bootp *)(pup + 1);

   /* see if it's full DHCP or plain bootp by looking for dhcp type option */
   opts = find_opt(DHOP_TYPE ,&bp->options[4]);
   if (opts && *opts == DHOP_TYPE)
   {
      dhcptype = *(opts+2);
   }

   switch(dhcptype)
   {
   case DHCP_REQUEST:
      client = sender;
      break;
   case DHCP_ACK:
      /* On ACK from the server install the target as the client */
      client = br_hostlookup((char*)bp->chaddr, iface);

      /*. If the MAC host is unknown, punt. */
      if(!client)
         return 0;

      /* See if we can get router from DHCP */
      brrt_fromdhcp(iface, sender, bp); 

      if((bp->ciaddr) && 
         (client->iphost->ipaddr != bp->ciaddr))
      {
         dprintf("dhcp ACK updating actual client IP from %u.%u.%u.%u",
            PASS_IPADDR(client->iphost->ipaddr) );
         dprintf(" to %u.%u.%u.%u\n", 
            PASS_IPADDR(bp->ciaddr) );

         client->iphost = br_iplookup(iface, bp->ciaddr);
         if(!client->iphost)
         {
            dtrap();
            return ENP_PARAM;
         }
      }
      break;
   case DHCP_OFFER:
      /* See if we can get router from DHCP */
      brrt_fromdhcp(iface, sender, bp); 
      return 0;
   default:
      /* If it's not the final phases of the assignment, return */
      return 0;
   }

   /* See if server is assigning a new IP address to existing host */
   opts = find_opt(DHOP_CADDR ,&bp->options[4]);
   if(opts)
   {

      getlong((char*)(opts + 2), &addr);
      addr = ntohl(addr);


      /* See if the DHCP address is different than the one currently 
       * assigned to this MAC address. This usually means the DHCP 
       * server is changing an existing DHCP address, but may also
       * happen when a Windows based Buster (or similar entity) gets 
       * an IP address and the Winsock stack has allready obtained 
       * one.
       */
      if(client->iphost && (client->iphost->ipaddr != addr))
      {
#ifdef REPORT_BRAINDAMAGE
         dprintf("DHCP sent replacment ipaddr=%u.%u.%u.%u, ", 
            PASS_IPADDR(addr));
         dprintf("for old addr=%u.%u.%u.%u\n", 
            PASS_IPADDR(client->iphost->ipaddr)); 
#endif /* REPORT_BRAINDAMAGE */

         /* Great. Fracking great. Microsoft Windoze machines (at least my Win2K)
          * send DHCP REQ packets even if they have a static, permanent IP 
          * address. The DHCP REQ is asking for a local ip address (192.168.1.134)
          * Which is NOT the assigned address (192.168.1.200). Etheral shows 
          * the DHCP server (a Linksys WiFi router) sending back an ACK, for
          * the requested address, however the Win2K machine ignores the ACK 
          * and merrily goes on using 192.168.1.200. How is Buster supposed 
          * to know which IP address the Bozo Windows system ended up using?
          * 
          * My workaround is to record the REQ/ACK address and the time it 
          * was received. Subsequesnt recipts of IP datagrams from this host
          * will tell us if the IP address changed or not.
          */
         client->dhcp_addr = addr;

         /* Clear the duplicate IP address bit. The logic to resolve
          * client->dhcp_addr will set it again if required.
          */
         client->h_flags &= ~H_DUPLIP;
      }
      else  /* IP is just being refreshed, not changed */
      {
         /* if DUP flag is set, re-check this. */
         if( client->h_flags & H_DUPLIP)
         {
            struct host_entry * tmp;
            int   hashval;

            /* Start by clearing the bit, set it again below if needed */
            client->h_flags &= ~H_DUPLIP;

            /* search the MAC hosts list for duplicate IP addr */
            hashval = HASH_HOST(client->macaddr);
            for(tmp = hosttab[hashval]; tmp; tmp = tmp->next)
            {
               if(tmp->iface != client->iface)  /* skip other nets */
                  continue;
               if(tmp == client)     /* skip client's own entry */
                  continue;
               if(tmp->iphost->ipaddr == addr)
               {
                  client->h_flags |= H_DUPLIP;
                  break;
               }
            }
         }
      }
   }

   if(!client->iphost)
   {
      dtrap();
      /* This can happen when Windows-hosted Busters return 
       * from hibernation. Is this a problem??? -JB-
       */
      return ENP_LOGIC;
   }

   /* extract IP information */
   opts = &bp->options[4]; /* set options pointer */
   end = opts + DHCP_OPTSIZE;

   while (opts <= end)
   {
      switch (*opts++)
      {
      case DHOP_TYPE:   /* silently ignore, skip data */
      case DHOP_CLIENT:
      case DHOP_CADDR:
      case DHOP_DUMPF:
      case DHOP_TIMES:
      case DHOP_SERVER:
      case DHOP_REQLIST:
         opts += ((*opts) + 1);
         break;
      case DHOP_USERCLASS:
         opts += ((*opts) + 1);
         break;
      case DHOP_PAD:    /* silently ignore, no data */
         break;
      case DHOP_END:
         return 0;   /* only good exit point */

      case DHOP_NAME:
         optlen = *opts++;
         nameset(DHOP_NAME, dhcptype, client, (char*)opts, optlen);
         opts += optlen;
         break;
      case DHOP_SNMASK:
         opts++;
         /* do we need to keep this? later... */
         opts += 4;
         break;
      case DHOP_ROUTER:
         optlen = *opts++;
         getlong((char*)opts, &addr);
         /* do we need to keep this? later... */
         opts += optlen;
         break;
      case DHOP_LEASE:
         opts++;
         getlong((char*)opts, &addr);
         client->ip_lease = addr;
         if(client->ip_lease != -1)
            client->ip_exp = cticks + (client->ip_lease * TPS);
         opts += 4;
         break;
      case DHOP_DNSRV:
         optlen = *opts++;

         /* See if need a DNS server. */
         while(optlen > 3)
         {
            /* Extract address and save it in dns list. */
            ip_addr dns = *(long*)opts;   /* leave in NET Endian */
            br_adddnssvr(dns);

            opts += 4;  /* skip tonext 4 byte IP addr */
            optlen -= 4;
         }
         opts += optlen;   /* Handle any odd len damage */
         break;
      case DHOP_CLASSID:
         optlen = *opts++;
         nameset(DHOP_CLASSID, dhcptype, client, (char*)opts, optlen);
         opts += optlen;
         break;
      case DHOP_DOMAIN:
         optlen = *opts++;
         nameset(DHOP_DOMAIN, dhcptype, client, (char*)opts, optlen);
         opts += optlen;
         break;
      case DHOP_DNSNAME:
         optlen = *opts++;
         nameset(DHOP_DNSNAME, dhcptype, client, (char*)opts, optlen);
         opts += optlen;
         break;
      default:
         br_log(0, "unknown DHCP option from %u.%u.%u.%u: %d, len %d\n", 
            PASS_IPADDR(client->iphost->ipaddr), 
            (int)(*(opts-1)), 
            (int)(*opts) );
         opts += ((*opts) + 1);
         break;
      }
   }
   bad_dhcp.parm = pip;
   alarm_general(&bad_dhcp, &baddhcp, "malformed DHCP option list\n");

   return ENP_PARAM;
}

/*
Copyright (C) 2003-2008 by Praemio Inc. All rights reserved.

Author: John Bartas - jbartas@praemio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


