/* brtcp.c
 *
 *  TCP spoofing code. 
 *
 * Copyright 2003-2008 by Praemio Inc. All rights reserved.
 * Author: John Bartas - jbartas@praemio.com
 * This software is released under the terms of the LGPL.
 * See LGPL Notice at end of this file; Full license in ../license.txt
 *
 * ROUTINES: 
 * br_spooftcp(), br_hookfile(), br_tcpreset(), 
 * bt_iptest(), ipe_lookup(), bt_porttest(), cpt_lookup(), 
 * killport(), br_tcpblock(), br_htmscan(), br_csize(), 
 * br_serverhttp(), br_clienthttp(), br_pop3(),  *
 *
 */

#include "busterport.h"
#include "brnet.h"
#include "ip.h"
#include "arp.h"
#include "brdb.h"


#include "brdb.h"


/****** File Notes *******
 *
 * - Calling sequence of internal routines:
 *
 *       br_tcpblock()----------+
 *       |        |             |
 *       | +------|-br_html()   |
 *       | |      |  |          V
 *       | |      |  |     killport()
 *       | |      |  |      |
 *       V V      V  V      V
 * br_hookfile()  br_tcpreset()
 *       |        |
 *       V        V
 *     br_spooftcp()
 *
 *
 * - All routines except killport() may be called from other files.
 *
 * - Any routine calling br_hookfile(), br_tcpreset() or br_tcpspoof()
 *   as a result of a ctl (struct ctdesc) should manage the ctl counter 
 *   and timer.
 */

unshort http_proxyport = 0;   /* user settable if proxy is in service */


/* Portability call to trigger visual block notification */
extern   void  br_flash(int itterate, int level);


/* br_spooftcp() - Send a single spoofed TCP packet into the connection passed.
 * 
 * Builds a tcp/ip header for the passed connection. IP header is at 
 * pkt->pk_buff + MaxLnh, TCP header immediatly follows IP. We don't do TCP 
 * options. TCP content (if any) is copied in and the packet sent.
 *
 * INPUT:
 * conn - our connection structure
 * content - TCP data to insert (copied!!!)
 * clength length of TCP content to allow in data field
 * local - boolean, true if sending to local host
 * flags - tcp header flags field.
 *
 * RETURNS: 0 if send OK, else ENP error.
 */

#define  TCPIP_HDRSIZE  (sizeof(struct ip) + sizeof(struct tcphdr))

int
br_spooftcp(struct conn_entry * conn, 
            u_char * content, 
            int clength, 
            int local, 
            unshort flags)
{
   struct tcphdr *   tcpp;
   struct ip *       pip;
   PACKET            pkt;
   ip_addr           nexthop;
   struct eth_hdr *   eth_hdr;
   NET               ifp;
   struct endpoint * him;     /* endpoint being reset */
   struct endpoint * me;      /* endpoint I am spoofing */
   int      err;


   /* If we're in "host only" mode, and the local host is not
    * our Windows host machine, then don't block.
    */
   if(br_cfg.cflags & BC_HOSTONLY)
   {
      if(nets[conn->ce_iface]->n_ipaddr != conn->loc.ce_addr)
      {
         return 0;   /* not an error */
      }
   }

   if(conn->ce_type != H_TCP)
      return ENP_LOGIC;

   if(local)
   {
      him = &conn->loc;
      me = &conn->rem;
   }
   else
   {
      me = &conn->loc;
      him = &conn->rem;
   }

   /* limit our packet sending to 3 packets per kill attempt. */
   if((local == FALSE) &&
      (flags & TH_RST) &&
      (him->sp_resets++ > 3))
   {
      return 0;
   }

   /* build the TCP spoof header */
   pkt = pk_alloc(MAC_HDRLEN + 44 + clength);
   if(!pkt)
      return ENP_NOBUFFER;
   MEMSET(pkt->pk_buff, 0, pkt->pk_blen);

   pip = (struct ip *)(pkt->pk_buff + MAC_HDRLEN + 2);   /* 2 alignment bytes */
   tcpp = (struct tcphdr *)(pip + 1);

   tcpp->th_sport = htons(me->ce_port);    /* ports are saved in local endian */
   tcpp->th_dport = htons(him->ce_port);
   tcpp->th_seq = htonl(me->ce_seq);
   me->ce_seq += clength;
   tcpp->th_flags = flags;

   /* include ACK of his next expected sequence number */
   tcpp->th_ack = htonl(him->ce_seq);
  
   /* if not resetting, fill in ACK and WIN fields */
   if((flags & TH_RST) == 0)
   {
      /* tmp - use weird size window to help ID these packets with sniffer */
      tcpp->th_win = htons(4444);
   }
   tcpp->th_doff = 0x50;

   pip->ip_src = me->ce_addr;
   pip->ip_dest = him->ce_addr;

   ifp = pkt->pk_net = nets[conn->ce_iface];
   pkt->pk_prot = (char*)tcpp;
   pkt->pk_plen = 20 + clength;
   mk_iphdr(pkt, ip_id++, 6 /* TCP */, 0, clength + TCPIP_HDRSIZE);
   if(clength)
      MEMCPY(tcpp + 1, content, clength);

   pkt->pk_prot -= sizeof(struct tcphdr);
   pkt->pk_plen += sizeof(struct tcphdr);
   tcpp->th_sum = tcp_cksum(pip);

   /* send direct if we can, via ARP if we can't */
   if((him->ce_mac) && (me->ce_mac))
   {
      /* paste in Ethernet header */
      eth_hdr = (struct eth_hdr *)(pkt->pk_prot - 14);
      MEMCPY(eth_hdr->e_dst, him->ce_mac->macaddr, 6);
//      MEMCPY(eth_hdr->e_src, me->ce_mac->macaddr, 6);

      /* WIFI spoofing works better if we use our real MAC - 
       * Linksys (and maybe other) WIFI home routers won't forward
       * packets from the "wrong" MAC address.
       */
      MEMCPY(eth_hdr->e_src, nets[conn->ce_iface]->n_macaddr, 6);
      eth_hdr->e_type = htons(0x0800);
      pkt->pk_prot -= 14;
      pkt->pk_plen += 14;
      err = ifp->n_send(pkt);
   }
   else
   {
      if((pip->ip_dest & ifp->n_mask) == 
         (ifp->n_ipaddr & ifp->n_mask))
      {
         nexthop = pip->ip_dest;
      }
      else
      {
         if(pkt->pk_net->n_router == 0)
         {
            br_user_alert("Unable to send TCP pkt - no router");
            return ENP_NO_ROUTE;
         }
         nexthop = pkt->pk_net->n_router;
      }
      err = send_via_arp(pkt, nexthop);
   }

   if(flags & TH_RST)
   {
      if(local)
         conn->ce_flags |= CEF_RESETLOCAL;
      else
         conn->ce_flags |= CEF_RESETREMOTE;
   }

   br_flash(10, 2);  /* Start LED or systray icon flash */

   return err;
}

/* br_hookfile() - Insert HTML content into a passed connection to redirect a 
 * browser looking for a blocked web page. 
 * 
 * If the page was blocked because the path name portion of the URL contained a 
 * blocked path, then "path" is a pointer to the node in path_tree database.
 * If "path" is NULL then the page was blocked because of the remopte IP address
 * in the conn structure passed.
 *
 * Return: zero if success, else errors are the same as with
 * br_spooftcp()
 */

char     br_hookbuf[512];

/* Code to do the html header & close/redirect script.
 *
 * This javascript insert a test to see the the window is a popup
 * and if so tries to close it.
 * It then checks to see if window is an iframe, and if so 
 * redirects it to the iframe blocker page
 * Else it loads the full-screen blocker page.
 */
 char hookcode[] = {"HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n"
"<html><head><meta http-equiv=\"Pragma\" content=\"no-cache\"></head>"
"<SCRIPT LANGUAGE=\"JavaScript\">\r\n"
"if(window.opener != null) window.close(self);\r\n"
"</script><SCRIPT LANGUAGE=\"JavaScript\">\r\n"
"if(window.self == window.top) "
" location.href=\"http://%u.%u.%u.%u:%d/blankpage.htm?conn=%u\";\r\n"
" else \r\n"
" location.href=\"http://%u.%u.%u.%u:%d/iframebk.htm?conn=%u\";\r\n"
"</SCRIPT></html>"};

char * jsonly = &hookcode[50];   /* shoud point to "<SCRIPT */

int
br_hookfile(struct conn_entry * conn,     /* connection to intercept */
            int local)                    /* Bool, TRUE if local end gets content */
{
   unshort  flags;
   int      err = 0;

   /* If the remote host has sent more than 1 pkt (the syn-ack) then inject
    * script to redirect the browser away from the offensive page 
    */
   if(conn->rem.ce_outpkts > 1)
   {
      /* pass connection to blankpage.htm or iframebk.htm */
      sprintf(br_hookbuf, jsonly, 
         PASS_IPADDR( nets[conn->ce_iface ]->n_ipaddr ), http_port, conn,
         PASS_IPADDR( nets[conn->ce_iface ]->n_ipaddr ), http_port, conn);
   }
   else  /* No data packets from remote host yet, replace whole page */
   {
      /* pass connection to blankpage.htm or iframebk.htm */
      sprintf(br_hookbuf, hookcode, 
         PASS_IPADDR( nets[conn->ce_iface ]->n_ipaddr ), http_port, conn,
         PASS_IPADDR( nets[conn->ce_iface ]->n_ipaddr ), http_port, conn);
   }

   flags = TH_ACK|TH_PUSH|TH_FIN;
   err = br_spooftcp(conn, (u_char*)br_hookbuf, 
      sizeof(br_hookbuf), local, flags);
   if(err)
      return err;

   return err;
}


/* send reset to one end of a TCP connection.
 *
 * PARM2: int local - true to rest local node, FALSE to reset remote. 
 *
 */

int
br_tcpreset(struct conn_entry * conn, int local)
{
   int      err;

   if(local)
      dtrap();
   err = br_spooftcp(conn, NULL, 0, local, TH_RST);
   if(err)
      return err;

   return 0;
}


/* btree compare routine for IP address ranges. 
 * The IP address is passed in LOCAL endian 
 */

int
bt_iptest(struct btnode * n, u_long ipaddr)
{
   struct control_ip * cip;
   cip = (struct control_ip *)n->data;

   if( ipaddr < cip->ci_ipaddr)
      return KEY_LT;
   else if( ipaddr > cip->ci_ipaddr)
      return KEY_GT;
   else
      return KEY_EQ;
}


/* ipe_lookup() routine - 
 * The IP address is passed in LOCAL endian. 
 */

struct control_ip *
ipe_lookup(ip_addr ipaddr) 
{
   struct btnode * n;

   TH_LOCK_MUTEX(control_ip_mutex);
   n = bt_access((struct btnode *)ip_tree, ipaddr, bt_iptest );
   TH_UNLOCK_MUTEX(control_ip_mutex);
   if(n)
      return (struct control_ip *)n->data;
   else
      return NULL;
}

/* binary tree compare routine for port range. 
 * The port is passed in LOCAL endian in the low 2 bytes of "key" 
 */

int
bt_porttest(struct btnode * n, u_long key)
{
   struct control_port * cpt;

   cpt = (struct control_port *)n->data;

   if((unshort)key < cpt->cp_port)
      return KEY_LT;
   else if((unshort)key > cpt->cp_port)
      return KEY_GT;
   else
      return KEY_EQ;
}

/* lookup a port-only control structure. 
 * port is passed in LOCAL endian 
 */

struct control_port *
cpt_lookup(unshort port, int transport)
{
   struct btnode * n;

   if(transport == CPF_TCP)
      n = bt_access((struct btnode *)tcpp_tree, (u_long)(port), bt_porttest );
   else if(transport == CPF_UDP)
      n = bt_access((struct btnode *)udpp_tree, (u_long)(port), bt_porttest );
   else
   {
      dtrap();       /* bogus parameter */
      return NULL;
   }


   if(!n)
      return NULL;

   /* make sure port control entry applies to the passed transport layer type */
   if( ((struct control_port *)(n->data))->cp_flags & transport)
      return (struct control_port *)n->data;
   else
      return NULL;
}

/* Utility to kill a connection because of port, and log it. */

static void
killport(struct conn_entry * conn, unshort port)
{
   br_tcpreset(conn, 0);
   br_tcpreset(conn, 1);

   br_log(AF_BLOCK, 
      "port %u blocked; connection %s<->%s, ports %u<->%u was cut.\n", 
      port, br_getnametext(conn->loc.ce_addr), br_getnametext(conn->rem.ce_addr), 
      conn->loc.ce_port, conn->rem.ce_port );

   return;
}

/* called for every TCP packet. This is the TCP control function -
 * checks for restricted IP addrs, and ports.
 *
 * PARAM2: remote - TRUE if conn->rem is sender
 *
 * returns 0 if connection is not reset,
 * returns 1 if connection is reset at one end,
 * returns 2 if connection is reset at both ends
 */

int
br_tcpblock(struct conn_entry * conn, int remote)
{
   struct control_ip * ipe;
   struct control_port * cpt;
   struct endpoint * sender;
   struct endpoint * receiever;

   if(remote)
   {
      sender = &conn->rem;
      receiever = &conn->loc;
   }
   else
   {
      sender = &conn->loc;
      receiever = &conn->rem;
   }

   /* see if remote port is blocked */
   if(((cpt = cpt_lookup(conn->rem.ce_port, CPF_TCP)) != NULL) &&
      (cpt->cp_flags & CPF_REMOTE))
   {
      /* If access is allowed (ipe is whitelisted), just return */
      if(ctl_getaccess(&cpt->ctl) > 0)
         return 0;

      conn->block.type = CTF_TYPEPORT;
      conn->block.ptr = (void*)cpt;

      /* port is blocked, reset both ends of connection */
      killport(conn, conn->rem.ce_port);

      cpt->ctl.actions++;
      cpt->ctl.lasttick = cticks;
      return 2;
   }

   /* see if local port is restricted or blocked */
   if(((cpt = cpt_lookup(conn->loc.ce_port, CPF_TCP)) != NULL) &&
      (cpt->cp_flags & CPF_LOCAL))
   {
      /* If access is allowed (ipe is whitelisted), just return */
      if(ctl_getaccess(&cpt->ctl) > 0)
         return 0;

      /* See if this block only applies to remote senders */
      if(cpt->cp_flags & CPF_RSONLY)
      {
         /* If the sender is local, return w/o blocking */
         if( islocal_ip(conn->ce_iface, sender->ce_addr))
            return 0;
      }
      conn->block.type = CTF_TYPEPORT;
      conn->block.ptr = (void*)cpt;

      /* port is blocked, reset both ends of connection */
      killport(conn, conn->loc.ce_port);

      cpt->ctl.actions++;
      cpt->ctl.lasttick = cticks;
      return 2;
   }

   /* see if IP address/port combo is blocked. */
   if((ipe = ipe_lookup(htonl(sender->ce_addr))) != NULL)
   {
      if((ipe->ci_port == 0) ||     /* port 0 is wildcard */
         (ipe->ci_port == sender->ce_port))
      {
         /* If access is allowed (ipe is whitelisted), just return */
         if(ctl_getaccess(&ipe->ctl) > 0)
            return 0;
         if(ipe->ctl.timeout)
         {
            if(ipe->ctl.timeout <= cticks)   /* has access timer expired? */
            {
               ipe->ctl.timeout = 0;         /* clear it, fall to tcp reset */
               goto block_reset;
            }
            else
               return 0;
         }

block_reset:
         if(!ipe->ctl.offense)
         {
            dtrap();    /* can this happen? */
            return -1;
         }

         if((ipe->ctl.offense->code & featuremask)==0)
         {
            /* User does not have feature to block this item. */
            return 0;
         }

         // record control which triggered blocking
         conn->block.type = CTF_TYPEIP;
         conn->block.ptr = (void*)ipe;

         br_tcpreset(conn, 0);   /* reset remote */

         /* Always be nice to the local side */
         br_hookfile(conn, 1);
         conn->ce_flags |= CEF_CLOSEDWIN;

         ipe->ctl.actions++;
         ipe->ctl.lasttick = cticks;

         br_log(AF_BLOCK, 
            "IP address %u.%u.%u.%u blocked; connection %s<->%s, ports %u<->%u (%s).\n",
            PASS_IPADDR(sender->ce_addr), br_getnametext(conn->loc.ce_addr),
            br_getnametext(conn->rem.ce_addr), conn->loc.ce_port, 
            conn->rem.ce_port, ipe->ctl.offense->desc );

         return 2;
      }
   }
   return 0;
}

u_long   htscanned = 0;

void
br_htmscan(struct conn_entry * conn, u_char * data, int dlen)
{
   u_char *    cp;
   u_char *    end;
   int         proximity;
   struct vword * wordp;

   /* Make sure there is enough content to filter */
   if(dlen < 4)
      return;

   /* Some people like sex, so make sure feature is enabled. */
   if((br_cfg.cflags & BC_CONTENT) == 0)
      return;

   end = data + dlen;      /* end of input buffer */
   cp = data;              /* start of input buffer */
   while((*cp < 'A') || (*cp > 'z'))
      cp++;

   /* scan input looking for weighted words */
   while(cp < end)
   {
      /* cp points to a word */
      wordp = word_lookup((char*)cp);
      if(wordp && ((wordp->ctl.action_mask & CTF_WHITELIST) == 0))
      {
         wordp->hits++;
         if(wordp->weight == WT_SHOWSTOPPER)
            goto kill_conn;

         /* see this word is close to the last vocab hit */
         proximity = conn->ce_wcount - conn->ce_vlast;
         if(proximity < 6)
            proximity = (6 - proximity);
         else
            proximity = 1;
         conn->ce_weight += (wordp->weight << proximity);
         conn->ce_vlast = conn->ce_wcount;
      }

      conn->ce_wcount++;      /* bump document word count */

      /* Find end of this word */
      while((*cp >= 'A') && (*cp <= 'z'))
         cp++;

      /* Find next word */
      while((*cp < 'A') || (*cp > 'z'))
      {
         if(*cp == '<')    /* skip tag contents */
         {
            /* ...except for "meta" tags, which often have keywords */
            if((*(cp+1) == 'm') && (*(cp+1) == 'e'))
               continue;
            while(*cp != '>')    /* end of tag */
            {
               cp++;
               if(cp >= end)
                  break;
            }
            continue;
         }
         cp++;
      }
   }

   /* see if we should censor the page/URL/Site */
   if((conn->ce_wcount > 20) &&
      ((conn->ce_weight * 5) > conn->ce_wcount))
   {
      struct control_dn * dnp;
      struct control_ip * cip;
      struct control_ip * tmp;
      char * url;

kill_conn:

      /* kill it */
      br_tcpreset(conn, TRUE);
      br_tcpreset(conn, FALSE);

      /* add IP address to black list */
      cip = new_ipe(conn->rem.ce_addr, NULL, CIF_USER, &local_comm );
      if(!cip)
         return;     /* alloc failed */

      cip->ctl.offense = mot_bynum(OF_CONTENT);

      conn->block.type = CTF_TYPEIP;
      conn->block.ptr = (void*)cip;

      /* get text for page and log this event */
      if(conn->data[0])
         url = conn->data;
      else
         url = br_getnametext(conn->rem.ce_addr);

      br_log(AF_BLOCK, 
         "page \"%s\" blocked, due to content, score:%d words:%d\n", 
         url, conn->ce_weight, conn->ce_wcount);

      /* See about blacklisting domain name. Try to find the 
       * server's domain name in the URL data buffer.
       */
      if(conn->data[0])
      {
         cp = (u_char*)conn->data;
         cp = (u_char*)strchr((char*)cp, ':');
         if(cp)
         {
            cp += 3;    /* now points to start of name */
            end = cp;
            while((*end > ' ') && (*end != '?') && (*end != '/'))
               end++;
            *end = 0;
            dnp = new_dn((char*)cp, &local_comm );

            MEMCPY(&dnp->ctl, &cip->ctl, sizeof(cip->ctl) );
 
            /* see if IP block is already associated with DNS entry */
            for(tmp = dnp->cu_cip; tmp; tmp = tmp->ci_dnlist)
            {
               if(cip == tmp)
                  break;
            }
            if(!tmp)    /* cip was not in associated IP list */
               br_insertcip(dnp, cip);
         }
      }
   }
   
   return;
}

/* br_csize() - process a chuck-size field in the http stream.
 *
 * If the file in in chunked mode, track length widgets. If
 * we fumble this logic, we will "blow chunks".
 *
 * Returns pointer to start of data chunk after chunk widget.
 */

u_char *
br_csize(u_char * http, struct conn_entry * conn)
{
   int csize;     /* size of chunk */

   if((conn->ce_flags & CEF_HTTPCHUNKED) == 0)
      return http;      /* not chunked == do nothing */

   /* sum the ascii hex dugits into csize */
   csize = 0;
   while(*http != '\r')
   {
      if((*http >= '0') && (*http <= '9'))
      {
         csize <<= 4;   /* shift up for next digit */
         csize += (*http++ - '0');
      }
      else if((*http >= 'A') && (*http <= 'F'))
      {
         csize <<= 4;   /* shift up for next digit */
         csize += (*http++ - ('A' - 10));
      }
      else if((*http >= 'a') && (*http <= 'f'))
      {
         csize <<= 4;   /* shift up for next digit */
         csize += (*http++ - ('a' - 10));
      }
      else if(*http++ == ' ')
         break;
      else
      {
         /* we blew chunks - bogus http stream? */
         conn->ce_clen = 0;
         return(http);
      }
   }
   conn->ce_clen = csize;  /* save size of coming chunk */
   if(csize & 0xFF000000)
   {
      dtrap();    /* really??? */
   }
   return (http + 2);      /* bump past "\r\n" */
}


/* br_serverhttp() 
 *
 * Called from brrx for all datagrams with a source port of 80 
 */

void
br_serverhttp(struct conn_entry * conn, u_char * data, int dlen)
{
   u_char * http;      /* pointer to http header in packet */
   u_char * end;       /* end of http header in packet */

   http = data;

   if((conn->ce_flags & CEF_HTTPREPDONE) == 0)
   {
      /* scan the http header, looking for some key fields */
      end = data + dlen;
      while(http < end)
      {
         switch (*http)
         {
         case 'C':
            switch (*(http + 8))
            {
            case 'T':   /*  Content-Type: */
               conn->ce_flags &= ~(CEF_ISTEXT|CEF_ISIMAGE);
               if(wordcmp((char*)(http + 14), "text") == 0)
                  conn->ce_flags |= CEF_ISTEXT;
               else if(wordcmp((char*)(http + 14), "image") == 0)
                  conn->ce_flags |= CEF_ISIMAGE;
               http += 19;    /* put ptr near end of field */
               break;
            case 'L':   /*  Content-Length: */
            case 'l':   /*  Content-length: */
               conn->ce_clen = atoi((char*)(http + 16));
               http += 17;    /* put ptr near end of field */
               break;
            }
            break;
         case 'T':     /* Transfer-Encoding: */
            if(wordcmp((char*)(http), "Transfer"))
               break;
            http += 19;
            if(wordcmp((char*)http, "Chunked") == 0)
               conn->ce_flags |= CEF_HTTPCHUNKED;
            break;
         }

         /* find next end-of-arg "\r\n" */
         while(*http++ > '\r')
            ;
         http++;    /* bump pointer past "\r\n" bytes */
         /* Don't walk off the end of the data. */
         if(http >= end)
            break;

         /* end of header ("\r\n\r\n") ? */
         if((*http == '\r') && (*(http + 1) == '\n'))
         {
            conn->ce_flags |= CEF_HTTPREPDONE;
            http += 2;

            /* Don't walk off the end of the data. */
            if(http >= end)
               break;

            http = br_csize(http, conn);     /* check chunks */
            break;      /* exit while() loop */
         }
      }
      dlen -= (http - data);
   }

   /* If this file is text AND we're done with http header, 
    * then check the data for blocked content. 
    */
   if( (conn->ce_flags & (CEF_ISTEXT | CEF_HTTPREPDONE)) ==
      (CEF_ISTEXT | CEF_HTTPREPDONE))
   {
      /* If the file has a length given by HTTP, then pass that
       * length to the scan routine; otherwise pass the whole dgram.
       */
      if((conn->ce_clen) && 
         (dlen > conn->ce_clen))
      {
         br_htmscan(conn, http, conn->ce_clen);
      }
      else
      {
         br_htmscan(conn, http, dlen);
      }
   }

   /* handle chunks & persistence */
   if(conn->ce_clen)
   {
      u_char * end;

      conn->ce_clen -= dlen;

      /* see if file or chunk ended in this datagram */
      if(conn->ce_clen == 0)
      {
         /* see if it looks like there's another chunk */
         end = http + conn->ce_clen;   /* end of file or chunk */

         if((conn->ce_flags & CEF_HTTPCHUNKED) &&
            (*end == '\r') && 
            (*(end+1) == '\n'))
         {
            end += 2;
            end = br_csize(end, conn);     /* check chunks */
         }
         else     /* no more chunks */
         {
            conn->ce_clen = 0;
         }

         /* Clear out file information */
         conn->ce_flags &= ~(CEF_FILE_FLAGS);
      }
      else if(conn->ce_clen < 0)
      {
//         dtrap();    /* We blew chunks */
         conn->ce_clen = 0;
      }
   }

   return;
}

void
br_clienthttp(struct conn_entry * conn, u_char * data, int dlen)
{
   u_char * http;      /* pointer to http header in packet */
   u_char * end;       /* end of http header in packet */
   u_char * filename;  /* GET or POST arg */
   u_char * ucp;
   u_long   hword;     /* word from http packet */
   char *   url;
   int      i;
   struct b_path * path;
   struct btnode * node;
   int      err;

   http = data;
   hword = htonl((*(long*)http) | 0x20202020);   /* uppercase */

   if(conn->ce_flags & CEF_HTTPREQDONE)
   {
      /* this connection is handling multiple requests */
      conn->ce_flags |= CEF_MULTIREQ;
      conn->ce_flags &= ~CEF_HTTPREQDONE;
   }

   /* check for GET, POST, and replies */
   if(hword == 0x67657420)          /* 'get ' */
      conn->ce_flags |= CEF_HTTPGET;
   else if(hword == 0x706f7374)     /* 'post' */
      conn->ce_flags |= CEF_HTTPGET;
   else if(hword == 0x68747470)     /* 'http' */
   {
      /* see if this is an OK reply to a GET for favicon */
      if((conn->rem.ce_iphost) && 
         (conn->rem.ce_iphost->flags & IH_FAVIGET))
      {
         ucp = http + 9;   /* point to reply code */
         if(*ucp <= '3')   /* 3xx or less means favicon is available */
            conn->rem.ce_iphost->flags |= IH_FAVICON;
      }
      return;
   }
   else
      return;

   /* Fall to here if this is a GET or POST command */

   /* save URL client asked for */
   if(conn->data[0] == 0)
   {
      url = br_getnametext(conn->rem.ce_addr);   /* get host portion */
      sprintf(conn->data, "http://%s/", url);   /* paste into buffer */

      /* build filename portion of URL */
      if(conn->ce_flags & CEF_HTTPGET)
         url = (char*)http + 3;  /* point past "GET" */
      else
         url = (char*)http + 4;  /* point past "POST" */
      while(*url == ' ')         /* point past spaces to file name */
         url++;
      for(i = strlen(conn->data); i < MAXDATALEN; i++)
      {
         if(*url++ <= ' ')       /* end of file name? */
         {
            conn->data[i] = 0;   /* null terminate */
            break;               /* end of copy loop */
         }
         conn->data[i] = *url;   /* copy file name into URL */
      }
   }

   http += 4;
   filename = http;
   hword = htonl(*(long*)(http)) & 0x002020ff;

   /* scan the http header, looking for some key fields */
   http += 4;
   end = data + dlen;
   while(http < end)
   {
      hword <<= 8;
      hword += (*http++);
      switch (hword)
      {
      case 0x6570743a:     /* Acc "ept:" */
         http += 5;
         hword = 0; 
         for(i = 0; i < 4; i++)
         {
             hword <<= 8;
             hword = (*(http + i));
         }
         if(hword == 0x696D6167)    /* 'imag' */
            conn->ce_flags |= CEF_ISIMAGE;
         while((*http != '\n') && (http < end))
            http++;
         break;
      case 0x656e743a:     /* User "agent:" */
         http += 5;
         while(*http != '\n')
            http++;
         break;
      case 0x0d0a0d0a:     /* /r/n/r/n - end of http header */
         end = http;
         conn->ce_flags |= CEF_HTTPREQDONE;
         break;
      }
   }

   if(conn->rem.ce_iphost)
   {
      /* check for "GET favicon.ico" requests */
      if((*(filename + 1)== 'f' ) &&
         (memcmp(filename, "/favicon.ico", 11) == 0 ))
      {
         conn->rem.ce_iphost->flags |= IH_FAVIGET;
      }
      else
      {
         conn->rem.ce_iphost->flags &= ~IH_FAVIGET;
      }
   }

   /* scan file name for blacklisted paths */
   for(ucp = filename; *ucp > ' '; ucp++)
   {
      node = bt_access(path_tree, (u_long)ucp, bt_pathtest);
      if(node)
      {
         /* get path info from tree node tructure */
         path = (struct b_path *)node->data;

         /* See if database path must be at start of URL path */
         if(((path->flags & BP_LEADWILD) == 0) &&
            (ucp != filename))
         {
            continue;   /* db path not at start of string */
         }

         /* See if database path must be at end of URL path */
         if(((path->flags & BP_TAILWILD) == 0) &&
            (*(ucp + path->length) < ' '))
         {
            continue;
         }

         /* See if access is allowed */
         if(ctl_getaccess(&path->ctl) > 0)
            continue;

         // record control which triggered blocking
         conn->block.type = CTF_TYPEPATH;
         conn->block.ptr = (void*)path;

         br_tcpreset(conn, 0);      /* reset to remote host */

         /* send the "close window" script to the local browser */
         err = br_hookfile(conn, 1);
         conn->ce_flags |= CEF_CLOSEDWIN;
         path->ctl.actions++;
         path->ctl.lasttick = cticks;

         if(err)
         {
            br_log(0, "br_html: %d\n", err);
            return;
         }
         br_log(AF_BLOCK, 
            "URL matching \"%s\" blocked, connection between %s<->%s, ports %u<->%u was cut.\n", 
            path->pathname, br_getnametext(conn->loc.ce_addr),
            br_getnametext(conn->rem.ce_addr), conn->loc.ce_port, conn->rem.ce_port );

         break;
      }
      if(ucp > end)
         break;      /* saftey */
   }

   return;
}

/*
Copyright (C) 2003-2008 by Praemio Inc. All rights reserved.

Author: John Bartas - jbartas@praemio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/



