/* brudpio.c - winsock version.
 * 
 * Copyright 2003-2008 by Praemio Inc. All rights reserved.
 * Author: John Bartas - jbartas@praemio.com
 * This software is released under the terms of the LGPL.
 * See LGPL Notice at end of this file; Full license in ../license.txt
 *
 * Map datagram UDP IO to local UDP stack API. This usually means supporting
 * the silly sockets "connection" model.
 * 
 * This includes UDP socket open and send routine. Receives on open 
 * sockets are handled via callbacks from a UDP thread. 
 *
 * ROUTINES: udpso_bysock(), udp_getempty(), udp_socksend(), 
 * ROUTINES: udp_sockopen(), 
 *
 * There is no close routine since Buster doesn't need one (yet)
 *
 */

#include "busterport.h"
#include "brnet.h"
#include "ip.h"
#include "brdb.h"


/* Task structure for the UDP RX callback task */

TH_OBJECT(to_udpcb);
TH_PROTO(udpcb_handler);

struct br_thread udp_thread = 
{
   &to_udpcb,
   "UDP callback",
   udpcb_handler,
   APP_PRIORITY,
   APP_STACK_SIZE,
};

static   int   udp_thread_started = FALSE;


#define  NUMUDPSOCKS    32
#define  NONE_EMPTY     -1

struct udpsock
{
   long     socket;     /* winsock-ish socket for this port combo */
   ip_addr  ipaddr;     /* remote IP address (0 == wildcard) */
   unshort  fport;      /* remote (foreign) port (0 == wildcard) */
   unshort  lport;      /* local port (0 == wildcard) */
   int(*callback)(PACKET pkt, void * Id);
   void *   Id_value;   /* session Identifier */
   int      inuse;      /* nonzero ifentry is in use */
   u_long   activity;   /* cticks of last activity */
} udpsocks[NUMUDPSOCKS];

struct udpsock *
udpso_bysock(long socket)
{
   int   sockx;      /* index into udpsocks[] */

   for(sockx = 0; sockx < NUMUDPSOCKS; sockx++)
   {
      if(udpsocks[sockx].socket == socket)
         return (&udpsocks[sockx] );
   }
   return NULL;
}

/* udp_getempty() - Get index of first empty udpsocks, -1 if none */

int
udp_getempty()
{
   int   sockx;      /* index into udpsocks[] */
   int   idle;

   for(sockx = 0; sockx < NUMUDPSOCKS; sockx++)
   {
      if(udpsocks[sockx].inuse == FALSE)   /* Empty? */
      {
         memset(&udpsocks[sockx], 0, sizeof(struct udpsock));
         udpsocks[sockx].inuse = TRUE;
         udpsocks[sockx].socket = INVALID_SOCKET;
         return sockx;
      }
   }
//   dtrap();       /* Should never happen */

   /* It happens... */
   /* Find getting most longest idle udpsocks[] entry which was 
    * created by a udp_socksend() call.
    */

   idle = NUMUDPSOCKS;
   for(sockx = 0; sockx < NUMUDPSOCKS; sockx++)
   {
      /* ipaddr is zero if this is a listen socket - don't take those */
      if(udpsocks[sockx].ipaddr == 0)
      {
         continue;
      }

      if((idle == NUMUDPSOCKS) || 
         (udpsocks[sockx].activity < udpsocks[idle].activity))
      {
         idle = sockx;
      }
   }

   if(idle >= NUMUDPSOCKS)      /* no send sockets? */
   {
      return -1;     /* NO sockets left */
   }

   dprintf("udp_getempty: recycling socket %u.%u.%u.%u:%u\n",
      PASS_IPADDR(udpsocks[idle].ipaddr), udpsocks[idle].fport);

   udpsocks[idle].inuse = FALSE;
   socketclose(udpsocks[idle].socket);
   udpsocks[idle].socket = INVALID_SOCKET;
   udpsocks[idle].activity = cticks;

   return idle;
}


/* udp_socksend() - send a datagram on a socket.
 *
 * Returns 0 if OK, or ENP_ error code.
 *
 */

int
udp_socksend(ip_addr target, unshort fport, PACKET pkt)
{
   int   sockx;      /* index into udpsocks[] */
   int   err;
   long  tmpsock;
   struct sockaddr_in sin;

   /* Find the socket table entry(if any) for this port combo */
   for(sockx = 0; sockx < NUMUDPSOCKS; sockx++)
   {
      if((udpsocks[sockx].fport == fport) && 
         ((udpsocks[sockx].ipaddr == 0) ||
          (udpsocks[sockx].ipaddr == target)))
      {
         break;
      }
   }

   if(sockx >= NUMUDPSOCKS)
   {
      sockx = udp_getempty();
      if(sockx < 0)
      {
         pk_free(pkt);
         return ENP_RESOURCE;
      }
      udpsocks[sockx].fport = fport;
      udpsocks[sockx].ipaddr = target;
   }
   
   /* Get here if udpsocks[sockx] has the port number we were passed.
    * If we don't already have a socket for this port combo, make one.
    */
   if(udpsocks[sockx].socket == INVALID_SOCKET)
   {
      tmpsock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
      if(tmpsock == INVALID_SOCKET)
      {
         dtrap();
         pk_free(pkt);
         return ENP_RESOURCE;
      }

      err = 1;
      err = setsockopt(tmpsock, SOL_SOCKET, SO_BROADCAST,
         (void *)&err, sizeof(err));
      if(err)
      {
         dtrap();
         err = sock_errno(tmpsock);   /* for debug */
         dprintf("udp_socksend: setsockopt(broadcast) error %d\n", err);
         pk_free(pkt);
         return ENP_RESOURCE;
      }

      udpsocks[sockx].socket = tmpsock;
   }

   /* socket is open and ready to send */
   memset(&sin, 0, sizeof(sin));
   sin.sin_port = htons(fport);
   sin.sin_addr.s_addr = target;   /* already net endian */
   sin.sin_family = AF_INET;
   err = sendto(udpsocks[sockx].socket, pkt->pk_prot, pkt->pk_plen, 0, 
      (struct sockaddr*)&sin, sizeof(sin));

   pk_free(pkt);

   if(err < 0)
   {
      /* Don't dtrap during initial startup period */
      if((br_gettime() - br_lastresume) > SYS_STARTUP_LAG)
      {
         err = sock_errno(udpsocks[sockx].socket);
         dtrap();
      }
      return ENP_RESOURCE;
   }

   return 0;
}

#ifdef NOTDEF
 WSAENOTSOCK       /*      (WSABASEERR+38)  */
 WSAEADDRINUSE     /*       (WSABASEERR+48) */
 WSAEADDRNOTAVAIL  /*       (WSABASEERR+49) */
 WSAEHOSTUNREACH         (WSABASEERR+65) 0x2751
#endif

/* udp_sockopen() - Create a UDP socket and start a listen.
 *
 * wildcard (0) is OK for fhost and fport.
 *
 * IP and port Parameters are in local endian.
 *
 * Returns socket if OK, else system's SOCKET_ERROR code on error.
 */

long
udp_sockopen(ip_addr fhost, 
             unshort lport, 
             unshort fport, 
             int(*callback)(PACKET pkt, void * Id),  
             void * Id_value )
{
   int   sockx;      /* index into udpsocks[] */
   int   err;
   long  tmpsock;
   struct sockaddr_in sin;

   tmpsock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
   if(tmpsock == SOCKET_ERROR)
   {
      dtrap();
      return tmpsock;
   }

   memset(&sin, 0, sizeof(sin));
   sin.sin_port = htons(lport);
   sin.sin_addr.s_addr = htonl(fhost);
   sin.sin_family = AF_INET;

   err = bind(tmpsock, (struct sockaddr*)&sin, sizeof(sin) );
   if(err)
   {
      err = sock_errno(tmpsock);   /* for debug */
      dprintf("udp_socksend: bind error %d\n", err);
      return SOCKET_ERROR;
   }

   sockx = udp_getempty();
   if(sockx < 0)
      return SOCKET_ERROR;

   udpsocks[sockx].socket = tmpsock;
   udpsocks[sockx].lport = lport;
   udpsocks[sockx].fport = fport;
   udpsocks[sockx].socket = tmpsock;
   udpsocks[sockx].callback = callback;
   udpsocks[sockx].Id_value = Id_value;

   if(udp_thread_started == FALSE)
   {
      err = TH_NEWTASK(&udp_thread);
      if(err)
      {
         dprintf("Can't start UDP RX thread");
         return -1;
      }
      udp_thread_started = TRUE;
   }

   return tmpsock;
}


/* We create one thread to manage receiving on the various UDP 
 * sockets. This simple threadblocks in select waiting for 
 * receiveddatagrams, and returns when one shows up. The received
 * datagrams are passedto their session's respective upcall 
 * routines. This design allowsus to support several low-overhead
 * UDP applications (DN updates,interbucter messaging, DNS lookups)
 * with a single thread.
 */


TH_ENTRY(udpcb_handler)
{
   int      i;
   int      err;
   int      count;
   int      sinsize;
   long     sock;
   long     highsock;   /* Highest value socket, for select */
   PACKET   pkt;
   fd_set   fd_recv;   /* sock list for recv */
   struct sockaddr_in sin;
   struct timeval timeout;
   struct udpsock * udpso;


   for(;;)
   {
      /* Build a select structure from the UDP sessions array */
      FD_ZERO(&fd_recv);
      highsock = 0;
      for(i = 0; i < NUMUDPSOCKS; i++)
      {
         if((udpsocks[i].socket != INVALID_SOCKET) &&
            (udpsocks[i].socket != 0))
         {
            FD_SET(udpsocks[i].socket, &fd_recv);
            if(udpsocks[i].socket > highsock)
                highsock = udpsocks[i].socket;
         }
      }
      timeout.tv_sec = 5;
      timeout.tv_usec = 0;
      count = select(highsock, &fd_recv, NULL, NULL, &timeout);
      if(count > 0)
      {
         /* handle the received dgrams */
         for(i = 0; i < NUMUDPSOCKS; i++)
         {
            /* See if this socket has a received datagram ready */
            sock = udpsocks[i].socket;
            if( !FD_ISSET(sock, &fd_recv) )
               continue;

            pkt = pk_alloc(largebufsize);
            if(!pkt)
            {
               dprintf("udpcb_handler: pk_alloc faled\n");
               continue;
            }
            sinsize = sizeof(sin);
            err = recvfrom(sock, pkt->pk_buff, pkt->pk_blen, 0, 
               (struct sockaddr*)&sin, &sinsize);
            if(err < 0)
            {
               dtrap();
               err = sock_errno(sock);
               pk_free(pkt);
               continue;
            }

            udpso = udpso_bysock(sock);   /* look up session */
            /* If this session not found, or session has no callback 
             * routine, then dump the packet and continue.
             */
            if((udpso == NULL) ||
               (udpso->callback == NULL))
            {
               pk_free(pkt);
               continue;
            }

            /* Get here if we have a received packetand a callback */
            pkt->pk_prot = pkt->pk_buff;
            pkt->pk_plen = err;
            err = udpso->callback(pkt, udpso->Id_value);
            if(err)
            {
               pk_free(pkt); /* need to do cleanup */
            }
         }
      }
      else if(count < 0)
      {
         int   err;
         dtrap();
         err = sock_errno(sock);
         dprintf("udpcb_handler: select error %d\n", err);
      }

   }
   TH_RETURN_OK();
}
TH_EXIT("udpcb_handler")


/*
Copyright (C) 2003-2008 by Praemio Inc. All rights reserved.

Author: John Bartas - jbartas@praemio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


