/* brupdate.c
 *
 *  per-packet update routines. 
 *
 * Copyright 2003-2008 by Praemio Inc. All rights reserved.
 * Author: John Bartas - jbartas@praemio.com
 * This software is released under the terms of the LGPL.
 * See LGPL Notice at end of this file; Full license in ../license.txt
 */


#include "busterport.h"
#include "brnet.h"
#include "arp.h"
#include "ip.h"
#include "dhcpdef.h"
#include "brdb.h"
#include "brdns.h"


#ifdef WINDOWS_HOSTED
struct ip_hostentry  * winsock_ip;
#endif   /* WINDOWS_HOSTED */

/* Called when there may have been new DHCP assignments */
int   br_newdhcpip(struct host_entry * machost, ip_addr ipaddr);

/* Highest and lowest addresses of a conneciton in heap memory */
struct conn_entry * conn_lowest = (struct conn_entry *)0xFFFFFFFF;
struct conn_entry * conn_highest = (struct conn_entry *)0x0;

/* sumname(char * name) - checksum passed name, for change detection */

static int
sumname(char * name)
{
   int sum = 0;
   while(*name >= ' ')
      sum += *name++;
   return sum;
}


/* br_hipcheck(struct host_entry * tmp) - Check IP addressing in passed host */
int
br_hipcheck(struct host_entry * machost)
{
   ip_addr tmp;

   /* Debug: did we get one IP assigned to both slots? */
   if(machost->iphost->ipaddr == machost->ipaddr2)
   {
      dtrap();

      /* Send ARP to this hosts alleged IP, reply may help cleanup.*/
      send_arp_net(nets[machost->iface], machost->iphost->ipaddr);

      /* Get rid of #2, merge data into #1 */
      machost->ipaddr2 = 0;
      machost->from_ip += machost->from_ip2;
      machost->from_ip2 = 0;
      return 1;
   }

   /* If secondary IP count outweighs primary (by a lot). and the secondary
    * is on the local net then swap 'em 
    */
   if((machost->from_ip2 > (machost->from_ip << 3)) &&
      (islocal_ip(machost->iface, machost->from_ip2)))
   {
      dtrap();

      /* Send ARP to both alleged IPs */
      send_arp_net(nets[machost->iface], machost->iphost->ipaddr);
      send_arp_net(nets[machost->iface], machost->ipaddr2);

      tmp = machost->iphost->ipaddr;   /* save actual address */
      machost->iphost = br_iplookup(machost->iface, machost->ipaddr2);  /* Get IP host struct */
      machost->ipaddr2 = tmp;
   }

   return 0;
}

/* br_swapipip2() - Swap hosts IP info with info 
 * for host->ipaddr2 
 */

void
br_swapipip2(struct host_entry * host)
{
   int      tmpct;
   ip_addr  tmpip;

  /* swap addresses */
   tmpip = host->ipaddr2;
   host->ipaddr2 = host->iphost->ipaddr;  /* old primary -> secondary */
   host->iphost = br_iplookup(host->iface, tmpip);    /* install new primary */
   if(host->iphost == NULL)               /* make sure it's known */
      host->iphost = iphost_new(host->iface, tmpip);
   host->iphost->mac_host = host;         /* set back pointer */

   /* swap counters too */
   tmpct = host->from_ip2;
   host->from_ip2 = host->from_ip;
   host->from_ip = tmpct;

   return;
}

 
/*
 * br_hostupdate() - this routine is called to process potential new 
 * host entries. It must first check the host database for a duplicate
 * and make a new entry if duplicate is not found.
 *
 * returns host_entry pointer if OK, NULL if error (no mem)
 */

struct host_entry *
br_hostupdate(char * macaddr, ip_addr ipaddr, int type, int iface)
{
   struct host_entry * host;

   /* See if we already know about this host */
   host = br_hostlookup(macaddr, iface);

   /* If unknown host, make a new entry for it */
   if(host == NULL)
   {
      host = new_host(iface, macaddr, ipaddr, type, -1);
      if(!host)      /* no memory- serious error */
         return NULL;

      host->lan = br_lanlookup(brstates[iface].current_lan);

      /* For new hosts, just mark lan info as stale and return */ 
      brstates[iface].evflags |= EV_LANSTALE;
      return host;
   }

   /* Fall to here if "host" was already known. */
   /* See if host is using a newly assigned DHCP address */
   if(host->dhcp_addr)
      br_newdhcpip(host, ipaddr);

   host->h_flags &= ~H_HIDDEN;   /* un-hide host if it was idle */

   /* If different IP addrs are in use on this mac, and the mac is NOT
    * known as the router, then record the secondary IP address.
    */
   if((host->iphost->ipaddr != ipaddr) &&
      (host != brstates[iface].br_router))
   {
      /* See if the first IP address has zero packets sent
       * from it. This can happen if old MAC host info was read
       * from a DB file after hibernation, but the IP address changed 
       * while the system was off-line.
       */
      if(host->from_ip == 0)
      {
         host->iphost = br_iplookup(host->iface, ipaddr);  /* install new primary */
         if(host->iphost == NULL)             /* make sure it's known */
         {
            host->iphost = iphost_new(iface, ipaddr);
            host->iphost->mac_host = host;        /* set back pointer */
         }
         else if(host->iphost->mac_host != host)
         {
            host->iphost->mac_host = host;  /* sanitize IP's MAC pointer */
         }
         brstates[iface].evflags |= EV_LANSTALE;
      }

      /* see if the IP address is MACs second IP address */
      if((ipaddr == host->ipaddr2) || (host->ipaddr2 == 0))
      {
         ip_addr mask = nets[iface]->n_mask;
         ip_addr net = mask & nets[iface]->n_ipaddr;

         /* Primary is set, and passed IP
          * does not match it. If the new IP is on our net 
          * and the existing primary is not
          * then install the new one as the primary IP.
          * Else set the new one as secondary.
          */
         if(((host->iphost->ipaddr & mask) != net) &&
            ((ipaddr & mask) == net) )
         {
            host->ipaddr2 = host->iphost->ipaddr;  /* old primary -> secondary */
            host->iphost = br_iplookup(iface, ipaddr);    /* install new primary */
            if(host->iphost == NULL)               /* make sure it's known */
               host->iphost = iphost_new(iface, ipaddr);
            host->iphost->mac_host = host;         /* set back pointer */
         }
         else if((ipaddr == host->ipaddr2) &&
            (host->from_ip2 > host->from_ip))
         {
            br_swapipip2(host);
         }
         else  /* new IP no better than old, make it secondary */
         {
            /* Set flag if both IP addresses are local. We already
             * know the old IP is local, so just test the new one:
             */
            if((ipaddr & nets[iface]->n_mask) == 
               (nets[iface]->n_ipaddr & nets[iface]->n_mask))
            {
               host->h_flags |= H_MULTILOCAL;
            }

            host->ipaddr2 = ipaddr;
            host->from_ip2++;
         }
         brstates[iface].evflags |= EV_LANSTALE;
      }
      else if(ipaddr != 0) /* didn't match either IP address */
      {
         host->h_flags |= H_MULTIIP;   /* host uses Multiple IPs */
         host->from_other++;

         /* If this MAC has a primary IP address which matches the 
          * configured routers then set it as router with a high 
          * confidence level, otherwise set this as router with 
          * a very low confidence level 
          */
         if(host->iphost && 
            (host->iphost->ipaddr == nets[iface]->n_router))
         {
            br_setrouter(iface, host, 215);
         }
         else
         {
            br_setrouter(iface, host, 25);
         }
         brstates[iface].evflags |= EV_LANSTALE;
      }
   }
   else if(ipaddr != 0)
   {
      host->from_ip++;
   }

   if(type == H_NETBEUI)
   {
      /* more testing to filter Samba? later */
      host->h_flags |= H_WINDOWS; /* Assume this is a windows machine. */
   }

   if((host->lan == NULL) && (brstates[iface].current_lan[0]))
      host->lan = br_lanlookup(brstates[iface].current_lan);

   return host;
}


/* br_newdhcpip() - See if a MAC host's IP address has been newly 
 * assigned. by DHCP.
 *
 * machost - MAC host being updated
 * ipaddr - Sender's IP  address, from IP header in 
 *          the updating packet
 *
 *
 * Returns TRUE if it is new Address from DHCP, else FALSE if not.
 */

int
br_newdhcpip(struct host_entry * machost, ip_addr ipaddr)
{
   struct ip_hostentry * iphost;

   /* Skip cases where there is no pending new DHCP address */   
   if(machost->dhcp_addr == 0)
      return FALSE;

   /* Skip cases where there is no IP host */   
   if(machost->iphost == NULL)
      return FALSE;

   /* See if host is using same old address */
   if(machost->iphost->ipaddr == ipaddr)
   {
      /* Yes, host ignored the last DHCP ACK. Clear lease: */
      machost->ip_lease = 0;

      /* See comment in brrx.c where dhcp_addr is set. */
      machost->dhcp_addr = 0;    /* Clear proposed new address */

      return FALSE;
   }

   /* See if host is using the newly assigned DHCP address */
   if(machost->dhcp_addr == ipaddr)
   {
      /* disconnect old IP address, install new one. The old IP
       * host object will be cleaned up by the timers.
       * ...later -JB-
       */
      machost->iphost->flags |= IH_DHCPEXPIRED;
      br_log(0, "DHCP reassigned %u.%u.%u.%u to %u.%u.%u.%u\n", 
         PASS_IPADDR(machost->iphost->ipaddr), PASS_IPADDR(ipaddr) );

      /* Since we're about to change the Ip address, clear the 
       * "duplicate IP" flag. If the new one is a duplicate it 
       * will be (re)flagged in iphost_new().
       */
      machost->h_flags &= ~H_DUPLIP;

      /* Make a new IP address. Dont try to recycle the old one, it
       * may contain stale information.
       */
      iphost = iphost_new(machost->iface, ipaddr);
      if(!iphost)
      {
         dtrap();    /* out of heap? */
         return FALSE;
      }
      machost->iphost = iphost;
      machost->iphost->flags |= (IH_LOCAL | IH_DHCP);
      machost->dhcp_addr = 0;
   }

   /* See if this is us */
   if(MAC_EQ(machost->macaddr, nets[machost->iface]->n_macaddr))
   {
      nets[machost->iface]->n_ipaddr = ipaddr;

      br_log(0, "DHCP reassigned our IP to %u.%u.%u.%u\n", 
         PASS_IPADDR(ipaddr) );
   }

   brstates[machost->iface].evflags |= EV_LANSTALE;
   return TRUE;
}


/* br_setrouter() - set the passed host as a router */

/* time of last "non-local" message, so we don't send out hundreds of 
 * these a second 
 */
static u_long   last_nlmessage;
 
void
br_setrouter(int iface, 
             struct host_entry * machost, 
             int confidence)
{
   int changedlan = FALSE;

   brstates[iface].evflags |= EV_ROUTERSET;

   if(machost->iphost == NULL)
   {
      dtrap();
      return;
   }
   if(islocal_ip(iface, machost->iphost->ipaddr) == FALSE)
   {
      /* We often get here in early startup because all the packets from 
       * the routers MAC have been from remote IP addresses, and thus we have
       * received no info about the routers real(local) IP address. To try to
       * fix this, ARP for the pre-existing iface router. 
       */
      if(nets[iface]->n_router)
         send_arp_net(nets[iface], nets[iface]->n_router);
      if( (last_nlmessage + (2*TPS)) < cticks)
      {
         dprintf("setrouter: non-local iphost passed, machost %s, confidence: %d\n",
            br_prtmac(machost->macaddr), confidence);
         last_nlmessage = cticks;
      }

      return;
   }
   if(br_isbcast(iface, machost->iphost->ipaddr))
   {
      /* arp for the pre-existing iface router. */
      if(nets[iface]->n_router)
         send_arp_net(nets[iface], nets[iface]->n_router);
      dprintf("setrouter: broadcast iphost passed, machost %s, confidence: %d\n",
         br_prtmac(machost->macaddr), confidence);

      return;
   }

   /* If the new submission has a higher level of confidence than
    * the current router, and the new one is not the current 
    * one, then install the new one.
    */
   if(confidence > brstates[iface].br_rtconf)
   {
      /* if passed router is different than existing one, install it */
      if(brstates[iface].br_router != machost)
      {
         /* switch routers flags on new and old routers */
         if(brstates[iface].br_router)    /* Clear existing router? */
            brstates[iface].br_router->h_flags &= ~H_ROUTER;

         machost->h_flags |= H_ROUTER;    /* Flag New router */

         /* printf about if we're installing new or replacing existing */
         if(brstates[iface].br_router)    /* router already set? */
         {
            dprintf("Changing router on net %d", iface );
            dprintf(" - New IP: %u.%u.%u.%u, old IP: %u.%u.%u.%u\n",
               PASS_IPADDR(machost->iphost->ipaddr),
               PASS_IPADDR(brstates[iface].br_router->iphost->ipaddr));
            dprintf("   New router MAC %s, old MAC %s\n",
               br_prtmac(machost->macaddr),
               br_prtmac(brstates[iface].br_router->macaddr));
            changedlan = !MAC_EQ(machost->macaddr, 
               brstates[iface].br_router->macaddr);
         }
         else
         {
            dprintf("Initial set of router on net %d; IP: %u.%u.%u.%u, mac: %s\n",
               iface, PASS_IPADDR(machost->iphost->ipaddr),
               br_prtmac(machost->macaddr));
         }
         brstates[iface].br_router = machost;
         if(changedlan)
         {
            /* Get here if we're on a new network */
            br_lanchanged();
         }
      }

      /* Update the confidence level */
      dprintf("New router confidence: %d  old: %d\n",
         confidence, brstates[iface].br_rtconf); 

      brstates[iface].br_rtconf = confidence;

      /* We may have been unable to read LAN info because we
       * didn't know our router at init time, so try again. 
       */
      if(lan_file_read == 0)
         br_readlans();
      if( !(brstates[br_ifview].evflags & EV_NETNAMED))
         br_checklan();
   }
   return;
}

void
br_clearrouter(int iface)
{
   dprintf("br_clearrouter: iface %d\n", iface);
   brstates[iface].br_router = NULL;   /* clear router info */
   brstates[iface].br_rtconf = 0;
   brstates[iface].evflags &= ~EV_ROUTERSET;

   return;
}

/* find or make an IP address host DB entry for the IP address passed */

struct ip_hostentry *
br_ipupdate(ip_addr ipaddr,      /* IP address of connection, in net endian */
      struct conn_entry * conn,  /* connection with IP address */
      char * mac)                /* MAC address packet came from */
{
   struct ip_hostentry * iphost;

   /* filter out NULL IP addresses */
   if(ipaddr == 0)
      return NULL;

   /* See if host already exists in list */
   iphost = br_iplookup(conn->ce_iface, ipaddr);

   /* If not, make new entry. Fill in from passed info */
   if(!iphost)
   {
      struct host_entry * machost;

      iphost = iphost_new(conn->ce_iface, ipaddr);
      if(!iphost)
      {
         dtrap();
         return NULL;
      }

      /* associate the IP host with a MAC address. */
      machost = br_hostlookup(mac, conn->ce_iface);
      if(machost == NULL)
      {
         /* This can happen on early DHCP packets the dest MAC is 
          * broadcast, and it's not in the table yet.
          */
         if(conn->ce_type == H_UDP)
         {
            switch(conn->rem.ce_port)
            {
            case BOOTP_SERVER_PORT:
            case BOOTP_CLIENT_PORT:
               machost = new_host(conn->ce_iface, mac, ipaddr, H_DHCP, -1);
               break;
            case 127:
            case 128:
            case 129:
               machost = new_host(conn->ce_iface, mac, ipaddr, H_NETBEUI, -1);
            default:
               machost = new_host(conn->ce_iface, mac, ipaddr, H_UDP, -1);
            }
         }
         if(machost == NULL)
         {        
            dtrap();    /* out of heap? */
            return NULL;
         }
         /* Do some sanity testing - if this is dest MAC DHCP, then 
          * it should be broadcast.
          */
         if(*mac != (char)0xFF)
         {
            dtrap();    /* Is this a problem? */
         }
      }

      /* machost variable is now set */

      /* Set special case flag for broadcast, multicast IP addresses */
      if(br_isbcast(conn->ce_iface, ipaddr))
      {
         iphost->flags |= IH_BROADCAST;
         strcpy(iphost->dnsname, "Broadcast");
      }
      else if((htonl(ipaddr) & 0xE0000000) == 0xE0000000)
      {
         /* flag as multicast IP addresses */
         iphost->flags |= IH_MULTICAST;
         strcpy(iphost->dnsname, "Multicast");
      }
      else     /* new IP is not broadcast or mcast */
      {
         iphost->mac_host = machost;   /* set IP to MAC pointer */

         /* Assigning the MAC to IP pointer is a bit trickier, since a 
          * single MAC host (especially the router) can be the source of 
          * many IP addresses.
          */
         if(machost->iphost == NULL)
         {
            /* do a sanity check that this is our MAC entry */
            if(machost->ipaddr2 != nets[conn->ce_iface]->n_ipaddr)
            {
               dtrap();    /* can this happen?? */
            }
            machost->iphost = iphost;  /* Assign winsock IP as primary MAC IP */
         }

         /* If the IP address does not match either MAC IP address, then 
          * it's a good bet the MAC device is a router.
          */
         if((machost->iphost->ipaddr != ipaddr) && 
            (machost->ipaddr2 != ipaddr))
         {
            br_setrouter(conn->ce_iface, machost, 75);   /* low confidence */
         }
      }
   }

   /* This is called exactly once for each host upon connection creation, 
    * so bump the IP hosts connection count. Call alerter if too high.
    */
   if(++iphost->conns > (u_long)br_cfg.alarm_connperhost)
      alarm_general(&conn_limit, &iphost->cflood, iphost);

   return iphost;
}



/* mini "ports" value structure to overlay on the part of the
 * transport header which we want to use as port values. For
 * TCP and UDP this is the first 4 bytes of the header, however
 * for ICMP (ping) it is the 5th thru 8th bytes.
 */
struct ports
{
   unshort src;
   unshort dest;
};

/* nb_decodename(u_char * nbname,  u_char * hostname) 
 *
 * Turns Netbeui "half-ascii" gibberish into a textual name. 
 * Algorithm from RFC1001.
 *
 * Returns number of chars in output buffer.
 *
 */

int
nb_decodename(u_char * haname,   /* IN - the netbios gibberish */
              u_char * hostname) /* out - the decrypted name */
{
   u_char   bytetmp;
   u_char   bytecode;
   int      len;

   bytecode = *haname++;     /* these always start with 0x20 */

   for(len = 0; len < (DNSNAMELEN - 1); len++)
   {
      bytetmp = *haname++ - 'A';
      bytecode = bytetmp << 4;
      bytetmp = *haname++ - 'A';
      bytecode |= bytetmp;
      if(bytecode <= 0x20)    /* Space ends the half-ascii name */
         break;
      *hostname++ = bytecode;
   }

   *hostname = 0;       /* null terminate output */
   return len;
}

/* br_extractNBname()
 *
 * Update name (and maybe other info) from Windoze NetBios packets 
 */

void
br_extractNBname(int iface, u_char * nbdata, struct endpoint * sender)
{
   int      namesum;
   struct host_entry *     machost;
   struct ip_hostentry *   iphost;
   
   /* Get MAC info of sender */
   machost = br_hostlookup((char*)sender->ce_mac->macaddr, iface);
   if(!machost)
   {
      dtrap();
      return;
   }

   /* Verfify IP sender's IP address info matched MAC sender */
   iphost = sender->ce_iphost;
   if(!iphost)
   {
      dtrap();
      return;
   }

   /* The Bozos in Redmond strike again. 
    *
    * Multihomed Windows systems (laptops w/ WiFi and Ethernet) send Netbios 
    * bogons with the MAC address of the sending iface and the IP address of 
    * the other iface. While technically not illegal, this is a really bad 
    * idea. We process these here. 
    */

   if(machost->iphost != iphost)
   {
      machost->h_flags |= H_MULTIHOMED;
      iphost = machost->iphost;
   }

   /* Don't overwrite higher priority (user assigned) names */
   if((iphost->mac_host->h_flags & H_NAMEMASK) > H_NETBIOSNAME)
      return;

   namesum = sumname(machost->locname);
   nb_decodename((u_char*)nbdata, (u_char*)machost->locname);
   if(namesum != sumname(machost->locname))
   {
      dprintf("Netbios (re)named  %s", print_mac(machost->macaddr));
      dprintf(" as %s\n", machost->locname);

      if(*machost->locname == '*')
         dtrap();

      machost->h_flags &= ~H_NAMEMASK;
      machost->h_flags |= H_NETBIOSNAME;
      brstates[iface].evflags |= EV_LANSTALE;
   }

   return;
}


u_long br_connrtt(struct conn_entry * conn)
{
   struct timeval tv;
   u_long   u_sec;

   gettimeofday( &tv ); 
   if(tv.tv_sec < conn->ce_seqtime.tv_sec)
      dtrap();

   u_sec = (tv.tv_sec - conn->ce_seqtime.tv_sec) * 1000000;
   u_sec += tv.tv_usec - conn->ce_seqtime.tv_usec;

   return u_sec;
}

/* br_ckforgery()
 *
 * Check received reset or fin for ISP forgery.
 *
 * If reset or fin is for timed sequence number or later, then
 * check it;s rount trip time. If it's 2x faster than average
 * then treat it as a forgery and post an alert.
 */

void br_ckforgery(struct conn_entry * conn, 
                  struct ip * pip, 
                  struct tcphdr * tcp)
{
   u_long rtt;

   if(conn->ce_timedseq == 0)    /* NN - Not being timed? */
      return;

   /* Don't fuss about our own forgeries! */
   if(conn->ce_state == TCPS_KILLED)
      return;

   /* Make sure reset/fin seq is not older than timed seq */ 
   if( htonl(tcp->th_ack) < conn->ce_timedseq)
      return;

   /* Ignore if source is local. */
   if(islocal_ip(conn->ce_iface, pip->ip_src))
      return;

   /* calulate time since we sent timed seq */
   rtt = br_connrtt(conn);

   if(rtt < (conn->ce_rtt/2))
   {
      if(tcp->th_flags & TH_RST)
      {
         conn->ce_flags |= CEF_FAKERESET;
         alarm_general(&forged_tcp, &conn->rem.ce_iphost->forgery, conn);
      }
//     else
//     {
//         conn->ce_flags |= CEF_FAKEFIN;
//         alarm_general(&forged_tcp, &conn->rem.ce_iphost->forgery, conn);
//     }

   }

   return;
}


/* br_cnupdate() - Update connection table. 
 *
 * All headers should still be in network endian.
 *
 * This is expected to be called at ISR context for each received packet.
 *
 * returns connection associated with packet.
 *
 * Also, *remotesource is TRUE if sender is remote node; FALSE if sender 
 * is local node or sender's location (local/remote) is not known
 */

struct conn_entry *
br_cnupdate(struct ip * pip,  /* packet IP header */
            char * macaddrs,  /* packets MAC addresses *dest, src) */
            void * trhead,    /* packet transport layer header */
            int type,         /* packet type (H_TCP, etc) */
            int iface,        /* Interface pkt was received on */
            int * remotesource)/* OUT - TRUE if packet is from remote node. */
{
   int      hashval;
   unshort  stateflags;       /* TCP header state bits (ACK, FIN, etc.) */
   char *   localmac;         /* MAC address of local machine */
   char *   remotemac;        /* MAC address of "remote" machine */
   struct conn_entry * conn;
   struct conn_entry * freecn;
   struct ports icmpports;    /* fake ports for ICMP */
   struct ports * ports;      /* pointer to ports, for UDP and TCP */
   struct tcphdr * tcp;
   struct udp * pup;
   struct icmp * icmp;
   struct endpoint * local;
   struct endpoint * remote;
   struct endpoint * sender;

   switch(type)
   {
   case  H_TCP:
      ports = (struct ports *) trhead;
      break;
   case H_UDP:
      ports = (struct ports *) trhead;
      break;
   case H_ICMP:
      {
      unshort icmptype;
      /* Use icmp type field as surrogate for ports. For paired types 
       * (echo request, echo reply) we assign the request to the initiator
       * and the reply to the replier. This prevents a sequence of pings
       * from showing up as multiple connections.
       */
      icmptype = (unshort)(((struct icmp *)trhead)->type);
      switch(icmptype)
      {
      case ECHOREQ:
         icmpports.src = htons(ECHOREQ);
         icmpports.dest = htons(ECHOREP);
         break;
      case ECHOREP:
         icmpports.src = htons(ECHOREP);
         icmpports.dest = htons(ECHOREQ);
         break;
      case TIMEREQ:
         icmpports.src = htons(TIMEREQ);
         icmpports.dest = htons(TIMEREP);
         break;
      case TIMEREP:
         icmpports.src = htons(TIMEREP);
         icmpports.dest = htons(TIMEREQ);
         break;
      }
      ports = &icmpports;
      break;
      }
   case H_OTHER:
      /* This may never happen, but don't leave ports == null. just
       * in case....
       */
      ports = &icmpports;
      dtrap();    /* anything to track here? later -JB- */
      break;
   default:
      dtrap();    /* unknow type == programming error */
      return NULL;
   }
      
   /* Find hash slot for connection entry. If it does not already exist,
    * we will need to create it.
    */
   hashval = HASH_CE(pip->ip_src, pip->ip_dest, 
         htons(ports->src), htons(ports->dest) );

   for(conn = ce_table[hashval]; conn; conn = conn->next )
   {
      if(conn->ce_type != type)
         continue;

      if(conn->loc.ce_addr == pip->ip_src)
      {
         /* check for outbound packet */
         remote = &conn->rem;
         local = &conn->loc;
         if((remote->ce_addr == pip->ip_dest) &&
            (local->ce_port == htons(ports->src)) &&
            (remote->ce_port == htons(ports->dest)))
         {
            break;
         }
      }
      else if(conn->loc.ce_addr == pip->ip_dest)
      {
         /* check for inbound packet */
         remote = &conn->rem;
         local = &conn->loc;
         if((remote->ce_addr == pip->ip_src) &&
            (local->ce_port == htons(ports->dest)) &&
            (remote->ce_port == htons(ports->src)))
         {
            break;
         }
      }
   }

   /* if no connection was found then we need to make one */
   if(conn == NULL)
   {
      /* If we've hit our connection table size limit, then
       * generate an alarm and return 
       */
      if(brstates[iface].total_conns >= br_cfg.max_connections)
      {
         alarm_general(&conn_limit, &gen_cflood, NULL);
         return NULL;
      }
      conn = CN_ALLOC(sizeof(*conn));
      if(!conn)
      {
         dtrap();
         return NULL;
      }

      /* Record high & low watermarks of connection addresses. 
       * This allows web page CGI code to pass connections as 
       * an address (in decimal, yet) and then verify that the 
       * conneciton address is still valid when it's returned 
       * in a web form later. This check prevents timed out 
       * connections and perhaps some malicious attacks.
       */
      if(conn_highest < conn)
         conn_highest = conn;
      if(conn_lowest > conn)
         conn_lowest = conn;

      brstates[iface].total_conns++;
      remote = &conn->rem;
      local = &conn->loc;
      conn->hashx = hashval;

      /* add new connection to hash table */
      if(ce_table[hashval] == NULL)    /* bucket is empty */
         ce_table[hashval] = conn;     /* add at head of bucket */
      else     /* Add at the end of the bucket's linked list */
      {
         freecn = ce_table[hashval];
         while(freecn->next)
            freecn = freecn->next;
         freecn->next = conn;
      }

      conn->ce_type = type;      /* set connection type */
      conn->ce_iface = iface;    /* record iface it was found on */
      conn->ce_start = cticks;   /* birth time */

      /* figure our which IP address to treat as the "local" address */

      /* See is we can figure out which way the packet's going - local to inet,
       * inet to local, or local to local. If the sender's IP is on our subnet
       * Then we assume the sender is the "local" node. The
       * dest IP gets set up as "remote", even if he too is on our subnet.
       */
      if((pip->ip_src & nets[iface]->n_mask) == 
         (nets[iface]->n_ipaddr & nets[iface]->n_mask))
      {
         /* Source is local */
         local->ce_addr = pip->ip_src;
         local->ce_port = htons(ports->src);
         remote->ce_addr = pip->ip_dest;
         remote->ce_port = htons(ports->dest);
         localmac = macaddrs + 6;
         remotemac = macaddrs;

         /* if dest is also local, set flag for both local */
         if((pip->ip_dest & nets[iface]->n_mask) == 
            (nets[iface]->n_ipaddr & nets[iface]->n_mask))
         {
            conn->ce_flags |= CEF_LOCALNODES;
         }
         /* If ICMP, then since this is the first packet, sender is client */
         if(type == H_ICMP)
            conn->ce_flags |= CEF_LOCALCLIENT;
      }
      else 
      {
         /* source is non-local. */
         local->ce_addr = pip->ip_dest;   /* Assume dest is local */
         local->ce_port = htons(ports->dest);
         remote->ce_addr = pip->ip_src;    /* source is foreign */
         remote->ce_port = htons(ports->src);
         localmac = macaddrs;     /* set local (dest) mac address */
         remotemac = macaddrs + 6;
         if(type == H_ICMP)
            conn->ce_flags |= CEF_LOCALSERVER;
      }

#ifdef NOTDEF
      if((remote->ce_addr == 0x6601a8c0) && 
         ((conn->ce_flags & CEF_LOCALNODES) == 0))
      {
         dtrap();    /* tmp debug -JB- */
      }
#endif
      /* lookup sender */
      local->ce_mac = br_hostlookup(localmac, conn->ce_iface);

      if(!local->ce_mac) 
      {
         br_hostupdate(localmac, local->ce_addr, type, iface);
         local->ce_mac = br_hostlookup(localmac, conn->ce_iface); /* try again */
      }

      /* find or create IP host DB entry for src IP address */
      if((pip->ip_src != 0) && (local->ce_addr == pip->ip_src))
      {
         struct ip_hostentry * ip2;

         local->ce_iphost = br_ipupdate(pip->ip_src, conn, localmac);
         if(local->ce_mac != local->ce_iphost->mac_host)
         {
            /* The sending IP address does not match the one we have 
             * assigned as the primary IP of the sending MAC address.
             * This is mostly Microsoft stupidity (see note below) but
             * Can happen in rare other cases. Check and see if we need 
             * to reassign the secondary IP as the primary.
             */
            if((pip->ip_src == local->ce_mac->ipaddr2) &&
               (local->ce_mac->from_ip2 > local->ce_mac->from_ip))
            {
               dtrap();
               ip2 = br_iplookup(iface, local->ce_mac->ipaddr2);

               /* Got the IP entry for the "better" ip2. Swap it into 
                * the MAC structure as the primary IP. 
                */
               if(ip2)
               {
                  br_swapipip2(local->ce_mac);
               }
            }
         }
      }
      else
         remote->ce_iphost = br_ipupdate(pip->ip_src, conn, remotemac);

      /* If destination IP is broadcast, then find/create a pseudo  "broadcast" host
       * and set that as the remote IP host.
       */
      if(br_isbcast(iface, pip->ip_dest))
      {
         remote->ce_iphost = br_ipupdate(pip->ip_dest, conn, remotemac);
      }

#ifdef NOTDEF
      /* The code below was added in the early days, but 
       * has run into yet another ugly windows stunt. The code was 
       * intended to keep the MAC host entry pointing to the latest 
       * IP info. But Windoze machines with both WiFi and Ethernet 
       * connections have been caught broadcasting Netbios packets
       * with the source MAC of the Ethernet iface, but the source 
       * IP address of the WiFi iface. The idiots in Redmond strike 
       * again, managing to screw up tables on anyone (like us) who's
       * paying attention to this stuff. Workaround: try deleting the 
       * code below. But when, then should we change IP info for
       * for non-DHCP Windows ifaces? later... -JB- 5/28/06
       */
      if(local->ce_iphost)
         local->ce_iphost->mac_host = local->ce_mac;
#endif /* NOTDEF */

      /* If foreign machine is also on local net, set it's info */
      if((remote->ce_addr & nets[iface]->n_mask) == 
         (nets[iface]->n_ipaddr & nets[iface]->n_mask))
      {
         remote->ce_mac = br_hostlookup(remotemac, conn->ce_iface);  /* lookup other MAC */
      }
      else
      {
         remote->ce_mac = NULL;
      }
   }
   else     /* connection struct already existed */
   {
      /* These are already set, but Microsnot C can't figure that out. Set
       * them again to stop compiler warnings 
       */
      remote = &conn->rem;
      local = &conn->loc;

      /* set local & remote MAC addresses */
      if(local->ce_addr == pip->ip_src)
      {
         localmac = macaddrs + 6;     /* set local (src) mac address */
         remotemac = macaddrs;
      }
      else  /* source is remote */
      {
         localmac = macaddrs;     /* set local (dest) mac address */
         remotemac = macaddrs + 6;

         remote->ce_mac = br_hostlookup(remotemac, conn->ce_iface);  /* lookup other MAC */
         if(!remote->ce_mac)
         {
            br_hostupdate(remotemac, remote->ce_addr, type, iface);
            remote->ce_mac = br_hostlookup(remotemac, conn->ce_iface);
         }
         /* if the remote IP is not on the subnet, then the MAC is probably
          * a router. 
          */
         if(nets[iface]->n_mask &&        /* make sure IP info is set up */
            ((remote->ce_addr & nets[iface]->n_mask) != 
            (nets[iface]->n_ipaddr & nets[iface]->n_mask)))
         {
            if(remote->ce_mac)
               remote->ce_mac->h_flags |= H_MULTIIP;
         }
      }
   }
   /* Now  we have found or created a connection - update it */

   if(local->ce_addr == pip->ip_src)   /* local is source */
   {
      *remotesource = FALSE;
      local->ce_outpkts++;        /* this is outbound packet */
      local->ce_outbytes += htons(pip->ip_len);

      /* If the first packet was from the remote, and this is the second 
       * packet, we may not yet have a local iphost set in the connection
       * structure. If this is the case, then get one now. Arguably this 
       * test should be done in the "connection already existed" case
       * above, but doing it here lets us recover if we ran out of memory
       * during a previous try. maybe the conn->ce_iphost should default
       * to a global static placeholder host struct if memory runs out 
       * during br_ipupdate() - look into this later -JB-
       */
      if(local->ce_iphost == NULL)
      {
         local->ce_iphost = br_ipupdate(pip->ip_src, conn, localmac);
         if(local->ce_mac != local->ce_iphost->mac_host)
         {
            dprintf("br_cnupdate: br_ipupdate mismatch\n");
            dprintf("   pkt local mac %s (%s), IP %u.%u.%u.%u \n",
               br_prtmac(local->ce_mac->macaddr), 
               br_getmacname(local->ce_mac),
               PASS_IPADDR(pip->ip_src) );

            dprintf("   IP mac %s (%s), IP %u.%u.%u.%u \n",
               br_prtmac(local->ce_iphost->mac_host->macaddr), 
               br_getmacname(local->ce_iphost->mac_host),
               PASS_IPADDR(local->ce_iphost->ipaddr) );

            /* This can happen if DHCP changes an IP from one machine 
             * to another and the agent misses the change (powered 
             * off, etc). If The other MAC (not the one sending this 
             * packet) has already assumed another IP address, then
             * disconnect the pip->ip_src entry.
             */
            if(local->ce_iphost->mac_host->iphost->ipaddr != 
               local->ce_iphost->ipaddr)
            {
               dprintf("reassigning iphost %u.%u.%u.%u to MAC %s\n",
                  PASS_IPADDR(local->ce_iphost->ipaddr),
                  br_prtmac(local->ce_mac->macaddr) );
               local->ce_iphost->mac_host = local->ce_mac;
            }
         }
      }
      else
         local->ce_iphost->tot_pkts++;
   }
   else        /* this is inbound packet */
   {
      *remotesource = TRUE;
      remote->ce_outpkts++;
      remote->ce_outbytes += htons(pip->ip_len);
      /* We know the remote IP machine exists since it sent the packet; so
       * make sure we have an IP host entry for it.
       */
      if(remote->ce_iphost == NULL)
         remote->ce_iphost = br_ipupdate(pip->ip_src, conn, remotemac);
      else
         remote->ce_iphost->tot_pkts++;
   }

   /* Update the activity timestamp on the IP hosts */
   conn->ce_time = cticks;
   if(local->ce_iphost)
   {
      local->ce_iphost->tm_lastpkt = cticks;
   }
   if(remote->ce_iphost)
   {
      remote->ce_iphost->tm_lastpkt = cticks;

      /* If remote port is low value, then remote host is a server (at
       * least sometimes), so tag it as a server 
       */
      if((remote->ce_port < 1024) && 
         (conn->ce_type != H_ICMP))
      {
         remote->ce_iphost->flags |= IH_SERVER;
      }
   }
   else
   {
      /* Get here if remote host is not set yet. If it's the source, find/create
       * it and set it. 
       */
      if(pip->ip_src == remote->ce_addr)
      {
         dtrap();    /* can this still happen? -JB- */
         remote->ce_iphost = br_ipupdate(pip->ip_src, conn, remotemac);
      }
   }

   /* update IP connection  NN stats */
   conn->ce_sec_pkts++;
   conn->ce_sec_bytes += pip->ip_len;

   if(*remotesource)
   {
      if(conn->ce_hops == 0)
         conn->ce_hops = pip->ip_time;
      else if(conn->ce_hops > pip->ip_time)
         conn->ce_hops--;
      else if(conn->ce_hops < pip->ip_time)
         conn->ce_hops++;
   }

   switch(type)
   {
      case H_TCP:

         tcp = (struct tcphdr *)trhead;
         /* update sender's next expected sequence number */
         if(local->ce_addr == pip->ip_src)   /* local machine is sender? */
         {
            conn->loc.ce_seq = htonl(tcp->th_seq) + 
               htons(pip->ip_len) - (20 + ((tcp->th_doff & 0xf0) >> 2));
            if(tcp->th_flags & TH_SYN)
               conn->loc.ce_seq++;
         }
         else
         {
            conn->rem.ce_seq = htonl(tcp->th_seq) + 
               htons(pip->ip_len) - (20 + ((tcp->th_doff & 0xf0) >> 2));
            if(tcp->th_flags & TH_SYN)
               conn->rem.ce_seq++;
         }

         /* guesstimate the state based on flag bits and what we already know */
         stateflags = tcp->th_flags & (TH_SYN | TH_ACK | TH_FIN | TH_RST);

         switch(stateflags)
         {
         case TH_SYN:
            /* Mark sender as client */
            if(local->ce_addr == pip->ip_src)   /* local machine is sender? */
               conn->ce_flags |= CEF_LOCALCLIENT;  /* set local==client flag */
            else     /* foreign machine is sender */
               conn->ce_flags |= CEF_LOCALSERVER;  /* set local==server flag */

            if(conn->ce_state < TCPS_SYN_SENT)
               conn->ce_state = TCPS_SYN_SENT;

            /* On SYN packets, look for potential problems */
            switch(conn->rem.ce_port)
            {
            /* MS Netbios crap has no business coming in over the router. If
             * this is such a packet, flag remote host as a worm.
             */
            case 136:   /* damn, they wasted a lot of ports :-( */
            case 137:
            case 138:
            case 139:
            case 445:
            case 500:
               if(conn->rem.ce_iphost)
                  strcpy(conn->rem.ce_iphost->imagefile, "msworm.gif");
               break;
            default:
               break;
            }
            break;
         case (TH_SYN | TH_ACK):
            if((conn->ce_flags & CEF_SERVERMASK) == 0)   /* server not set? */
            {
               /* SYN/ACKs are usually sent by servers responding to SYNs */
               if(local->ce_addr == pip->ip_src)   /* local machine is sender? */
                  conn->ce_flags |= CEF_LOCALSERVER;  /* set local==server flag */
               else
                  conn->ce_flags |= CEF_LOCALCLIENT;  /* set local==client flag */
            }
            if(conn->ce_state < TCPS_SYN_RECEIVED)
               conn->ce_state = TCPS_SYN_RECEIVED;

            /* On SYN/ACK packets, assign host image based on common services */
            switch(conn->rem.ce_port)
            {
            case POP3_PORT:     /* remote is POP3 server */
               if(conn->rem.ce_iphost->imagefile[0] == 0)
                  strcpy(conn->rem.ce_iphost->imagefile, "mailbox1.gif");
               break;
            case SMTP_PORT:     /* remote is SMTP server */
               if((conn->rem.ce_iphost->imagefile[0] == 0) ||
                  (strcmp(conn->rem.ce_iphost->imagefile, "mailbox1.gif") == 0))
               {
                  strcpy(conn->rem.ce_iphost->imagefile, "mailbox2.gif");
               }
               break;
            default:
               break;
            }
         case TH_ACK:
            if(*remotesource)
            {
               if(conn->ce_timedseq != 0)    /* NN - Being timed? */
               {
                  /* Packet form remote machine, see if this is ack 
                   * for timed sequence. 
                   */
                  if( htonl(tcp->th_ack) >= conn->ce_timedseq)
                  {
                     /* calulate round trip time */
                     u_long rtt = br_connrtt(conn);

                     if(conn->ce_rtt == 0)
                        conn->ce_rtt = rtt;
                     else if(rtt < conn->ce_rtt)
                        conn->ce_rtt--;
                     else if(rtt > conn->ce_rtt)
                        conn->ce_rtt++;

                     conn->ce_timedseq = 0;  /* Clear timed sequence */
                  }
               }
            }
            else  /* packet is from local machine */
            {
               if(conn->ce_timedseq == 0)    /* NN - Not currently being timed? */
               {
                  conn->ce_timedseq = htonl(tcp->th_seq);   /* Start timing */
                  gettimeofday( &conn->ce_seqtime );
               }
            }

            if(conn->ce_state == TCPS_ESTABLISHED)
               break;      /* fast path for most packets */
            if(conn->ce_state < TCPS_ESTABLISHED)
               conn->ce_state = TCPS_ESTABLISHED;
            else if(conn->ce_state >= TCPS_LAST_ACK)   /* ACK of last FIN */
               conn->ce_state = TCPS_CLOSED;
            break;
         case (TH_ACK | TH_FIN):
         case TH_FIN:
            if(*remotesource)
            {
               conn->ce_rmfins++;            /* NN variable */
//               br_ckforgery(conn, pip, tcp);
            }
            else
               conn->ce_lofins++;            /* NN variable */

            if(conn->ce_state <= TCPS_ESTABLISHED)
               conn->ce_state = TCPS_CLOSE_WAIT;
            else if(conn->ce_state == TCPS_CLOSE_WAIT)
               conn->ce_state = TCPS_CLOSING;
            else if(conn->ce_state == TCPS_CLOSING)
               conn->ce_state = TCPS_LAST_ACK;
            break;
         case (TH_RST | TH_FIN | TH_ACK):
         case (TH_RST | TH_FIN):
         case TH_RST:
            if(*remotesource)
            {
               conn->ce_rmresets++;          /* NN variable */
               br_ckforgery(conn, pip, tcp);
            }
            else
               conn->ce_loresets++;          /* NN variable */

            if(conn->ce_state != TCPS_KILLED)
               conn->ce_state = TCPS_RESET;     /* State made up */
            break;
         }
         break;
      case H_UDP:
         pup = (struct udp *)trhead;
         /* for UDP, the end with the higher port number is (usually) the client */
         if(local->ce_port >= remote->ce_port)
            conn->ce_flags |= CEF_LOCALCLIENT;  /* set local==client flag */
         else
            conn->ce_flags |= CEF_LOCALSERVER;

         /* get iphost for the IP who sent this */
         if(pip->ip_src == local->ce_addr)
            sender = &conn->loc;
         else
            sender = &conn->rem;

         /* See if sender is on local net */
         if(islocal_ip(conn->ce_iface, sender->ce_addr))
         {
            struct msworm_hdr *     namehdr;
            struct netbios_dgram *  nbhdr;
            unshort                 opcode;


            /* Check for Microsnot Windoze Netbios name packets. If this packet 
             * is one of those, then see if we can extract Windows name. 
             */
            switch(conn->loc.ce_port)
            {
            case 137:      /* potential NBNS name query reply */
               if(sender->ce_mac)
                  sender->ce_mac->h_flags |= H_WINDOWS;

               namehdr = (struct msworm_hdr *)(pup + 1);
               opcode = htons(namehdr->flags) & 0xF800;

               if((opcode == 0x8000) &&    /* name query response */
                  (htons(namehdr->qrrs) == 0) && 
                  (strncmp((char*)(&namehdr->data), " CKAAAAA", 8)))
               {
                  br_extractNBname(iface, &namehdr->data, sender); 
               }
               break;

            case 138:      /* Netbios datagram service port */
               if(sender->ce_mac)
                  sender->ce_mac->h_flags |= H_WINDOWS;

               nbhdr = (struct netbios_dgram *)(pup + 1);

               /* three types of datagram have the sender name 
                * embedded in the header 
                */
               if((nbhdr->type == 16) ||
                  (nbhdr->type == 17) ||
                  (nbhdr->type == 18))
               {
                  br_extractNBname(iface, &nbhdr->data, sender); 
               }
               break;

            default:
               break;
            }
         }
         else     /* sender is NOT local */
         {
            /* Certain UDP packets from outside get special images */
            if(conn->rem.ce_iphost)     /* make sure we have a host set */
            {
               unshort  opcode;
               struct   msworm_hdr * worm;  /* potential worm, via Microsoft */
               char *   imagefile;

               imagefile = NULL;    /* use NULL as flag */

               switch(conn->loc.ce_port)
               {
               case 111:     /* Alleged "RPC" (carries popups) */
               case 445:     /* Microsoft "POPUP" packet */
                  imagefile = "mspopmsg.gif";
                  break;
               case 135:     /* Various Microsoft Netbeui packets */
               case 136:
               case 137:
               case 138:
                  worm = (struct msworm_hdr * )(pup + 1);
                  opcode = htons(worm->flags);
                  if((opcode & 0x7000) == 0)
                     imagefile = "msworm.gif";
                  break;
               default:
                  break;
               }
               /* If no IP host image is set, and we have a candidate, set it */
               if((conn->rem.ce_iphost->imagefile[0] == 0) && 
                  (imagefile != NULL))
               {
                  strcpy(conn->rem.ce_iphost->imagefile, imagefile);
               }
            }
         }
         break;
      case H_ICMP:
         icmp = (struct icmp *)trhead;
         /* Count packet & save header */
         if(*remotesource)
         {
            if(remote->ce_iphost)
               remote->ce_iphost->icmp_pkts++;
            MEMCPY(remote->app, icmp, sizeof(struct icmp) );
         }
         else
         {
            if(local->ce_iphost)
               local->ce_iphost->icmp_pkts++;
            MEMCPY(local->app, icmp, sizeof(struct icmp) );
         }

         /* ICMPs from outside get "spy" image */
         if((icmp->type == ECHOREQ) && 
            ((conn->ce_flags & CEF_LOCALNODES) == 0) &&
            (conn->rem.ce_iphost))     /* make sure we have a host set */
         {
            /* See if remote host is doing a ping scan */
            if(conn->rem.ce_iphost->icmp_pkts > conn->rem.ce_outpkts)
            {
               /* Yikes! he's scanning us.... */
               strcpy(conn->rem.ce_iphost->imagefile, "msworm.gif");
               alarm_general(&remote_scan, &conn->rem.ce_iphost->scan, 
                  conn->rem.ce_iphost);
            }
         }
         break;
      case H_OTHER:
         dtrap();    /* anything to track here? later -JB- */
         break;
      default:
         dtrap();
         break;
   }

   return conn;
}


/* br_cleardbs() - Utility routine to clear all the in-ram 
 * runtime database information - connections, hosts, and IP 
 * addresses. This does NOT affect the blocklists.
 */

void
br_cleardbs()
{
   int   i;
   struct ip_hostentry * iphost;
   struct conn_entry * conn;
   struct host_entry * host;
   int   rxstate[MAXNETS];       /* for saving iface receiver states */

   /* disable the receiver(s) */
   for(i = 0; i < (int)ifNumber; i++)
   {
      rxstate[i] = brstates[i].rx_disabled;
      brstates[i].rx_disabled = TRUE;
   }

   /* Make sure we're is not in received packet logic */
   if(br_inreceiver)
   {
      // dtrap(); - yes, this does happen. -JB-
      while(br_inreceiver)
      {
         TH_SLEEP(1);
      }
   }

   /* clear the connections ram database first, since the 
    * delete code references iphosts 
    */
   for(i = 0; i < HASH_CONN_ENTRIES; i++)
   {
      conn = ce_table[i];
      while(conn)
         conn = conn_delete(conn, i);
   }


   /* clear the IP hosts ram database */
   for(i = 0; i < HASH_IPHOSTS; i++)
   {
      iphost = ip_hosttab[i];
      while(iphost)
      {
         /* clear back pointer in MAC host, if any */
         if((iphost->mac_host != NULL) &&
            (iphost->mac_host->iphost == iphost))
         {
            iphost->mac_host->iphost = NULL;
         }
         iphost = iphost_delete(iphost, i);
      }
   }

   /* clear the MAC hosts ram database */
   for(i = 0; i < HASH_HOSTS; i++)
   {
      host = hosttab[i];
      while(host)
      {
         host = host_delete(host, i);
      }
   }

   /* restore the receiver(s) state */
   for(i = 0; i < (int)ifNumber; i++)
   {
      brstates[i].rx_disabled = rxstate[i];
   }
   return;
}

/*
Copyright (C) 2003-2008 by Praemio Inc. All rights reserved.

Author: John Bartas - jbartas@praemio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/

