/* brutils.c
 *
 * Misc utilities
 *
 * Copyright 2003-2008 by Praemio Inc. All rights reserved.
 * Author: John Bartas - jbartas@praemio.com
 * This software is released under the terms of the LGPL.
 * See LGPL Notice at end of this file; Full license in ../license.txt
 */

#include "busterport.h"
#include "brnet.h"
#include "ip.h"
#include "brdb.h"



/* nextarg()
 *
 * return next arg in string passed. Args are cstrings with 
 * whitespace delimiting args. If pointer passed is last 
 * arg then a pointer to the NULL terminator is returned.
 *
 * Returns pointer to next arg in string
 */

char *
nextarg(char * argp)
{
   while (*argp > ' ')
      argp++; /* scan past current arg */
   while ((*argp <= ' ') && (*argp > 0))
      argp++;   /* scan past whitespace */
   return(argp);
}


/* ishexdigit()
 *
 * Return TRUE if passed digit is hex, else return FALSE
 */

int
ishexdigit(char chr)
{
   if (isdigit(chr))    /* Get the 0-9 cases */
      return TRUE;

   /* Fall to here to check letters */
   chr |= 0x20;       /* Make lowercase */
   if ((chr >= 'a') && (chr <= 'f'))
      return TRUE;
   else
      return FALSE;
}



/* hexnibble()
 * 
 * Assume passed hexdigit is a nibble (bit bit hex), conver to 
 * a value and retuenit
 */

unsigned hexnibble(char hexdigit)
{
   if (hexdigit <= '9')
      return (hexdigit-'0');

   hexdigit &= ~0x20;   /* uppercase */
   return (hexdigit-'A')+10 ;
}



/* atoh()
 *
 * Return the value for the passed hex digits
 */

u_long
atoh(char * input)
{
   u_long   retval = 0;
   char *   cp;
   char     digit;

   cp = input;

   /* skip leading spaces and tabs */
   while (*cp == ' ' || *cp == 9)
      cp++;

   while((digit = *cp++) != 0)
   {
      if (digit == 'x')
         continue;
      if(!ishexdigit(digit))
          break;
      retval = (retval << 4) + hexnibble(digit);
   }

   return retval;
}


u_long
atox(char * hextext)
{
    u_long   retval;
    u_char   nibble;

    if(!hextext)
        return 0;

    retval = 0;

    while(ishexdigit(*hextext))
    {
        nibble = (u_char)(*hextext++);
        if(nibble <= '9')
            nibble -= '0';
        else if(nibble & 0x40)  /* lowercase letter? */
            nibble -= ('a' - 10);
        else              /* uppercase letter */
            nibble -= ('A' - 10);

        retval += nibble;

        if(retval & 0xF0000000)    /* retval is full */
            break;

        /* If more digits, shift retval 4 bits and loop */
        if(ishexdigit(*hextext))
            retval <<= 4;
    }

    return retval;
}


/* hex2bin()
 *
 * Convert long hexadecimal text string into binary sequence. Passed
 * count is number of bytes in "input", which should be even. Output
 * should be at must 1/2/ the size of input
 *
 * Returns 0 if OK else error code 
 */

int
hex2bin(char * input, unsigned char * output, int count)
{
   int      i;
   char *   cp = input;

   for (i = 0; i < count; i++)
   {
      if (ishexdigit(*cp))
         output[i] = (unsigned char)(hexnibble(*cp) << 4);
      else
         return -1;
      cp++;
      if (ishexdigit(*cp))
         output[i] |= (unsigned char)hexnibble(*cp);
      else
         return -1;
      cp++;
   }
   return 0;
}

/* hexdigit() - helper routine to convert binary bytes into hexadecimal 
 * text. Accepts a nibble as input, 
 *
 * Returns a character 0-9 or A-F for display
 */

char
hexdigit(u_char byte)
{
   if(byte <= 9)
      return (byte + '0');

   /* else byte is A-F */
   return (byte + ('A' - 0x0A) );
}


/* getlong() 
 *
 * Exctract a 32-bit value from 4 bytes of an input stream. This
 * coult be a cast like "long l = *(long*)input" except that the
 * the location of input may not be 32-bit aligned - problem for
 * some processors.
 *
 * Returns pointer to next char after long. 
 * Returned long value is left in out, 
 */

char * 
getlong(char * input, u_long * out)
{
   u_char * uchars = (u_char*)input;

   *out = (u_long)(*uchars++) << 24;
   *out |= (u_long)(*uchars++) << 16;
   *out |= (u_long)(*uchars++) << 8;
   *out |= (u_long)(*uchars++);
   return ( (char*)uchars);
}


/* getshort()
 * 
 * 16 bit version of getlong()
 */

char * 
getshort(char * input, u_short * out)
{
   *out = (u_long)(*input++) << 8;
   *out += (u_long)(*input++);
   return (input);
}

/* br_checkip() 
 * 
 * check "input" for valid dot notation IP address, return IP in 
 * "out" if format is OK. This is meant for parsing code from user
 * typed input, so it should be careful with checking.
 *
 * Return NULL if OK or error string describing problem.
 */

char *
br_checkip(ip_addr * out, char * input)
{
    int         octet;  /* Counter, 1-4 octets */
    unsigned    value;
    char *      cp;
    u_long      ipval;

    ipval = 0;
    cp = input;

    /* Note - try for 5 values; however finding more than 4 
     * is error. Correct loop exit if via "!cp".
     */
    for(octet = 1; octet < 5; octet++)
    {
        value = atoi(cp);
        if((value == 0) && (*cp != '0'))
            return ("All chars must be digits or dots");

        if(value > 255)
            return ("All values must be 0-255");

        ipval = (ipval << 8) + value;

        cp = strchr(cp, '.');
        if(!cp)
            break;
        cp++;
    }

    if((octet > 4) || (octet < 2))
        return ("Must be 1 to 3 dots");

    /* If there are missing ocets (e.g. "10.1") then shift the high and low 
     * values to allow the traditional shorthand.
     */
    if(octet == 2)
    {
        ipval = (ipval & 0x000000FF) + ((ipval & 0x0000FF00) << 16);
    }
    else if(octet == 3)
    {
        ipval = (ipval & 0x000000FF) + ((ipval & 0x00FFFF00) << 8);
    }

    *out = htonl(ipval);
    return NULL;        /* No error */
}


/* br_printdigest(() - convert MD5 digest to printable ascii */

void
        br_printdigest(char * text, u_char * digest, int digest_len)
{
    int   i, j;
    j = 0;
    for(i = 0; i < digest_len; i++)
    {
        text[j++] = (digest[i] & 0x000f) + 'A';
        text[j++] = (digest[i] >> 4) + 'J';
    }
    text[j] = 0;       /* Null terminate string */
    return;
}

/* tcp_cksum() 
 * 
 * non-destructive TCP checksum verification. Returns corret checksum of 
 * passed unfragmented IP/TCP packet. 
 * 
 * From old non-copyrighted NetPort software code.
 */

unshort
tcp_cksum(struct ip * pip)
{
   unshort  iphlen;     /* length of IP header, usually 20 */
   unshort  tcplen;     /* length of TCP header and data */
   char *   tcpdata;    /* pointer to end of tcp data, for 0 padding */
   char     oddchar;    /* tmp save of odd char we 0 for pad */
   unshort  oldsum;
   unshort  newsum;
   struct tcphdr * tp;

   iphlen = (unshort)ip_hlen(pip);
   tcplen = htons(pip->ip_len) - iphlen;

   tp = (struct tcphdr*)ip_data(pip);   /* get TCP header */
   oldsum = tp->th_sum;                 /* Save passed checksum */

   /* The TCP checksum spec. says to Init TCP header cksum field to 0 before
   * summing. We also need to inclue in the sum a protocol type (always 6
   * for TCP). the Ip length, and the IP addresses. Nominally this is done 
   * by constructing a "pseudo header" and prepending it to the tcp header
   * and data to be summed. Instead, we add the type & length (they 
   * cannot overflow a 16 bit field) and put them in the cksum field. 
   * We include the IP addresses by passing them to the lower level 
   * fast sum routine. This results in their values being factored into 
   * the sum and the cksum field contributes zero.
   */
   tp->th_sum = htons(tcplen + 6);

   /* zero pad odd sized packets for checksumming */
   if (tcplen & 1)   
   {
       tcpdata = ((char*)pip) + iphlen + tcplen;    /* end of packet */
       oddchar = *tcpdata;
       *tcpdata = 0;  /* zero out pad byte */
       tcplen++;      /* bump length to pass to cksum() */
   }
   else
   {
        tcpdata = NULL; /* Use as flag */
        oddchar = 0;
   }

   /* Pass a pointer to the beginning of the IP address area into the IP header
    * the the low level sum routine. Add the size of these two IP addresses to
    * the length, and convert the length to 16 bit words.
    */
   newsum = ~cksum( ((char*)tp) - 8, (tcplen + 8)/2 );

   /* restore what we clobbered */
   tp->th_sum = oldsum;   /* put back passed checksum */
   if (tcpdata)
      *tcpdata = oddchar;  /* restore odd byte if we zeroed it */

   return newsum;
}

static char printed_addr[24];

char *  print_mac(u_char * macaddr)
{
    sprintf( printed_addr, "%02X %02X %02X-%02X %02X %02X", 
             (unsigned)(*macaddr),
             (unsigned)(*(macaddr+1)),
             (unsigned)(*(macaddr+2)),
             (unsigned)(*(macaddr+3)),
             (unsigned)(*(macaddr+4)),
             (unsigned)(*(macaddr+5)) );

    return printed_addr;
}

/* br_pwdigest() - Return the MD5 string for a passed password. */

#include "brmd5.h"

static MD5_CTX user_digest;
static char hexdigest[MAXUSERLENGTH + 2];

#if(MAXUSERLENGTH < 32)
#error MAXUSERLENGTH must be 32 or more
#endif

char *
br_pwdigest(char * password, u_long mask)
{

    MD5Init(&user_digest);
    MD5Update(&user_digest, (u_char*)password, strlen(password) );
    MD5Update(&user_digest, (u_char*)(&mask), 4);
    MD5Final(&user_digest);

    /* convert digest to printable ascii */
    br_printdigest(hexdigest, user_digest.digest, 16 );

    return (hexdigest);
}


/* base64_encode()/decode() - found on the net copyrighted
 * under the LGPL, which means we MUST NOT MODIFY IT!!!! 
 */

/*
 *  RFC 1521 base64 encoding/decoding
 *
 *  Copyright (C) 2006-2007  Christophe Devine
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License, version 2.1 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *  MA  02110-1301  USA
 */
#define ERR_BASE64_INVALID_CHARACTER    ENP_PARAM
#define ERR_BASE64_BUFFER_TOO_SMALL     ENP_RESOURCE

static const int base64_enc_map[64] =
{
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
    'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
    'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd',
    'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
    'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x',
    'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7',
    '8', '9', '+', '/'
};

static const int base64_dec_map[128] =
{
    127, 127, 127, 127, 127, 127, 127, 127, 127, 127,
    127, 127, 127, 127, 127, 127, 127, 127, 127, 127,
    127, 127, 127, 127, 127, 127, 127, 127, 127, 127,
    127, 127, 127, 127, 127, 127, 127, 127, 127, 127,
    127, 127, 127,  62, 127, 127, 127,  63,  52,  53,
    54,  55,  56,  57,  58,  59,  60,  61, 127, 127,
    127,  64, 127, 127, 127,   0,   1,   2,   3,   4,
    5,   6,   7,   8,   9,  10,  11,  12,  13,  14,
    15,  16,  17,  18,  19,  20,  21,  22,  23,  24,
    25, 127, 127, 127, 127, 127, 127,  26,  27,  28,
    29,  30,  31,  32,  33,  34,  35,  36,  37,  38,
    39,  40,  41,  42,  43,  44,  45,  46,  47,  48,
    49,  50,  51, 127, 127, 127, 127, 127
};

/*
 * Encode a buffer into base64 format
 */
int base64_encode( unsigned char *dst, int *dlen,
                   unsigned char *src, int  slen )
{
    int i, n;
    int C1, C2, C3;
    unsigned char *p;

    if( slen == 0 )
        return( 0 );

    n = ( slen << 3 ) / 6;

    switch( ( slen << 3 ) - ( n * 6 ) )
    {
        case  2: n += 3; break;
        case  4: n += 2; break;
        default: break;
    }

    if( *dlen < n + 1 )
    {
        *dlen = n + 1;
        return( ERR_BASE64_BUFFER_TOO_SMALL );
    }

    n = ( slen / 3 ) * 3;

    for( i = 0, p = dst; i < n; i += 3 )
    {
        C1 = *src++;
        C2 = *src++;
        C3 = *src++;

        *p++ = base64_enc_map[( C1 >> 2 ) & 0x3F];
        *p++ = base64_enc_map[((( C1 &  3 ) << 4) + ( C2 >> 4 )) & 0x3F];
        *p++ = base64_enc_map[((( C2 & 15 ) << 2) + ( C3 >> 6 )) & 0x3F];
        *p++ = base64_enc_map[C3 & 0x3F];
    }

    if( i < slen )
    {
        C1 = *src++;
        C2 = ((i + 1) < slen) ? *src++ : 0;

        *p++ = base64_enc_map[( C1 >> 2 ) & 0x3F];
        *p++ = base64_enc_map[((( C1 & 3 ) << 4) + ( C2 >> 4 )) & 0x3F];
        *p++ = ((i + 1) < slen) ?
                base64_enc_map[((( C2 & 15 ) << 2)) & 0x3F] : '=';

        *p++ = '=';
    }

    *dlen = p - dst;

    return( *p = 0 );
}

/*
 * Decode a base64-formatted buffer
 */
int base64_decode( unsigned char *dst, int *dlen,
                   unsigned char *src, int  slen )
{
    int i, j, n;
    unsigned long x;
    unsigned char *p;

    for( i = j = n = 0; i < slen; i++ )
    {
        if( ( slen - i ) >= 2 &&
              src[i] == '\r' && src[i + 1] == '\n' )
            continue;

        if( src[i] == '\n' )
            continue;

        if( src[i] == '=' && ++j > 2 )
            return( ERR_BASE64_INVALID_CHARACTER );

        if( src[i] > 127 || base64_dec_map[src[i]] == 127 )
            return( ERR_BASE64_INVALID_CHARACTER );

        if( base64_dec_map[src[i]] < 64 && j != 0 )
            return( ERR_BASE64_INVALID_CHARACTER );

        n++;
    }

    if( n == 0 )
        return( 0 );

    n = ( ( n * 6 ) + 7 ) >> 3;

    if( *dlen < n )
    {
        *dlen = n;
        return( ERR_BASE64_BUFFER_TOO_SMALL );
    }

    for( j = 3, n = x = 0, p = dst; i > 0; i--, src++ )
    {
        if( *src == '\r' || *src == '\n' )
            continue;

        j -= ( base64_dec_map[*src] == 64 );
        x  = ( x << 6 ) | ( base64_dec_map[*src] & 0x3F );

        if( ++n == 4 )
        {
            n = 0;
            *p++ = (unsigned char) ( x >> 16 );
            if( j > 1 ) *p++ = (unsigned char) ( x >> 8 );
            if( j > 2 ) *p++ = (unsigned char )  x;
        }
    }

    *dlen = p - dst;

    return( 0 );
}

/*
Copyright (C) 2003-2008 by Praemio Inc. All rights reserved.

Author: John Bartas - jbartas@praemio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


