/* brwhois.c
 *
 * Buster "whois" client
 *
 * Copyright 2003-2008 by Praemio Inc. All rights reserved.
 * Author: John Bartas - jbartas@praemio.com
 * This software is released under the terms of the LGPL.
 * See LGPL Notice at end of this file; Full license in ../license.txt
 *
 *
 */

#include "busterport.h"
#include "brnet.h"
#include "ip.h"
#include "brdb.h"
#include "brdns.h"


/* Array of useable whois servers, set from NV and other 
 * configuration data sources. The servers should be in
 * the array in the order we prefer their results.
 */
struct whoserver whoservers[WHOSERVERS];

static int   whois_tryconn(int i);   /* predecl */


/* br_whois(ip_addr ipaddr)
 * 
 * Does a "whopis" lookup on the passed IP address and returns the
 * result as a monolithic text string.
 *
 * Parameter iptype is set of user wants only IP info (not name)
 *
 * Returns negative number if error, else index of whoserver[]
 * with the best result.
 * If there is any error then "*errmsg" is set to a pointer 
 * to problem desctiption.
 */

int
br_whois(ip_addr hostaddr, char ** errmsg, int iptype)
{
   int      i;
   int      err;
   int      failed;              /* counter of failed sessions */
   int      servers = 0;         /* active whois servers */
   int      typeflag;            /* type flag bit for fast test */
   char *   cp;                  /* scratch */
   u_long   whotimeout;          /* ctick when we quite waiting */
   ip_addr  srvaddr;             /* whois server IP address */
   struct ip_hostentry * iphost; /* IP host to look up */
   char     ipbuf[16];           /* buffer for text IP address */
   char     domain[DNSNAMELEN];  /* name of host to whois */
   int      favorite = -1;       /* server (index) we prefer */
   int      empty = -1;          /* index of empty server slot */


   /* Make sure at least the first server is set */
   if(whoservers[0].dnsname[0] == 0)
   {
      *errmsg = "No Whois servers available";
      return -1;
   }
   *errmsg = NULL;    /* assume no error */

   iphost = br_ipanyif(hostaddr);
   if(!iphost || (iphost->dnsname[0] == 0))
   {
      /* Can't do a name lookup with no host, or no
       * DNS name, so force IP lookup. 
       */
      iptype = TRUE;
   }
   else
   {
      /* after this either iptype or dnsname is set */
      cp = strrchr(iphost->dnsname, '.');   /* find last dot */
      if(!cp)
         iptype = TRUE;
      else
      {
         cp--;
         while((*cp != '.') && (cp > iphost->dnsname))
            cp--;
         if(*cp == '.')
            cp++;
         strncpy(domain, cp, MAXDNSNAME);
      }
   }

   /* set flag for fast checking of server flags */
   if(iptype)
      typeflag = WH_IPTYPE;
   else
      typeflag = WH_NAMETYPE;

   /* fire up sessions on all the whois servers */
   for(i = 0; i < WHOSERVERS; i++)
   {
      /* skip unset and wrong-type server entries */
      if((whoservers[i].flags & typeflag) == 0)
      {
         if((whoservers[i].flags == 0) && (empty == -1))
            empty = i;
         continue;
      }
   
      if(favorite == -1 )
         favorite = i;
      whoservers[i].state = WHO_DNSING;
      whoservers[i].received = 0;
      servers++;
   }

   if(servers == 0)
   {
      *errmsg = "No whois servers configured";
      return -1;
   }
   /* run all the whois clients simultaineously for a few seconds */
   whotimeout = cticks + (TPS * 5);
   while(whotimeout > cticks)
   {
      failed = 0;

      for(i = 0; i < WHOSERVERS; i++)
      {
         /* skip unset and wrong-type server entries */
         if((whoservers[i].flags & typeflag) == 0)
            continue;

         switch(whoservers[i].state)
         {
         case WHO_DNSING:
            /* first get IP address for whois server */
            srvaddr = br_dnslookup(whoservers[i].dnsname);
            if(srvaddr == 0)
            {
               whoservers[i].state = WHO_FAILED;
               break;
            }
            else
            {
               /* DNS OK, start open of TCP connection */
               whoservers[i].ipaddr = srvaddr;
               whoservers[i].socket = socket(AF_INET, SOCK_STREAM, 0);
               if(whoservers[i].socket == 0)
               {
                  dtrap();
                  whoservers[i].state = WHO_FAILED;
                  break;
               }
               err = br_nbiosocket(whoservers[i].socket);
               if(err == -1)
               {
                  dtrap();
                  whoservers[i].state = WHO_FAILED;
                  break;
               }
               whoservers[i].state = whois_tryconn(i);
            }
            break;
         case WHO_CONNECTING:
            /* see if the connection is open yet */
            whoservers[i].state = whois_tryconn(i);
            break;
         case WHO_CONNECTED:
            /* we should probably poll select at this point, but it's 
             * easier to just try the send and check the error code 
             * to see if the connection is still forming.
             */
            if(iptype)
               sprintf(ipbuf, "%u.%u.%u.%u\r\n", PASS_IPADDR(hostaddr));
            else
               sprintf(ipbuf, "%s\r\n", domain);

            err = send(whoservers[i].socket, ipbuf, strlen(ipbuf), 0);
            if(err < 0)
            {
               err = sock_errno(whoservers[i].socket);
               whoservers[i].state = WHO_FAILED;
               socketclose(whoservers[i].socket);
               whoservers[i].socket = 0;
            }
            else     /* query sent OK, wait for answer */
            {
               whoservers[i].state = WHO_ANSWERING;
            }
            break;
         case WHO_ANSWERING:
            err = recv(whoservers[i].socket, 
               whoservers[i].whobuf + whoservers[i].received, 
               WHOBUFSIZE - whoservers[i].received, 0);
            if(err == 0)   /* end of data */
            {
               char * regname;

               whoservers[i].state = WHO_DONE;
               whoservers[i].whobuf[whoservers[i].received] = 0;  /* null term */
               socketclose(whoservers[i].socket);
               whoservers[i].socket = 0;
               /* see if this result gave us a registrar server */
               regname = strstr(whoservers[i].whobuf, "Whois Server:");
               if( regname && 
                  (empty != -1) && 
                  (whoservers[empty].flags == 0))
               {
                  MEMSET(&whoservers[empty], 0, sizeof(whoservers[empty]));
                  regname += 14;    /* point past "Whois Server:" */
                  strncpy(whoservers[empty].dnsname, regname, DNSNAMELEN);

                  cp = whoservers[empty].dnsname;
                  while (*cp > ' ')
                     cp++;
                  *cp = 0;    /* null terminate string */

                  /* set up a new session to this server */
                  whoservers[empty].state = WHO_DNSING;
                  whoservers[empty].flags = typeflag;
                  whoservers[empty].whoport = 43;
                  favorite = empty;
               }
               /* If this was our preferred server, stop now */
               if(i == favorite)
                  goto whois_done;
            }
            else if(err < 0)  /* hard error */
            {
               err = sock_errno(whoservers[i].socket);
               if(err == EWOULDBLOCK)
                  break;
               else
                  whoservers[i].state = WHO_FAILED;
            }
            else     /* received some data */
            {
               whoservers[i].received += err;
            }
            break;
         case WHO_FAILED:
            failed++;
            break;
         case WHO_DONE:
            /* just wait for the other servers */
            break;
         default:
            dtrap();
            break;
         }
      }
      /* If all the sessions have failed, don't waste time here */
      if(failed >= servers)
         break;
      TH_SLEEP(1);
   }

whois_done:
   /* close any stray sockets */
   for(i = 0; i < WHOSERVERS; i++)
   {
      if(whoservers[i].socket)
      {
         socketclose(whoservers[i].socket);
         whoservers[i].socket = 0;
      }
   }

   if(empty != -1)
      whoservers[empty].flags = 0;     /* free tmp entry */

   /* return answer from the best server that responded */
   if((favorite != -1) &&
      (whoservers[favorite].state == WHO_DONE))
   {
      return (favorite);
   }
   for(i = 0; i < WHOSERVERS; i++)
   {
      if(whoservers[i].state == WHO_DONE)
         return (i);
   }

   /* fall to here if no server worked */
   *errmsg = "No whois servers responded";
   return -1;
}


/* whois_tryconn() - helper routine to start & poll whois
 * tcp connections to a server indicated by passed index.
 *
 * Returns state code for whois session.
 */

static int
whois_tryconn(int i)
{
   int err;
   struct sockaddr_in sin;

   sin.sin_addr.s_addr = whoservers[i].ipaddr;
   sin.sin_port = htons(whoservers[i].whoport);
   sin.sin_family = AF_INET;
   err = connect(whoservers[i].socket, (struct sockaddr*)&sin, sizeof(sin) );

   /* This return handling is heavly modified to support Winsock. 
    * Hopefully it still works on real sockets; if not it may need 
    * to be ifdeffed. 7/05. -JB-
    */
   if(err < 0)
   {
      err = sock_errno(whoservers[i].socket);
      if((err == EINPROGRESS) ||
         (err == EALREADY) ||
         (err == EINVAL) ||
         (err == EWOULDBLOCK))
      {
         return WHO_CONNECTING;
      }
      else if(err == EISCONN)   /* per MS Winsock spec */
      {
         return WHO_CONNECTED;
      }
      else
      {
         return WHO_FAILED;
      }
   }

   return WHO_CONNECTED;
}

/*
Copyright (C) 2003-2008 by Praemio Inc. All rights reserved.

Author: John Bartas - jbartas@praemio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


