/*
 * FILENAME: btree.c
 *
 * Copyright 2003-2008 by Praemio Inc. All rights reserved.
 * Author: John Bartas - jbartas@praemio.com
 * This software is released under the terms of the LGPL.
 * See LGPL Notice at end of this file; Full license in ../license.txt
 *
 * Binary tree code, heavily based on the public domain
 * "avltree.c" file. The original avltree.c is not copyrighted and is
 * explicitly assigned to the public domain. Code is based on the
 * ancient Algorithm documented by Knuth.
 *
 * This version is modified to assume the left/right flag and pointers are at
 * the beginning of the structure to be filed, rather than a seperate
 * memory item.
 */


#include "busterport.h"

#include "btree.h"


/* FUNCTION: bt_rotleft()
 * 
 *  bt_rotleft: perform counterclockwise rotation
 *
 *  Parameters:
 *
 *    n           Address of a pointer to a node
 *
 * PARAM1: struct btnode * n
 *
 * RETURNS: 
 */

void
bt_rotleft(struct btnode ** n)
{
	struct btnode * tmp = *n;

	*n = (*n)->right;
	tmp->right = (*n)->left;
	(*n)->left = tmp;
}



/* FUNCTION: bt_rotright()
 * 
 *  bt_rotright: perform clockwise rotation
 *
 *  Parameters:
 *
 *    n           Address of a pointer to a node
 *
 * PARAM1: struct bt_node *n
 *
 * RETURNS: 
 */

void
bt_rotright(struct btnode ** n)
{
	struct btnode * tmp = *n;

	*n = (*n)->left;
	tmp->left = (*n)->right;
	(*n)->right = tmp;
}



/* FUNCTION: bt_leftgrown()
 * 
 *  bt_leftgrown: helper function for bt_insert
 *
 *  Parameters:
 *
 *    n           Address of a pointer to a node. This node's left 
 *                subtree has just grown due to item insertion; its 
 *                "skew" flag needs adjustment, and the local tree 
 *                (the subtree of which this node is the root node) may 
 *                have become unbalanced.
 *
 *  Return values:
 *
 *    TREE_OK          The local tree could be rebalanced or was balanced 
 *                from the start. The parent activations of the bt_insert 
 *                activation that called this function may assume the 
 *                entire tree is valid.
 *
 *    TREE_BALANCE     The local tree was balanced, but has grown in height.
 *                Do not assume the entire tree is valid.
 *
 * PARAM1: struct btnode *n
 *
 * RETURNS: 
 */

enum BT_RES
bt_leftgrown(struct btnode ** n)
{
   struct btnode * rtp = *n;

   switch (rtp->skew) 
   {
   case BT_LEFT:
      if (rtp->left->skew == BT_LEFT) 
      {
         rtp->skew = rtp->left->skew = BT_NONE;
         bt_rotright(n);
      }
      else 
      {
         switch (rtp->left->right->skew) 
         {
         case BT_LEFT:
            rtp->skew = BT_RIGHT;
            rtp->left->skew = BT_NONE;
            break;

         case BT_RIGHT:
            rtp->skew = BT_NONE;
            rtp->left->skew = BT_LEFT;
            break;

         default:
            rtp->skew = BT_NONE;
            rtp->left->skew = BT_NONE;
         }
         rtp->left->right->skew = BT_NONE;
         bt_rotleft(&(rtp->left));
         bt_rotright(n);
      }
      return TREE_OK;

   case BT_RIGHT:
      rtp->skew = BT_NONE;
      return TREE_OK;

   default:
      rtp->skew = BT_LEFT;
      return TREE_BALANCE;
   }
}



/* FUNCTION: bt_rightgrown()
 * 
 *  bt_rightgrown: helper function for bt_insert
 *
 *  See bt_leftgrown for details.
 *
 * PARAM1: struct btnode *n
 *
 * RETURNS: 
 */

enum BT_RES
bt_rightgrown(struct btnode ** n)
{
   struct btnode * rtp = *n;

   switch (rtp->skew) 
   {
   case BT_LEFT:     
      rtp->skew = BT_NONE;
      return TREE_OK;

   case BT_RIGHT:
      if (rtp->right->skew == BT_RIGHT) 
      { 
         rtp->skew = rtp->right->skew = BT_NONE;
         bt_rotleft(n);
      }
      else 
      {
         switch (rtp->right->left->skew) 
         {
         case BT_RIGHT:
            rtp->skew = BT_LEFT;
            rtp->right->skew = BT_NONE;
            break;

         case BT_LEFT:
            rtp->skew = BT_NONE;
            rtp->right->skew = BT_RIGHT;
            break;

         default:
            rtp->skew = BT_NONE;
            rtp->right->skew = BT_NONE;
         }
         rtp->right->left->skew = BT_NONE;
         bt_rotright(&(rtp->right));
         bt_rotleft(n);
      }
      return TREE_OK;

   default:
      rtp->skew = BT_RIGHT;
      return TREE_BALANCE;
   }
}



/* FUNCTION: bt_insert()
 * 
 *  bt_insert: insert a node into the AVL tree.
 *
 *  Parameters:
 *
 *    n           Address of a pointer to a node.
 *
 *    d           Item to be inserted.
 *
 *  Return values:
 *
 *    TREE_ERROR    The datum provided could not be inserted, either due 
 *                to AVLKEY collision (the tree already contains another
 *                item with which the same AVLKEY is associated), or
 *                due to insufficient memory.
 *
 *    TREE_OK       None of the subtrees of the node that n points to 
 *                has grown, the AVL tree is valid.
 *
 *    TREE_BALANCE  One of the subtrees of the node that n points to 
 *                has grown, the node's "skew" flag needs adjustment,
 *                and the AVL tree may have become unbalanced.
 *
 * That latter two return values are distinguished so that when avlinsert
 * recursively calls itself, the number returned tells the parent activation
 * if the AVL tree may have become unbalanced.
 *
 *
 * PARAM1: struct btnode ** parent
 * PARAM2: struct btnode *  d
 *
 * RETURNS: 
 */

enum BT_RES
bt_insert(struct btnode ** parent,
         struct btnode * new,
         u_long new_key,
         int (* cmpfunc)(struct btnode *, u_long key) )
{
   enum BT_RES tmp;
   enum BT_REL eval;

#ifdef NPDEBUG
   if(new->data == NULL)
   {
      dtrap();
   }
#endif   /* NPDEBUG */

   if (!(*parent))
   {
      *parent = new;
      new->left = new->right = NULL;
      new->skew = BT_NONE;
      return TREE_BALANCE;
   }

   eval = cmpfunc(*parent, new_key);
   if (eval == KEY_LT)
   {
      if ((tmp = bt_insert(&(*parent)->left, new, new_key, cmpfunc)) == TREE_BALANCE) 
      {
         return bt_leftgrown(parent);
      }
      return tmp;
   }
   if (eval == KEY_GT) 
   {
      if ((tmp = bt_insert(&(*parent)->right, new, new_key, cmpfunc)) == TREE_BALANCE) 
      {
         return bt_rightgrown(parent);
      }
      return tmp;
   }
   return TREE_ERROR;
}



/* FUNCTION: bt_leftshrunk()
 * 
 *  bt_leftshrunk: helper function for bt_remove and bt_findlowest
 *
 *  Parameters:
 *
 *    n           Address of a pointer to a node. The node's left
 *                subtree has just shrunk due to item removal; its
 *                "skew" flag needs adjustment, and the local tree
 *                (the subtree of which this node is the root node) may
 *                have become unbalanced.
 *
 *   Return values:
 *
 *    TREE_OK          The parent activation of the bt_remove activation
 *                that called this function may assume the entire
 *                tree is valid.
 *
 *    TREE_BALANCE     Do not assume the entire tree is valid.
 *
 * PARAM1: struct btnode **n
 *
 * RETURNS: 
 */

enum BT_RES
bt_leftshrunk(struct btnode ** n)
{
   struct btnode * rtp = *n;

   switch (rtp->skew) 
   {
   case BT_LEFT:
      rtp->skew = BT_NONE;
      return TREE_BALANCE;

   case BT_RIGHT:
      if (rtp->right->skew == BT_RIGHT) 
      {
         rtp->skew = rtp->right->skew = BT_NONE;
         bt_rotleft(n);
         return TREE_BALANCE;
      }
      else if (rtp->right->skew == BT_NONE) 
      {
         rtp->skew = BT_RIGHT;
         rtp->right->skew = BT_LEFT;
         bt_rotleft(n);
         return TREE_OK;
      }
      else 
      {
         switch (rtp->right->left->skew) 
         {
         case BT_LEFT:
            rtp->skew = BT_NONE;
            rtp->right->skew = BT_RIGHT;
            break;

         case BT_RIGHT:
            rtp->skew = BT_LEFT;
            rtp->right->skew = BT_NONE;
            break;

         default:
            rtp->skew = BT_NONE;
            rtp->right->skew = BT_NONE;
         }
         rtp->right->left->skew = BT_NONE;
         bt_rotright(&(rtp->right));
         bt_rotleft(n);
         return TREE_BALANCE;
      }

   default:
      rtp->skew = BT_RIGHT;
      return TREE_OK;
   }
}


/* FUNCTION: bt_rightshrunk()
 * 
 *  bt_rightshrunk: helper function for bt_remove and bt_findhighest
 *
 *  See bt_leftshrunk for details.
 *
 * PARAM1: struct btnode **n
 *
 * RETURNS: 
 */

enum BT_RES
bt_rightshrunk(struct btnode ** n)
{
   struct btnode * rtp = *n;

   switch (rtp->skew) 
   {
   case BT_RIGHT:
      rtp->skew = BT_NONE;
      return TREE_BALANCE;

   case BT_LEFT:
      if (rtp->left->skew == BT_LEFT) 
      {
         rtp->skew = rtp->left->skew = BT_NONE;
         bt_rotright(n);
         return TREE_BALANCE;
      }
      else if (rtp->left->skew == BT_NONE) 
      {
         rtp->skew = BT_LEFT;
         rtp->left->skew = BT_RIGHT;
         bt_rotright(n);
         return TREE_OK;
      }
      else 
      {
         switch (rtp->left->right->skew) 
         {
         case BT_LEFT:
            rtp->skew = BT_RIGHT;
            rtp->left->skew = BT_NONE;
            break;

         case BT_RIGHT:
            rtp->skew = BT_NONE;
            rtp->left->skew = BT_LEFT; 
            break;

         default:
            rtp->skew = BT_NONE;
            rtp->left->skew = BT_NONE;
         }
         rtp->left->right->skew = BT_NONE;
         bt_rotleft(&(rtp->left));
         bt_rotright(n);
         return TREE_BALANCE;
      }

   default:
      rtp->skew = BT_LEFT;
      return TREE_OK;
   }
}


/* FUNCTION: bt_findhighest()
 * 
 *  bt_findhighest: replace a node with a subtree's highest-ranking item.
 *
 *  Parameters:
 *
 *    target      Pointer to node to be replaced.
 *
 *    tree        Address of pointer to subtree.
 *
 *    res         Pointer to variable used to tell the caller whether
 *                further checks are necessary; analog to the return
 *                values of bt_leftgrown and bt_leftshrunk (see there). 
 *
 *
 * PARAM1: struct btnode *target
 * PARAM2: struct btnode **n
 * PARAM3: enum BT_RES *res
 *
 * RETURNS: 
 *
 *    1           A node was found; the target node has been replaced.
 *
 *    0           The target node could not be replaced because
 *                the subtree provided was empty.
 *
 */

int
bt_findhighest(struct btnode * target, struct btnode ** tree, enum BT_RES * res)
{
   struct btnode * tmp;

   *res = TREE_BALANCE;
   if (!(*tree)) 
   {
      return 0;
   }
   if ((*tree)->right) 
   {
      if (!bt_findhighest(target, &(*tree)->right, res)) 
      {
         return 0;
      }
      if (*res == TREE_BALANCE) 
      {
         *res = bt_rightshrunk(tree);
      }
      return 1;
   }

   /* Fall to here if "*tree" is the highest item in the sub tree.
    * Rather than getting rid of the target node and adjusting 
    * all the pointers around it to the sub-node, we move the 
    * data pointer from the highest sub-node into the target
    * node, and get rid of the sub-node - it's less work.
    */

   /* Sanity check target/data back pointer */
   if( *target->data != target)
      dtrap();
   
   target->data = (*tree)->data; /* move highest sub-node data into target */
   *(target->data) = target;     /* Fix data item's back pointer */

   target->data = (*tree)->data; /* move highest sub-node data into target */
   tmp = *tree;            /* save node to delete */
   *tree = (*tree)->left;  /* unlink node to delete */
   BT_FREE(tmp);            /* delete */

   return 1;
}


/* FUNCTION: bt_findlowest()
 * 
 *  bt_findlowest: replace node with a subtree's lowest-ranking item.
 *
 *  See bt_findhighest for the details.
 *
 * PARAM1: struct btnode *target
 * PARAM2: struct btnode **n
 * PARAM3: enum BT_RES *res
 *
 * RETURNS: 
 */

int
bt_findlowest(struct btnode * target, struct btnode **tree, enum BT_RES * res)
{
   struct btnode * tmp;

   *res = TREE_BALANCE;
   if (!(*tree))
   {
      return 0;
   }
   if ((*tree)->left)
   {
      if (!bt_findlowest(target, &(*tree)->left, res)) 
      {
         return 0;
      }
      if (*res == TREE_BALANCE) 
      {
         *res =  bt_leftshrunk(tree);
      }
      return 1;
   }

   /* Check target/data back pointer */
   if( *target->data != target)
      dtrap();
   
   target->data = (*tree)->data; /* move highest sub-node data into target */
   *(target->data) = target;

   tmp = *tree;            /* save node to delete */
   *tree = (*tree)->right; /* unlink node to delete */
   BT_FREE(tmp);            /* delete */

   return 1;
}



/* FUNCTION: bt_remove()
 * 
 *  bt_remove: remove an item from the tree.
 *
 *  Parameters:
 *
 *    n           Address of a pointer to a node.
 *
 *    key         bt_KEY of item to be removed.
 *
 *
 * PARAM1: struct btnode **n
 * PARAM2: int key
 *
 * RETURNS: 
 *
 *   TREE_OK        None of the subtrees of the node that n points to
 *                has shrunk, the bt_ tree is valid.
 *
 *   TREE_BALANCE   One of the subtrees of the node that n points to
 *                has shrunk, the node's "skew" flag needs adjustment,
 *                and the bt_ tree may have become unbalanced.
 *
 *   TREE_ERROR     The tree does not contain an item yielding the
 *                bt_KEY value provided by the caller.
 *
 * If return is not TREE_ERROR then the item has been removed. The exact
 * value of nonzero yields if of no concern to user code; when
 * bt_remove recursively calls itself, the number returned tells
 * the parent activation if the AVL tree may have become unbalanced;
 * specifically:
 */

enum BT_RES
bt_remove(struct btnode **n,
         u_long key,
         int (* cmpfunc)(struct btnode *, u_long))
{
   enum BT_RES tmp;
   enum BT_REL avlrel;

#ifdef NPDEBUG
   if (!n || !(*n)) 
   {
      dtrap();
      return TREE_ERROR;
   }
#endif /* NPDEBUG */

   avlrel = cmpfunc(*n, key);
   /* recurse if value to delete is on left side of tree (lower) */
   if (avlrel == KEY_LT)
   {
      if ((tmp = bt_remove(&(*n)->left, key, cmpfunc)) == TREE_BALANCE) 
      {
         return bt_leftshrunk(n);
      }
      return tmp;
   }
   /* recurse if value to delete is on right side of tree (higher) */
   if (avlrel == KEY_GT)
   {
      if ((tmp = bt_remove(&(*n)->right, key, cmpfunc)) == TREE_BALANCE) 
      {
         return bt_rightshrunk(n);
      }
      return tmp;
   }

   /* get to here if *n contains value to delete */

   /* find node with which to replace n. If n has a left (lower)
    * branch, move it up to  replace n; else use right.
    */
   if ((*n)->left)
   {
      if (bt_findhighest(*n, &(*n)->left, &tmp)) 
      {
         if (tmp == TREE_BALANCE) 
         {
            tmp = bt_leftshrunk(n);
         }
         return tmp;
      }
   }
   if ((*n)->right)
   {
      if (bt_findlowest(*n, &(*n)->right, &tmp)) 
      {
         if (tmp == TREE_BALANCE) 
         {
            tmp = bt_rightshrunk(n);
         }
         return tmp;
      }
   }
   BT_FREE(*n);
   *n = NULL;
   return TREE_BALANCE;
}


/* FUNCTION: bt_access()
 * 
 *  Retrieve the btree entry for key value
 *
 * PARAM1: struct btnode * n - Pointer to the root node.
 * PARAM2: u_long key      - key for node to be returned.
 *
 * RETURNS: pointer to tree entry containing passed value if
 *          found, else NULL if not found.
 *
 */

struct btnode *
bt_access(struct btnode * n,
      u_long key,
      int (* cmpfunc) (struct btnode *, u_long ))
{
   enum BT_REL avlrel;

   if (!n)
   {
      return NULL;
   }

#ifdef NPDEBUG
   if(n->data == NULL)
   {
      dtrap();
   }
#endif   /* NPDEBUG */

   avlrel = cmpfunc(n, key);
   if (avlrel == KEY_LT )
   {
      return bt_access(n->left, key, cmpfunc);
   }
   if (avlrel == KEY_GT )
   {
      return bt_access(n->right, key, cmpfunc);
   }
   return (n);
}


/* FUNCTION: bt_depthfirst()
 * 
 *  bt_depthfirst: depth-first tree traversal.
 *
 *  Parameters:
 *
 *    n          Pointer to the root node.
 *
 *    f          Worker function to be called for every node.
 *
 *    param      Additional parameter to be passed to the
 *               worker function
 *
 *    depth      Recursion depth indicator. Allows the worker function
 *               to determine how many levels the node being processed
 *               is below the root node. Can be used, for example,
 *               for selecting the proper indentation width when
 *               bt_depthfirst ist used to print a tree dump to
 *               the screen.
 *
 *               Most of the time, you will want to call bt_depthfirst
 *               with a "depth" value of zero.
 *
 * PARAM1: struct btnode *n
 * PARAM2: AVLWORKER *f
 * PARAM3: int param
 * PARAM4: int depth
 *
 * RETURNS: 
 */

void
bt_depthfirst(
   struct btnode * n,
   void(*func)(struct btnode * n, long param, int depth),
   long param,
   int depth)
{
   struct btnode * save_right;
   if(!n)
      return;

   bt_depthfirst(n->left, func, param, depth + 1);
   save_right = n->right;     /* In case func deletes n */
   (*func)(n, param, depth);
   bt_depthfirst(save_right, func, param, depth +1);
}

/*
Copyright (C) 2003-2008 by Praemio Inc. All rights reserved.

Author: John Bartas - jbartas@praemio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


